/************************************************************************/
/*			 	CellMatCopy				*/
/************************************************************************/
/* SCCS INFO @(#)CellMatCopy.c 1.2         last modified P. Strazdins, 99/02/25
 *	first created P. Strazdins DCS ANU 95/09
 */
/* general scaled matrix copy routine with arbitrary row/column strides.
 * compilation: if macro VANILLA defined, `vanilla' versions of the routines
 *              are used (slower, but useful for debugging or vectorization.
 *		if macro NO_ALIGN defined, create version where 
 *		double word alignment in A or B cannot be assumed.
 *			(should be compiled with appropriate non-align. flags)
 *              if macro BLAS defined, Fortran BLAS is used whenever possible;
 */ 
/************************************************************************/
/*#define APP*/

#include "CellMatCopy.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#include "MachineParams.h"

#ifdef NO_ALIGN
# define MatCopy MatCopyW
# define MatAdd  MatAddW
# undef BLAS
# define inline                                 /* so cc can compile this */
#endif
#include "CellMatrix.h"

#ifdef BLAS
# include "FtnBLAS.h"
#endif
 


void MatCopy(M, N, A, alpha, B)
  int M, N; FDATA *alpha; pLocMat A, B;
{ int i; 
  int RIncA = pLM_RInc(A), 		CIncA = pLM_CInc(A), 
      RIncB = pLM_RInc(B), 		CIncB = pLM_CInc(B);
  int RInc = MAX(RIncA, RIncB),		CInc = MAX(CIncA, CIncB);
  int sN, N0;  
  FDATA *Ax = pLM_ADR(A), *Bx = pLM_ADR(B);

#if 0
  { int AOk = pLM_IncOK(M, N, A), BOK = pLM_IncOK(M, N, B);
    if (!AOk || !BOK) {
      printf("AOk=%d, BOK=%d B=%d,%d=%d,%d\n", AOk, BOK, pLM_RInc(B), pLM_CInc(B), RIncB, CIncB);
      printf("BOK=%d %d %d %d\n", LM_IncOK(M, N, *B),
      	      IncOK(M, N, LM_RInc(*B), LM_CInc(*B)), IncOK(M, N, pLM_RInc(B), pLM_CInc(B)),
	      IncOK(M, N, RIncB, CIncB));
    } 	      
  }  
#endif
    
  COMMON_ASSERT_TR(
    (TraceStr, "(%dx%d,a%d %d %d,%s,a%d %d %d)", M, N,
     FD_ADR(Ax), RIncA, CIncA, pFD2Str(alpha), FD_ADR(Bx), pLM_RInc(B), pLM_CInc(B)),
     IncOK(M, N, RIncA, CIncA) && IncOK(M, N, RIncB, CIncB), 0, "MatCopy", 4, TRACE_COMPUTE
  );

#ifdef ULTRA 
  if (RInc>1 && CInc>1 && M>MIN_VECSIZE && N>1 && MIN(RIncA,RIncB)==1 &&
      MIN(CIncA,CIncB)==1) {	/* transpose - choose longest dimension as 
      	 			   inner, irrespective of the strides */
    if (M > N) { 	 			   
      SWAP(int, M, N);            SWAP(int, RInc, CInc);  
      SWAP(int, RIncA, CIncA);    SWAP(int, RIncB, CIncB);                    	
    }
  } else if (M_INNER(M, N, RInc, CInc))                            
#elif defined(APP)
  if (M_INNER(M, N, RIncA, CIncA))
#else
  if (M_INNER(M, N, RInc, CInc))
#endif  
  {  	
    SWAP(int, M, N);		SWAP(int, RInc, CInc);
    SWAP(int, RIncA, CIncA);	SWAP(int, RIncB, CIncB);
  }
  
  N0 = N; sN = 0;

  if (M>1 && CInc>1 && N>MIN_VECSIZE /*ie. a local X-pose on a proper matrix*/
      && TLBSIZE) {
    /* nb assumes RInc>1; gnerally RInc >= CInc in any case */
    N = MAX(TLBSIZE/2, CACHESIZE/2/CInc);
    N = MIN(N0, N); 
  }  

#if 1
/*printf("M=%d,N=%d, dN=%d\n", M, N0, N);        */
#endif
  do {   				/* partition to reduce TLB misses */
#ifdef VANILLA
  if (alpha==NULL) alpha = &FD_1;
  for (i=0; i < M; i++) {
    int j; register FDATA *ax = Ax, *bx = Bx;
    FD_ALPHA_REGS(*alpha);
    for (j=0; j < N; j++) {
      FD_ALPHA_MULSET(*bx, =, *ax);
      ax += CIncA; bx += CIncB;
    }
    Ax += RIncA; Bx += RIncB;
  } /* for (i...) */
#elif defined(BLAS)
  for (i=0; i < M; i++) {
    FTN_COPY(&N, Ax, &CIncA, Bx, &CIncB);
    if (! pFD_UNITY(alpha))
      FTN_SCAL(&N, alpha, Bx, &CIncB);          	
    Ax += RIncA; Bx += RIncB;
  } /* for (i...) */  
#else
  { register int CIncA2, CIncB2, CIncA3, CIncB3, CIncA4, CIncB4;
    register FDATA *ax, *bx;
    FD_register a, t0, t1, t2, t3, b0, b1, b2, b3;  
    /* this forces gcc to map t0-t3 to FPU registers, gen. faster for alpha=1*/ 
    register int j, n;

     if (N >= 4) {			
       CIncA2=CIncA+CIncA, CIncA3=CIncA2+CIncA, CIncA4=CIncA2+CIncA2;
       CIncB2=CIncB+CIncB, CIncB3=CIncB2+CIncB, CIncB4=CIncB2+CIncB2;
     }
    
    if (pFD_UNITY(alpha)) {
      n=8*(N/8);	
      for (i=0; i<M; i++) {
        ax = Ax; bx = Bx;
        j=0;
        bx -= CIncB4;
        for (j=-n; j <0; j+=8) {
          t0 = *ax;
          t1 = *(ax+CIncA);
          t2 = *(ax+CIncA2);
          t3 = *(ax+CIncA3);
          ax += CIncA4; 
          b0 = *ax;
          b1 = *(ax+CIncA);
          bx += CIncB4;
          b2 = *(ax+CIncA2);
          b3 = *(ax+CIncA3);
          *bx = t0; *(bx+CIncB) = t1;
          ax += CIncA4; 
          *(bx+CIncB2) = t2; *(bx+CIncB3) = t3;
          bx += CIncB4;
          *bx = b0; *(bx+CIncB) = b1;
          *(bx+CIncB2) = b2; *(bx+CIncB3) = b3;
        } /* for (j...) */
        bx += CIncB4;
        for (j=n; j < N; j++) {
          *bx = *ax;
          ax += CIncA; bx += CIncB;
        }
        Ax += RIncA; Bx += RIncB;
      } /* for (i...)*/
    } else {
     a = *alpha;
     n=4*(N/4);	
     if (n >= 4) n = n-4;               /* avoid accessing beyond bounds */
     for (i=0; i<M; i++) {
        ax = Ax; bx = Bx;
#ifndef ULTRA
        if (n >= 4) {
                t0 = *ax;
                t1 = *(ax+CIncA);
                                        FD_MULSET(b0, =, a, t0);
                t2 = *(ax+CIncA2);
                                        FD_MULSET(b1, =, a, t1);
          j=0;
          do {                          /*4 way loop unrolling*/
                t3 = *(ax+CIncA3);
                                        FD_MULSET(b2, =, a, t2);
	     ax += CIncA4;
                                                        *bx = b0;
                t0 = *(ax);
                                        FD_MULSET(b3, =, a, t3);
                                                        *(bx+CIncB) = b1;
             j += 4;
                t1 = *(ax+CIncA);
                                        FD_MULSET(b0, =, a, t0);
                                                        *(bx+CIncB2) = b2;
 
                t2 = *(ax+CIncA2);
                                        FD_MULSET(b1, =, a, t1);
                                                        *(bx+CIncB3) = b3;
             bx += CIncB4;
          } while (j < n);
        } /* if (n....) */
#else
        if (n >= 4) {
/*             bx -= CIncB4; */
                t0 = *ax;
                t1 = *(ax+CIncA);
                t2 = *(ax+CIncA2);
          j=0;
          do {                          /*4 way loop unrolling*/
                t3 = *(ax+CIncA3);
/*             bx += CIncB4;*/
                                        FD_MULSET(b0, =, a, t0);
	     ax += CIncA4;
                                                        *bx = b0;
                t0 = *(ax);
                                        FD_MULSET(b1, =, a, t1);
                                                        *(bx+CIncB) = b1;
             j += 4;
                t1 = *(ax+CIncA);
                                        FD_MULSET(b2, =, a, t2);
                                                        *(bx+CIncB2) = b2;
 
                t2 = *(ax+CIncA2);
                                        FD_MULSET(b3, =, a, t3);
                                                        *(bx+CIncB3) = b3;
           bx += CIncB4;
         } while (j < n);
/*          bx += CIncB4;*/
        } /* if (n....) */
#endif        
        for (j=n; j < N; j++) {   
          t0 = *(ax);
                                        FD_MULSET(b0, =, a, t0);
                                                        *bx = b0;
          ax += CIncA; bx += CIncB;
        }
        Ax += RIncA; Bx += RIncB;
      } /* for (i...) */
    } /* if (*alpha...) */
  }    
#endif /*VANILLA*/
  N0 -= N;
  if (N0>0) {
    sN += N;	
    Ax = pLM_ADR(A)+sN*CIncA;	  	Bx = pLM_ADR(B)+sN*CIncB;
    N = MIN(N0, N);
  } 
  } while (N0>0);    
} /* MatCopy() */


void  MatAdd(M, N, A, alpha, B)
  int M, N; FDATA *alpha; pLocMat A, B;
{ int i; 
  int RIncA=pLM_RInc(A), CIncA=pLM_CInc(A), RIncB=pLM_RInc(B), CIncB=pLM_CInc(B);
  FDATA *Ax = pLM_ADR(A), *Bx = pLM_ADR(B);
  int sN, N0, CInc;
  
  COMMON_ASSERT_TR(
    (TraceStr, "(%dx%d,a%d %d %d,%s,a%d %d %d)", M, N,
     FD_ADR(Ax), RIncA, CIncA, pFD2Str(alpha), FD_ADR(Bx), RIncB, CIncB),
     IncOK(M, N, RIncA, CIncA) && pLM_IncOK(M, N, B), 0, "MatAdd", 4, TRACE_COMPUTE
  );

  if (M_INNER(M, N, RIncB, CIncB)){
    SWAP(int, M, N);	SWAP(int, RIncA, CIncA);	SWAP(int, RIncB, CIncB);
  }    

  N0 = N; sN = 0; 			CInc = MAX(CIncA, CIncB);
  if (M>1 && CInc>1 && N>MIN_VECSIZE /*ie. a local X-pose on a proper matrix*/
      && TLBSIZE) {
    /* nb assumes RInc>1; generally RInc >= CInc in any case */
    N = MAX(TLBSIZE/2, CACHESIZE/2/CInc);  
    N = MIN(N0, N);
  }
    
  do {                                  /* partition to reduce TLB misses */
              
  if (alpha==NULL) alpha = &FD_1;
  if (FD_EQ(*alpha, FD_0))
    return;
#ifdef VANILLA
  else { 
    FD_ALPHA_REGS(*alpha); 
    for (i=0; i < M; i++) {
      int j; register FDATA *ax = Ax, *bx = Bx; 
      for (j=0; j < N; j++) {
        FD_ALPHA_SCALEADD(*bx, *bx, *ax);
        ax += CIncA; bx += CIncB;
      }
      Ax += RIncA; Bx += RIncB;
    } /* for (i...) */
  }
#elif defined(BLAS)
  else
    for (i=0; i < M; i++) {
      FTN_AXPY(&N, alpha, Ax, &CIncA, Bx, &CIncB); 	
      Ax += RIncA; Bx += RIncB;
    } /* for (i...) */  
#else
  else if (!FD_EQ(*alpha, FD_1) && !FD_EQ(*alpha, FD_M1)) {  
    /* series of proper axpys */
    LocMat alpham; 		LM_Init(alpham, alpha, 1, 1);
    TRACE_LEVEL_SHIFT(2);
    for (i=0; i<M; i++) {      	
# if !defined(NO_ALIGN) && (FDATATYPE < COMPLEX)
      Axpy1(N, alpha, Ax, CIncA, Bx, CIncB);
# else 
      int j; FDATA *ax = Ax, *bx = Bx;	FD_ALPHA_REGS(*alpha); 
      for (j=0; j < N; j++) {
	FD_ALPHA_MULSET(*bx, +=, *ax);
	ax += CIncA; bx += CIncB;
      }	   
# endif	
      Ax += RIncA; Bx += RIncB;
    } /* for (i...) */     
    TRACE_LEVEL_SHIFT(-2);
  } else 
  { register int CIncA2=CIncA+CIncA, CIncA3=CIncA2+CIncA, CIncA4=CIncA2+CIncA2;
    register int CIncB2=CIncB+CIncB, CIncB3=CIncB2+CIncB, CIncB4=CIncB2+CIncB2;
    register int n = 4*(N/4), j;
    FD_register t0, t1, t2, t3, t0b, t1b, t2b, t3b, b0, b1, b2, b3;
    if (FD_EQ(*alpha,FD_1)) 
      for (i=0; i<M; i++) {
        register FDATA *ax = Ax, *bx = Bx, *blx = Bx;
#ifndef ULTRA
        if (n > 0) {
		t0 = *(ax);		t0b = *(bx);
		t1 = *(ax+CIncA);	t1b = *(bx+CIncB);
				FD_SETOP(b0, t0b, +, t0);
		t2 = *(ax+CIncA2);	t2b = *(bx+CIncB2);
				FD_SETOP(b1, t1b, +, t1);
          for (j=4; j < n; j+= 4) {	
		t3 = *(ax+CIncA3);	t3b = *(bx+CIncB3);
	    ax += CIncA4; 
				FD_SETOP(b2, t2b, +, t2);
							*(bx) = b0;
		t0 = *(ax);		t0b = *(bx+CIncB4);
				FD_SETOP(b3, t3b, +, t3);
							*(bx+CIncB) = b1;
	    bx += CIncB2;
		t1 = *(ax+CIncA);	t1b = *(bx+CIncB3);
				FD_SETOP(b0, t0b, +, t0);
							*(bx) = b2;
		t2 = *(ax+CIncA2);	t2b = *(bx+CIncB4);
				FD_SETOP(b1, t1b, +, t1);
							*(bx+CIncB) = b3;
	    bx += CIncB2;
	  } /* for (j...) */ 
		t3 = *(ax+CIncA3);	t3b = *(bx+CIncB3);
		ax += CIncA4; 
				FD_SETOP(b2, t2b, +, t2);
							*(bx) = b0;
				FD_SETOP(b3, t3b, +, t3);
							*(bx+CIncB) = b1;
		bx += CIncB2;
							*(bx) = b2;
							*(bx+CIncB) = b3;					
		bx += CIncB2;
	}
#else
        if (n > 0) {
		t0 = *(ax);		t0b = *(blx);
		t1 = *(ax+CIncA);	t1b = *(blx+CIncB);
		t2 = *(ax+CIncA2);	t2b = *(blx+CIncB2);
	  for (j=-n+4; j<0; j+=4) {
		t3 = *(ax+CIncA3);  	t3b = *(blx+CIncB3); 	ax += CIncA4; 
				FD_SETOP(b0, t0b, +, t0); 	blx+=CIncB4;
						*(bx) = b0;  
		t0 = *(ax);		t0b = *(blx);
				FD_SETOP(b1, t1b, +, t1);
						*(bx+CIncB) = b1;
		t1 = *(ax+CIncA);	t1b = *(blx+CIncB);
				FD_SETOP(b2, t2b, +, t2);
						*(bx+CIncB2) = b2; 
		t2 = *(ax+CIncA2);	t2b = *(blx+CIncB2);
				FD_SETOP(b3, t3b, +, t3);
						*(bx+CIncB3) = b3; bx += CIncB4;
	  } /* for (j...) */
		t3 = *(ax+CIncA3);	t3b = *(blx+CIncB3);
		ax += CIncA4; 
				FD_SETOP(b0, t0b, +, t0);
						*(bx) = b0;
				FD_SETOP(b1, t1b, +, t1);
						*(bx+CIncB) = b1;
				FD_SETOP(b2, t2b, +, t2);
						*(bx+CIncB2) = b2;
				FD_SETOP(b3, t3b, +, t3);
						*(bx+CIncB3) = b3; bx += CIncB4;
	}
#endif	
        for (j=n; j < N; j++) {
          t0 = *(ax); 		t0b = *(bx);  	FD_SETOP(b0, t0b, +, t0); 
          ax += CIncA; 	*(bx) = b0;	bx += CIncB;
        }
        Ax += RIncA; Bx += RIncB;
      } /* for (i...) */
    else if (FD_EQ(*alpha,FD_M1)) 
      for (i=0; i<M; i++) {
        register FDATA *ax = Ax, *bx = Bx, *blx = Bx;
#ifndef ULTRA
        if (n > 0) {
		t0 = *(ax);		t0b = *(bx);
		t1 = *(ax+CIncA);	t1b = *(bx+CIncB);
				FD_SETOP(b0, t0b, -, t0);
		t2 = *(ax+CIncA2);	t2b = *(bx+CIncB2);
				FD_SETOP(b1, t1b, -, t1);
          for (j=4; j < n; j+= 4) {	
		t3 = *(ax+CIncA3);	t3b = *(bx+CIncB3);
	    ax += CIncA4; 
				FD_SETOP(b2, t2b, -, t2);
							*(bx) = b0;
		t0 = *(ax);		t0b = *(bx+CIncB4);
				FD_SETOP(b3, t3b, -, t3);
							*(bx+CIncB) = b1;
	    bx += CIncB2;
		t1 = *(ax+CIncA);	t1b = *(bx+CIncB3);
				FD_SETOP(b0, t0b, -, t0);
							*(bx) = b2;
		t2 = *(ax+CIncA2);	t2b = *(bx+CIncB4);
				FD_SETOP(b1, t1b, -, t1);
							*(bx+CIncB) = b3;
	    bx += CIncB2;
	  } /* for (j...) */ 
		t3 = *(ax+CIncA3);	t3b = *(bx+CIncB3);
		ax += CIncA4; 
				FD_SETOP(b2, t2b, -, t2);
							*(bx) = b0;
				FD_SETOP(b3, t3b, -, t3);
							*(bx+CIncB) = b1;
		bx += CIncB2;
							*(bx) = b2;
							*(bx+CIncB) = b3;					
		bx += CIncB2;
        }
#else
        if (n > 0) {
		t0 = *(ax);		t0b = *(blx);
		t1 = *(ax+CIncA);	t1b = *(blx+CIncB);
		t2 = *(ax+CIncA2);	t2b = *(blx+CIncB2);
	  for (j=-n+4; j<0; j+=4) {
		t3 = *(ax+CIncA3);  	t3b = *(blx+CIncB3); 	ax += CIncA4; 
				FD_SETOP(b0, t0b, -, t0); 	blx+=CIncB4;
						*(bx) = b0;  
		t0 = *(ax);		t0b = *(blx);
				FD_SETOP(b1, t1b, -, t1);
						*(bx+CIncB) = b1;
		t1 = *(ax+CIncA);	t1b = *(blx+CIncB);
				FD_SETOP(b2, t2b, -, t2);
						*(bx+CIncB2) = b2; 
		t2 = *(ax+CIncA2);	t2b = *(blx+CIncB2);
				FD_SETOP(b3, t3b, -, t3);
						*(bx+CIncB3) = b3; bx += CIncB4;
	  } /* for (j...) */
		t3 = *(ax+CIncA3);	t3b = *(blx+CIncB3);
		ax += CIncA4; 
				FD_SETOP(b0, t0b, -, t0);
						*(bx) = b0;
				FD_SETOP(b1, t1b, -, t1);
						*(bx+CIncB) = b1;
				FD_SETOP(b2, t2b, -, t2);
						*(bx+CIncB2) = b2;
				FD_SETOP(b3, t3b, -, t3);
						*(bx+CIncB3) = b3; bx += CIncB4;

        }
#endif 		
        for (j=n; j < N; j++) {
          t0 = *(ax); 		t0b = *(bx);  	FD_SETOP(b0, t0b, -, t0); 
          ax += CIncA; 		*(bx) = b0;	bx += CIncB;
        }
        Ax += RIncA; Bx += RIncB;
      } /* for (i...) */
  } /* alpha =+/-1 */       
#endif /*VANILLA*/
  N0 -= N;
  if (N0>0) {
    sN += N;
    Ax = pLM_ADR(A)+sN*CIncA;           Bx = pLM_ADR(B)+sN*CIncB;
    N = MIN(N0, N);
  }
  } while (N0>0);
} /*  MatAdd() */
