/************************************************************************/
/*       	CellMatCopy: FDATA Matrix add/copy procedures     	*/
/************************************************************************/
/* SCCS INFO @(#)CellMatCopy.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 97/02
*/
/* Set of scaled matrix copy & add routines used for DBLAS.
 */
#ifndef __CellMatCopy_INCL__
#define __CellMatCopy_INCL__  

#include "FDATASetUp.h"
#include "LocMat.h"
#include "CellVecCopy.h"		/* for backward compatability */

/********************* exported symbols  ***********************************/

#ifndef CELLBLAS

#if FDATATYPE==SINGLE
#define MatCopy		DSMatCopy   
#define MatAdd		DSMatAdd   
#elif FDATATYPE==DOUBLE
#define MatCopy		DDMatCopy   
#define MatAdd		DDMatAdd   
#elif FDATATYPE==COMPLEX
#define MatCopy		DCMatCopy   
#define MatAdd		DCMatAdd   
#else
#define MatCopy		DZMatCopy   
#define MatAdd		DZMatAdd   
#endif

#else  /* rename to avoid linker clash */

#if FDATATYPE==SINGLE
#define MatCopy         SMatCopy   
#define MatAdd          SMatAdd   
#elif FDATATYPE==DOUBLE
#define MatCopy         DMatCopy   
#define MatAdd          DMatAdd   
#elif FDATATYPE==COMPLEX
#define MatCopy         CMatCopy
#define MatAdd          CMatAdd
#else
#define MatCopy         ZMatCopy
#define MatAdd          ZMatAdd
#endif

# endif /*CELLBLAS*/

void MatCopy STDC_PARAMS(
        (int M, int N, pLocMat A, FDATA *alpha, pLocMat B)
);
/* pre:  pLM_IncOK(M, N, A) && pLM_IncOK(M, N, B)
 * post: B[i,j] = (alpha)*A[i,j], 0<=i<M, 0<=j<N
 * optimized for M > N, N > M, *alpha = +1.0
 */

void MatAdd STDC_PARAMS(
        (int M, int N, pLocMat A, FDATA *alpha, pLocMat B)
);
/* pre:  pLM_IncOK(M, N, A) && pLM_IncOK(M, N, B)
 * post: B[i,j] += (alpha)*A[i,j], 0<=i<M, 0<=j<N
 * optimized for M > N, N > M, *alpha = +/- 1.0
 */

#endif /*__CellMatCopy_INCL__*/
