#define INLINE_BLAS
/************************************************************************/
/*       	CellMatScale: FDATA Matrix procedures for cells    	*/
/************************************************************************/
/* SCCS INFO @(#)CellMatScale.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 93/03/23 
*/
/* Set of support matrix routines used for DBLAS, independent of distributed
 * memory issues.
 * Most of these can be implemented via level 1 BLAS, although for convenience
 * and efficiency, they are implemented as proper matrix operations,
 * using local matrix dimensions and strides.
 * These refer to local cell view of matrices only with generalized row-column 
 * strides (see LocMat.h).
 */
 
#ifndef __CellMatScale_INCL__
#define __CellMatScale_INCL__  

#include "FDATASetUp.h"
#include "LocMat.h"

/********************* exported symbols  ***********************************/

#ifndef CELLBLAS

#if FDATATYPE==SINGLE
#define MatZero		DSMatZero   
#define MatScale	DSMatScale  
#elif FDATATYPE==DOUBLE
#define MatZero		DDMatZero   
#define MatScale	DDMatScale  
#elif FDATATYPE==COMPLEX
#define MatZero		DCMatZero   
#define MatScale	DCMatScale  
#else
#define MatZero		DZMatZero   
#define MatScale	DZMatScale  
#endif

#else	/* rename routines to avoid name clash */

#if FDATATYPE==SINGLE
#define MatZero		SMatZero   
#define MatScale	SMatScale  
#elif FDATATYPE==DOUBLE
#define MatZero		DMatZero   
#define MatScale	DMatScale  
#elif FDATATYPE==COMPLEX
#define MatZero		CMatZero   
#define MatScale	CMatScale  
#else
#define MatZero		ZMatZero   
#define MatScale	ZMatScale  
#endif

#endif /*CELLBLAS*/

#if defined(BLAS) 
  /* define Fortran BLAS procedure names called */
# include "FtnBLAS.h"  
#endif /*BLAS*/

/********************** matrix routines *********************************/
    
void MatZero STDC_PARAMS( (int M, int N, pLocMat A) );
/* pre:  pLM_IncOK(M, N, A)
 * post: A[i,j] = 0 for 0<=i<M, 0<=j<N
 */

void MatScale STDC_PARAMS( (int M, int N, FDATA *alpha, pLocMat A ) );
/* pre:  pLM_IncOK(M, N, A)
 *	 A[i,j] = a(i,j)  for 0<=i<M, 0<=j<N and some function a
 * post: A[i,j] = *alpha * a(i,j) for 0<=i<M, 0<=j<N
 * nb. optimized for *alpha == 0.0, 1.0
 */

#endif /*__CellMatScale_INCL__*/

