/************************************************************************/
/*       	CellMatrix: FDATA Matrix procedures for cells    	*/
/************************************************************************/
/* SCCS INFO @(#)CellMatrix.c 1.4         last modified P. Strazdins, 99/06/04
   first created: P. Strazdins 93/03/23 
*/
/* Set of support matrix routines used for DBLAS, independent of distributed
 * memory issues.
 * compilation: if macro VANILLA defined, `vanilla' versions of the routines
 *              are used (slower, but useful for debugging or vectorization.
 *              if macro NO_PART is defined, suppress partitoning methods
 *              wherever possible    
 *              if macro BLAS defined, Fortran BLAS is used whenever
 *              possible; if BLAS_MM defined, do for only BLAS matrix-mult calls
 */

#include "BLASOptions.h"
#include "FDATAOps.h"
#include "FDAssert.h"

#include "CellMatrix.h"

#include "MachineParams.h"
	
#ifndef OLD_PARTMM

FDATA *PackMMAlphaB = NULL;

FDATA *PackMMOpdsP(M, N, Np, K, alphaf, A, B)
  int M, N, Np, K; FDATA **alphaf; pLocMat *A, *B; 
{ int PackB = 0, /*LM_PackUseful(K, N, **B, M), */
      PackA = 0; /*LM_PackUseful(M, K, **A, N);*/
  int Partition = (Np < N);    
  FDATA *alphapack = MM_SCALE? *alphaf: NULL;
  static LocMat A1, B1; 
  FDATA *WS = NULL, *Ax = NULL, *Bx = NULL;
  
  TRACE_ENTRY;
  if (K > 0) {
    if (CopyEfficient(N)) 
      PackA = Partition?  LM_MMStridePack(0, **A, M, K): 
      			  LM_PackUseful_(M, K, **A);       
    if (CopyEfficient(M)) 
      PackB = Partition? LM_MMStridePack(1, **B, N, K): 
     			  LM_PackUseful_(K, N, **B);
     
    if (!PackA && !PackB && !pFD_UNITY(alphapack)) {
      if (M<N)
        PackA=1;
      else
        PackB=1;  	
    }   
  }

  Np = MIN(N, Np);
  if (PackA && PackB && MatMalloc_FT("PackMMOpnds", 0, &WS, K*(M+Np))) 	
    Bx = WS+M*K, Ax = WS;
  else if (PackA && (!PackB || M<Np) && 
    MatMalloc_FT("PackMMOpnds", 0, &WS, K*M))
    Ax = WS;
  else if (PackB && MatMalloc_FT("PackMMOpnds", 0, &WS, K*Np))   
    Bx = WS;

  PackMMAlphaB = NULL;
  if (Bx != NULL) {
    if (Ax==NULL || M<N)
      PackMMAlphaB = alphapack, alphapack = NULL;	
    if (!MM_INNER_K)	
      LM_Init(B1, Bx, Np, 1);
    else
      LM_Init(B1, Bx, 1, K);
    MatCopy(K, Np, *B, PackMMAlphaB, &B1);
    *B = &B1;	
    if (PackMMAlphaB)   			
      *alphaf = NULL;
  } 
  if (Ax != NULL) {
    if (!MM_INNER_K)
      LM_Init(A1, Ax, 1, M);
    else
      LM_Init(A1, Ax, K, 1);       	  
    MatCopy(M, K, *A, alphapack, &A1);
    *A = &A1;
    if (alphapack)     	
      *alphaf = NULL;
  }  
  TRACE_EXIT;
  return(WS);
} /* PackMMOpdsP() */

#else

FDATA *PackMMOpds(M, N, K, alphaf, A, B)
  int M, N, K; FDATA **alphaf; pLocMat *A, *B; 
{ int PackB = LM_PackUseful(K, N, **B, M), 
      PackA = LM_PackUseful(M, K, **A, N);
  static LocMat A1, B1; 
  FDATA *WS = NULL, *Ax = NULL, *Bx = NULL;
  int MustScale = MM_SCALE && !pFD_UNITY(*alphaf);

  if (PackA && PackB && MatMalloc_FT("MatPartMult", 0, &WS, K*(M+N))) {
    Bx = WS+M*K;
    Ax = WS;	
  } else if (((PackA && K<=M) || (!PackB && M<N && MustScale) && K<=N))
    	      && MatMalloc_FT("MatPartMult", 0, &WS, K*M))  {
    Ax = WS;  
  } else if ((PackB || (MustScale && K<=N && M>0)) 
  	     && MatMalloc_FT("MatPartMult", 0, &WS, K*N)) {
    Bx = WS;
  }  
  if (Bx != NULL) {
    LM_Init(B1, Bx, N, 1); 
    MatCopy(K, N, *B, *alphaf, &B1);
    *B = &B1;	    	*alphaf = NULL;
  } 
  if (Ax != NULL) {
    LM_Init(A1, Ax, 1, M); 	  
    MatCopy(M, K, *A, *alphaf, &A1);
    *A = &A1;     	*alphaf = NULL;
  }  

  return(WS);
} /* PackMMOpds() */

#endif /* defined(OLD_PARTMM) */


#if defined(BLAS) || defined(BLAS_MM)

# ifndef INLINE_BLAS 
void MatPartMult(M, N, K, alpha, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta; pLocMat A, B, C;
{ MatMult_f(M, N, K, alpha, A, B, beta, C);
}
	
void MatMult(M, N, K, alpha, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta; pLocMat A, B, C;
{ MatMult_f(M, N, K, alpha, A, B, beta, C);
}	
# endif

#elif !defined(OLD_PARTMM)

void MatPartMult(M, N, K, alpha, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta; pLocMat A, B, C;
{ 
#if defined( VANILLA ) || defined(NO_PART)
  MatMult(M, N, K, alpha, A, B, beta, C);
#else
  int k, Np;
  LocMat A_, B_, C_; 			/* used to prevent side effects *A,..*/
 
COMMON_ASSERT_TR_E(
   (TraceStr, "(%dx%dx%d,%s,a%d %d %d,a%d %d %d,%s,a%d %d %d)", 
    M, N, K, pFD2Str(alpha), FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A),
    FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), pFD2Str(beta),
    FD_ADR(pLM_ADR(C)), pLM_RInc(C), pLM_CInc(C)), 
   (N==0 || pLM_IncOKUnit(M, K, A)) && (M==0 || pLM_IncOKUnit(K, N, B)) &&
   pLM_IncOKUnit(M, N, C) && 0<=M && 0<=N && 0<=K, 
   MatOverlap(M, K, A, M, N, C) + 2*MatOverlap(K, N, B, M, N, C),
   "MatPartMult", 2, TRACE_COMPUTE
  );   
  if (K==0) {
    MatScale(M, N, beta, C);
    TRACE_RETURN;
  }   

  Np = OUTER_PARTITION_WIDTH(K);
  if (Np>=MAX(M, N)) {			/* reduce overheads for small comp'n */
    FDATA *WS = PackMMOpdsP(M, N, Np, K, &alpha, &A, &B);   
    MatMult(M, N, K, alpha, A, B, beta, C); 
    if (WS != NULL)
      MatFree(WS);
    TRACE_RETURN;
  }         	

  if (LM_Inc(MM_INNER_K,*C)!=1) { /* transpose to optimize memory refs on C */
    SWAP(int, M, N);	
    LM_Copy(1, B_, *A);	LM_Copy(1, A_, *B);	LM_Copy(1, C_, *C);
    A = &A_;		B = &B_;		C = &C_;
  } else { 			/* prevent side effects on *A, *B */
    LM_Copy(0, A_, *A); LM_Copy(0, B_, *B);
    A = &A_;		B = &B_;
  }  

  k=0;
  while (k < K) {
    int dK = K-k, j;
    FDATA *alpha_ = alpha;
    FDATA *WS = NULL; 			int CpyB;
    LocMat Bj;				LM_Copy(0, Bj, *B);

    if (2*dK > 3*INNER_VECSIZE)
      dK = INNER_VECSIZE;      
#if 1      
    if (CACHESIZE && TLBSIZE && (dK > TLBSIZE / 2) && 
        ((N < 4 && pLM_CInc(A)*dK > CACHESIZE) || 
        (M < 4 && pLM_RInc(B)*dK > CACHESIZE))) {
       /* matrix-vector like multiply, with non-unit stride across K-dim of matrix*/
       dK = TLBSIZE / 2; 		/* minimize TLB/cache misses */
    }   
#endif
    Np = OUTER_PARTITION_WIDTH(dK);
    WS = PackMMOpdsP(M, N, Np, dK, &alpha_, &A, &B);   
    CpyB = pLM_ADR(B)!=LM_ADR(Bj);


    for (j=0; j < N; j+=Np) {
      int n = MIN(Np, N-j);
      LocMat Cj;			 pLocSubMat_(&Cj, C, 0, j);
      if (j>0 && CpyB)
        MatCopy(dK, n, &Bj, PackMMAlphaB, B);
      MatMult(M, n, dK, alpha_, A, CpyB? B: &Bj, beta, &Cj);
      pLocSubMat(&Bj, 0, n);
    } /* for (j...) */
    if (WS != NULL)
      MatFree(WS);

    beta = NULL;			/* must be unity for next iteration */
    /* ensure A, B point back to orig. structures; PackMMOpdsP() may have changed them  */
    A = &A_;				B = &B_;
    pLocSubMat(A, 0, dK);		pLocSubMat(B, dK, 0);	
    k += dK;
  } /* while (k<K) */

#endif /*VANILLA*/
  TRACE_EXIT;
  
} /* MatPartMult() */

#else 

void MatPartMult(M, N, K, alpha, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta; pLocMat A, B, C;
{ 
#if defined( VANILLA ) || defined(NO_PART)
  MatMult(M, N, K, alpha, A, B, beta, C);
#else
  int k; int mn = MIN(M, N); 
  LocMat A_, B_, C_; 			/* used to prevent side effects *A,..*/
 
COMMON_ASSERT_TR_E(
   (TraceStr, "(%dx%dx%d,%s,a%d %d %d,a%d %d %d,%s,a%d %d %d)", 
    M, N, K, pFD2Str(alpha), FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A),
    FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), pFD2Str(beta),
    FD_ADR(pLM_ADR(C)), pLM_RInc(C), pLM_CInc(C)), 
   (N==0 || pLM_IncOKUnit(M, K, A)) && (M==0 || pLM_IncOKUnit(K, N, B)) &&
   pLM_IncOKUnit(M, N, C) && 0<=M && 0<=N && 0<=K, 
   MatOverlap(M, K, A, M, N, C) + 2*MatOverlap(K, N, B, M, N, C),
   "MatPartMult", 2, TRACE_COMPUTE
  );   

  if (!(CopyEfficient(mn) && PartitionUseful(M, N, K))) {
    /* small computation - quick fall through */
    FDATA * WS = PackMMOpds(M, N, K, &alpha, &A, &B);
    MatMult(M, N, K, alpha, A, B, beta, C);
    if (WS != NULL)
      MatFree(WS);
    TRACE_EXIT;
    return;
  }
      
  if (M < N) {			/* make A (M) largest, B (N) smallest */
    SWAP(int, M, N);	
    LM_Copy(1, A_, *A);	LM_Copy(1, B_, *B);	LM_Copy(1, C_, *C);
    A = &B_;		B = &A_;		C = &C_;
  } else {
    LM_Copy(0, A_, *A);	LM_Copy(0, B_, *B);	
    A = &A_;		B = &B_;		
  }      

  k=0;
  while (k < K) {
    int dK; 
    int DoPartition = 0;
    FDATA *WS = NULL; int Mw, Ldw;	/* partitioning workspace parameters */
 
    dK = K - k;
    if (2*dK > 3*INNER_VECSIZE && PartitionUseful(M, N, dK))
      dK = INNER_VECSIZE;

    if (CopyEfficient(mn) && PartitionUseful(M, N, dK)) {
      /* try partitioning over M & N; a time-space tradeoff */
      Ldw = Mw = (CACHESIZE/2) / dK;   
      DoPartition = MatMalloc("MatPartMult", 0, &WS, CACHESIZE*((mn+Ldw-1)/Ldw));

      if (! DoPartition) {/* partition fairly efficiently with 1 CACHESIZE WS */
        Ldw = Mw = (CACHESIZE - MINOR_OUTER_VECSIZE*dK)/dK;
        DoPartition = - MatMalloc("MatPartMult", 0, &WS, CACHESIZE);
      }
    }
  
    if (! DoPartition ) {
      FDATA *alphaf = alpha; pLocMat Af = A, Bf = B;	
      WS = PackMMOpds(M, N, dK, &alphaf, &Af, &Bf);
      MatMult(M, N, dK, alphaf, Af, Bf, beta, C);
    } else if (DoPartition == -1) {
      int i, j, m, n; LocMat WA, WB, Aik, Bkj, Cij;
      LM_Init(WA, WS, 1, MINOR_OUTER_VECSIZE);
      LM_Init(WB, WS + MINOR_OUTER_VECSIZE*dK, Ldw, 1);
      for (j=0; j < N; j += Mw) {
        n = MIN(Mw, N-j);	    	 
        MatCopy(dK, n, pLocSubMat_(&Bkj, B, 0, j), alpha, &WB);
        pLocSubMat_(&Aik, A, 0, 0);	pLocSubMat_(&Cij, C, 0, j);
        for (i=0; i < M; i += MINOR_OUTER_VECSIZE) {
      	  m = MIN(MINOR_OUTER_VECSIZE, M-i);
          MatCopy(m, dK, &Aik, NULL, &WA);
          MatMult(m, n, dK, NULL, &WA, &WB, beta, &Cij);
          LM_SubMat(Aik, m, 0);		LM_SubMat(Cij, m, 0); 
        }
      } /* for (i...) */ 
    } else {   
      int i, m, n, j=0, jInc=Mw; LocMat WA, WB, Aik, Bkj, Cij; 
      LM_Init(WA, WS+(CACHESIZE/2), 1, Ldw);
      LM_Init(WB, WS, Ldw, 1);
      for (i=0; i < M; i+= Mw) {
        m = MIN(Mw, M-i);	
        MatCopy(m, dK, pLocSubMat_(&Aik, A, i, 0), NULL, &WA);
        for (; (0 <= j) && (j < N); j += jInc) {
          n = MIN(Mw, N-j);
          if (i==0) 	
            MatCopy(dK, n, pLocSubMat_(&Bkj, B, 0, j), alpha, &WB);
            MatMult(m, n, dK, NULL, &WA, &WB, beta, pLocSubMat_(&Cij, C, i, j));
            LM_ADR(WB) += (jInc>0)? CACHESIZE: -CACHESIZE;
        } /* for (j...) */
        j -= jInc;			jInc = -jInc; 
        LM_ADR(WB) += (jInc>0)? CACHESIZE: -CACHESIZE;
      } /* for (i...) */
    } /* if (DoPartition) */

    if (WS != NULL)
      MatFree(WS);

    k += dK;
    beta = NULL;		/* must be unity for next iteration */
    pLocSubMat(A, 0, dK);		pLocSubMat(B, dK, 0);	
  } /* while (k<K) */
#endif /*VANILLA*/
  TRACE_EXIT;
} /* MatPartMult() */
#endif /* defined(OLD_PARTMM) */  
 

#ifndef BLAS
#if defined(VANILLA) 

void MatMult(M, N, K, alpha, A, B, beta, C)
  int M, N, K; FDATA *alpha, *beta;  pLocMat A, B, C;
{ 
  int i, j, k; FDATA *Ax = pLM_ADR(A), *Cx = pLM_ADR(C);
  COMMON_ASSERT_TR_E(
   (TraceStr, "(%dx%dx%d,%s,a%d %d %d,a%d %d %d,%s,a%d %d %d)", 
    M, N, K, pFD2Str(alpha), FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A),
    FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), pFD2Str(beta),
    FD_ADR(pLM_ADR(C)), pLM_RInc(C), pLM_CInc(C)),
   (N==0 || pLM_IncOKUnit(M, K, A)) && (M==0 || pLM_IncOKUnit(K, N, B)) &&
   pLM_IncOKUnit(M, N, C) && 0<=M && 0<=N && 0<=K, 
   MatOverlap(M, K, A, M, N, C) + 2*MatOverlap(K, N, B, M, N, C),
   "MatMult", 3, TRACE_COMPUTE
  );

  if (alpha==NULL) alpha = &FD_1;	if (beta==NULL) beta = &FD_1;		 
  for (i = 0; i < M; i++) {
    FDATA *b = pLM_ADR(B), *cx = Cx;	
    for (j=0; j < N; j++) {
      FDATA *ax=Ax, *bx=b; FDATA c = FD_0;
      for (k=0; k < K; k++) {
        FD_MULSET(c, +=, *ax, *bx);
        ax += pLM_CInc(A); bx += pLM_RInc(B);
      }  
      FD_MULSET(*cx, =, *cx, *beta);
      FD_MULSET(*cx, +=, c, *alpha);
      b += pLM_CInc(B);  cx += pLM_CInc(C);
    } /* for (j...) */
    Cx += pLM_RInc(C); Ax+= pLM_RInc(A); 
  } /* for (i...) */ 
  TRACE_EXIT;
} /* MatMult() */
# else 
#   include "CellMatMult.c"
# endif /*VANILLA*/
#endif
