/************************************************************************/
/*  CellSplitUnpack: matrix copy routines from a packed `split' source  */
/*		     (as can occur on AP1000 ring-buffer wraparound)	*/
/************************************************************************/

/* SCCS INFO %Z%%M% %I%         last modified P. Strazdins, %E%
   first created: P. Strazdins 95/05
 */

#include "FDATASetUp.h"

#if (FDATATYPE==SINGLE) || (FDATATYPE==COMPLEX)
# define SplitMatUnpack       		SSplitMatUnpack
# define SplitBlkMatStrideUnpack	SSplitBlkMatStrideUnpack
#else
# define SplitMatUnpack       		DSplitMatUnpack
# define SplitBlkMatStrideUnpack	DSplitBlkMatStrideUnpack
#endif

void SplitMatUnpack(STDC_PARAMS(
	int Add, int N, int size, int s, FD_BASE *Ax1, int size1, FD_BASE *Ax2, 
	FD_BASE *alpha, FD_BASE *Bx, int RIncB, int CIncB)
);

/* pre:  size1 <= size
 * post: let A = strip(s,M,N,concat(Ax1[0..size1-1]:Ax2[0..size-size1])),
 *       M=size/N, a = Add? *alpha: FD_1;
 *       as for (Add? MatCopy: MatAdd)(M, N, {A, N, 1}, a, {Bx, RIncB, CIncB})
 *       strip(a,s,M,N,A) = concat(A[0..N-1]:strip(a',s,M-1,N, A+N1))
 *       where  N1 = ((N + a + COMM_UNIT-1)/COMM_UNIT)*COMM_UNIT - a + a'
 *              a' = (a+s)%COMM_UNIT
 * notes:the strip() operation arises when A has been formed by communication
 *       of a vector of length M, element size N and stride S st. S%COMM_UNIT=s.
 *       with the unit communication being COMM_UNIT*sizeof(FDATA),
 *       surplus words at the either end of each element may be sent
 *       and must be now be ignored.
 */    

void SplitBlkMatStrideUnpack(STDC_PARAMS(
	int Add, int N, int size, FD_BASE *Ax1, int size1, FD_BASE *Ax2, 
	int RIncA, FD_BASE *alpha, FD_BASE *B, int RB0, int RB, int RBInc, 
	int CB0, int CB, int CBInc)
);
/* pre:   size % N == 0 && RIncA >= N && size >= size1;
 * post:  as for BlkMatStrideUnpack(Add, M, N, {A, RIncA, 1}, &FD_1,
 *                                  B, RB0, RB, RBInc, CB0, CB, CBInc);
 *        where A = concat(Ax1[0..size1-1], Ax2[0..size-size1-1]), M = size/N
 */

