/************************************************************************/
/*       CellTriMatrix: Triangular Output Matrix procedures for cells  	*/
/************************************************************************/
/* SCCS INFO @(#)CellTriMatrix.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 93/03/23 
*/
/* set of non-communication matrix routines used for DBLAS.
 * These routines can be run on cells separately, but the operations
 * assume block matrix storage on a P[0] x P[1] mesh.
 */
#ifndef __CellTriMatrix_INCL__
#define __CellTriMatrix_INCL__  

#include "FDATASetUp.h"
#include "LocMat.h"
#include "DistMat.h"
#include "BLASOptions.h"

/********************* exported symbols  ***********************************/

#ifndef CELLBLAS

#if FDATATYPE==SINGLE
#define TriMatScale  		DSTriMatScale
#define TriMatMult		DSTriMatMult
#define TriMatPartMult 		DSTriMatPartMult
#elif FDATATYPE==DOUBLE
#define TriMatScale  		DDTriMatScale
#define TriMatMult		DDTriMatMult
#define TriMatPartMult 		DDTriMatPartMult
#elif FDATATYPE==COMPLEX
#define TriMatScale  		DCTriMatScale
#define TriMatMult		DCTriMatMult
#define TriMatPartMult 		DCTriMatPartMult
#else
#define TriMatScale  		DZTriMatScale
#define TriMatMult		DZTriMatMult
#define TriMatPartMult 		DZTriMatPartMult
#endif

#else	/* rename to avoid name clash */

#if FDATATYPE==SINGLE
#define TriMatScale  		STriMatScale
#define TriMatMult		STriMatMult
#define TriMatPartMult 		STriMatPartMult
#elif FDATATYPE==DOUBLE
#define TriMatScale  		DTriMatScale
#define TriMatMult		DTriMatMult
#define TriMatPartMult 		DTriMatPartMult
#elif FDATATYPE==COMPLEX
#define TriMatScale  		CTriMatScale
#define TriMatMult		CTriMatMult
#define TriMatPartMult 		CTriMatPartMult
#else
#define TriMatScale  		ZTriMatScale
#define TriMatMult		ZTriMatMult
#define TriMatPartMult 		ZTriMatPartMult
#endif

#endif /* CELLBLAS */

void TriMatScale STDC_PARAMS(
	(int M, int N, FDATA *alpha, int FormC, DistMat C)
);
/* pre:  UpLoOK(FormC) && LM_IncOK(M, N, DM_Loc(C))
 *       DM_Loc(C) is an MxN local triangular block matrix of form FormC
 * post: DM_Loc(C) *= *alpha 
 */

void TriMatMult STDC_PARAMS(
	(int M, int N, int K, FDATA *alpha, pLocMat A, pLocMat B, 
	 FDATA *beta, int FormC, DistMat C)
);
/* pre:  pLM_IncOKUnit(M, K, A) && pLM_IncOKUnit(K, N, B) && 
 *       LM_IncOKUnit(M, N, DM_Loc(C)) && UpLoOK(FormC) && 0 <= M, N, K &&
 *	 !Overlap(M, K, A, M, N, DM_Loc(C)) && !Overlap(K, N, B, M, N, DM_Loc(C))
 *       DM_Loc(C) is local MxN block triangular matrix of form FormC
 * post: DM_Loc(C) = (*alpha)A*B +(*beta)DM_Loc(C);  A is MxK, B is KxN,
 * notes:generally C would be the cell view of a global N' x N' symm, matrix.
 *       in this case,  M:N ~ C.D[0].P:C.D[1].P, or more precisely:
 *              N' = L2Gd(M, C.D[0]) = L2Gd(N, C.D[1])
 *       Only in this case is performance particularly important.
 *       
 */

void TriMatPartMult STDC_PARAMS(
	(int M, int N, int K, FDATA *alpha, pLocMat A, pLocMat B, 
	 FDATA *beta, int FormC, DistMat C)
);
/* pre/post:  as for TriMatMult()
 * notes:see notes for TriMatMult(); this routine also applies
 *	 partitioning methods suitable for large M, N, K
 */

#endif /*__CellTriMatrix_INCL__*/
