/************************************************************************/
/*			 	CellVecCopy				*/
/************************************************************************/
/* SCCS INFO @(#)CellVecCopy.c 1.4         last modified P. Strazdins, 99/07/06
 *	first created P. Strazdins DCS ANU 95/09
 */
/* vcort copy/adds routine with arbitrary  strides.
 * compilation: if macro VANILLA defined, `vanilla' versions of the routines
 *              are used (slower, but useful for debugging or vectorization.
 *              if macro BLAS defined, Fortran BLAS is used whenever
 *              possible; 
 */ 
/************************************************************************/
#include "CellVecCopy.h"
#include "FDATAOps.h"
#include "FDAssert.h"

#ifdef NO_ALIGN
# define VecCopy VecCopyW
# define VecAdd  VecAddW
# undef BLAS
# undef BLAS_MM
# define inline                                 /* so cc can compile this */
#endif

#include "CellMatrix.h"

#ifdef BLAS
# include "FtnBLAS.h"
#endif
 

#if !defined(BLAS) || !defined(INLINE_BLAS)

void VecCopy(N, A, IncA, B, IncB)
  int N, IncA, IncB; FDATA *A, *B;
{ int i, n; 

  COMMON_ASSERT_TR(
    (TraceStr, "(%d,a%d %d,a%d %d)", N, FD_ADR(A), IncA, FD_ADR(B), IncB),
     1, 0, "VecCopy", 6, TRACE_COMPUTE
  );

#ifdef BLAS
  FTN_COPY(&N, A, &IncA, B, &IncB);
#else
  i=0; n=0;
#ifndef VANILLA     
   n=8*(N/8);	
   if (i < n) {
     register int IncA2=IncA+IncA, IncA3=IncA2+IncA, IncA4=IncA2+IncA2;
     register int IncB2=IncB+IncB, IncB3=IncB2+IncB, IncB4=IncB2+IncB2;
     FD_register t0, t1, t2, t3, b0, b1, b2, b3;  
     do {
          t0 = *A;
          t1 = *(A+IncA);
          i+=8;
          t2 = *(A+IncA2);
          t3 = *(A+IncA3);
          A += IncA4; 
          b0 = *A;
          b1 = *(A+IncA);
          b2 = *(A+IncA2);
          b3 = *(A+IncA3);
          *B = t0; *(B+IncB) = t1;
          A += IncA4; 
          *(B+IncB2) = t2; *(B+IncB3) = t3;
          B += IncB4;
          *B = b0; *(B+IncB) = b1;
          *(B+IncB2) = b2; *(B+IncB3) = b3;
          B += IncB4;          
    } while (i < n);

#if (FDATATYPE < COMPLEX)    
    if (ForceAlgVers==-1 && (t0+t1+t2+t3 > b0+b1+b2+b3)) {
      /*existance of 2nd condition forces gcc to put t0-b3 in SPARC FPU registers !!!*/
        CommonAssert(1, 0, "VecSwap", 10, TRACE_ALL);
    }
#endif
  } /* if (i< n)*/ 
#endif VANILLA    
  for (i=n; i < N; i++) {
    *B = *A;
    A += IncA; B += IncB;
  }
#endif /*not defined BLAS*/
} /* VecCopy() */

#endif /* BLAS */


#if !defined(BLAS) || !defined(INLINE_BLAS)
void VecAdd(N, A, IncA, B, IncB)
  int N, IncA, IncB; FDATA *A, *B;
{ int i, n; 

  COMMON_ASSERT_TR(
    (TraceStr, "(%d,a%d %d,a%d %d)", N, FD_ADR(A), IncA, FD_ADR(B), IncB),
     1, 0, "VecAdd", 6, TRACE_COMPUTE
  );

#ifdef BLAS
  FTN_AXPY(&N, &FD_1, A, &IncA, B, &IncB);
#else
  i=0; n=0;
# ifndef VANILLA     
   n = 4*(N/4);
#ifndef ULTRA   	
   if (n > 0) {
     register int IncA2=IncA+IncA, IncA3=IncA2+IncA, IncA4=IncA2+IncA2;
     register int IncB2=IncB+IncB, IncB3=IncB2+IncB, IncB4=IncB2+IncB2;
     FD_register t0, t1, t2, t3, t0b, t1b, t2b, t3b, b0, b1, b2, b3;  
    
		t0 = *(A);		t0b = *(B);
		t1 = *(A+IncA);	t1b = *(B+IncB);
				FD_SETOP(b0, t0b, +, t0);
		t2 = *(A+IncA2);	t2b = *(B+IncB2);
				FD_SETOP(b1, t1b, +, t1);
     for (i=4; i<n; i+=4) {
		t3 = *(A+IncA3);	t3b = *(B+IncB3);
	     A += IncA4; 
				FD_SETOP(b2, t2b, +, t2);
							*(B) = b0;
		t0 = *(A);		t0b = *(B+IncB4);
				FD_SETOP(b3, t3b, +, t3);
							*(B+IncB) = b1;
	     B += IncB2;
		t1 = *(A+IncA);	t1b = *(B+IncB3);
				FD_SETOP(b0, t0b, +, t0);
							*(B) = b2;
		t2 = *(A+IncA2);	t2b = *(B+IncB4);
				FD_SETOP(b1, t1b, +, t1);
							*(B+IncB) = b3;
	     B += IncB2;
	  } /* for (i...) */
		t3 = *(A+IncA3);	t3b = *(B+IncB3);
		A += IncA4; 
				FD_SETOP(b2, t2b, +, t2);
							*(B) = b0;
				FD_SETOP(b3, t3b, +, t3);
							*(B+IncB) = b1;
		B += IncB2;
							*(B) = b2;
							*(B+IncB) = b3;
		B += IncB2;
  } /* if (n>0) */
#else
  if (n > 0) {
    register int IncA2=IncA+IncA, IncA3=IncA2+IncA, IncA4=IncA2+IncA2;
    register int IncB2=IncB+IncB, IncB3=IncB2+IncB, IncB4=IncB2+IncB2;
    FD_register t0, t1, t2, t3, t0b, t1b, t2b, t3b, b0, b1, b2, b3;
    register FDATA *blx = B; 
		t0 = *(A);		t0b = *(blx);
		t1 = *(A+IncA);	t1b = *(blx+IncB);
		t2 = *(A+IncA2);	t2b = *(blx+IncB2);
    for (i=-n+4; i<0; i+=4) {
		t3 = *(A+IncA3);  	t3b = *(blx+IncB3); 	A += IncA4; 
				FD_SETOP(b0, t0b, +, t0); 	blx+=IncB4;
						*(B) = b0;  
		t0 = *(A);		t0b = *(blx);
				FD_SETOP(b1, t1b, +, t1);
						*(B+IncB) = b1;
		t1 = *(A+IncA);	t1b = *(blx+IncB);
				FD_SETOP(b2, t2b, +, t2);
						*(B+IncB2) = b2; 
		t2 = *(A+IncA2);	t2b = *(blx+IncB2);
				FD_SETOP(b3, t3b, +, t3);
						*(B+IncB3) = b3; B += IncB4;
    } /* for (i...) */
		t3 = *(A+IncA3);	t3b = *(blx+IncB3);
		A += IncA4; 
				FD_SETOP(b0, t0b, +, t0);
						*(B) = b0;
				FD_SETOP(b1, t1b, +, t1);
						*(B+IncB) = b1;
				FD_SETOP(b2, t2b, +, t2);
						*(B+IncB2) = b2;
				FD_SETOP(b3, t3b, +, t3);
						*(B+IncB3) = b3; B += IncB4;						
  }
#endif	
  
#endif VANILLA    
  for (i=n; i < N; i++) {
      FD_ASSIGN(*B, +=, *A);
      A += IncA; B += IncB;
  } 
#endif /*not defined BLAS*/
} /* VecAdd() */

#endif /* BLAS */

