#define INLINE_BLAS
/************************************************************************/
/*       	CellVecCopy: FDATA Matrix add/copy procedures     	*/
/************************************************************************/
/* SCCS INFO @(#)CellVecCopy.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 97/02
*/
/* Set of vector copy & add routines used for DBLAS,
 */
#ifndef __CellVecCopy_INCL__
#define __CellVecCopy_INCL__  

#include "FDATASetUp.h"

/********************* exported symbols  ***********************************/
#ifndef CELLBLAS

#if FDATATYPE==SINGLE
#define VecCopy		DSVecCopy   
#define VecAdd		DSVecAdd   
#elif FDATATYPE==DOUBLE
#define VecCopy		DDVecCopy   
#define VecAdd		DDVecAdd   
#elif FDATATYPE==COMPLEX
#define VecCopy		DCVecCopy   
#define VecAdd		DCVecAdd   
#else
#define VecCopy		DZVecCopy   
#define VecAdd		DZVecAdd   
#endif

#else	/* rename to avoid linker clash */

#if FDATATYPE==SINGLE
#define VecCopy		SVecCopy   
#define VecAdd		SVecAdd   
#elif FDATATYPE==DOUBLE
#define VecCopy		DVecCopy   
#define VecAdd		DVecAdd   
#elif FDATATYPE==COMPLEX
#define VecCopy		CVecCopy   
#define VecAdd		CVecAdd   
#else
#define VecCopy		ZVecCopy   
#define VecAdd		ZVecAdd   
#endif

# endif /*CELLBLAS*/

#if defined(BLAS)
#include "FtnBLAS.h"

static inline void VecCopy_f(N, A, IncA, B, IncB)
  int N, IncA, IncB; FDATA *A, *B;
{ FTN_COPY(&N, A, &IncA, B, &IncB);
}

static inline void VecAdd_f(N, A, IncA, B, IncB)  
  int N, IncA, IncB; FDATA *A, *B;
{ FTN_AXPY(&N, &FD_1, A, &IncA, B, &IncB);    
}
#endif

#if defined(BLAS) && defined(INLINE_BLAS) 

#undef VecCopy
#undef VecAdd
#define VecCopy VecCopy_f
#define VecAdd VecAdd_f

#else
void VecCopy STDC_PARAMS( (int N, FDATA *A, int IncA, FDATA *B, int IncB) );
/* post: B[j] = A[j], 0<=j<N
 * note: low-latency routine
 */

void VecAdd STDC_PARAMS( (int N, FDATA *A, int IncA, FDATA *B, int IncB) );
/* post: B[j] += A[j], 0<=j<N
 * note: low-latency routine
 */
#endif /*BLAS*/

#endif /*__CellVecCopy_INCL__*/
