/************************************************************************/
/*       	CellVector: FDATA vector procedures for cells    	*/
/************************************************************************/
/* SCCS INFO @(#)CellVector.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 96/12/18 
*/
/* Set of support matrix routines used for DBLAS, independent of distributed
 * memory issues. A vector is represented as an address-stride pair.
 */
#ifndef __CellVector_INCL__
#define __CellVector_INCL__  

#include "FDATASetUp.h"

/********************* exported symbols  ***********************************/

#ifndef CELLBLAS

#if FDATATYPE==SINGLE
#define VecSwap		DSVecSwap
#define VecAbsMaxPos	DSVecAbsMaxPos
#define VecAbsSum	DSVecAbsSum	
#elif FDATATYPE==DOUBLE
#define VecSwap		DDVecSwap
#define VecAbsMaxPos	DDVecAbsMaxPos
#define VecAbsSum	DDVecAbsSum	
#elif FDATATYPE==COMPLEX
#define VecSwap		DCVecSwap
#define VecAbsMaxPos	DCVecAbsMaxPos
#define VecAbsSum	DCVecAbsSum	
#else
#define VecSwap		DZVecSwap
#define VecAbsMaxPos	DZVecAbsMaxPos
#define VecAbsSum	DZVecAbsSum	
#endif

#else 	/* rename to avoid name clash */

#if FDATATYPE==SINGLE
#define VecSwap		SVecSwap
#define VecAbsMaxPos	SVecAbsMaxPos
#define VecAbsSum	SVecAbsSum	
#elif FDATATYPE==DOUBLE
#define VecSwap		DVecSwap
#define VecAbsMaxPos	DVecAbsMaxPos
#define VecAbsSum	DVecAbsSum	
#elif FDATATYPE==COMPLEX
#define VecSwap		CVecSwap
#define VecAbsMaxPos	CVecAbsMaxPos
#define VecAbsSum	CVecAbsSum	
#else
#define VecSwap		ZVecSwap
#define VecAbsMaxPos	ZVecAbsMaxPos
#define VecAbsSum	ZVecAbsSum	
#endif

#endif

#ifdef BLAS             /* define Fortran BLAS procedure names called */
#include "FtnBLAS.h"
#endif /*BLAS*/

/************************ optimization-related macros ***********************/
 
void VecSwap STDC_PARAMS( (int n, FDATA A[], int IncA, FDATA B[], int IncB) );
/* post: A[0..n-1] <-> B[0..n-1]
 */
 
int VecAbsMaxPos STDC_PARAMS( (int m, FDATA A[], int IncA) );
/* post: returns 1st. 0 <= i < m: A[i] and |A[i]| >= |A[j]| for 0 <= j < m
 */

FD_BASE VecAbsSum STDC_PARAMS( (int m, FDATA A[], int IncA ) );
/* post: returns sum: 0 <= i < m: |A[i]|
 */   

#endif /*__CellVector_INCL__*/
