/************************************************************************/
/*         D_AMAX: Level 1 Distributed BLAS vector absolute maximum	*/
/************************************************************************/
/* SCCS INFO @(#)D_AMAX.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 96/01
 */

/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#include "FDMalloc.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

#include "CellMatrix.h"		/* MatScale(), MatCopy/Add(), MatMult() */
#include "FDCommun.h"

#include "CellVector.h"        	/* VecAbsMaxPos() */

int D_AMAX (N, X, Max) 
  int N; DistMat X; FDATA *Max;
{ FDATA amax[2];			int imax = -1;
  int ParamErr=0;
  FD_BASE MsgMaxAbsValTmp = MsgMaxAbsVal;

  if (!NoOuterErrCheck)   
    CheckDim(ParamErr, N, DM_Inc(1-Tr(X), X), DM_DM(X), 0x12);
    
  SetUpComm(X);
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s,_)", N, DM_TRACE(0, X)),
    ParamErr, "AMAX", 1
  );

  MsgMaxAbsVal = 0.0; 		
  /* suppress received element data element checking in FDATACommun;
     here we are sending index position in messages, which may be
     > expected value of any matrix element */

  if (d_P0(DM_DN(X))==0) {
    int nx = G2Ld(N, DM_DM(X));		int IncX = DM_Inc(DM_Trans(X), X);
    imax = VecAbsMaxPos(nx, LM_ADR(DM_Loc(X)), IncX);
    if (imax==-1) {			/* assert nx = 0 */
      amax[0] = FD_0;			imax = 0;
    } else  
      amax[0] = *(LM_ADR(DM_Loc(X)) + FAST_MUL2(imax, IncX));
    imax = L2Gd(imax, DM_DM(X));	/* convert to a global index */
    AbsMaxElt(DM_Trans(X), &amax[0], &imax);
    if (FD_EQ(amax[0], FD_0))			/* take top element as being the max */
      imax = 0;
    FD_SET_REAL(amax[1], (FD_BASE) imax);  
/*  BrdMat(DM_Trans(X), d_P0(DM_DM(X)), 1, 2, amax, 2);*/
  }
  if (d_P(DM_DN(X)) > 1)
    BrdMat(1-DM_Trans(X), d_P0(DM_DN(X)), 1, 2, amax, 2); 
  *Max = amax[0];    

  MsgMaxAbsVal = MsgMaxAbsValTmp;  	/* restore value for FDATACommun */

  TRACE_RETURN((int) FD_REAL(amax[1]));        
} /* D_AMAX() */
