/************************************************************************/
/*         D_ASUM: Level 1 Distributed BLAS vector absolute sum		*/
/************************************************************************/
/* SCCS INFO @(#)D_ASUM.c 1.3         last modified P. Strazdins, 99/06/21
   first created: P. Strazdins 96/01
 */
/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h" 
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#include "FDMalloc.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

#include "CellMatrix.h"		/* MatScale(), MatCopy/Add(), MatMult() */
#include "FDCommun.h"


#include "CellVector.h"			/* VecAbsSum() */

FD_BASE D_ASUM (N, X) 
  int N; DistMat X; 
{ FDATA asum = FD_0;
  int ParamErr = 0;

  if (!NoOuterErrCheck)  
    CheckDim(ParamErr, N, DM_Inc(1-Tr(X), X), DM_DM(X), 0x12);
   
  SetUpComm(X);
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s)", N, DM_TRACE(0,X)),
    ParamErr, "ASUM", 1
  );
  
  if (d_P0(DM_DN(X))==0) {
    int IncX = DM_Inc(DM_Trans(X), X);	LocMat Ap;
    FD_SET_REAL(asum, VecAbsSum(G2Ld(N, DM_DM(X)), LM_ADR(DM_Loc(X)), IncX)); 
    LM_Init(Ap, &asum, 1, 1);
    SumMat(DM_Trans(X), d_P0(DM_DM(X)), d_P(DM_DM(X)), 1, 1, &Ap);
    BrdMat(DM_Trans(X), d_P0(DM_DM(X)), 1, 1, &asum, 1);
  } 
  if (d_P(DM_DN(X)) > 1)
    BrdMat(1-DM_Trans(X), d_P0(DM_DN(X)), 1, 1, &asum, 1);      
  TRACE_RETURN(FD_REAL(asum));
} /* D_ASUM() */

