/************************************************************************/
/*         D_AXPY.c: Level 1 Distributed BLAS vector AXPY 		*/
/************************************************************************/
/* SCCS INFO @(#)D_AXPY.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 96/01
 */

/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#include "FDMalloc.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

#include "CellMatrix.h"		/* MatScale(), MatCopy/Add(), MatMult() */
#include "FDCommun.h"

static inline void SendVec(SwapDIJ, DI, DJ, nx, Xp)
  int SwapDIJ, DI, DJ, nx; LocMat Xp; 
/* post: send nx x 1 Xp to cell (SwapDIJ? (DI,DJ): (DJ:DI))   
 */
{ int IncX = LM_Inc(1, Xp);
  FDATA *b = LM_ADR(Xp);
  if (IncX != 1) {			/* must pack before sending */
    MatMalloc_FT("SendVec", 1, &b, nx);
    VecCopy(nx, LM_ADR(Xp), IncX, b, 1);
  }  
  if (SwapDIJ)
    SWAP(int, DI, DJ);
  SendMat(DI, DJ, 1, nx, b, nx, GENMSGTTYPE);  
  if (IncX != 1)
     MatFree(b);
} /* SendVec() */

#include "DistMatrix.h"            	/* TransMat(), RotMat()  */

void D_AXPY (N, beta, X, Y) 
  int N; FDATA beta; DistMat X; DistMat Y;
{ LocMat Xp, Yp;			BetaDecln();
  int ParamErr=0;                       int AlignN;

 if (!NoOuterErrCheck) {
    CheckAlign(ParamErr, AlignN, N, DM_Inc(1-Tr(X), X), DM_DM(X),
               DM_Inc(1-Tr(Y), Y), DM_DM(Y), 0x123);
    CheckParam(ParamErr, AlignN && P_Same(X, Y), 0x34);
 }    
  
  SetUpComm(X);				
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s,%s,%s)", N, pFD2Str(&beta), DM_TRACE(0, X), DM_TRACE(1, Y)),
    ParamErr, "AXPY", 1
  );
  
  BetaSet(beta, pBeta);

  LM_Copy(1-DM_Trans(X), Xp, DM_Loc(X));	/* Xp, Yp are 1 x n */
  LM_Copy(1-DM_Trans(Y), Yp, DM_Loc(Y));	

  if (DM_Trans(X)!=DM_Trans(Y)) 
    TransMat(1-DM_Trans(X), 1, 1, N, &Xp, DM_DN(X), DM_DM(X), 1,
             pBeta, &Yp, d_P0(DM_DM(Y)), d_P0(DM_DN(Y)), 1);
  else if (d_P0(DM_DN(X))==0 && d_P0(DM_DN(Y))==0 && 
	   d_P0(DM_DM(X))==d_P0(DM_DM(Y))) {
    AxpyMult(G2Ld(N, DM_DM(X)), pBeta, &Xp, &Yp);
  } else if (d_P0(DM_DN(X))==0 || d_P0(DM_DN(Y))==0) {
    int n = G2Ld(N, DM_DM(X));	
    if (d_P0(DM_DN(X))==0 && n>0) {
      SendVec(DM_Trans(X), d_P0(DM_DM(X))-d_P0(DM_DM(Y)), -d_P0(DM_DN(Y)), 
	      n, Xp);
    }
    if (d_P0(DM_DN(Y))==0) {
      int IncY = LM_Inc(1, Yp);
      int DI = d_P0(DM_DM(Y))-d_P0(DM_DM(X)),	DJ = -d_P0(DM_DN(X));
      int ny = DI==0? n: G2Ld(N, DM_DM(Y));
      if (DM_Trans(X))
        SWAP(int, DI, DJ);  
      if (IncY != 1 && ny>0) { 
        FDATA *b = NULL;	
        MatMalloc_FT("D_AXPY", 1, &b, ny);
        LM_Init(Xp, b, ny, 1);
        RecvMat(DI, DJ, 1, ny, LM_ADR(Xp), ny, GENMSGTTYPE);  
        MatAdd(1, ny, &Xp, pBeta, &Yp);
        MatFree(b);
      } else if (ny>0) 
        RecvBlkMat(DI, DJ, 1, 1, 0, 1, 1, ny, pBeta, LM_ADR(Yp), GENMSGTTYPE);
    }    
  }
  TRACE_EXIT;
} /* D_AXPY() */

