/************************************************************************/
/*         D_BLASCheck: Distributed BLAS error checking macros		*/
/************************************************************************/
/* SCCS INFO @(#)D_BLASCheck.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 97/02
 */
/* common macro definitions used to simplify error checking in outer level
 * DBLAS procedures
*/   

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

#ifndef _D_BLASCheck_INCL_
#define _D_BLASCheck_INCL_

#define UpperCase(c) (('a' <= (c) && (c) <= 'z')? (c)+'A'-'a': (c))

#define DimOK(M) ((M)>=0)

#define Tr(A)		DM_Trans(A)
#define DM_Inc(t, A)	LM_Inc(t, DM_Loc(A))
#define DM_P(t, C)	d_P(DM_D(t, C))

#define CheckParam(ParamErr, ParamCond, ParamNum) 		\
	(ParamErr) = (!(ParamCond) && (ParamErr)==0)? (ParamNum): (ParamErr);

#define CheckAlign(ParamErr, AlignN, N, MIncA, DNA, MIncB, DNB, ParamNums) \
	if (!ParamErr) {						\
	  (ParamErr) = (ParamNums);					\
	  AlignN = CommonDimAlignCheck(N, MIncA, DNA, MIncB, DNB, &(ParamErr));\
	}

#define CheckDim(ParamErr, N, MIncA, DNA, ParamNums) 			\
	if (!ParamErr) {						\
	  (ParamErr) = (ParamNums);					\
	  CommonDimAlignCheck(N, MIncA, DNA, 0, NULL, &(ParamErr));\
	}

#define P_OK(P_ADN, P_Aref)	((P_ADN)==(P_Aref) || (P_ADN)==1)

static inline int PAlignCheck(P_ADN, P_Aref, NAlign)
  int P_ADN, P_Aref, *NAlign;
{ int NAlignA = P_ADN==P_Aref;
  *NAlign &= NAlignA;
  return (NAlignA || P_ADN==1);
}
	
#define DM_TRACE(n, A)	PrintDistMat(A, TmpTraceStrs[n])

#define SetUpComm(C)	CommSetUp(d_P(DM_D0(C)), d_P(DM_D1(C)))

/*copy of parameter alpha etc needed for workaround apparent Sun Sys V cc bug*/
#define AlphaDecln()			FDATA Alpha, *pAlpha
#define BetaDecln()			FDATA Beta, *pBeta
#define AlphaBetaDecln()		AlphaDecln(); BetaDecln()
#define AlphaSet(alpha, pAlpha)		Alpha=alpha, pAlpha=&Alpha
#define BetaSet(beta, pBeta)		pBeta= FD_EQ(beta, FD_1)? NULL: (Beta=beta, &Beta)

#endif _D_BLASCheck_INCL_

