/************************************************************************/
/*         D_COPY: Level 1 Distributed BLAS vector copy			*/
/************************************************************************/
/* SCCS INFO @(#)D_COPY.c 1.1         last modified P. Strazdins, 98/12/24
   first created: P. Strazdins 96/01
 */

/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDAssert.h"
#include "FDMalloc.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

#include "CellMatrix.h"		/* MatScale(), MatCopy/Add(), MatMult() */
#include "FDCommun.h"

static inline void SendVec(SwapDIJ, DI, DJ, nx, Xp)
  int SwapDIJ, DI, DJ, nx; LocMat Xp; 
/* post: send nx x 1 Xp to cell (SwapDIJ? (DI,DJ): (DJ:DI))   
 */
{ int IncX = LM_Inc(1, Xp);
  FDATA *b = LM_ADR(Xp);
  if (IncX != 1) {			/* must pack before sending */
    MatMalloc_FT("SendVec", 1, &b, nx);
    VecCopy(nx, LM_ADR(Xp), IncX, b, 1);
  }  
  if (SwapDIJ)
    SWAP(int, DI, DJ);
  SendMat(DI, DJ, 1, nx, b, nx, GENMSGTTYPE);  
  if (IncX != 1)
     MatFree(b);
} /* SendVec() */

#include "DistMatrix.h"            	/* TransMat(), RotMat()  */
#include "SpreadMatrix.h"

void D_COPY (N, X, Y)
  int N; DistMat X; DistMat Y;
{ int ParamErr=0;                       int AlignN=1;
  LocMat Xp, Yp;

  if (!NoOuterErrCheck) {
    CheckAlign(ParamErr, AlignN, N, DM_Inc(1-Tr(X), X), DM_DM(X),
               DM_Inc(1-Tr(Y), Y), DM_DM(Y), 0x123);
    CheckParam(ParamErr, AlignN && (P_Same(X, Y) ||
    		(d_P(DM_DN(Y))==1 && 
    	         d_P(DM_DM(Y))==d_P(DM_D(DM_Trans(Y), X)))), 0x23);
  }
      
  SetUpComm(X);
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s,%s)", N, DM_TRACE(0,X), DM_TRACE(1,Y)),
    ParamErr, "COPY", 1
  );

  if (N==0) {
    TRACE_RETURN;
  }
    
  LM_Copy(1-DM_Trans(X), Xp, DM_Loc(X));/* Xp, Yp are 1 x n */
  LM_Copy(1-DM_Trans(Y), Yp, DM_Loc(Y));	

  if (d_P(DM_DN(Y)) < d_P(DM_D(1-DM_Trans(Y), X))) {
    LocMat sX;				int nsX;
    int SpreadToY = 0; 			/*=1 if a spread can go directly into C*/
    int ForceCopy = 0;	
    int RC = 1-DM_Trans(Y);
    LM_ADR(sX) = NULL;
    if (DM_Trans(Y))
      X = Trans(X);
    if (LM_Inc(1-RC, DM_Loc(Y))==1 && 
       (DM_Trans(X) || d_P0(DM_D(RC, X))==d_P0(DM_D(RC, Y)))) {
      	SpreadToY = 1,			ForceCopy = 1;
        LM_Copy(0, sX, DM_Loc(Y));   
    }  
    nsX = SpreadMat(RC, 1, N, X, &FD_1, DM_DM(Y),
                    &sX, ForceCopy, 1+d_B(DM_DN(X)), DM_DN(X));
    if (!SpreadToY) 
      VecCopy(ABS(nsX), LM_ADR(sX), 1, LM_ADR(Yp), LM_CInc(Yp)); 
    if (nsX > 0 && !SpreadToY)
      MatFree(LM_ADR(sX));			
  } else if (DM_Trans(X)!=DM_Trans(Y))
    TransMat(1-DM_Trans(X), 0, 1, N, &Xp, DM_DN(X), DM_DM(X), 1,
             NULL, &Yp, d_P0(DM_DM(Y)), d_P0(DM_DN(Y)), 1);
  else if (d_P0(DM_DN(X))==0 && d_P0(DM_DN(Y))==0 && 
	   d_P0(DM_DM(X))==d_P0(DM_DM(Y)))
    MatCopy(1, G2Ld(N, DM_DM(Y)), &Xp, NULL, &Yp);
  else if (d_P0(DM_DN(X))==0 || d_P0(DM_DN(Y))==0) {
    int n = G2Ld(N, DM_DM(X));	
    if (d_P0(DM_DN(X))==0 && n>0) 
      SendVec(DM_Trans(X), d_P0(DM_DM(X))-d_P0(DM_DM(Y)), -d_P0(DM_DN(Y)), n, Xp);
    if (d_P0(DM_DN(Y))==0) {
      int IncY = LM_Inc(1, Yp);
      int DI = d_P0(DM_DM(Y))-d_P0(DM_DM(X)),	DJ = -d_P0(DM_DN(X));
      int ny = DI==0? n: G2Ld(N, DM_DM(Y));
      FDATA *b = LM_ADR(Yp);			
      if (DM_Trans(X))
        SWAP(int, DI, DJ);
      if (ny > 0) {  
        if (IncY != 1)          		/* must receive into a buffer */
          MatMalloc_F("D_COPY", 1, &b, ny);
        RecvMat(DI, DJ, 1, ny, b, ny, GENMSGTTYPE);  
        if (IncY != 1) {
          VecCopy(ny, b, 1, LM_ADR(Yp), IncY);                  
          MatFree(b);
        }    
      }
    }    
  }   
  TRACE_EXIT;
} /* D_COPY() */


