/************************************************************************/
/*         D_DOT: Level 1 Distributed BLAS vector dot product		*/
/************************************************************************/
/* SCCS INFO @(#)D_DOT.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 96/01
 */

/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDAssert.h"
#include "FDMalloc.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

#include "CellMatrix.h"		/* MatScale(), MatCopy/Add(), MatMult() */
#include "FDCommun.h"

static inline void SendVec(SwapDIJ, DI, DJ, nx, Xp)
  int SwapDIJ, DI, DJ, nx; LocMat Xp; 
/* post: send nx x 1 Xp to cell (SwapDIJ? (DI,DJ): (DJ:DI))   
 */
{ int IncX = LM_Inc(1, Xp);
  FDATA *b = LM_ADR(Xp);
  if (IncX != 1) {			/* must pack before sending */
    MatMalloc_FT("SendVec", 1, &b, nx);
    VecCopy(nx, LM_ADR(Xp), IncX, b, 1);
  }  
  if (SwapDIJ)
    SWAP(int, DI, DJ);
  SendMat(DI, DJ, 1, nx, b, nx, GENMSGTTYPE);  
  if (IncX != 1)
     MatFree(b);
} /* SendVec() */

FDATA D_DOT (N, X, Y) 
  int N; DistMat X; DistMat Y;
{ LocMat Xp, Yp, Sp; 			FDATA prod; 
  int nx;
  int DI=d_P0(DM_DM(Y))-d_P0(DM_DM(X)), DJ=d_P0(DM_DN(Y))-d_P0(DM_DN(X));
  int ParamErr=0;                       int AlignN;

 if (!NoOuterErrCheck) {
    CheckAlign(ParamErr, AlignN, N, DM_Inc(1-Tr(X), X), DM_DM(X),
               DM_Inc(1-Tr(Y), Y), DM_DM(Y), 0x123);
    CheckParam(ParamErr, AlignN && P_Same(X, Y) && DM_Trans(X)==DM_Trans(Y),
               0x24);
 }
    
  SetUpComm(X);  
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s,%s)", N, DM_TRACE(0, X), DM_TRACE(1, Y)),
    ParamErr, "DOT", 1
  );
    
  nx = G2Ld(N, DM_DM(X));
  LM_Copy(1-DM_Trans(X), Xp, DM_Loc(X));	/* Xp is 1 x nx */
  if (d_P0(DM_DN(X))==0 && (DI!=0 || DJ!=0) && nx>0) 
    SendVec(DM_Trans(X), -DI, -DJ, nx, Xp);
  if (d_P0(DM_DN(Y))==0) {
    int ny = DI==0? nx: G2Ld(N, DM_DM(Y));
    FDATA *b=NULL;		
    if ((DI!=0 || DJ!=0) && ny>0) {
      MatMalloc_FT("D_DOT", 1, &b, ny);
      if (DM_Trans(Y))
        SWAP(int, DI, DJ);
      RecvMat(DI, DJ, 1, ny, b, ny, GENMSGTTYPE);
      LM_Init(Xp, b, ny, 1);
    }  
    LM_Copy(DM_Trans(Y), Yp, DM_Loc(Y));	/* Yp is ny x 1 */
    LM_Init(Sp, &prod, 1, 1);
    prod = FD_0;
    IPMult(&prod, ny, &Xp, &Yp);
    if (b!=NULL)
      MatFree(b);
    SumMat(DM_Trans(Y), d_P0(DM_DM(Y)), d_P(DM_DM(Y)), 1, 1, &Sp);
     if (d_P(DM_DM(Y)) > 1)
      BrdMat(DM_Trans(Y), d_P0(DM_DM(Y)), 1, 1, &prod, 1);
  } 
  if (d_P(DM_DN(Y)) > 1)
    BrdMat(1-DM_Trans(Y), d_P0(DM_DN(Y)), 1, 1, &prod, 1); 
  TRACE_RETURN(prod);  
} /* D_DOT() */

