/************************************************************************/
/*         D_GEMM: Level 3 Distributed BLAS general matrix multiply	*/
/************************************************************************/
/* SCCS INFO @(#)D_GEMM.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 93/03/23
 */
/* implementation file for Distributed Level-3 BLAS procedures; 
   for compilation, the macro FDATATYPE (having the values SINGLE or DOUBLE)
   must be defined:
	if FDATATYPE=SINGLE, FDATA=float
	if FDATATYPE=DOUBLE, FDATA=double
   Below, the underscores in the procedure name should be replaced by 'S',
   if FDATATYPE=SINGLE, and 'D', if FDATATYPE=DOUBLE. 
   eg. for FDATATYPE=SINGLE, D_GEMM becomes DSGEMM.

   See DBLAS.h for a general description of this module.
   
   gerneral (externally-defined) cpp FLAGs:
        FDATATYPE       SINGLE (default), DOUBLE, COMPLEX or DOUBLE_COMPLEX
        ASSERT, TRACE   turn assertion checking and procedure tracing on
                        in major internal procedure calls
*/   
#include "D_BLAS.h"
#include "D_BLASCheck.h"
#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDCommun.h"  			/* CommSetUp() */

#include "GenMult.h"

void D_GEMM(M, N, K, alpha, A, B, beta, C)
  int M; int N; int K;
  FDATA alpha; DistMat A; DistMat B; FDATA beta; DistMat C;
{ int AlignM, AlignN, AlignK;
  int ParamErr=0;
  AlphaBetaDecln();

  CheckAlign(ParamErr, AlignM, M, DM_Inc(1-Tr(A), A),   DM_DM(A), 
  	     DM_Inc(1-Tr(C), C),   DM_DM(C), 0x158);
  CheckParam(ParamErr, 
  	     PAlignCheck(d_P(DM_DM(A)), DM_P(Tr(A), C), &AlignM), 0x58);
  CheckAlign(ParamErr, AlignN, N, DM_Inc(Tr(B), B), DM_DN(B), 
  	     DM_Inc(Tr(C), C), DM_DN(C), 0x268);  
  CheckParam(ParamErr, 
             PAlignCheck(d_P(DM_DN(B)), DM_P(1-Tr(B), C), &AlignM), 0x68);
  CheckAlign(ParamErr, AlignK, K, DM_Inc(Tr(A), A), DM_DN(A), 
  	     DM_Inc(1-Tr(B), B),   DM_DM(B), 0x356);  
  CheckParam(ParamErr, 
   	     PAlignCheck(d_P(DM_DN(A)), DM_P(1-Tr(A), C), &AlignK), 0x58);
  CheckParam(ParamErr, 
   	     PAlignCheck(d_P(DM_DM(B)), DM_P(Tr(B), C), &AlignK), 0x68);

  if (ForceAlgVers==1) 		AlignM = 0;
  else if (ForceAlgVers==2) 	AlignN = 0;
  else if (ForceAlgVers==3) 	AlignK = 0;
           	     
  CheckParam(ParamErr, AlignM+AlignN+AlignK>=2, 0x568);
  
  SetUpComm(C);				
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%d,%d,%s,%s,%s,%s,%s)", 
      M, N, K, pFD2Str(&alpha), DM_TRACE(0, A), DM_TRACE(1, B), pFD2Str(&beta), DM_TRACE(2, C)),
    ParamErr, "GEMM", 0
  );

  if (AlignM+AlignN+AlignK==3) 		/* chose version simply on dim. size */	
    if (M < N && M < K)
      AlignM = 0;
    else if (N < M && N < K)
      AlignN = 0;         	
    else 
     AlignK = 0;  
  
  AlphaSet(alpha, pAlpha);	 	BetaSet(beta, pBeta);
  if (!AlignK)
    DistRankKMult(M, N, K, pAlpha, A, B, pBeta, C);
  else if (!AlignM)
    DistVecMatMult(M, N, K, pAlpha, A, B, pBeta, C);
  else 
    DistVecMatMult(N, M, K, pAlpha, Trans(B), Trans(A), pBeta, Trans(C));    
  TRACE_EXIT;
} /* D_GEMM() */
