/************************************************************************/
/*         D_GEMV: Level 2 Distributed BLAS matrix vector multiply	*/
/************************************************************************/
/* SCCS INFO @(#)D_GEMV.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 95/03
 */
/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDCommun.h"  			/* CommSetUp() */

#include "GenMult.h"

void D_GEMV (M, N, alpha, A, X, beta, Y)
  int M; int N; FDATA alpha; DistMat A;
  DistMat X; FDATA beta; DistMat Y;
{ int ParamErr=0;			int  AlignM, AlignN;
  AlphaBetaDecln();  

  if (!NoOuterErrCheck) {
    CheckAlign(ParamErr, AlignN, N, DM_Inc(Tr(A), A), DM_DN(A),
               DM_Inc(1-Tr(X), X), DM_DM(X), 0x245);
    CheckParam(ParamErr, AlignN && d_P(DM_DM(X))==DM_P(Tr(X), A) &&
                         P_OK(d_P(DM_DN(X)), DM_P(1-Tr(X), A)), 0x45);
    CheckAlign(ParamErr, AlignM, M, DM_Inc(1-Tr(A), A),  DM_DM(A),
               DM_Inc(1-Tr(Y), Y),   DM_DM(Y), 0x147);
    CheckParam(ParamErr, AlignM && d_P(DM_DM(Y))==DM_P(Tr(Y), A) &&
                         P_OK(d_P(DM_DN(Y)), DM_P(1-Tr(Y), A)), 0x47);
  }

  SetUpComm(A);
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%d,%s,%s,%s,%s,%s)", M, N, 
     pFD2Str(&alpha), DM_TRACE(0, A), DM_TRACE(1, X), pFD2Str(&beta), DM_TRACE(2, Y)),
     ParamErr, "GEMV", 1
  );  

  AlphaSet(alpha, pAlpha);              BetaSet(beta, pBeta);
  DistVecMatMult(1, M, N, pAlpha, Trans(X), Trans(A), pBeta, Trans(Y));
  TRACE_EXIT;
} /* D_GEMV() */

