/************************************************************************/
/*         D_NRM2: Level 1 Distributed BLAS vector norm			*/
/************************************************************************/
/* SCCS INFO @(#)D_NRM2.c 1.3         last modified P. Strazdins, 99/06/21
   first created: P. Strazdins 96/01
 */

/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#include "FDMalloc.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

#include "CellMatrix.h"		/* MatScale(), MatCopy/Add(), MatMult() */
#include "FDCommun.h"

FD_BASE D_NRM2 (N, X)
  int N; DistMat X;	
{ LocMat Xp, XpT, Sp;			FDATA prod;
  double sqrt();
  int ParamErr = 0;

  if (!NoOuterErrCheck) 
    CheckDim(ParamErr, N, DM_Inc(1-Tr(X), X), DM_DM(X), 0x12);
    
  SetUpComm(X);  
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s)", N, DM_TRACE(0, X)),
    ParamErr, "NRM2", 1
  );
  
  if (d_P0(DM_DN(X))==0) {
    int n = G2Ld(N, DM_DM(X));
    LM_Copy(DM_Trans(X), Xp, DM_Loc(X));	/* Xp is ny x 1 */
    LM_Copy(1-DM_Trans(X), XpT, DM_Loc(X));	/* XpT is 1 x ny */
    LM_Init(Sp, &prod, 1, 1);
    prod = FD_0; 
    IPMult(&prod, n, &XpT, &Xp);
    SumMat(DM_Trans(X), d_P0(DM_DM(X)), d_P(DM_DM(X)), 1, 1, &Sp);
    if (d_P(DM_DM(X)) > 1)
      BrdMat(DM_Trans(X), d_P0(DM_DM(X)), 1, 1, &prod, 1);
    FD_SET_REAL(prod, (FD_BASE) sqrt((double) FD_REAL(prod))); 
  } 
  if (d_P(DM_DN(X)) > 1)
    BrdMat(1-DM_Trans(X), d_P0(DM_DN(X)), 1, 1, &prod, 1); 
  TRACE_RETURN(FD_REAL(prod));  
} /* D_NRM2() */

