/************************************************************************/
/*         D_SCAL: Level 1 Distributed BLAS vector scale 		*/
/************************************************************************/
/* SCCS INFO @(#)D_SCAL.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 96/01
 */
/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#include "FDMalloc.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

#include "CellMatrix.h"		/* MatScale(), MatCopy/Add(), MatMult() */
#include "FDCommun.h"

void D_SCAL (N, beta, X) 
  int N; FDATA beta; DistMat X; 
{ int ParamErr=0;
  BetaDecln();

  if (!NoOuterErrCheck) {
    CheckDim(ParamErr, N, DM_Inc(1-Tr(X), X), DM_DM(X), 0x13);
  }
  
  SetUpComm(X);	
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s,%s)", N, pFD2Str(&beta), DM_TRACE(0, X)),
    ParamErr, "SCAL", 1
  );

  BetaSet(beta, pBeta);
  if (d_P0(DM_DN(X))==0) {
    LocMat Xp;
    LM_Copy(1-DM_Trans(X), Xp, DM_Loc(X));	
    MatScale(1, G2Ld(N, DM_DM(X)), pBeta, &Xp);
  }  	
       
  TRACE_EXIT;
} /* D_SCAL() */

