/************************************************************************/
/*         D_SWAP: Level 1 Distributed BLAS vector swap			*/
/************************************************************************/
/* SCCS INFO @(#)D_SWAP.c 1.2         last modified P. Strazdins, 99/03/19
   first created: P. Strazdins 96/01
 */
/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDAssert.h"
#include "FDMalloc.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"

#include "CellMatrix.h"		/* MatScale(), MatCopy/Add(), MatMult() */
#include "FDCommun.h"

#include "CellVector.h"			/* VecSwap() */

void D_SWAP (N, X, Y) 
  int N; DistMat X; DistMat Y;
{ int IncX = DM_Inc(DM_Trans(X), X),	IncY = DM_Inc(DM_Trans(Y), Y);
  int ParamErr=0;			int AlignN=1;

  if (d_P0(DM_DN(X))!=0 && d_P0(DM_DN(Y))!=0) 
    return;				/* fast exit */
 
  if (!NoOuterErrCheck) {
    CheckAlign(ParamErr, AlignN, N, DM_Inc(1-Tr(X), X), DM_DM(X),
               DM_Inc(1-Tr(Y), Y), DM_DM(Y), 0x123);
    CheckParam(ParamErr, AlignN && P_Same(X, Y) && DM_Trans(X)==DM_Trans(Y), 
	       0x23);
  }

  SetUpComm(X);
  ERROR_CHECK_TR(
    (TraceStr, "(%d,%s,%s)", N, DM_TRACE(0, X), DM_TRACE(1, Y)),
    ParamErr, "SWAP", 1
  );


  if (d_P0(DM_DN(X))==0 && d_P0(DM_DN(Y))==0 && d_P0(DM_DM(X))==d_P0(DM_DM(Y)))
    VecSwap(G2Ld(N, DM_DM(X)), LM_ADR(DM_Loc(X)), IncX, LM_ADR(DM_Loc(Y)), IncY);
  else if (d_P0(DM_DN(X))==0 || d_P0(DM_DN(Y))==0) {
    int DIy = d_P0(DM_DM(Y))-d_P0(DM_DM(X)), 
        DJy = -d_P0(DM_DN(Y))-d_P0(DM_DN(X)),
        DIx = -DIy, 			DJx = DJy;
    FDATA *b = NULL;			
    int nx = G2Ld(N, DM_DM(X)),		ny = DIy==0? nx: G2Ld(N, DM_DM(Y));
	      
    if (DM_Trans(X)) {
      SWAP(int, DIy, DJy);		SWAP(int, DIx, DJx);
    }
#ifndef GLOB_BLK_SEND
    if (IncX != 1 || IncY != 1)
      MatMalloc_FT("D_SWAP", 1, &b, MAX(nx,ny));
      
    if (d_P0(DM_DN(X))==0 && nx>0) {
      FDATA *Xp = LM_ADR(DM_Loc(X));	      
      if (IncX != 1)
        VecCopy(nx, Xp, IncX, b, 1);
      SendMat(DIx, DJx, 1, nx, IncX!=1? b: Xp, nx, GENMSGTTYPE); 
    }
    if (d_P0(DM_DN(Y))==0 && ny>0) {
      FDATA *Yp = LM_ADR(DM_Loc(Y));	
      if (IncY != 1) 
        VecCopy(nx, Yp, IncY, b, 1);
      SendMat(DIy, DJy, 1, ny, IncX!=1? b: Yp, ny, GENMSGTTYPE); 
    }

    if (d_P0(DM_DN(Y))==0 && ny>0) {
      FDATA *Yp = LM_ADR(DM_Loc(Y));	
      RecvMat(DIy, DJy, 1, ny, IncY==1? Yp: b, ny, GENMSGTTYPE);  
      if (IncY != 1)         
        VecCopy(ny, b, 1, Yp, IncY);                  
    }
    if (d_P0(DM_DN(X))==0 && nx>0) {
      FDATA *Xp = LM_ADR(DM_Loc(X));	
      RecvMat(DIx, DJx, 1, nx, IncX==1? Xp: b, nx, GENMSGTTYPE);  
      if (IncX != 1)         
        VecCopy(nx, b, 1, Xp, IncX);                  
    }
#else
    /* GLOB_BLK_SEND => cannot use Send(X), Send(Y); Recv(X); Recv(Y)
       without buffering both X and Y. A solution is to complete  
       the transfer of 1 vector before the other is started.
       The following pattern will generally leave the cell owning X completing 
       1st, which is more useful in its typical usage in applications like LU.
       As the tests (Inc{X,Y} != 1) are only valid
       on the cell owning {X,Y}, the order must be fixed
       (note that for performance, only the the case where both are unity or
       both non-unity is important).
    */

    if (d_P0(DM_DN(X))==0 && nx>0) {
      FDATA *Xp = LM_ADR(DM_Loc(X)), *Xs, *Xr;	
      MatMalloc_FT("D_SWAP", 1, &Xs, IncX==1? nx: 2*nx);      	
      Xr = IncX==1? Xp: Xs + nx;
      VecCopy(nx, Xp, IncX, Xs, 1);
      RecvMat(DIx, DJx, 1, nx, Xr, nx, GENMSGTTYPE);  
      SendMat(DIx, DJx, 1, nx, Xs, nx, GENMSGTTYPE); 
      if (IncX != 1)         
        VecCopy(nx, Xr, 1, Xp, IncX);                  
      MatFree(Xs);
    }

    if (d_P0(DM_DN(Y))==0 && ny>0) {
      FDATA *Yp = LM_ADR(DM_Loc(Y)), *Ys;	
      if (IncY != 1) {
        MatMalloc_FT("D_SWAP", 1, &Ys, ny);      	
        VecCopy(nx, Yp, IncY, Ys, 1);
      } else
        Ys = Yp; 
      SendMat(DIy, DJy, 1, ny, Ys, ny, GENMSGTTYPE); /*must be a blocking send*/
      RecvMat(DIy, DJy, 1, ny, Ys, ny, GENMSGTTYPE);  
      if (IncY != 1) {        
        VecCopy(ny, Ys, 1, Yp, IncY);                  
        MatFree(Ys);
      }  
    }
#endif    
 }
  TRACE_EXIT;
} /* D_SWAP() */

