/************************************************************************/
/*         D_SYMM: Level 3 Distributed BLAS symmetric matrix multiply	*/
/************************************************************************/
/* SCCS INFO @(#)D_SYMM.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 93/03/23
 */
/* see D_GEMM.c for common cpp flag description				*/

#include "D_BLAS.h"
#include "D_BLASCheck.h"
#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
 
#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDCommun.h"                   /* CommSetUp() */

#include "SYMMult.h"

void D_SYMM (UpLo, M, N, alpha, A, B, beta, C)
  char *UpLo; int M; int N; 
  FDATA alpha; DistMat A; DistMat B; FDATA beta; DistMat C;
{ char UpLoA = UpperCase(*UpLo); 
  int ParamErr=0;  			
  AlphaBetaDecln();
  int AlignM=1, AlignN=1;

  if (!NoOuterErrCheck) {      
    CheckParam(ParamErr, UpLoOK(UpLoA), 0x1);  
    CheckParam(ParamErr, Dim_BCompat(DM_D0(A), DM_D1(A)), 0x55);
    CheckAlign(ParamErr, AlignM, M, DM_Inc(1-Tr(A), A),   DM_DM(A),
               DM_Inc(1-Tr(C), C),   DM_DM(C), 0x258);
    CheckParam(ParamErr, AlignM && d_P(DM_DM(A))==DM_P(Tr(A), C) && 
    			           d_P(DM_DN(A))==DM_P(1-Tr(A), C), 0x58);
    CheckAlign(ParamErr, AlignM, M, DM_Inc(Tr(A), A),     DM_DN(A),
               DM_Inc(1-Tr(B), B),  DM_DM(B), 0x256);
    CheckParam(ParamErr, AlignM, 0x57);
    CheckAlign(ParamErr, AlignN, N, DM_Inc(Tr(B), B),     DM_DN(B),
               DM_Inc(Tr(C), C),    DM_DN(C), 0x368);
    CheckParam(ParamErr, 
               PAlignCheck(d_P(DM_DN(B)), DM_P(1-Tr(B), C), &AlignN), 0x68);
  }

  SetUpComm(C);		
  ERROR_CHECK_TR(
    (TraceStr, "(%c,%d,%d,%s,%s,%s,%s,%s)", 
     *UpLo, M, N, pFD2Str(&alpha), DM_TRACE(0, A), DM_TRACE(1, B), pFD2Str(&beta), DM_TRACE(2, C)),
    ParamErr, "SYMM", 0
  );

  AlphaSet(alpha, pAlpha);	 	BetaSet(beta, pBeta);
  SYMMult(M, N, pAlpha, UpLoA, A, B, pBeta, C);
  TRACE_EXIT;
} /* D_SYMM() */
