/************************************************************************/
/*         D_SYMV: Level 2 Distributed BLAS symmetric matrix vector mult*/
/************************************************************************/
/* SCCS INFO @(#)D_SYMV.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 95/03
 */
/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"
 
#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
 
#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDCommun.h"                   /* CommSetUp() */

#include "SYMMult.h"

void D_SYMV (UpLo, N, alpha, A, X, beta, Y)
  char  *UpLo; int N; FDATA alpha; DistMat A;
  DistMat X; FDATA beta; DistMat Y;
{ char UpLoA = UpperCase(*UpLo);
  int ParamErr=0;
  AlphaBetaDecln();
  int AlignM=1, AlignN=1;

  if (!NoOuterErrCheck) {
    CheckParam(ParamErr, UpLoOK(UpLoA), 0x1);
    CheckParam(ParamErr, Dim_BCompat(DM_D0(A), DM_D1(A)), 0x44);
    CheckAlign(ParamErr, AlignM, N, DM_Inc(1-Tr(A), A),   DM_DM(A),
               DM_Inc(1-Tr(Y), Y),   DM_DM(Y), 0x247);
    CheckParam(ParamErr, AlignM && d_P(DM_DM(A))==DM_P(Tr(A), Y) &&
                                   d_P(DM_DN(A))==DM_P(1-Tr(A), Y), 0x47);
    CheckAlign(ParamErr, AlignM, N, DM_Inc(Tr(A), A),     DM_DN(A),
               DM_Inc(1-Tr(X), X),  DM_DM(X), 0x45);
    CheckParam(ParamErr, AlignM, 0x45);
    CheckParam(ParamErr,
               PAlignCheck(d_P(DM_DN(X)), DM_P(1-Tr(X), Y), &AlignN), 0x57);
  }

  SetUpComm(Y);
  ERROR_CHECK_TR(
    (TraceStr, "(%c,%d,,%s,%s,%s,%s,%s)",
     *UpLo, N, pFD2Str(&alpha), DM_TRACE(0, A), DM_TRACE(1, X), pFD2Str(&beta), DM_TRACE(2, Y)),
    ParamErr, "SYMV", 0            
  );
    
  AlphaSet(alpha, pAlpha);              BetaSet(beta, pBeta);
  SYMMult(N, 1, pAlpha, UpLoA, A, X, pBeta, Y);
  TRACE_EXIT;

} /* D_SYMV() */
