/************************************************************************/
/*         D_SYR2.c: Level 2 Distributed BLAS symmetric rank-2 update	*/
/************************************************************************/
/* SCCS INFO @(#)D_SYR2.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 96/01
 */
/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDCommun.h"  			/* CommSetUp() */

#include "SYR2KMult.h"

void D_SYR2 (UpLo, N, alpha, X, Y, A)
  char *UpLo; int N; FDATA alpha; DistMat X; DistMat Y; DistMat A;
{ char UpLoA = UpperCase(*UpLo); 
  int ParamErr=0;
  AlphaBetaDecln();
  int AlignN=1;

  if (!NoOuterErrCheck) {
    CheckParam(ParamErr, UpLoOK(UpLoA), 0x1);
    CheckAlign(ParamErr, AlignN, N, DM_Inc(1-Tr(X), X),   DM_DM(X),
               DM_Inc(1-Tr(A), A),   DM_DM(A), 0x146);
    CheckParam(ParamErr, AlignN && d_P(DM_DM(X))==DM_P(Tr(X), A) &&
                         P_OK(d_P(DM_DN(X)), DM_P(1-Tr(X), A)), 0x46);
    CheckDim(ParamErr, N, DM_Inc(1-Tr(Y), Y), DM_DM(Y), 0x25);
    CheckParam(ParamErr, Dim_BCompat(DM_DM(Y), DM_DM(A)) &&
                         d_P(DM_DM(Y))==DM_P(Tr(Y), A) &&
                         P_OK(d_P(DM_DN(Y)), DM_P(1-Tr(Y), A)), 0x56);
    CheckDim(ParamErr, N, DM_Inc(Tr(A), A), DM_DN(A), 0x26);
    CheckParam(ParamErr, Dim_BCompat(DM_D0(A), DM_D1(A)), 0x66);
  }  

  SetUpComm(A);
  ERROR_CHECK_TR(   
    (TraceStr, "(%c,%d,%s,%s,%s,%s)", *UpLo, N, pFD2Str(&alpha), 
     DM_TRACE(0, X), DM_TRACE(1,Y), DM_TRACE(2, A)),
     ParamErr, "SYR2", 1
  ); 

  AlphaSet(alpha, pAlpha);
  DistSYR2KMult(N, 1, pAlpha, X, Y, NULL, UpLoA, A);
  TRACE_EXIT;
} /* D_SYR2() */

