/************************************************************************/
/*         D_SYR2: Level 3 Distributed BLAS symmetrix rank-2 update	*/
/************************************************************************/
/* SCCS INFO @(#)D_SYR2K.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 95/03
 */
/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"
#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDCommun.h"  			/* CommSetUp() */

#include "SYR2KMult.h"

void D_SYR2K (UpLo, N, K, alpha, A, B, beta, C)
  char *UpLo; int N; int K;
  FDATA alpha; DistMat A; DistMat B; FDATA beta; DistMat C;
{ char UpLoC = UpperCase(*UpLo); 
  int ParamErr=0;
  AlphaBetaDecln();
  int AlignM=1;

  if (!NoOuterErrCheck) {
    CheckParam(ParamErr, UpLoOK(UpLoC), 0x1);
    CheckParam(ParamErr, Tr(A)==Tr(B), 0x56);
    CheckAlign(ParamErr, AlignM, N, DM_Inc(1-Tr(A), A),   DM_DM(A),
               DM_Inc(1-Tr(C), C),   DM_DM(C), 0x258);
    CheckDim(ParamErr, K, DM_Inc(Tr(A), A), DM_DN(A), 0x35);
    CheckParam(ParamErr, AlignM &&
			 d_P(DM_DM(A))==DM_P(Tr(A), C) &&
			 P_OK(d_P(DM_DN(A)), DM_P(1-Tr(A), C)), 0x58); 
    CheckDim(ParamErr, N, DM_Inc(1-Tr(B), B),   DM_DM(B), 0x26);
    CheckDim(ParamErr, K, DM_Inc(Tr(B), B), DM_DN(B), 0x36);
    CheckParam(ParamErr, Dim_BCompat(DM_DM(B), DM_DM(C)) &&
			 d_P(DM_DM(B))==DM_P(Tr(B), C) &&
			 P_OK(d_P(DM_DN(B)), DM_P(1-Tr(B), C)), 0x68);
    CheckDim(ParamErr, N, DM_Inc(Tr(C), C), DM_DN(C), 0x28);
    CheckParam(ParamErr, Dim_BCompat(DM_D0(C), DM_D1(C)), 0x88);
  }  
    

  SetUpComm(C);  
  ERROR_CHECK_TR(   
    (TraceStr, "(%c,%d,%d,%s,%s,%s,%s,%s)", *UpLo, 
     N, K, pFD2Str(&alpha), DM_TRACE(0, A), DM_TRACE(1, B), pFD2Str(&beta), DM_TRACE(2, C)),
    ParamErr, "SYR2K", 0
  );
  
  AlphaSet(alpha, pAlpha);              BetaSet(beta, pBeta);   
  DistSYR2KMult(N, K, pAlpha, A, B, pBeta, UpLoC, C);
  TRACE_EXIT;
} /* D_SYR2K()*/

