/************************************************************************/
/*         D_SYRK: Level 3 Distributed BLAS symmetric rank-k update	*/
/************************************************************************/
/* SCCS INFO @(#)D_SYRK.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 93/03/23
 */
/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"
#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDCommun.h"  			/* CommSetUp() */

#include "SYRKMult.h"

void D_SYRK (UpLo, N, K, alpha, A, beta, C)
  char *UpLo; int N; int K;  FDATA alpha; DistMat A; FDATA beta; DistMat C;
{ char UpLoC = UpperCase(*UpLo); 
  int ParamErr=0;  			
  AlphaBetaDecln();
  int AlignM=1;

  if (!NoOuterErrCheck) {      
    CheckParam(ParamErr, UpLoOK(UpLoC), 0x1);  
    CheckAlign(ParamErr, AlignM, N, DM_Inc(1-Tr(A), A),   DM_DM(A),
               DM_Inc(1-Tr(C), C),   DM_DM(C), 0x257);
    CheckParam(ParamErr, AlignM && d_P(DM_DM(A))==DM_P(Tr(A), C) && 
    			 P_OK(d_P(DM_DN(A)), DM_P(1-Tr(A), C)), 0x57);
    CheckDim(ParamErr, K, DM_Inc(Tr(A), A), DM_DN(A), 0x35);
    CheckDim(ParamErr, N, DM_Inc(Tr(C), C), DM_DN(C), 0x27);
    CheckParam(ParamErr, Dim_BCompat(DM_D0(C), DM_D1(C)), 0x77);
  }

  SetUpComm(C);
  ERROR_CHECK_TR(
    (TraceStr, "(%c,%d,%d,%s,%s,%s,%s)", 
     *UpLo, N, K, pFD2Str(&alpha), DM_TRACE(0, A), pFD2Str(&beta), DM_TRACE(1, C)), 
    ParamErr, "SYRK", 0
  );

  AlphaSet(alpha, pAlpha);              BetaSet(beta, pBeta);
  DistSYRKMult(N, K, pAlpha, A, pBeta, UpLoC, C);
  TRACE_EXIT;
} /* D_SYRK()*/
