/************************************************************************/
/*         D_TRMM: Level 3 Distributed BLAS triangular mtrix multiply	*/
/************************************************************************/
/* SCCS INFO @(#)D_TRMM.c 1.3         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 93/03/23
 */
/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"
#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDCommun.h"  			/* CommSetUp() */

#include "TRSM.h"
#include "TriMatUpdate.h"

void D_TRMM( UpLo, Diag, M, N, alpha, A, B)
  char *UpLo; char *Diag; int M; int N; FDATA alpha; DistMat A;  DistMat B;
{ char UpLoA = UpperCase(*UpLo), 	DiagA = UpperCase(*Diag); 
  int AlignM=1,				AlignK = CompatBlk(DM_DM(B), DM_DN(A));
  int NoCommA = (DM_Trans(A)==DM_Trans(B)) && (d_P(DM_DN(A))==1);
  int ParamErr=0;  
  AlphaDecln();
  
  CheckAlign(ParamErr, AlignM, M, DM_Inc(1-Tr(A), A),   DM_DM(A),
             DM_Inc(1-Tr(B), B),  DM_DM(B), 0x367);
  if (!NoOuterErrCheck) {  
    CheckParam(ParamErr, UpLoOK(UpLoA), 0x1);  
    CheckParam(ParamErr, DiagOK(DiagA), 0x2);  
    CheckDim(ParamErr, M, DM_Inc(Tr(A), A), DM_DN(A), 0x36);
    CheckDim(ParamErr, N, DM_Inc(Tr(B), B), DM_DN(B), 0x47);
    CheckParam(ParamErr, (AlignM | AlignK) && 
    	       P_OK(d_P(DM_DM(A)), DM_P(Tr(A), B)) &&
               P_OK(d_P(DM_DN(A)), DM_P(1-Tr(A), B)), 0x67);
    CheckDim(ParamErr, M, DM_Inc(Tr(A), A), DM_DN(A), 0x36);
    CheckDim(ParamErr, N, DM_Inc(Tr(B), B), DM_DN(B), 0x47);
  }

  SetUpComm(B);
  ERROR_CHECK_TR(   
    (TraceStr, "(%c,%c,%d,%d,%s,%s,%s)",
    *UpLo, *Diag, M, N, pFD2Str(&alpha), DM_TRACE(0, A), DM_TRACE(1, B)),
    ParamErr, "TRMM", 0
  );

  if (DM_Trans(A)^DM_Trans(B))
    UpLoA = TransForm(UpLoA);
  AlphaSet(alpha, pAlpha); 
  if (ForceAlgVers==2) AlignM = 0;
  else if (ForceAlgVers==1) AlignK = 0;
  if (!AlignM || (!NoCommA && AlignK && 4*N<=M))
    TriMatUpdate(UpLoA, DiagA, M, N, pAlpha, A, B);    
  else
    TR_M(0, UpLoA, DiagA, M, N, pAlpha, A, B);
  TRACE_EXIT;
} /* D_TRMM() */

