/************************************************************************/
/*         D_TRMV: Level 2 Distributed BLAS triangular matrix vector mult*/
/************************************************************************/
/* SCCS INFO @(#)D_TRMV.c 1.1         last modified P. Strazdins, 98/12/24
   first created: P. Strazdins 95/03
 */

/* see D_GEMM.c for common cpp flag description                         */

#include "D_BLAS.h"
#include "D_BLASCheck.h"

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDCommun.h"  			/* CommSetUp() */

#include "TriMatUpdate.h"
#include "TRSM.h"

void D_TRMV (UpLo, Diag, N, A, X)
  char *UpLo; char *Diag; int N;
  DistMat A; DistMat X;
{ char UpLoA = UpperCase(*UpLo), 	DiagA = UpperCase(*Diag);
  int ParamErr=0;
  int AlignN=1;				
  int AlignM = CompatBlk(DM_DM(X), DM_DM(A));

  CheckAlign(ParamErr, AlignN, N, DM_Inc(Tr(A), A),   DM_DN(A),
             DM_Inc(1-Tr(X), X),  DM_DM(X), 0x345);
  if (!NoOuterErrCheck) {
    CheckParam(ParamErr, UpLoOK(UpLoA), 0x1);
    CheckParam(ParamErr, DiagOK(DiagA), 0x2);
    CheckDim(ParamErr, N, DM_Inc(1-Tr(A), A), DM_DM(A), 0x34);
    CheckDim(ParamErr, 1, DM_Inc(Tr(X), X), DM_DN(X), 0x5);
    CheckParam(ParamErr, (AlignN|AlignM) && d_P(DM_DM(A))==DM_P(Tr(A), X) &&
               		 DM_P(1-Tr(A), X)==d_P(DM_DN(A)), 0x45);
  }

  SetUpComm(A);
  ERROR_CHECK_TR(  
    (TraceStr, "(%c,%c,%d,%s,%s)", *UpLo, *Diag, 
     N, DM_TRACE(0, A), DM_TRACE(1,X)),
     ParamErr, "TRMV", 1
  );    

  if (DM_Trans(A)^DM_Trans(X))
    UpLoA = TransForm(UpLoA);
  if (!AlignN)
    TR_M(0, UpLoA, DiagA, N, 1, &FD_1, A, X);        
  else  
    TriMatUpdate(UpLoA, DiagA, N, 1, &FD_1, A, X);
  TRACE_EXIT;
} /* D_TRMV() */
