/************************************************************************/
/*	DistMat: distributed block-cyclic matrices definition		*/
/************************************************************************/
/* SCCS INFO @(#)DistMat.c 1.1         last modified P. Strazdins, 98/12/23
   first created: P. Strazdins 94/05/17
*/

#include "DistMat.h"
#include "MatDim.h"
#include "FDAssert.h"
#include "FDMalloc.h"

DistMat InitDimMat_(A, Trans, B_0, B_1, P0_0, P_0, P0_1, P_1, di, dj)
  DistMat A; int Trans, B_0, B_1, P0_0, P_0, P0_1, P_1, di, dj;
{ FDATA *Ax;
  pInitDim(DM_D0(A), 0, B_0, P0_0, P_0);
  pSubDim_(DM_D0(A), DM_D0(A), di);
  pInitDim(DM_D1(A), 0, B_1, P0_1, P_1);
  pSubDim_(DM_D1(A), DM_D1(A), dj);
  DM_SetTrans(Trans, A);
  return (A);
} /* InitDimMat_() */


DistMat SubMat_(Aij, A, i, j)
  DistMat Aij, A; int i, j;
{ if (DM_Trans(A)) 
    SWAP(int, i, j);    
  if (Aij != A) {
    *DM_Ptr(Aij) = *DM_Ptr(A);
    DM_SetTrans(DM_Trans(A), Aij);
  }
#define Ax LM_ADR(DM_Loc(Aij))
  if (i!=0) {
      if (i < 0)	
        i = G2LNSubDim(i, DM_D0(Aij));
      else
        i = G2LSubDim(i, DM_D0(Aij));        
      if (Ax!=NULL) {
        Ax += FAST_MUL1(LM_Inc(0, DM_Loc(Aij)), i);    
      }                      
    }  /* if (i!=0) */ 
  if (j!=0) {
      if (j < 0)	
        j = G2LNSubDim(j, DM_D1(Aij));
      else
        j = G2LSubDim(j, DM_D1(Aij));        
      if (Ax!=NULL) {
        Ax += FAST_MUL1(LM_Inc(1, DM_Loc(Aij)),  j);    
      }                      
    }  /* if (i!=0) */ 
  return (Aij);
} /* SubMat_() */

#define KEEP_BLK_RED

DistMat ReduceDimMat_( VA, RowCol, A )
  int RowCol; DistMat VA, A;
{ 
  DM_SetTrans(DM_Trans(A), VA);
  DM_Loc(VA) = DM_Loc(A);
  if (RowCol==-1) {
#ifdef KEEP_BLK_RED
    TransDim(DM_D(0, VA), DM_D(0, A), &UnitDim);
    TransDim(DM_D(1, VA), DM_D(1, A), &UnitDim);
#else
    MatDimUnit(DM_D0(VA));
    MatDimUnit(DM_D1(VA)); 
#endif    
    return(VA);
  }   	
  RowCol = RowCol^DM_Trans(A);
#ifdef KEEP_BLK_RED
  TransDim(DM_D(RowCol, VA), DM_D(RowCol, A), &UnitDim);
#else  
  MatDimUnit(DM_D(RowCol, VA));
#endif
  MatDimCp(DM_D(1-RowCol, VA), DM_D(1-RowCol, A));
  return (VA); 
} /* ReduceDimMat_() */


DistMat Spread2DistMat_(B, RowCol, Trans, DN, sB)
  DistMat B; int  RowCol, Trans; pMatDim DN; pLocMat sB;
{ DM_SetTrans(Trans, B);
  DM_Loc(B) = *sB;  
  MatDimUnit(DM_D(1-RowCol, B));
  MatDimCp(DM_D(RowCol, B), DN);
  return(B);
} /*Spread2DistMat_()*/
 

DistMat NewMat_(NewC, M, N, C, WS)
  int M, N; DistMat NewC, C; FDATA *WS;
{ int m = G2Ld(M, DM_DM(C)),    n = G2Ld(N, DM_DN(C));
  LocMat lC; 
  if (n==1) n=2;			/* avoid degenerate row-major strides */
  if (WS==NULL)
    MatMalloc("NewMat", 1, &WS, m*n);
  if (n==0) 
    n = 2;
  if (DM_Trans(C))  
    LM_Init(DM_Loc(NewC), WS, 1, n);
  else
    LM_Init(DM_Loc(NewC), WS, n, 1);
  MatDimCp(DM_D0(NewC), DM_D0(C));
  MatDimCp(DM_D1(NewC), DM_D1(C));
  DM_SetTrans(DM_Trans(C), NewC);
  return(NewC);
} /*NewMat_()*/


char *PrintDistMat( A, pA )
  DistMat A; char *pA; 
{ 
  sprintf(pA, DM_SPRINTF_FORMAT, DM_SPRINTF_PARAMS(A)); 
  return (pA);
} /* PrintDistMat() */ 
