/************************************************************************/
/*	DistMatrix: distributed matrix procedures used by DBLAS		*/
/************************************************************************/
/* SCCS INFO @(#)DistMatrix.c 1.4         last modified P. Strazdins, 99/07/06
   first created: P. Strazdins 93/15/01
 */
/* set of higher-level distributed matrix routines used by main DBLAS
 * parallel driver routines.   
 * see FDCommun.h for description of matrix notations.
 * Special compile switches:
 *	TEST_TREE: 	test tree based GatherMat()/MultSumMat()
 */
/*#define TEST_TREE*/

#include "DistMatrix.h"
#include "FDAssert.h"

#include "MatDim.h"
#include "MachineParams.h"
#include "FDCommun.h"
#include "CellMatrix.h"
#include "CellBlkMatrix.h"

#ifdef VPPLIB_STRIDE
#include "VPPStride.h"
#endif

#define OPTM(N)	( (N)==0? 1: (OPT_MSG_SIZE+(N)-1)/(N) )

#define REDUCE(m, B0, B, P)  /*try to reduce commun startup by dec m by<= 33%*/\
        if ((m)>1 && (P)>1) {						\
	  int m0 = SubMatB0(m, B0, B);					\
      	  if (3*m0 <= (m)) (m) -= m0;	/* end on block boundary */	\
	}

void RotMat(RowCol, dP0, m, D, n, A)
  int RowCol, dP0, m, n; pMatDim D; FDATA A[]; 
{ int P = GetNumCells(RowCol);		int m_opt = m<=1? m: OPTM(n);
  int B0 = BlkB0d(0, D);
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%d[%d %d %d]x%d,a%d)", RowCol, dP0,
     m, B0, d_B(D), d_P(D), n, FD_ADR(A)),
    RCOK(RowCol) && Dim_OK(D), 0, "RotMat", 3, TRACE_COMMUN
  );

  while (dP0 > P/2)
    dP0 -= P;
  while (-dP0 > P/2)
    dP0 += P;     
  if (dP0 == 0 || P==1 || n==0) {
    TRACE_RETURN;
  }  
  while (m > 0) { 		
#if defined(GLOB_BLK_SEND)
    LocMat lA, lAs, lAr; int sz;
#endif    
    int m1 = MIN(m, m_opt); 
    REDUCE(m1, B0, d_B(D), d_P(D));
#if defined(GLOB_BLK_SEND)
    sz = m1*n;
    LM_Init(lAs, NULL, n, 1);
    MatMalloc_F("RotMat", 1, &LM_ADR(lAs), 2*sz);
    LM_Init(lAr, LM_ADR(lAs)+sz, n, 1);    
    BlkMatStridePack(0, m1, n, pMkLocMat(&lA, A, n, 1), NULL, &lAs, 
    		     B0, d_B(D), d_P(D), 0, 1, 1);
    SendWaitMark();
    SendMat(RowCol==0? dP0: 0, RowCol==1? dP0: 0, m1, n, 
    	    LM_ADR(lAs), LM_RInc(lAs), ROTMSGTYPE);
    RecvMat(RowCol==0? -dP0: 0, RowCol==1? -dP0: 0, m1, n, 
    	    LM_ADR(lAr), LM_RInc(lAr), ROTMSGTYPE);
    BlkMatStrideUnpack(0, m1, n, &lAr, NULL, &lA, 
    		      B0, d_B(D), d_P(D), 0, 1, 1);    	    
    CompleteSends(-1);
    MatFree(LM_ADR(lAs));
#else
    SendBlkMat(RowCol==0? dP0: 0, RowCol==1? dP0: 0, 
  	       m1, B0, d_B(D), d_P(D), n, A, ROTMSGTYPE); 
    RecvBlkMat(RowCol==0? -dP0: 0, RowCol==1? -dP0: 0, 0, 
  	      m1, B0, d_B(D), d_P(D), n, NULL, A, ROTMSGTYPE); 
#endif     
    m -= m1;
    if (m > 0) {
      A += L2Gd_(m1, D, B0, 0)*n;   	B0 = SubMatB0(m1, B0, d_B(D));
    }  
  } /* while (m...) */
  TRACE_EXIT;  
} /* RotMat() */

#ifdef VPPLIB_STRIDE

#define RingComb StrideRingComb

#else 
static void RingComb(RowCol, Add, D, M, n, A)
  int RowCol, Add, M, n;  pMatDim D; FDATA A[];
/* pre:  RCOK(RowCol); Dim_OK(D); Inc(A) = {n,1}
 *	 (G2L(M, d_B0(D), d_B(D), j, d_P(D)) <= OPTM(n), for 0<=j<d_P(D)) 
 *	 || M <= d_B(D)*d_P(D)-d_B0(D); d_P(D) > 1 
 *	 all (d_P(D)) cells ins cope of RowCol are participating in this call; 
 *	 M, n and D, are consistent across all these cells; 
 * post: if Add, in cell (I+i)%P, the ith row block of A contains the sum of 
 *	 that in all cells in the scope, 0 <= i <= NBLKS(M, B0, B)
 *       else, the ith row block of A the value of that in cell (I+i)%P
 * notes:partitioning is done only for when M <= d_B(D)*d_P(D)-d_B0(D);
 */
{ int ms=0; 				int m_opt = Blk1d(M, D)? M: OPTM(n);
  FDATA *W = NULL;			
  LocMat lAj;				int i = 0;
#if !defined(GLOB_BLK_SEND)
  LocMat lW;
# define lW_ lW
#else
  LocMat lWs[2];
# define lW lWs[i%2]
# define lW_ lWs[(i+1)%2]
#endif    
  
  COMMON_ASSERT_TR(
    (TraceStr, "(%d,%d,%d %d:%d %d,%dx%d,a%d)", 
     RowCol, Add, d_B0(D), d_B(D), d_P0(D), d_P(D), M, n, FD_ADR(A)),
    RCOK(RowCol) && Dim_OK(D) && 
    		 ((G2Ld_P0(M, D, 0) <= m_opt && G2Ld_P0(M, D, 1) <= m_opt) 
   		 || M <= d_B(D)*d_P(D)-d_B0(D)), 0, "RingComb", 4, TRACE_COMMUN
  );
  
  if (!FAST_STRIDE_MSG && NBLKSd(M, D)>d_P(D)) {
    int m_mx = G2Ld_Mx(M, D);		m_mx = MIN(m_mx, m_opt);
#if !defined(GLOB_BLK_SEND)
    MatMalloc("RingComb", 0, &W, m_mx*n);	
#else
    MatMalloc("RingComb", 0, &W, 2*m_mx*n);	
    LM_Init(lW_, W+m_mx*n, n, 1);    
#endif    
    LM_Init(lW, W, n, 1);		LM_Init(lAj, NULL, n, 1);
  }

  while (ms < d_B(D)) { 
    int j = Add? CYCLIC_DEC(d_P0(D), d_P(D)): d_P0(D), 	
        B0j = BlkB0(j, d_B0(D)),
        m = G2Ld_P0(M, D, j) - ms;   
    FDATA *Aj = A+(BlkIxd_(j, D)+ms)*n;	LM_ADR(lAj) = Aj;	
    m = MIN(m, m_opt);
    i = 1;    				/* ensure this lW is the same as is sent at i=1*/
    if (W != NULL) 
      BlkMatStridePack(0, m, n, &lAj, NULL, &lW, B0j, d_B(D), d_P(D), 0, 1, 1);
    for (i=1; i < d_P(D); i++) {
      SendWaitMark();			/* ensure send does not block */
      if (m > 0) 
        if (W==NULL)
      	  SendBlkMat(1-RowCol, RowCol, m, B0j, d_B(D), d_P(D), n, Aj, SUMMSGTYPE);
      	else {
      	  SendMat(1-RowCol, RowCol, m, n, LM_ADR(lW), n, SUMMSGTYPE); 
      	  if (!Add && i>1) 
      	    BlkMatStrideUnpack(0, m, n, &lW, NULL, &lAj, B0j, d_B(D), d_P(D), 
      	    		       0, 1, 1);
      	}  
      j = CYCLIC_DEC(j, d_P(D));	B0j = BlkB0(j, d_B0(D));
      Aj = A+(BlkIxd_(j, D)+ms)*n;	LM_ADR(lAj) = Aj;
      m = G2Ld_P0(M, D, j) - ms;  	m = MIN(m, m_opt);     
      if (m > 0) 
        if (W==NULL)
          RecvBlkMat(RowCol-1, -RowCol, Add, m, B0j, d_B(D), d_P(D), 
        	     n, NULL, Aj, SUMMSGTYPE);
        else {
          RecvMat(RowCol-1, -RowCol, m, n, LM_ADR(lW_), n, SUMMSGTYPE);
          if (Add && i < d_P(D)-1)
            BlkMatStridePack(1, m, n, &lAj, NULL, &lW_, B0j, d_B(D), d_P(D), 
            		     0, 1, 1);    
        }	     
      CompleteSends(-1);    		/* ensure send of lW completed as
      					   receive of next itn will clobber it*/
    } /* for (i...) */
    i = d_P(D) - 1;			/* ensure this lW_ is the same as on last itn */
    if (W!=NULL)    
      BlkMatStrideUnpack(Add, m, n, &lW_, NULL, &lAj, B0j, d_B(D), d_P(D), 0, 1,1);
    ms += m_opt;
  } /* while (ms ...) */     
  if (W!=NULL)
    MatFree(W);
#undef lW
#undef lW_    
} /* RingComb() */
#endif /* !defined(VPPLIB_STRIDE) */

void GatherMat(RowCol, D, M, n, A)
  int RowCol, M, n; pMatDim D; pLocMat A; 
{ int nB = Blk1d(M, D)? 1: MAX(1, OPT_MSG_SIZE/(n*d_B(D)));
					/* max # of blocks in each b/c */
  int MB = NBLKSd(M, D);
  int Ms=0;				MatDim newD;
  int m_opt = M<=1? M: OPTM(n);
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d %d %d %d,%dx%d,a%d %d %d)", RowCol, d_B0(D), d_B(D), 
    d_P0(D), d_P(D), M, n, FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A)),
    RCOK(RowCol) && Dim_OK(D) &&
    pLM_CInc(A)==1 && (pLM_CInc(A)==1 || (d_P0(D)==0 && Blk1d(M,D))), 
    0, "GatherMat", 3, TRACE_COMMUN
  );
  if (d_P(D)==1 || M==0 || n==0) {
    TRACE_RETURN;
  }  

#ifndef TEST_TREE
  if (MB > d_P(D)/2) {
    FDATA * Ax = pLM_ADR(A);		
    while (M > 0) {
      int Mp = MIN(M, nB*d_B(D)*d_P(D)-d_B0(D));
      RingComb(RowCol, 0, D, Mp, pLM_RInc(A), Ax);/* nb. pLM_RInc(A) >= n*/
      M -= Mp;
      if (M > 0) {
        Ax += Mp * pLM_RInc(A);  		D = pSubDim_(&newD, D, Mp);
      }  
    } /* while (M...) */ 
    TRACE_RETURN;        
  }
#endif

  while (M > 0) {			/* partition over max # complete blks */
    int Mp = Blk1d(M, D)? M: MIN(M, nB*d_B(D)*d_P(D)-d_B0(D));
    int ms = 0;
    while (ms < MIN(Mp, d_B(D))) {	/* partition only if nB=1 & n*B>m_opt */
      int i, 		 		p0 = d_P0(D);
      for (i=0; i < d_P(D); i++) {
      	int m = G2Ld_P0(Mp, D, i) - ms;	m = MIN(m, m_opt);
        if (m<=0 && i>0) 		/* no more to b/c, as max m for i=0,1 */
          break; 
        if (m > 0)
          BrdBlkMat(RowCol, p0, m, BlkB0(i, d_B0(D))+ms, d_B(D), d_P(D), n, 
        	    pLM_Index(A, BlkIxd_(i, D)+ms+Ms, 0), 
        	    pLM_RInc(A)); 
        p0 = CYCLIC_DEC(p0, d_P(D));
      } /* for (i...) */
      ms += m_opt;
      if (ms<Mp && Blk1d(Mp, D)) 	/* single block needing multiple sends*/
        BarrierSync(RowCol);		/* wait for recv. cells to complete */
    } /* while (ms...) */
    M -= Mp;				Ms += Mp;
    if (M > 0) {
      D = pSubDim_(&newD, D, Mp);
    }  
  } /* while (M...) */
  TRACE_EXIT;  
} /* GatherMat() */


void MultSumMat(RowCol, D, M, n, A)
  int RowCol, M, n; pMatDim D; pLocMat A; 
{ int MB = NBLKSd(M, D);		MatDim newD;
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d %d %d %d,%dx%d,a%d %d %d)", RowCol, d_B0(D), d_B(D),
     d_P0(D), d_P(D), M, n, FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A)),
    RCOK(RowCol) && Dim_OK(D) && 
    ((pLM_RInc(A)>=n && pLM_CInc(A)==1) || (d_P0(D)==0 && MB==1)), 
    0, "MultSumMat", 3, TRACE_COMMUN
  );
  if (d_P(D)==1 || M==0 || n==0) {
    TRACE_RETURN;
  }

  if (M==1) {				/* optimize important special case */
    SumMat(RowCol, d_P0(D), d_P(D), M, n, A);	  
#ifdef TEST_TREE				/* force this version to test SumMat() perf*/
  } else if (1) {			/* only valid for B=1, if MB > d_P(D) */
#else
  } else if (MB <= d_P(D)/2) {
#endif  	
    int ms = 0;				int m_opt = OPTM(n);
    FDATA * Ax = pLM_ADR(A);		/* used to prevent side effect */
    while (ms < MIN(M, d_B(D))) {
      int j;				int B0 = d_B0(D);
      if (ms > 0)
        pLocSubMat(A, ms, 0); 
      for (j=0; j < MIN(MB, d_P(D)); j++) {/* sum cols of A via binary tree alg */
        int m = G2Ld_P0(M, D, j) - ms;
        m = MIN(m, m_opt);
#ifdef TEST_TREE
        pLM_RInc(A) *= d_P(D);        
#endif
        SumMat(RowCol, CYCLIC_SUB(d_P0(D), j, d_P(D)), d_P(D), m, n, A);
#ifdef TEST_TREE
        pLM_RInc(A) /= d_P(D);        
#endif  
        pLocSubMat(A, d_B(D)-B0, 0);	B0 = 0;       
      } /* for (j=...) */
      pLM_ADR(A) = Ax;			/* restore A */
      ms += m_opt;		
      if (MB==1 && ms < M)
        BarrierSync(RowCol);		/* wait for recv. cells to complete */
    } /* while (ms...) */
  } else {
    int nB = MAX(1, OPT_MSG_SIZE/(n*d_B(D)));	/* max # of blocks in each sum*/
    FDATA * Ax = pLM_ADR(A);		
    while (M > 0) {
      int Mp = MIN(M, nB*d_B(D)*d_P(D)-d_B0(D));
      RingComb(RowCol, 1, D, Mp, pLM_RInc(A), Ax);/* nb. pLM_RInc(A) >= n*/
      M -= Mp;
      if (M > 0) {
        Ax += Mp * pLM_RInc(A);  		D = pSubDim_(&newD, D, Mp);
      }  
    } /* while (M...) */ 
  }
  TRACE_EXIT;  
} /* MultSumMat() */


static int lQ = 1, lP = 1, lG = 1, lP_G = 1, lQ_G = 1; 
/* records values for last call SetGCDvals() */
			
static int SetGCDvals(P, Q, P_G, Q_G) 
  int P, Q, *P_G, *Q_G;
/* pre: P, Q >= 0;
 * post: returns GCD(P, Q), *P_G = P / G; *Q_G = P / G
 * notes: enables fast calculations of these quantities, taking into account
 * 	  special cases and values calculated in previous calls
 */
{ if (P!=lP || Q!=lQ) {
    lP = P;			lQ = Q;
    lG = GCD(P, Q);
    lP_G = lG==P? 1: FAST_DVD(P, lG);
    lQ_G = lG==Q? 1: FAST_DVD(Q, lG);
  }  	
  *P_G = lP_G; 			*Q_G = lQ_G;
  return (lG);
} /* SetGCDvals() */


void TransMat(RowCol, Add, M, N, A, DM, DN, MBIncA, alpha, B, NP0, MQ0, MBIncB)
  int RowCol, Add, M, N, MBIncA, MBIncB, NP0, MQ0;
  pLocMat A, B; pMatDim DM, DN; FDATA *alpha; 
{ int P = d_P(DM), Q = d_P(DN), P_G, Q_G;
  int G = SetGCDvals(P, Q, &P_G, &Q_G);
  int N_part;					
  MatDim newD; LocMat A_, B_;		/* used to prevent side effects */

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%dx%d,a%d %d %d (%d,%d) %d,%s,a%d %d %d (%d,%d) %d)",
     RowCol, Add, M, N, FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A), d_P0(DM), d_P0(DN),
     MBIncA, pFD2Str(alpha), FD_ADR(pLM_ADR(B)), pLM_RInc(B), pLM_CInc(B), NP0, MQ0, MBIncB),
    RCOK(RowCol) && 0<=NP0 && NP0<P && 0<=MQ0 && MQ0<=Q && M>=0 && N>=0 
    && MBIncB>0 && MBIncA>0, 0, "TransMat", 3, TRACE_COMMUN
  );
  
  if (N==1)
    N_part=1;
  else {  
    N_part = L2Glg(OPT_MSG_SIZE/G2Ld2_(M, DM, DN, 0, 0), 
    		   0, d_B(DN), 0, d_P(DM), d_lgB(DN), d_lgP(DM));
    N_part = MAX(N_part, 1);    	N_part=MIN(N_part, N);
  }  
 
 if (P_G==1 && Q_G==1 && N_part>=N) {
    /* optimize s/w overheads for an important special case (<=1 message/cell)*/
    int n = G2Ld(N, DN),		DMB0 = d_P0(DM)==0? d_B0(DM): 0; 
    int m = G2Ld(M, DM);
    SendWaitMark();			/* ensure sends are non-blocking */
    if (m>0 && n>0) {
      int P0 = d_P0(DN) - NP0,		Q0 = d_P0(DM) - MQ0;
      if (P0!=0 || Q0!=0)
        Send2BlkMat(RowCol? Q0: P0, RowCol? P0: Q0, m, n, alpha, A,
      		    DMB0, d_B(DM), MBIncA, 0, 1, 1);
      else		     
        BlkMatCopy(Add, m, n, A, alpha, B, DMB0, d_B(DM), MBIncA, MBIncB,
        	   0, 1, 1, 1);
    }
    n = G2Ld_P0(N, DN, NP0);		DMB0 = MQ0==0? d_B0(DM): 0;
    m = G2Ld_P0(M, DM, MQ0);		/* nb. d_P(DM)==P==Q==d_P(DN)  here */ 
    if (m>0 && n>0) {
      int P0 = MQ0 - d_P0(DM),   	Q0 = NP0 - d_P0(DN);
      if (P0!=0 || Q0!=0) {
         Recv2BlkMat(RowCol? Q0: P0, RowCol? P0: Q0, Add, m, n, B, 
      		     DMB0, d_B(DM), MBIncB, 0, 1, 1); 
      }		        	        	
    }   
    CompleteSends(-1);                    /* wait for sends to complete */
    TRACE_RETURN;	
  } 
  
  /* below, strategy is to optimize index calculations for P_G==1, Q_G==1 */
  while (N > 0) {			/*partition across M-dimension*/
    int dN = MIN(N, N_part);
    int i, j, m, n, DMB0, DNB0;
    LocMat Ai, Bi;
    int ir=-1, jr=-1;	/* vals of i,j for recv to self; can be proven unique */
    LocMat Aijs;			/* send misxnjs from Aijs to self */
    int mis, njs;
   
    m = G2Ld(M, DM); 			n = G2Ld(dN, DN); 
    DMB0 = d_P0(DM)==0? d_B0(DM): 0;	DNB0 = d_P0(DN)==0? d_B0(DN): 0; 
    LM_Copy(0, Ai, *A);
    SendWaitMark();			/* ensure sends are non-blocking */
    for (i=0; i < Q_G; i++) {
      int mi = G2LdB(m, DM, Q_G, DMB0, i);
      int DMB0i = i==0? DMB0: 0;
      LocMat Aij;			
      if (mi == 0) break;      		LM_Copy(0, Aij, Ai);
      for (j=0; j < P_G; j++) {      	
        /* local blk (i,j) = global (p,q) = (i*P+d_P0(DM), j*Q+d_P0(DN)) sent
         * to local blk (q/P,p/Q) of relative cell (P0,Q0) = (q%P-NP0, p%Q-MQ0)
         * notes: if Q_G==1, d_P0(DN)+j*Q < P, as d_P0(DN) < Q && j < P/Q; 
         *	  also i==0. Similarly for P_G==1. This motivates the use
         *	  of the macros below to reduce integer operation overhead.
         */
        int P0 = FAST_MODLT(d_P0(DN) + FAST_MULZ(j, Q), P) - NP0,	
            Q0 = FAST_MODLT(d_P0(DM) + FAST_MULZ(i, P), Q) - MQ0;
        int DNB0j = j==0? DNB0: 0;
        int nj = G2LdB(n, DN, P_G, DNB0, j);	
        if (nj == 0) break;
        if (P0!=0 || Q0!=0)
          Send2BlkMat(RowCol? Q0: P0, RowCol? P0: Q0, mi, nj, alpha, &Aij, 
      		      DMB0i, d_B(DM), FAST_MUL1(Q_G, MBIncA), DNB0j, d_B(DN), 
      		      P_G); 
        else {
          ir = FAST_DIVLT(d_P0(DM) + FAST_MULZ(i, P), Q), 		
          jr = FAST_DIVLT(d_P0(DN) + FAST_MULZ(j, Q), P);
      	  mis = mi; 			njs = nj;
      	  Aijs = Aij;			
        }			  
        if (j < P_G-1)
          pLocSubMat(&Aij, 0, d_B(DN)-DNB0j);
      } /* for (j...) */
      if (i < Q_G-1)
        pLocSubMat(&Ai, FAST_MUL1(MBIncA, d_B(DM))-DMB0i, 0);	
    } /* for (i...) */    

    if (N-dN>0)
      A = pLocSubMat_(&A_, A, 0, n);

    DMB0 = MQ0==0? d_B0(DM): 0;	 	DNB0 = NP0==0? d_B0(DN): 0; 
    if (ir != -1) { /* performe a send to self (optimize by memory copy) */
      LocMat Bijr;
      pLocSubMat_(&Bijr, B, BlkIx(ir,DMB0,d_B(DM)*MBIncB), 
      			    BlkIx(jr,DNB0,d_B(DN)));
      BlkMatCopy(Add, mis, njs, &Aijs, alpha, &Bijr, 
                 ir>0? 0: DMB0, d_B(DM), FAST_MUL1(MBIncA, Q_G), 
                 FAST_MUL1(MBIncB, P_G), jr>0? 0: DNB0, d_B(DN), P_G, Q_G);
    }

    m = G2Ld2_(M,  DM, DN, d_B0(DM), MQ0);	
    n = G2Ld2_(dN, DN, DM, d_B0(DN), NP0);
    
    LM_Copy(0, Bi, *B);  
    for (i=0; i < P_G; i++) {
      int mi = G2LdB(m, DM, P_G, DMB0, i);	
      int DMB0i = i==0? DMB0: 0;		
      LocMat Bij;			
      if (mi == 0) break;		LM_Copy(0, Bij, Bi);
      for (j=0; j < Q_G; j++) {    
        int P0 = FAST_MODLT(MQ0 + FAST_MULZ(i, Q), P) - d_P0(DM),	
            Q0 = FAST_MODLT(NP0 + FAST_MULZ(j, P), Q) - d_P0(DN); 
        int DNB0j = j==0? DNB0: 0;
        int nj = G2LdB(n, DN, Q_G, DNB0, j);	
        if (nj == 0) break;
        if (P0!=0 || Q0!=0)
          Recv2BlkMat(RowCol? Q0: P0, RowCol? P0: Q0, Add, mi, nj, &Bij, 
      		      DMB0i, d_B(DM), FAST_MUL1(P_G, MBIncB), DNB0j, d_B(DN), Q_G);
        else {
          COMMON_ASSERT_TR(
             (TraceStr, "self recv: exp %dx%d @b(%d,%d), act %dx%d @b(%d,%d)\n",
	      mis, njs, ir, jr, mi, nj, i, j),    
             ir==i && jr==j && njs==nj && mis==mi, 0, "TransMat", 6, TRACE_COMMUN
          );    
        }
        if (j < Q_G-1)
          pLocSubMat(&Bij, 0, d_B(DN)-DNB0j);
      } /* for (j...) */
      if (i < P_G-1)
        pLocSubMat(&Bi, FAST_MUL1(MBIncB, d_B(DM))-DMB0i, 0);	
    } /* for (i...) */    
    CompleteSends(-1);			/* wait for sends to complete */
    /* notes: on GLOB_BLK_SEND, if the send require packing, they must
     * be in Send2BlkMat() must ensure each send is from an individual buffer.
     * if no packing required (unlikely), A and B must be separate
     * buffers to ensure correctness.
     */      					        
    N -= dN; 	
    if (N > 0) {
      NP0 = SubMatP0(dN, d_B0(DN), d_B(DN), NP0, P);	
      DN = pSubDim_(&newD, DN, dN);
      B = pLocSubMat_(&B_, B, 0, n);
    }
  } /* while (N...) */  
  TRACE_EXIT;
} /* TransMat() */


static int SolveLC(x, y, c)
 int x, y, c;
/* pre:  GCD(x,y)=1; x,y>0; -x<c<y
 * post: returns a where  ax = c modulo y and 0<=a<y
 * note: most often, MIN(x, y) = 1  
 */
{ int a=0; 				int ax = 0;
  if (y==1 || c==0)			/* optimizations */
    return(0);
  if (x==1)
    return (c);
  while (c < 0) {			/* ensure 0 <= c < y */
    c += y;
  }  	  
  while (ax % y != c) { 		/* general algorithm */
    a++;
    ax += x;
  }  
  return (a);
} /* SolveLC */   


void TransReSpreadMat(RowCol, K, N, A, DN, alpha, DQ, sA, NoCopysA)
  int RowCol, K, N, NoCopysA;  pMatDim DN, DQ;
  pLocMat A, sA; FDATA *alpha;
{ int Q = d_P(DQ),			Q0 = d_P0(DQ);
  int P = d_P(DN), P_G, Q_G, 		G = SetGCDvals(P, Q, &P_G, &Q_G);
  int n = G2Ld2(N, DN, DQ),       	NB0 = Q0==0? d_B0(DN): 0;
  int ni, NB0i;				/* b/c A[:,0..ni-1: NB0i:d_B(DN):Q_G]*/
  LocMat Ap; 				FDATA *W = NULL;
  int k, K_part;  			/* unit of partitioning b/c in K dim */
  int CopyAp = 0;			/* whether pack(A) needed before send */
  LocMat A_;				/* prevent side effects on A */

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%dx%d,a%d %d %d (%d),%s,%d %d,a%d %d %d)", 
     RowCol, K, N, FD_ADR(pLM_ADR(A)), pLM_RInc(A), pLM_CInc(A), 
     d_P0(DN), pFD2Str(alpha), Q0, Q, FD_ADR(pLM_ADR(sA)), pLM_RInc(sA), pLM_CInc(sA)),
    RCOK(RowCol) && 0<=Q0 && Q0<=Q && K>=0 && N>=0 && 
    (P==1 || pLM_Inc(1,sA)==1) &&
    (!NoCopysA || ((Blk1(N, d_B0(DN), d_B(DN)) || P==Q) && d_P0(DN)==Q0%P) && 
     		  (P==1 || LM_Contiguous(0, K, n, *A))),
    0, "TransReSpreadMat", 3, TRACE_ALL
  );
  
  if (n==0) {				/* note: n is same across all P cells */
    TRACE_RETURN;
  }  	 

  if (K==1) 
    K_part = 1;
  else {  
    K_part = OPTM(G2LdB(N, DN, P*Q_G, 0, 0));
    K_part = MAX(K_part, 1);		K_part = MIN(K, K_part);
  }
  
  if (d_P0(DN)==Q0 || (d_P0(DN) + Q - Q0)%G == 0) { 	
    /* will b/c from this cell :
     * local blk i corresp. global blk i*P+d_P0(DN) = j*Q+Q0 for some j */
    int i = d_P0(DN)==Q0? 0: SolveLC(P_G, Q_G, (Q0 - d_P0(DN))/G);
    if (i > 0)		
      A = pLocSubMat_(&A_, A, 0, BlkIxd(i, DN));
    COMMON_ASSERT(!NoCopysA || LM_Eq(*A, *sA), 0, 
    		  "TransReSpreadMat: cpy", 6, TRACE_COMPUTE);
    if (Q_G==1)				/* optimization; assert i=0 */
      ni = P_G==1? n: G2Ld(N, DN);
    else  	  
      ni = G2LdB(N, DN, P*Q_G, d_B0(DN), i*d_P(DN)+d_P0(DN));	
    NB0i = BlkB0d(i, DN);
    if ((Q==G || Blk1(N, d_B0(DN), d_B(DN))) && LM_Contiguous(0, K, ni, *A)) { 
      /*send from A direct*/
      LM_Copy(0, Ap, *A);		CopyAp=0;
    } else if ((P==G || Blk1(N, d_B0(DN), Q*d_B(DN))) && LM_Inc(1, *sA)==1) {	
      /* sole sender; can send from sA after initialized. assert ni=n, i=0 */  
      LM_Copy(0, Ap, *sA);		CopyAp=1-NoCopysA;
      NoCopysA=1;  
    } else {
      MatMalloc_F("TransReSpreadMat", 1, &W, FAST_MUL2(ni, K_part));  	
      LM_Init(Ap, W, ni, 1);		CopyAp=1;
    }
  } 
 
  for (k=0; k<K; k+=K_part) {
    int j, 				
        P0 = CYCLIC_SUB(d_P0(DN), FAST_MODLT(Q0, P), P);	
    int dK = MIN(K_part, K-k);		int NB0j = NB0;
    
    if (CopyAp)  	  
      BlkMatStridePack(0, dK, ni, A, alpha, &Ap, 0, 1, 1, NB0i, d_B(DN), Q_G);
    if (P==G && k>0)		
      BarrierSync(RowCol);
      
    for (j=0; j<P_G; j++) {
      int nj = G2LdB(n, DN, P_G, NB0, j);
      LocMat sA_;
      pLocMat sAr = (k+j==0)? sA: 
      			      pLocSubMat_(&sA_, sA, k, BlkIx(j, NB0, d_B(DN))); 
      if (nj==0) break;     		
      if (P0==0) {		
        COMMON_ASSERT(ni==nj && (P==1 || dK==1 || LM_Ld(Ap)==nj), 0, 
        	      "TransReSpreadMat: brd", 6, TRACE_COMMUN);
        if (P>1)
          BrdMat(RowCol, P0, dK, nj, LM_ADR(Ap), LM_Ld(Ap));
        if (!NoCopysA)
          BlkMatStrideUnpack(0, dK, nj, &Ap, NULL, sAr, 0, 1, 1, 
          		     NB0j, d_B(DN), P_G);
      } else
        if (P_G==1)			/* can reduce s/w overheads */
          BrdMat(RowCol, P0, dK, nj, pLM_ADR(sAr), pLM_RInc(sAr));
        else
          RecvBrd2BlkMat(RowCol, P0, dK, nj, sAr, 0, 1, 1, NB0j, d_B(DN), P_G);
      P0 = CYCLIC_SUB(P0, FAST_MODEQ(Q, P), P);			NB0j = 0;
    } /* for (j...) */
    if (k + dK < K) {
      if (W == NULL)			/* Ap points within A or sA */
        pLocSubMat(&Ap, dK, 0);
      A = pLocSubMat_(&A_, A, dK, 0); 
    }   	
  } /* for (k...) */   

  if (W != NULL)
    MatFree(W);    
  TRACE_EXIT;
} /* TransReSpreadMat() */
