/************************************************************************/
/*                     FDATAOps.h		                        */
/************************************************************************/
/* SCCS INFO @(#)FDATAOps.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 91/12
*/
/* header file for BLAS Floating Point Data Operations (macros)
	the FDATA data type should only be directly accessed through
	these macros; in this way, C code can be indepenent of FDATA  
 
  	assumes FDATATYPE, FDATA, FD_BASE are already defined 
*/
/************************************************************************/

#ifndef __FDATAOps_INCL__

#define __FDATAOps_INCL__

#include "FDATASetUp.h"

#if (FDATATYPE==SINGLE || FDATATYPE==DOUBLE)

/* the FD_ALPHA macros are designed to be more efficient than their
   (equivalent) FD_MULSET() & FD_SCALE() counterparts; 
   to use in a procedure, declare and initialize a register variable 
   FD_ALPHA using FD_ALPHA_REGS(); then  FD_ALPHA_SCALE() &  
   FD_ALPHA_MULSET() can be used within that procedure
*/

#define FD_ALPHA_REGS(a) 	register FD_BASE FD_ALPHA=a;
	/* FDATA a; */

#define FD_ALPHA_SCALE(c) 	{ (c) *=  FD_ALPHA; }
	/* FDATA c; */

#define FD_ALPHA_MULSET(c,op,b) { (c) op  FD_ALPHA * (b); }
	/* FDATA c, b; 
	   op is =, +=, *=. -= */
                
#define FD_ALPHA_SCALEADD(c, a, b) { (c) = (a) + FD_ALPHA * (b); }
	/* FDATA c, b; */
                
#define FD_MULSET(c,op,a,b) 	{ (c) op (a) * (b); }
	/* FDATA c, a, b; 
	   op is =, +=, -= */

#define FD_SCALE(c,a) 		{ (c) *= (a); }
	/* FDATA c, a; */

#define FD_RECIP(c,a) 		{ (c) = FD_1 / (a); }
	/* FDATA c, a; */
                       
#define FD_DVDSET(c, a, b) 	{ (c) = (a) / (b); }
	/* FDATA c, a, b; */
                       
#define FD_NEG(c,a) 		(c) = -(a) 
	/* FDATA c, a; */

#define FD_ASSIGN(c,op,a) 	(c) op (a)
	/* FDATA c, a; 
	   op is =, +=, -= */

#define FD_SETOP(c, b, op, a) 	(c) = (b) op (a)
	/* FDATA c, a, b; 
	   op is +, - */

#define FD_SET_REAL(c, s) 	(c) = (s)
	/* FDATA c; FD_BASE s; */

#define FD_SET_IMAG(c, s) 
	/* FDATA c; FD_BASE s; */

#define FD_EQ(a,b) 		((a) == (b))
	/* int FD_EQ(); 
	   FDATA a, b; */ 

#define FD_ABS(a) 		((a) < 0.0? -(a): (a))
	/* FD_BASE FD_ABS();
	   FDATA a; */

#define FD_REAL(a)              (a)
        /* FD_BASE FD_REAL();
           FDATA a; */

#define FD_IMAG(a)              (0.0)
        /* FD_BASE FD_IMAG();
           FDATA a; */

#define FD_MKVAL(r, i)		(r)

#define FD_ASSIGN_SC(a, op, r, i)	(a) op (r)		

#else 	/* define equivalent complex operations */


#define FD_ALPHA_REGS(a) 	register FD_BASE FD_ALPHAr=(a).r, \
					 FD_ALPHAi=(a).i;
        /* FDATA a; */

#define FD_ALPHA_SCALE(c) 	{ register FD_BASE cr_=(c).r, ci_=(c).i; \
        			  (c).r = cr_*FD_ALPHAr - ci_*FD_ALPHAi; \
                		  (c).i = ci_*FD_ALPHAr + cr_*FD_ALPHAi; }
        /* FDATA c; */
        
#define FD_ALPHA_MULSET(c,op,b) { register FD_BASE br_=(b).r, bi_=(b).i;  \
       		 		  (c).r op br_*FD_ALPHAr - bi_*FD_ALPHAi; \
                		  (c).i op bi_*FD_ALPHAr + br_*FD_ALPHAi; }
        /* FDATA c, b;
           op is =, +=, -= */
                   
#define FD_ALPHA_SCALEADD(c, a, b) { register FD_BASE br_=(b).r, bi_=(b).i; \
       		 		  (c).r = (a).r + br_*FD_ALPHAr - bi_*FD_ALPHAi; \
                		  (c).i = (a).i + bi_*FD_ALPHAr + br_*FD_ALPHAi; }
	/* FDATA c, a, b;  */


#define FD_MULSET(c,op,a,b) 	{ register FD_BASE ar_=(a).r, br_=(b).r, \
						   ai_=(a).i, bi_=(b).i; \
				  (c).r op ar_*br_ - ai_*bi_; 		 \
                		  (c).i op ai_*br_ + ar_*bi_; 		 }
	/* FDATA c, a, b; 
	   op is =, +=, -= */
                		  
#define FD_SCALE(c,a) 		{ register FD_BASE ar_=(a).r, cr_=(c).r, \
					 	   ai_=(a).i, ci_=(c).i; \
        			  (c).r = ar_*cr_ - ai_*ci_;		 \
                		  (c).i = ai_*cr_ + ar_*ci_;		 }
	/* FDATA c, a; */

#define FD_RECIP(c,a) 	 	{ register FD_BASE ar_=(a).r, ai_=(a).i, t_; \
        			  t_ = 1.0 / (ar_*ar_ + ai_*ai_);	 \
                		  (c).r = ar_ * t_;			 \
                        	  (c).i = - ai_ * t_; 			 }
	/* FDATA c, a; */

#define FD_DVDSET(c, b, a) 						\
 				{ register FD_BASE ar_=(a).r, ai_=(a).i, t_, br_ = (b).r; \
        			  t_ = 1.0 / (ar_*ar_ + ai_*ai_);	 \
                		  (c).r = (br_*ar_ + (b).i*ai_) * t_;	 \
                        	  (c).i = ((b).i*ar_ - br_*ai_) * t_; 	 }
	/* FDATA c, a, b; */
                        
#define FD_NEG(c,a) 		(c).r = -(a).r, (c).i = -(a).i
	/* FDATA c, a; */

#define FD_ASSIGN(c,op,a) 	(c).r op (a).r, (c).i op (a).i
	/* FDATA c, a; */

#define FD_SETOP(c, b, op, a) 	(c).r = (b).r op (a).r, (c).i = (b).i op (a).i
	/* FDATA c, a, b; 
	   op is +, - */

#define FD_SET_REAL(c, s) 	(c).r = (s)
	/* FDATA c; FD_BASE s; */

#define FD_SET_IMAG(c, s) 	(c).i = (s)
	/* FDATA c; FD_BASE s; */

#define FD_EQ(a,b) 		((a).r == (b).r && (a).i == (b).i)
	/* int FD_EQ(); 
	   FDATA a, b; */ 

#define FD_ABS(a) 		(((a).r < 0.0? -(a).r: (a).r) + \
				 ((a).i < 0.0? -(a).i: (a).i)) 
	/* FD_BASE FD_ABS();
	   FDATA a; */

#define FD_REAL(a)              ((a).r)
        /* FD_BASE FD_REAL();
           FDATA a; */

#define FD_IMAG(a)              ((a).i)
        /* FD_BASE FD_IMAG();
           FDATA a; */

#define FD_MKVAL(r, i)		{r,i}	/* can only by used to init static vars */

#define FD_ASSIGN_SC(a, op, re, im)	(a).r op (re),  (a).i op (im)	

#endif /*(FDATAYPE...)*/

#endif /*__FDATAOps_INCL__*/
