/************************************************************************/
/*                     FDATASetUp.h		                        */
/************************************************************************/
/* SCCS INFO @(#)FDATASetUp.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 91/12
*/
/* header file for BLAS-3 Floating Point Data file 
   	assumes FDATATYPE is already defined 
  	written by P. Strazdins, from 11/12/91				*/
/************************************************************************/

#ifndef __FDATASetUp_INCL__ 

#define __FDATASetUp_INCL__

#include "GenMacros.h"
#include "IntOps.h"

#define SINGLE 		0	/* code for FDATA=float			*/
#define DOUBLE 		1	/* code for FDATA=double                */
#define COMPLEX 	2	/* code for FDATA=complex               */
#define DOUBLE_COMPLEX 	3	/* code for FDATA=double complex        */

/*  FD_BASE is base type of floating point data */
#if (FDATATYPE==SINGLE || FDATATYPE==COMPLEX)
#define FD_BASE float
#else
#define FD_BASE double
#endif

#ifndef FDATA			/* dont redefine FDATA & numeric constants */
/* FDATA is numeric type of floating point data */
# if (FDATATYPE==SINGLE || FDATATYPE==DOUBLE)
#  define FDATA FD_BASE
# else
typedef struct FD_complex {  
		FD_BASE r, 	/* real part				*/
			i; 	/*imaginary part			*/
		} FD_complex_struct;
#  define FDATA FD_complex_struct
# endif
#endif

typedef FDATA *FDATAPtr; 

#if (FDATATYPE==SINGLE)		/* needed as cannot use sizeof(FDATA) in #if */
# define FD_SIZE 4
# define LG_FD_SIZE 2
#elif (FDATATYPE==DOUBLE || FDATATYPE==COMPLEX)
# define FD_SIZE 8
# define LG_FD_SIZE 3
#else
# define FD_SIZE 16
# define LG_FD_SIZE 4
#endif /*FDATATYPE*/

#if (FDATATYPE==SINGLE)
# define FD_1		SFD_1
# define FD_M1		SFD_M1
# define FD_0		SFD_0
# define pFD2Str	SpFD2Str
# define Str2FD		SStr2FD
#elif (FDATATYPE==DOUBLE)
# define FD_1		DFD_1
# define FD_M1		DFD_M1
# define FD_0		DFD_0
# define pFD2Str	DpFD2Str
# define Str2FD		DStr2FD
#elif (FDATATYPE==COMPLEX)
# define FD_1		CFD_1
# define FD_M1		CFD_M1
# define FD_0		CFD_0
# define pFD2Str	CpFD2Str
# define Str2FD		CStr2FD
#else
# define FD_1		ZFD_1
# define FD_M1		ZFD_M1
# define FD_0		ZFD_0
# define pFD2Str	ZpFD2Str
# define Str2FD		ZStr2FD
#endif

extern FDATA FD_1 /*1.0*/, FD_M1 /*-1.0*/, FD_0 /*0.0 */;

#define FD_ADR(A) (((int) (A))/sizeof(FDATA))

/* convention that a pointer to a scaling factor represents unity if NULL */
#if (FDATATYPE<COMPLEX)
# define pFD_UNITY(pa) 		((pa) == NULL || (*(pa) == FD_1))
# define FD_register		FDATA register 
#else
# define pFD_UNITY(pa) 		((pa) == NULL || (((pa)->r == FD_1.r) && ((pa)->i == FD_1.i)))
# define FD_register		FDATA 
#endif
	/* int pFD_UNITY(); 
	   FDATA *pa; */ 
	   
#define pFD_VAL(pa)		((pa) == NULL? FD_1: *(pa))

char *pFD2Str STDC_PARAMS((FDATA *v));
/* post: returns a pointer to a string containing the text formm of v
 *	 in the case of complex, the 2 parts are separated by a ','
 *	 v==NULL is assumed to represent a value of unity
 */

FDATA Str2FD STDC_PARAMS((char *str));
/* post: returns the FDATA value that can be scanned from str
 *	 in the case of complex, the 2 parts assumed to be separated by a ','
 *	 if no valid number can be scanned, 0.0 is returned
 */
#endif /*__FDATASetUp_INCL__*/

