/************************************************************************/
/*       	FDMalloc: FDATA workspace memory allocation 		*/
/************************************************************************/
/* SCCS INFO @(#)FDMalloc.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 97/02
*/
/* essentially a front-end to the standard malloc library.
 * Due to serious call overheads observed in the implementation 
 * of this library, this module is required to improve 
 * performance for small workspaces.
 */

#ifndef __FDMalloc_INCL__
#define __FDMalloc_INCL__  

#include "FDATASetUp.h"

/********************* exported symbols  ***********************************/

#if FDATATYPE==SINGLE
#define MatAllocInit	DSMatAllocInit
#define MatFreeBuf	DSMatFreeBuf
#define MatMalloc_f	DSMatMalloc_f
#define MatFree		DSMatFree
#elif FDATATYPE==DOUBLE
#define MatAllocInit	DDMatAllocInit
#define MatFreeBuf	DDMatFreeBuf
#define MatMalloc_f	DDMatMalloc_f
#define MatFree		DDMatFree
#elif FDATATYPE==COMPLEX
#define MatAllocInit	DCMatAllocInit
#define MatFreeBuf	DCMatFreeBuf
#define MatMalloc_f	DCMatMalloc_f
#define MatFree		DCMatFree
#else
#define MatAllocInit	DZMatAllocInit
#define MatFreeBuf	DZMatFreeBuf
#define MatMalloc_f	DZMatMalloc_f
#define MatFree		DZMatFree
#endif

/****************************************************************/  

/* the following refer to a `pre-malloced' workspace W, from which 
 * `small' requests for memory will be serviced where posible.
 * `Small' means the size < FastWSSize.
 * Default values:  
 *	FastWSSize = FastWSSize_DEF
 *	|W| = 4*FastWSSize
 */

#define FastWSSize_DEF	512	

void MatMallocInit STDC_PARAMS( (int ParamType, int ParamVal) );
/* pre:  ParamType IN {0, 1}; ParamVal > 0
 * post: if ParamType=0, |W| = ParamVal
 *	 if ParamType=1, FastWSSize = ParamVal
 * notes: MallocInit(0, _) will only have effect if currently no unfreed
 *        allocations are from W
 */

void MatFreeBuf STDC_PARAMS( (void) );
/* post:  storage associated with W is freed.
 * notes: any unfreed blocks in W must not be used after this call 
 */


#define FD_ALLOC_FALSE	0
#define FD_ALLOC_TRUE	(~FD_ALLOC_FALSE)

/* MatMalloc() is recommended interface from outside the DBLAS;
 * of for inside the DBLAS wherever s/w overheads is not an issue,
 * ie. Size is always large 
 */
#define MatMalloc(Where, CrucialAlloc, A, Size)			\
	MatMalloc_f(Where, CrucialAlloc, A, Size, 		\
		    FD_ALLOC_FALSE, FD_ALLOC_FALSE)
		    
/* MatMalloc_F() is for fast allocation where Size may be < FastWSSize,
 * ie. a vector or small matrix 
 */	   	   
#define MatMalloc_F(Where, CrucialAlloc, A, Size)         	\
	MatMalloc_f(Where, CrucialAlloc, A, Size, 		\
		    FD_ALLOC_TRUE, FD_ALLOC_FALSE)

/* MatMalloc_FT() is for fast allocation where Size may be < |W|
 * and it can be guaranteed the coresponding MatFree() will be called
 * before any future malloc.
 */ 			   
#define MatMalloc_FT(Where, CrucialAlloc, A, Size)         	\
	MatMalloc_f(Where, CrucialAlloc, A, Size, 		\
		    FD_ALLOC_TRUE, FD_ALLOC_TRUE)
			   
int MatMalloc_f STDC_PARAMS( 
	(char *Where, int CrucialAlloc, FDATA **A, int Size,
	int UseFastMalloc, int ImmedFree) 
);
/* pre:  UseFastMalloc, UseFastMalloc IN {FD_ALLOC_FALSE, FD_ALLOC_TRUE}
 * post: returns whether (Size>0) FDATA elements allocated at *A;
 *	 if allocation fails, an error messgage is printed; 
 *	 if CrucialAlloc, DBLAS_EXIT() is called also.
 * notes:if UseFastMalloc==FD_ALLOC_TRUE, fast allocation methods
 *	 will be used if there is sufficient space insside the workspace.
 *	
 */ 

void MatFree STDC_PARAMS( (FDATA *A) );
/* pre:  a matrix (of base address) A was allocated from MatMalloc()
 * post: frees matrix which begins at A
 */ 

#endif /*__FDMalloc_INCL__*/
