/************************************************************************/
/*       	GenAssert: FDATA checking and debugging module	*/
/************************************************************************/
/* SCCS INFO @(#)GenAssert.c 1.1         last modified P. Strazdins, 98/12/24
   first created: P. Strazdins 96/01/23 
*/
/* general assertion checking routines and and tracing parameters; also error
 * handler.
 * Here, every call to assertion checking routine CommonAssert() can also be 
 * traced, according to its trace type and level. 
 */
#include <stdlib.h>		/*exit()*/
#include <stdio.h>		/*printf()*/
#include "GenAssert.h"

/********************* assertion checking code *****************************/

#define ASSERT_FAIL_EXIT_CODE 	12
#define ERROR_EXIT_CODE	13

int TraceType = TRACE_ALL;	/* categories of traces to be printed */ 
int TraceLevel = -1;    	/* tracing is off */
int BaseTraceLevel = 0;
int CurTraceDepth = 0;
char * CallerStack[MAX_CALL_DEPTH];
/* print traces for calls to CommonAssert() with trace level <= TraceLevel */ 

char TraceStr[256];	 	/* string to print out assertion traces on */
static char TmpTraceStr[3][64]; 
char *TmpTraceStrs[] = {
	&TmpTraceStr[0][0],	&TmpTraceStr[1][0], 	&TmpTraceStr[2][0]
};
	
int AssertCId[2] = {-1, -1}; 	/* suppress error msg / trace prefixes */
void PrintLinePrefix()
{ if (AssertCId[0] >= 0)
    printf("%d.%d: ", AssertCId[0], AssertCId[1]);
}

void CommonAssert_(Cond, ErrFlag, CallerPrefix, Caller, CallTrLevel, CallTrType)
  int Cond, ErrFlag, CallTrLevel, CallTrType;
  char *CallerPrefix, *Caller; 
{ int i;
  if (TraceLevel >= BaseTraceLevel+CallTrLevel && TraceLevel >= CurTraceDepth &&
      (TraceType == TRACE_ALL || TraceType == CallTrType)) {
    PrintLinePrefix();
    for (i=0; i < CurTraceDepth; i++)
      printf("  ");
    printf("%s%s%s\n",  CallerPrefix, Caller, TraceStr);
  }
  if (! Cond || ErrFlag) {
    PrintLinePrefix();
    printf("%s%s%s : assert fail:  ",  CallerPrefix, Caller, TraceStr);
    if (ErrFlag) {
      if (CallTrType == TRACE_COMMUN)
        printf("communication error=%d ", ErrFlag);
      else if (CallTrType == TRACE_MALLOC)
        printf("failed memory allocation %d words", ErrFlag); 
      else if (CallTrType == TRACE_COMPUTE)
        printf("matrix overlap error (%d)", ErrFlag);
      else
        printf("error flag (%d) raised", ErrFlag);
    } 
    printf("\n");
    PrintLinePrefix();
    printf("	Call Trace:");
    for (i=0; i < MIN(MAX_CALL_DEPTH, CurTraceDepth); i++)
      printf(" %s", CallerStack[i]);
    printf("\n");
    exit(ASSERT_FAIL_EXIT_CODE);
  }
  TraceStr[0] = 0;   
} /*CommonAssert_()*/


void Error_ ( CallerPrefix, Caller, ParamNum ) 
  char *CallerPrefix, *Caller; int ParamNum;
{ int p, d=0, n=0;			/* ParamNum = 0x(p)(d)(n) */
  if (ParamNum < 0)
    p = -ParamNum;
  else { 
    p = ParamNum>>8;
    d = (ParamNum>>4)%16 ;	
    n = ParamNum%16;
    if (p==0) 
      p=d, d=n, n=0;
    if (p==0) 
      p=d, d=n, n=0;
  }    
  PrintLinePrefix();
  printf("*** Error detected in procedure %s%s, parameter %d\n",
 	  CallerPrefix, Caller, p);
  if (d > 0) {
    PrintLinePrefix();
    if (d==p)
      printf("***	(special constraints for parameter not met)\n");	
    else {  
      printf("***	(in relation to parameter %d", d);
      if (n > 0)
         printf(" and parameter %d", n);
      printf(")\n");
    }        	
  }	    
  PrintLinePrefix();
  printf("*** Values of parameters to call follow:\n");
  PrintLinePrefix();
  printf("***	%s\n", TraceStr);
  exit(ERROR_EXIT_CODE);
} /* Error_() */

/********************* miscellaneous variables *****************************/

int PrintRes = 0;		/* suppress matrix printing */
int ForceAlgVers = 0;		/* do not force any particular alg. version */
int NoOuterErrCheck = 0;	/* library outer level error checking is on */
