/************************************************************************/
/* GetBlockFactor: calulate blocking factors for blocked algorithms     */
/************************************************************************/
/* SCCS INFO GetBlockFactor.h@(#)         	last modified P. Strazdins,  98/12/24
   first created: P. Strazdins 94/10/11					   
*/

#ifndef __GetBlockFactor_INCL__
#define __GetBlockFactor_INCL__

#include "GenMacros.h"
#include "MatDim.h"

#define GenGetW	GGenGetW


int GenGetW STDC_PARAMS(
  (int N, double w_scale, int w_min, int w_max, pMatDim WDim, int j, int jInc)
);
/* pre:  N is the size of column j  of an algorithm where j changes with dj
 *	 WDim is dimension relative to row 0 of the columns.
 * post: returns largest blocking factor w such that:
 *	   MIN(w_max, w_min) <= w <= w_max
 *	 for w' = 2* (int) (w_scale * sqrt(N/2)), if w_min <= w' <= w_max,
 *	 w' will be used as an initial value for w, with the following 
 *	 strategies to be applied: 
 *	   1. w is to be chosen so that
 *		j+jInc*w
 *	      is on a block boundary (wrt WDim) (suppress by setting WDim.B=1)
 *	   2. w is to be chosen so that  j+jInc*w is on a cell-block boundary
 *	      (suppress by setting d_P(WDim)=1).
 *	 these are only applied if the resulting value of w satisifes:
 *	    MAX(w_min, 3w'/4) <= w <= MIN(1.5*w', w_max)
 */

#endif /*__GetBlockFactor_INCL__*/
