/************************************************************************/
/*                     MachineParams		                        */
/************************************************************************/
/* SCCS INFO @(#)MachineParams.h 1.3         last modified P. Strazdins, 99/07/06
   first created: P. Strazdins 92/01
*/
/* header file for cell architecture dependent constants,
 * currently defaults appropriate for  Fujitsu AP1000. 
 * 
 * compilation: 
 *	if TUNE is not defined, parameters are defined in this file as
 *	constants. This enables compiler to optimize thier use better.
 *	Otherwise, they are defined to be elements of an exported parameter 
 *	vector which is dynamically settable (initial values being the
 *	corresponding constants). This is useful for performance tuning, 
 *	and also testing/debugging.
 */	
/************************************************************************/

#ifndef __MachineParams_INCL__
#define __MachineParams_INCL__

#include "FDATASetUp.h"

#if FDATATYPE==SINGLE
#define MachParamVec	DSMachParamVec
#define MachParamVal	DSMachParamVal
#elif FDATATYPE==DOUBLE
#define MachParamVal    DDMachParamVal
#define MachParamVec	DDMachParamVec
#elif FDATATYPE==COMPLEX
#define MachParamVal    DCMachParamVal
#define MachParamVec	DCMachParamVec
#else
#define MachParamVal    DZMachParamVal
#define MachParamVec	DZMachParamVec
#endif

extern int MachParamVec[];
int MachParamVal STDC_PARAMS( ( char *ParamMnem) );
/* post: converts MachineParam paramemter menemonics to their corresp.
 *	 indices, ie.
 *	 if *ParamMnem = one of CACHESIZE_MNEM, ..., PAR_OUTER_VECSIZE_MNEM
 *	 returns the _NUM index corresp  to *ParamMnem; else returns -1
 */ 
 
					/* define indexes for each parameter */
#define CACHESIZE_NUM			0
#define CACHE_ASSOC_NUM			1
#define CACHE_LINE_SIZE_NUM		2
#define OPT_MSG_SIZE_NUM		3
#define MAX_DBLAS_WS_SIZE_NUM		4		
#define MIN_VECSIZE_NUM			5
#define INNER_VECSIZE_NUM		6
#define OUTER_VECSIZE_NUM		7
#define MINOR_OUTER_VECSIZE_NUM		8
#define K0_MINOR_OUTER_VECSIZE_NUM 	9
#define MINOR_OUTER_VECSIZE_K0_NUM	10
#define PAR_INNER_VECSIZE_NUM		11
#define PAR_INNER_VECSIZE_TRSM_NUM	12
#define PAR_OUTER_VECSIZE_NUM		13
#define PAR_INNER_VECSIZE_TRSM_MIN_NUM 	14
#define LG_MEM_BANKS_NUM		15
#define MM_INNER_K_NUM			16
#define FAST_STRIDE_MSG_NUM		17
#define MM_SCALE_NUM			18
#define TLBSIZE_NUM			19

#define NUM_MACH_PARAMS			20	


#define CACHESIZE_MNEM			"CS"
#define CACHE_ASSOC_MNEM		"CA"
#define CACHE_LINE_SIZE_MNEM		"CL"
#define OPT_MSG_SIZE_MNEM		"MsgS"
#define MAX_DBLAS_WS_SIZE_MNEM		"WS"
#define MIN_VECSIZE_MNEM                "MVS"		
#define INNER_VECSIZE_MNEM		"IVS"
#define OUTER_VECSIZE_MNEM		"OVS"
#define MINOR_OUTER_VECSIZE_MNEM	"mOVS"
#define K0_MINOR_OUTER_VECSIZE_MNEM 	"kmOVS"
#define MINOR_OUTER_VECSIZE_K0_MNEM 	"mOVSk"
#define PAR_INNER_VECSIZE_MNEM		"pIVS"
#define PAR_INNER_VECSIZE_TRSM_MNEM	"pIVST"
#define PAR_OUTER_VECSIZE_MNEM		"pOVS"
#define PAR_INNER_VECSIZE_TRSM_MIN_MNEM "pIVSTm"
#define LG_MEM_BANKS_MNEM		"MemB"
#define MM_INNER_K_MNEM			"MMInK"
#define FAST_STRIDE_MSG_MNEM		"fSMsg"
#define MM_SCALE_MNEM			"MMSc"
#define TLBSIZE_MNEM			"TLBS"

#ifdef TUNE		/* define parameters as elts of the parameter vector */
#define MachParamSet(ParamNum, ParamVal)				\
	{ if (0<=(ParamNum) && (ParamNum) < NUM_MACH_PARAMS)		\
	    MachParamVec[ParamNum] = (ParamVal); }		

#define CACHESIZE		MachParamVec[CACHESIZE_NUM]
#define CACHE_ASSOC		MachParamVec[CACHE_ASSOC_NUM]
#define CACHE_LINE_SIZE		MachParamVec[CACHE_LINE_SIZE_NUM]
#define OPT_MSG_SIZE		MachParamVec[OPT_MSG_SIZE_NUM]
#define MAX_DBLAS_WS_SIZE	MachParamVec[MAX_DBLAS_WS_SIZE_NUM]
#define MIN_VECSIZE		MachParamVec[MIN_VECSIZE_NUM]
#define INNER_VECSIZE		MachParamVec[INNER_VECSIZE_NUM]
#define OUTER_VECSIZE		MachParamVec[OUTER_VECSIZE_NUM]
#define MINOR_OUTER_VECSIZE	MachParamVec[MINOR_OUTER_VECSIZE_NUM]
#define K0_MINOR_OUTER_VECSIZE	MachParamVec[K0_MINOR_OUTER_VECSIZE_NUM]
#define MINOR_OUTER_VECSIZE_K0	MachParamVec[MINOR_OUTER_VECSIZE_K0_NUM]
#define PAR_INNER_VECSIZE	MachParamVec[PAR_INNER_VECSIZE_NUM]
#define PAR_INNER_VECSIZE_TRSM	MachParamVec[PAR_INNER_VECSIZE_TRSM_NUM]
#define PAR_OUTER_VECSIZE	MachParamVec[PAR_OUTER_VECSIZE_NUM]
#define PAR_INNER_VECSIZE_TRSM_MIN	\
				MachParamVec[PAR_INNER_VECSIZE_TRSM_MIN_NUM]
#define LG_MEM_BANKS		MachParamVec[LG_MEM_BANKS_NUM]
#define MM_INNER_K		MachParamVec[MM_INNER_K_NUM]
#define FAST_STRIDE_MSG		MachParamVec[FAST_STRIDE_MSG_NUM]
#define MM_SCALE		MachParamVec[MM_SCALE_NUM]
#define TLBSIZE			MachParamVec[TLBSIZE_NUM]

#else			/* define default values of parameters as constants */
#define MachParamSet(ParamNum, ParamVal) /* no effect */

/* cache parameters; if none exists, CACHESIZE should bet set to infinity */
#ifndef CACHE
# define CACHE			16	/* (primary) cache size in Kbytes */	
#endif 

#ifndef CACHESIZE
# define CACHESIZE 		(CACHE * 1024 / FD_SIZE)
#endif

#ifndef CACHE_ASSOC
# define CACHE_ASSOC		1	/* 1 for direct-mapped; n for n-way assoc */
#endif

#ifndef CACHE_LINE_SIZE
# define CACHE_LINE_SIZE 	16	/* cache line size in bytes */
#endif

/* there may be an optimum or maximum message size; if none exists, set to inf.*/
#ifndef OPT_MSG_SIZE
# define OPT_MSG_SIZE 		116777216 /*16384*/		
#endif			

/* can restrict size of workspace allocated within any DBLAS call.
 * This is only useful in an application already running out of heap memory.
 * Restriction  may result in loss of speed however. 
 * Note: currently not effective.
 */
#ifndef MAX_DBLAS_WS_SIZE
# define MAX_DBLAS_WS_SIZE	116777216 /* 2^24, ie effectively infinity */
#endif

/* the following parameters refer to the (minimum) vector size for optimal
 * cell performance, based on a cell matrix multiply C = A*B, C is M x N. 
 * `OUTER' [`INNER'] refers to the outer [inner] dimension of a matrix multiply
 * (eg M or N) [(eg. K )]. 
 *
 * On a scalar cell, [INNER,OUTER]_VECSIZE would rever to the cache level 
 * of the memory hierarchy (with 2*INNER_VECSIZE*OUTER_VECSIZE ~ CACHESIZE
 * and OUTER_VECSIZE <= INNER_VECSIZE)
 * and MINOR_OUTER_VECSIZE referring to the register level.
 * On a vector cell, using an GAXPY-based multiply, OUTER_VECSIZE 
 * [MINOR_OUTER_VECSIZE] would refer to the length [number] of C vectors held in vector registers.
 * These parameters are used for rectangular/triangular cell MM partitioning. 
 *
 * MINOR_OUTER_VECSIZE may change for low K, ie K < K0_MINOR_OUTER_VECSIZE, 
 * (ie. for scalar cells, when the cell MM switches from inner-product
 * to rank-1 update based); in this case, MINOR_OUTER_VECSIZE_K0 should be set
 * to this different value.
 * 
 * All the above parameters are dependent on the implementation of cell BLAS
 * being used. If these are not precisely known, over-estimating will probably
 * yield better performance than underestimating. In any case,  experimentation
 * can be used to determine the best values.
 *
 * The prefix `PAR' refers to the partitioning size on the outer-level DBLAS
 * algorithms, which generally occurs over the smallest of the global 
 * dimensions. For most situations, they will be the same as the corresp.
 * macors without the prefix.
 */

#ifndef INNER_VECSIZE		/* ~ sqrt( CACHESIZE / 2) */
#if (CACHESIZE >= 32768)
# define INNER_VECSIZE 128
#elif (CACHESIZE >= 16384)
# define INNER_VECSIZE 88
#elif (CACHESIZE >= 8192)
# define INNER_VECSIZE 64
#elif (CACHESIZE >= 4096)
# define INNER_VECSIZE 44
#elif (CACHESIZE >= 2048)
# define INNER_VECSIZE 32
#elif (CACHESIZE >= 1024)
# define INNER_VECSIZE 24
#elif (CACHESIZE >= 512)
# define INNER_VECSIZE 16
#else
# define INNER_VECSIZE 16
#endif /*CACHESIZE*/
#endif /*INNER_VECSIZE*/

#ifndef OUTER_VECSIZE		
# define OUTER_VECSIZE 		(CACHESIZE / 2 / INNER_VECSIZE) 
					/* suitable for scalar cells only */
#endif 

#ifndef MINOR_OUTER_VECSIZE
#define MINOR_OUTER_VECSIZE	8	/* prob. valid for 16-32 reg scalar */
#endif

#ifndef K0_MINOR_OUTER_VECSIZE
#define K0_MINOR_OUTER_VECSIZE	4	/* prob. valid for 16-32 reg scalar */
#endif

#ifndef MINOR_OUTER_VECSIZE_K0
#define MINOR_OUTER_VECSIZE_K0	6	/* 6-8  prob. valid for 16-32 reg scalar */
#endif

#ifndef PAR_INNER_VECSIZE	/* partitioning for || Matrix Mult routines */
# define PAR_INNER_VECSIZE	INNER_VECSIZE
#endif

/* useful for TRSM proedures to have a seperately tunable parameter */
#ifndef PAR_INNER_VECSIZE_TRSM		
# define PAR_INNER_VECSIZE_TRSM	PAR_INNER_VECSIZE
#endif

#ifndef PAR_INNER_VECSIZE_TRSM_MIN /* minimum useful blocking factor for TRSM */
#define PAR_INNER_VECSIZE_TRSM_MIN	4
#endif

#ifndef PAR_OUTER_VECSIZE
# define PAR_OUTER_VECSIZE	OUTER_VECSIZE
#endif

#ifndef MIN_VECSIZE		/* min vector size for efficient iteration */
# define MIN_VECSIZE 		16
#endif

#ifndef LG_MEM_BANKS		/* lg_2 of # of interleaved mem banks */
# define LG_MEM_BANKS		0
#endif

#ifndef MM_INNER_K		/* = 1 if innermost matrix mult loop in K dim.
				  (the case for most scalar architectures) 
				   = 0, otherwuse */
# define MM_INNER_K 		1
#endif

#ifndef FAST_STRIDE_MSG		/* = 1 iff underlying communication library
				  can do simple stride commun. without penalty 
				*/
# define FAST_STRIDE_MSG	0
#endif


#ifndef MM_SCALE		/* = 1 iff there is likely to be a gain
				   in scaling input operands to matrix multiply,
				   even if this means copying it */
# define MM_SCALE MM_INNER_K /* 0 for PerfLib blas */
#endif 				  

#ifndef TLBSIZE			/* # entries in the TLB; set to 0 if no TLB*/
# define TLBSIZE		64 
#endif

#endif /*TUNE*/

#endif /*__MachineParams_INCL__*/
