/************************************************************************/
/* SYMMult: Symmetric Matrix Multiply alg				*/
/************************************************************************/
/* SCCS INFO @(#)SYMMult.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 97/04
*/
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *              if macro TRACE defined, parameter value tracing enabled
 */


#ifndef __SYMMult_INCL__
#define __SYMMult_INCL__
   
#include "LocMat.h"
#include "DistMat.h"

#if FDATATYPE==SINGLE
# define SYMMult	DSSYMMult
#elif FDATATYPE==DOUBLE
# define SYMMult	DDSYMMult
#elif FDATATYPE==COMPLEX
# define SYMMult	DCSYMMult
#else
# define SYMMult	DZSYMMult
#endif


void SYMMult STDC_PARAMS( 
  	(int M, int N, FDATA *alpha, char UpLo, DistMat A, DistMat B, 
  	 FDATA *beta, DistMat C)
);
/* pre:  all cells are executing SYMMult with consistent
 *       UpLo, Diag M, N, alpha, A, B, beta, & C;
 *       UpLoOK(UpLo); 0<= M, N
 * post: performs C := *alpha A B + *beta C
 *       where A is a symmetric matrix of form UpLo with MxN C
 * notes: efficient for N<<M; most efficient for DM_Trans(B)==DM_Trans(A)
 */

#endif /*__SYMMult_INCL__*/


