/************************************************************************/
/* SYMMult: DBLAS Symmetric Matrix Multiply alg				*/
/************************************************************************/
/* SCCS INFO @(#)SYMMult.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 97/04
*/

#include "SYMMult.h"

#include "BLASOptions.h"
#include "FDATAOps.h"
#include "FDAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "MachineParams.h"
#include "CellMatrix.h"
#include "CellTri1Matrix.h"
#include "SpreadMatrix.h"

#define ADM DM_DM(A)
#define ADN DM_DN(A)
#define BDM DM_DM(B)
#define BDN DM_DN(B)  
#define CDM DM_DM(C)
#define CDN DM_DN(C)
#define lA DM_Loc(A)
#define lB DM_Loc(B)
#define lC DM_Loc(C)

void SYMMult(M, N, alpha, UpLo, A, B, beta, C)
  int M, N; FDATA *alpha, *beta; char UpLo; DistMat A, B, C; 
{ int n;                         	
  int TrA = DM_Trans(A),			U;
  DistMatStore BN, CN;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%d,%s,%c,%d a%d %d %d,%d a%d %d %d,%s,%d a%d %d %d)",
     M, N, pFD2Str(alpha), UpLo, DM_Trans(A), FD_ADR(LM_ADR(lA)), LM_RInc(lA), 
     LM_CInc(lA), DM_Trans(B), FD_ADR(LM_ADR(lB)), LM_RInc(lB), LM_CInc(lB),
     pFD2Str(beta), DM_Trans(C), FD_ADR(LM_ADR(lC)), LM_RInc(lC), LM_CInc(lC)),
     CompatBlk(ADM, ADN) && M>=0 && N>=0 , 0, "SYMMult", 1, TRACE_ALL
  ); 
 
  if (M==0 || N==0) {
    TRACE_RETURN;
  }

  U = UpLo==Upper;  
  if (TrA)
    UpLo = TransForm(UpLo);
  n=0;
  while (n < N) {
    int mA /*= G2Ld(M, ADM/ADN)*/, 	kA /*= G2Ld(M, ADN/ADM)*/;
    FDATA ScAlpha = *alpha;
    int dN = MIN(PAR_OUTER_VECSIZE, N-n);
    DistMat_RVAlloc(3);			DistMat BT;
    DistMat C1 = !U? SubMat(C, 1, 0): C,A1 = SubMat(A, 1-U, U);
    DistMat Bt = TrA==0? B: Trans(B), 	Ct = TrA!=0? C: Trans(C),
    	    C1t = TrA!=0? Trans(C1): C1;	
    LocMat BsN, BTsN, CsN, CTsN;  	/* spread workspaces of B, C */
    int BsNn, BTsNn, CsNn, CTsNn;

    LM_ADR(BsN) = NULL;
    BsNn = SpreadMat(1-TrA, dN, M, Bt, &ScAlpha, DM_DM(A), &BsN, 0, M, DM_DN(A));
    kA = ABS(BsNn);

    LM_ADR(CTsN) = NULL;
    CTsNn = InitReduceWS(TrA, dN, M, Ct, beta, DM_DN(A), &CTsN, 0, M, DM_DM(A));
    mA = ABS(CTsNn);

    LM_Trans(TrA, BsN);			/* ensure BsN is kA x dN  */
    LM_Trans(1-TrA, CTsN);		/* ensure CTsN is mA x dN */
    Tri1MatPartMult(mA, dN, kA, &ScAlpha, UpLo, NonUnit, Trans(A), 
		    &BsN, NULL, &CTsN); 
    LM_Trans(1-TrA, CTsN);		/* restore CTsN */
    LM_Trans(TrA, BsN);			

    ReduceMat(TrA, dN, M, Ct, beta, DM_DN(A), &CTsN, M, DM_DM(A));
    if (CTsNn>0)
      MatFree(LM_ADR(CTsN));

    LM_ADR(BTsN) = NULL; 
    BT = Spread2DistMat(TrA, 1, DM_DM(A), &BsN);
    BTsNn = SpreadMat(TrA, dN, M, BT, &ScAlpha, DM_DN(A), &BTsN, 0, M, DM_DM(A));
    if (!U)
      kA = G2Ld(M-1, DM_DN(A1));
    else				/* use a faster method to compute this*/
      kA = ABS(BTsNn) - (d_P0(DM_DN(A))==0);/*nb. ABS(BTsNn)=G2Ld(M,DM_DN(A))*/

    LM_ADR(CsN) = NULL;
    CsNn = InitReduceWS(1-TrA, dN, M-1, C1t, NULL, DM_DM(A1), &CsN, 
    		        0, M-1, DM_DN(A1)); 
    mA = ABS(CsNn);
    
    LM_Trans(1-TrA, BTsN);		/* ensure BTsN is kA x dN ... */
    if (d_P0(DM_DN(A))==0 && U) 	/* ... and aligned with DM_DN(A1) */
      pLocSubMat(&BTsN, 1, 0);
    LM_Trans(TrA, CsN);			/* ensure CsN is mA x dN */
    Tri1MatPartMult(mA, dN, kA, &ScAlpha, UpLo, NonUnit, A1, &BTsN, NULL, &CsN);
    LM_Trans(TrA, CsN);			/* restore CsN */
    if (d_P0(DM_DN(A))==0 && U) 	/* restore BTsN */
      pLocSubMat(&BTsN, -1, 0); 

    ReduceMat(1-TrA, dN, M-1, C1t, NULL, DM_DM(A1), &CsN, M-1, DM_DN(A1));
    if (CsNn>0)  
      MatFree(LM_ADR(CsN));
    if (BTsNn>0)
      MatFree(LM_ADR(BTsN));

    if (BsNn>0)
      MatFree(LM_ADR(BsN));
      
    n += dN;
    if (n < N) {
      B = SubMat_(&BN, B, 0, dN);	      C = SubMat_(&CN, C, 0, dN);
    }
  } /* while (n...) */
  TRACE_EXIT;
} /*SYMMult()*/
