/************************************************************************/
/* SYR2KMult: DBLAS Symmetric Rank-2K Multiplication procedures		*/
/************************************************************************/
/* SCCS INFO @(#)SYR2KMult.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 95/04/02
*/
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *              if macro TRACE defined, parameter value tracing enabled
 *		if macro PRINTMAT define, matrices are printed for debugging
 */
#ifndef __SYR2KMult_INCL__
#define __SYR2KMult_INCL__

#include "FDATASetUp.h"
#include "DistMat.h"

#if FDATATYPE==SINGLE
# define DistSYR2KMult	DSDistSYR2KMult
#elif FDATATYPE==DOUBLE
# define DistSYR2KMult	DDDistSYR2KMult
#elif FDATATYPE==COMPLEX
# define DistSYR2KMult	DCDistSYR2KMult
#else
# define DistSYR2KMult	DZDistSYR2KMult
#endif

void DistSYR2KMult STDC_PARAMS( 
	(int N, int K, FDATA *alpha, DistMat A, DistMat B, 
	 FDATA *beta, char FormC, DistMat C)
);
/* pre:  all cells are executing DistRankKMult() with 
 *	 M, N, K, *alpha, *beta,  FormC parameters and (D[0..1], Trans) 
 * 	 fields for A, B & C consistent across all cells;
 *	 M, N, K >= 0; DM_Trans(A)==DM_Trans(B) && UpLoOK(FormC)
 * post: performs global operation  C = (*beta)*C + (*alpha)*(AB^T + BA^T),
 *	 where C is a triangular matrix of form FormC,
 * notes: efficient for K << N
 */

#endif /*__SYR2KMult_INCL__*/

