/************************************************************************/
/* SYRKMult: DBLAS Symmetric Rank K Matrix Multiplication procedure	*/
/************************************************************************/
/* SCCS INFO @(#)SYRKMult.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 97/02
*/

#include "SYRKMult.h"

#include "BLASOptions.h"
#include "FDATAOps.h"
#include "FDAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDMalloc.h"
#include "MachineParams.h"
#include "CellTriMatrix.h"
#include "SpreadMatrix.h"

#define lA DM_Loc(A)
#define lC DM_Loc(C) 
#define CDM DM_DM(C)
#define CDN DM_DN(C)


void DistSYRKMult(N, K, alpha, A, beta, FormC, C)
  int N, K; FDATA *alpha, *beta; 
  DistMat A, C; char FormC;
{ int k, MAXdK; 			/* max size of spread across K-dim */
  int TransA = DM_Trans(A);
  DistMatStore AK;
  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%dx%d,%s,%d a%d %d %d,%s,%d %c a%d %d %d)",
     N, K, pFD2Str(alpha), DM_Trans(A), FD_ADR(LM_ADR(lA)), LM_RInc(lA), LM_CInc(lA),
     pFD2Str(beta), DM_Trans(C), FormC, FD_ADR(LM_ADR(lC)), LM_RInc(lC), LM_CInc(lC)),
    N>=0 && K>=0, 0, "DistSYRKMult", 1, TRACE_ALL
  );

  if (DM_Trans(C)^TransA) {
    DM_Transpose(C);    		FormC = TransForm(FormC);
  }

  if (!pFD_UNITY(beta)) 
    TriMatScale(G2Ld(N, DM_DM(C)), G2Ld(N, DM_DN(C)), beta, FormC, C);

  if (K==0 || (FD_EQ(*alpha, FD_0)) || N==0) {
    TRACE_RETURN;  
  }
  
  MAXdK=MIN(PAR_INNER_VECSIZE, K);

  k=0;  
  while (k < K) { 			/* C += alpha A[:,k..+dK] A^[:,k..+dK]*/
    int dK = MIN(MAXdK, K-k);
    int m, n;				/* local sizes of C */
    LocMat AsK, BsK;			/* spread workspaces of A, B */
    int AsKn, BsKn;
    DistMat At = TransA? Trans(A): A;
    DistMat B;				DistMat_RVAlloc(1);
            
    LM_ADR(AsK) = NULL;
    AsKn = SpreadMat(1-TransA, dK, N, At, &FD_1, CDM, &AsK, 0, N, CDN);
    m = ABS(AsKn);
    
    LM_ADR(BsK) = NULL;
    B = Spread2DistMat(TransA, 1, CDM, &AsK); 	
    BsKn = SpreadMat(TransA, dK, N, B, &FD_1, CDN, &BsK, 0, N, CDM);    
    n = ABS(BsKn);

    LM_Trans(TransA, AsK);		/* make m x dK */
    LM_Trans(TransA, BsK);		/* make dK x n */	
    TriMatPartMult(m, n, dK, alpha, &AsK, &BsK, NULL, FormC, C);

    if (BsKn>0)  
      MatFree(LM_ADR(BsK));  
    if (AsKn>0) 
      MatFree(LM_ADR(AsK));
    if (PrintRes)
      MatPrint("C", MIN(PrintRes, m), MIN(PrintRes, n), LocMatTr(Trans(C), lC));
    k += dK;
    if (k < K) {
      A = SubMat_(&AK, A, 0, dK);
    } 
  } /* while (k...) */

  TRACE_EXIT;
} /*DistSYRKMult*/


