/************************************************************************/
/* SYRKMult: DBLAS Symmetric Rank-2K Multiplication procedures		*/
/************************************************************************/
/* SCCS INFO @(#)SYRKMult.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 95/04/02
*/
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *              if macro TRACE defined, parameter value tracing enabled
 *		if macro PRINTMAT define, matrices are printed for debugging
 */
#ifndef __SYRKMult_INCL__
#define __SYRKMult_INCL__

#include "FDATASetUp.h"
#include "DistMat.h"

#if FDATATYPE==SINGLE
# define DistSYRKMult	DSDistSYRKMult
#elif FDATATYPE==DOUBLE
# define DistSYRKMult	DDDistSYRKMult
#elif FDATATYPE==COMPLEX
# define DistSYRKMult	DCDistSYRKMult
#else
# define DistSYRKMult	DZDistSYRKMult
#endif

void DistSYRKMult STDC_PARAMS( 
	(int N, int K, FDATA *alpha, DistMat A, 
	FDATA *beta, char FormC, DistMat C)
);
/* pre:  all cells are executing DistRankKMult() with 
 *	 N, K, *alpha, *beta,  FormC parameters and (D[0..1], Trans) 
 * 	 fields for A,& C consistent across all cells;
 *	 N, K>= 0; UpLoOK(FormC)
 * post: performs global operation  C = (*beta)*C + (*alpha)*(AA^T),
 *	 where C is a triangular matrix of form FormC,
 * notes: efficient for K << N
 */

#endif /*__SYRKMult_INCL__*/

