/************************************************************************/
/* TRSM: DBLAS update B with (inverse of) triangular matrix A		*/
/************************************************************************/
/* SCCS INFO @(#)TRSM.c 1.3         last modified P. Strazdins, 99/11/17
   first created: P. Strazdins 93/15/01
*/

#include "TRSM.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#include "BLASOptions.h"
 
#include "LocMat.h"
#include "MatDim.h"
#include "MachineParams.h"
#include "CellMatrix.h"
#include "FDCommun.h"
#include "SpreadMatrix.h"
#include "GetBlockFactor.h"

pLocMat TRSMsBUnScal = NULL;
static LocMat TRSMsBUnScal_;

void SpreadL2TR_M(
	DoTRSM, UpLo, Diag, RowCol, w, w_loc, n, BDM, AsMScale, AsM, BsM, 
	B, brdB
)
  int DoTRSM, UpLo; char Diag; int RowCol, w, w_loc, n; 
  pMatDim BDM; FDATA *AsMScale; pLocMat AsM, BsM, B;
  int brdB;
{ int iw;
  int i_loc; 				MatDim BDM_; 
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d%c,%d,%dx%d,%d %d:%d %d,%s,a%d %d %d,a%d %d %d,a%d %d %d,%d)",
    DoTRSM, UpLo, Diag, RowCol, w, n, d_B0(BDM), d_B(BDM), d_P0(BDM), d_P(BDM), 
    pFD2Str(AsMScale), FD_ADR(pLM_ADR(AsM)), pLM_RInc(AsM), pLM_CInc(AsM), 
    FD_ADR(pLM_ADR(BsM)), pLM_RInc(BsM), pLM_CInc(BsM), 
    FD_ADR(pLM_ADR(B)),   pLM_RInc(B),   pLM_CInc(B), brdB),
    w>=1 && n>=0 && (brdB || Blk1d(w, BDM)), 0, "SpreadL2TR_M", 2, TRACE_ALL
  );
  if (n==0) {
    TRACE_RETURN;
  }  
  TRACE_LEVEL_SHIFT(2);
  
  BDM = pSubDim_(&BDM_, BDM, (DoTRSM^UpLo)? 0: w-1);
  i_loc = (DoTRSM^UpLo)? 0: (w_loc - (d_P0(BDM)==0));  
  
  for (iw=0; iw < w; iw++) {
    /* invariant: i_loc = G2Ld(i, BDM_pre) */	
    int i = (DoTRSM^UpLo)? iw: w-iw-1;	/* global relative index */
    int BDMiP0 = d_P0(BDM);   		/* cell-offset of diag cell */
    LocMat AsMi, BsMi, Bi, Bid;
    int i_locd, m;
    pLocSubMat_(&Bi, B, i_loc, 0);	pLocSubMat_(&BsMi, BsM, i, 0);
    if (DoTRSM && BDMiP0==0 && Diag!=Unit) {
      FDATA Aii_inv; 
      if (TRSMsBUnScal != NULL) 
         VecCopy(n, LM_ADR(Bi), LM_CInc(Bi), 
         	 LM_Index(TRSMsBUnScal_, i, 0), LM_CInc(TRSMsBUnScal_));
      FD_DVDSET(Aii_inv, pFD_VAL(AsMScale), *pLM_Index(AsM, i_loc, i));
      FD_NEG(Aii_inv, Aii_inv);
      MatScale(1, n, &Aii_inv, &Bi);
    }       
    if (DoTRSM && brdB) {
      FDATA *Bx = LM_ADR(BsMi);
      if (BDMiP0==0) {
        if (LM_Inc(1, Bi)==1)		/* optimization: send directly from B */
	  Bx = LM_ADR(Bi);
	else if (LM_ADR(Bi)!=LM_ADR(BsMi)) 
	  VecCopy(n, LM_ADR(Bi), LM_CInc(Bi), LM_ADR(BsMi), LM_CInc(BsMi));
      } else {
        COMMON_ASSERT(LM_ADR(Bi)!=Bx, 0, "SpreadL2TR_M:BsM=B", 6, TRACE_COMMUN);
      }
      BrdMat(RowCol, BDMiP0, 1, n, Bx, n);
      if (BDMiP0==0 && LM_ADR(Bi)!=LM_ADR(BsMi) && LM_Inc(1, Bi)==1)
	VecCopy(n, LM_ADR(Bi), LM_CInc(Bi), LM_ADR(BsMi), LM_CInc(BsMi));
      if (PrintRes)
        MatPrint("Bi", 1, MIN(PrintRes, n), BsMi);      	
    } 

    /* B[i1..i2,:] -= A[i1..12,i]*B[i,:], i1..i2 = UpLo? 0..i-1: i+1..w-1 */
    i_locd = i_loc;
    if (!UpLo && BDMiP0==0) {		/* advance past diagonal element */
      i_locd++;		
      pLocSubMat(&Bi, 1, 0);
    }  		
    pLocSubMat_(&AsMi, AsM, UpLo? 0: i_locd, i);
    m = UpLo? i_locd: w_loc - i_locd;
    if (m > 0) {
      R1Mult(m, n, AsMScale, &AsMi, &BsMi, UpLo? B: &Bi);
    }
    if (!DoTRSM && BDMiP0==0 && Diag!=Unit) 
      MatScale(1, n, pLM_Index(AsM, i_loc, i), pLocSubMat_(&Bi, B, i_loc, 0));
          	      
    i_loc += (DoTRSM^UpLo)? G2LSubDim(1, BDM): G2LSubDimM1(BDM);
  } /* for (iw...) */
  TRACE_LEVEL_SHIFT(-2);
  TRACE_EXIT;
} /* SpreadL2TR_M() */ 


int TRSM_CacheB=0;
DistMat TRSMsB;
static DistMatStore TRSMsB_;

DistMat TRMMsB = NULL;

static void TRSM_InitCacheSpreadB(M, N, B)
 int M; int N; DistMat B;
{ 
  if (Blk1d(M, DM_DM(B)) || TRSM_CacheB==0) 
    TRSMsB = B, TRSM_CacheB=0;
  else 
    TRSMsB = NewMat_(&TRSMsB_, M, N, ReduceDimMat_(&TRSMsB_, 0, B), NULL);
} /* TRSM_InitCacheSpreadB */

#define lA 	DM_Loc(A)
#define lB 	DM_Loc(B)
#define BDM 	DM(B)
#define DM(B) 	DM_D(  Side, B)
#define BDN 	DM_D(1-Side, B)

void TR_M(DoTRSM, UpLo, Diag, M, N, alpha, A, B)
  int DoTRSM; char UpLo, Diag; int M, N;
  FDATA *alpha; DistMat A, B;
{ int j, jF, dj;			/* iterate over j: (;j!=jF;j+=dj) */
  int Side = DM_Trans(B);		/* 0 = `Left'; 1 = `Right' */
  int UpLoTr = ((UpLo==Upper)^DM_Trans(A))^Side;
  int mB = G2Ld(M, DM_DM(B)),		nB = G2Ld(N, DM_DN(B));
  int PackB = 0;			LocMat oldlB;	/* used for packing B*/

  DistMatStore sA_, sB_;		DistMat sA = &sA_, sB = &sB_;
  FDATA *wsA = NULL, *wsB = NULL; 

  char BLACSBrdTopChar;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%c%c,%dx%d,%s,%d a%d %d %d,%d a%d %d %d)",
     DoTRSM, UpLo, Diag, M, N, pFD2Str(alpha), DM_Trans(A), FD_ADR(LM_ADR(lA)), LM_RInc(lA), 
     LM_CInc(lA), DM_Trans(B), FD_ADR(LM_ADR(lB)), LM_RInc(lB), LM_CInc(lB)),
    M>=0 && N>=0 && UpLoOK(UpLo) && DiagOK(Diag), 0, "TR_M", 1, TRACE_ALL
  );

  TRSM_InitCacheSpreadB(M, N, B);
  if (M==0 || N==0) {
    TRSM_CacheB = 0;	
    TRACE_RETURN;
  } 
  if (M==1 && Diag==Unit /* A is identity matrix */ || FD_EQ(*alpha, FD_0)) { 
    LocMat lBt;				LM_Copy(Side, lBt, lB);
    MatScale(G2Ld(M, BDM), G2Ld(N, BDN), alpha, &lBt);
    TRSM_CacheB = 0;	
    TRACE_RETURN;
  }

  PackB = 0;
  if (mB>1 && nB>=MIN_VECSIZE && M>=8 && 
      LM_Inc(1-Side, DM_Loc(B)) > 2*mB &&
      LM_Inc(1-Side, DM_Loc(B))*nB > CACHESIZE) {
    /* packing B should be worthwhile */
    FDATA *newB; 
    LM_Copy(Side, oldlB, DM_Loc(B));	/* oldB is mB x nB */
    PackB = MatMalloc("TR_M", 0, &newB, mB*nB);
    if (PackB) {
      LM_Init(DM_Loc(B), newB, nB, 1);	/* now mB x nB */
      MatCopy(mB, nB, &oldlB, NULL, DM_pLoc(B)); 
      LM_Trans(Side,  DM_Loc(B));		
    } 
  }

  if (M <= 2*PAR_INNER_VECSIZE) {
    /* pre-spread to reduce communication startups */	
    if (/*!Blk1d(M, DM_DM(A))*/ 1) {
      LocMat AsM; int AsMn;	
      LM_ADR(AsM) = NULL;	
      AsMn = SpreadMat(1-Side, M, M, Side? Trans(A): A, &FD_1, 
      		       BDM, &AsM, 0, N, BDN);
      if (AsMn > 0) 
      	wsA = LM_ADR(AsM);
      sA = Spread2DistMat_(sA, Side, 0, BDM, &AsM);
      A = Side? Trans(sA): sA;
    }
    if (!Blk1d(M, BDM) && !DoTRSM && TRMMsB==NULL) {
      LocMat BsM; int BsMn;
      LM_Copy(Side, BsM, DM_Loc(B));
      MatScale(mB, nB, alpha, &BsM);	/* scale all of B now */
      alpha = NULL;			/* prevent re-scaling later */
      LM_ADR(BsM) = NULL;
      BsMn = SpreadMat(Side, M, N, Side? Trans(B): B, &FD_1,
      		       BDN, &BsM, 0, M, BDM);
      if (BsMn > 0) 
      	wsB = LM_ADR(BsM);
      sB = Spread2DistMat_(sB, 1-Side, 0, BDN, &BsM);
      TRMMsB = Side? Trans(sB) : sB;          
    }
  }

  BLACSBrdTopChar = BrdTopChar[1-Side];
  BrdTopChar[1-Side] = 'I';		/* can use increasing-ring broadcast */

  if (UpLoTr==DoTRSM) { 		/* backward column/row iteration */  
    j = M-1; jF = -1; dj = -1; 
  } else {          			/* forward column/row iteration */
    j = 0;   jF = M;  dj =  1; 
  }

  while (j != jF) { 
    /* update B[j0..j1,:] by A[j0..j1,jL..jR];  [jL..jR] = spread of jL..jR */
    DistMat AjL, BjL;			/* SubMat(A, j0, jL), SubMat(B, jL, 0)*/
    DistMat_RVAlloc(2);
    int w, jL, jR;			/* w is blocking factor; jL+w-1 = jR  */
    int m; 				/* local length in AsM for level 3 mm */
    int w_loc;				/* local length for w */
    LocMat AsM, 			/* init. A[j0..j1, [jL..jR]] 	      */
	   BsM;                    	/* B[[jL..jR], :]		      */
    int AsMn, BsMn;    			/* = workspace lengths for AsM,BsM*/
    FDATA AsMScale = DoTRSM? FD_M1: FD_1;
    pLocMat AsMr1, 			/* A[jL..jR, [jL..jR]]		      */
   	    AsMmm; 			/* A[jL..jR, [jL..jR]]		      */
    LocMat AsM_,
	   lBr1, 			/* B[jL..jR, :]  		*/
	   lBmm;			/* B[j0mm..j1mm, :] 		*/
    int j0, j1/*, j0mm , j1mm*/;	/* see defn below; m ~ j1mm-j0mm+1 */
    MatDim BDMj0;			/* SubDim(BDM, j0) */

    w = GenGetW(UpLoTr? j+1: M-j, 1.0, PAR_INNER_VECSIZE_TRSM_MIN,  
    		MIN(ABS(jF-j), PAR_INNER_VECSIZE_TRSM), BDM, j, dj); 
    jL = dj>0? j: j-w+1;		jR = dj>0? j+w-1: j;
    if (UpLoTr) {
      j0 = 0; 		j1 = jR;	/* j0mm = 0;		j1mm = jL-1; */
    } else {
      j0 = jL; 		j1 = M-1; 	/* j0mm = jR+1;		j1mm = M-1; */
    }      

    BjL = SubMat(B, jL, 0);		
    if (Side)
      DM_Transpose(BjL); 		/* no transpose in SpreadMat() */
    w_loc = G2Ld(w, DM(BjL));
    LM_Copy(Side, lBr1, DM_Loc(BjL));	/* lBr1 is w_loc x n */
    if (UpLoTr) {
      LM_Copy(Side, lBmm, lB);		/* lBmm is m x nB, above lBr1 */
    } else {
      LM_Copy(Side, lBmm, DM_Loc(BjL));	
      pLocSubMat(&lBmm, w_loc, 0); 	/* lBmm is m x nB, below lBr1 */
    }  
    pSubDim_(&BDMj0, BDM, j0); 
    
    AjL = SubMat(A, j0, jL); 
    if (Side)                      	/* SpreadMat() requires transpose ...*/
      DM_Transpose(AjL);   		/* ... iff DM_Trans(A)==Side	     */
    LM_ADR(AsM) = NULL;
    AsMn = SpreadMat(1-Side, w, j1-j0+1, AjL, &AsMScale, &BDMj0, &AsM, 0, N, BDN);
    m = ABS(AsMn) - w_loc;

    LM_Trans(Side, AsM);		/* AsM is (m+w_loc) x w */
    if (UpLoTr) {
      AsMr1 = pLocSubMat_(&AsM_, &AsM, m, 0); 
      AsMmm = &AsM;
    } else {
      AsMr1 = &AsM;
      AsMmm = pLocSubMat_(&AsM_, &AsM, w_loc, 0);
    }  	  				/* AsMmm m x w, A[j0mm..j1mm, [jL..jR]]*/

    LM_ADR(BsM) = NULL;
    if (!DoTRSM) {
      if (!pFD_UNITY(alpha))	
        MatScale(w_loc, G2Ld(N, BDN), alpha, &lBr1);
      if (TRMMsB!=NULL) {
      	LM_Copy(Side, BsM, DM_Loc(TRMMsB)); 
      	pLocSubMat(&BsM, jL, 0);
      	BsMn = -G2Ld(N, BDN);		/* <= 0 to prevent freeing here */
      } else {
        BsMn = SpreadMat(Side, w, N, BjL, &FD_1, BDN, &BsM, 0, j1-j0+1, &BDMj0);
        LM_Trans(Side, BsM);		/* BsM is w x nB */	
      }  
      nB = ABS(BsMn);
      if (m > 0)
        MatPartMult(m, nB, w, NULL, AsMmm, &BsM, NULL, &lBmm);
    } else {
      if (TRSM_CacheB) {
      	LM_Copy(Side, BsM, DM_Loc(TRSMsB));
      	pLocSubMat(&BsM, jL, 0);
	BsMn = -G2Ld(N, BDN);           /* <= 0 to prevent freeing here */
      } else {	              	        
        BsMn = InitSpreadWS(Side, w, N, BjL, BDN, &BsM, TRSM_CacheB, j1-j0+1, 
      			    &BDMj0);
        LM_Trans(Side, BsM);		/* BsM is w x nB */
      }
      nB = ABS(BsMn);
      if (FD_EQ(AsMScale, FD_M1))
        MatScale(m+w_loc, w, &AsMScale, &AsM);
    }

    /* level 3 MM B[j0mm..j1mm,:] -= A[j0mm..j1,[jL..jR]] * B[[jL..jR],:]; 
       tri. update B[jL..jR,:] by A[jL..jR, [jL..jR]], B[[jL..jR],:]	*/ 
    if (TRSMsBUnScal != NULL)
      pLocSubMat_(&TRSMsBUnScal_, TRSMsBUnScal, jL, 0);      
    SpreadL2TR_M(DoTRSM, UpLoTr, Diag, Side, w, w_loc, nB, DM(BjL), 
    		 NULL, AsMr1, &BsM, &lBr1, 
    		 TRSM_CacheB | (1-Blk1d(j1-j0+1, &BDMj0)));

    if (DoTRSM) {
      if (m > 0)
        MatPartMult(m, nB, w, NULL, AsMmm, &BsM, NULL, &lBmm);
      if (!pFD_UNITY(alpha))	
        MatScale(w_loc, nB, alpha, &lBr1);
    }
    if (BsMn>0 && !TRSM_CacheB) 
      MatFree(LM_ADR(BsM));
    if (AsMn>0)
      MatFree(LM_ADR(AsM));      
    else if (FD_EQ(AsMScale, FD_M1))
      MatScale(m+w_loc, w, &AsMScale, &AsM);      
    j += w * dj;  
  } /* while (j...) */
  
  if (PackB) {
     LM_Trans(Side,  DM_Loc(B));		/* make mB x nB */	
     MatCopy(mB, nB, DM_pLoc(B), NULL, &oldlB); /* unpack */
     MatFree(LM_ADR(DM_Loc(B)));
     LM_Copy(Side, DM_Loc(B), oldlB);		/* restore B */
  }
  
  MatFree(wsA);					MatFree(wsB);
  BrdTopChar[1-Side] = BLACSBrdTopChar;
 
  if (DoTRSM)
    TRSM_CacheB = 0;
  else  
    TRMMsB = NULL;
    
   
  TRACE_EXIT;
} /* TR_M() */  

#undef lA 
#undef lB 
#undef BDM 
#undef BDN 
#undef DM
