/************************************************************************/
/* TRSM: DBLAS update B with (inverse of) triangular matrix A		*/
/************************************************************************/
/* SCCS INFO @(#)TRSM.h 1.3         last modified P. Strazdins, 99/11/17
   first created: P. Strazdins 93/15/01
*/
/* compilation: macro FDATATYPE must be defined;
 *              if macro ASSERT defined, assertion checking turned on
 *              if macro TRACE defined, parameter value tracing enabled
 *              if macro PRINTMAT defined, matrices are printed for debugging
 */
#ifndef __TRSM_INCL__
#define __TRSM_INCL__ 

#include "FDATASetUp.h" 
#include "DistMat.h"

#if (FDATATYPE==SINGLE)
#define SpreadL2TR_M	DSSpreadL2TR_M
#define TR_M		DSTR_M
#define TRSM_CacheB	DSTRSM_CacheB
#define TRSMsB		DSTRSMsB
#define TRMMsB		DSTRMMsB
#define TRSMsBUnScal	DSTRSMsBUnScal
#elif FDATATYPE==DOUBLE
#define SpreadL2TR_M	DDSpreadL2TR_M
#define TR_M		DDTR_M
#define TRSM_CacheB	DDTRSM_CacheB
#define TRSMsB		DDTRSMsB
#define TRMMsB		DDTRMMsB
#define TRSMsBUnScal	DDTRSMsBUnScal
#elif FDATATYPE==COMPLEX
#define SpreadL2TR_M	DCSpreadL2TR_M
#define TR_M		DCTR_M
#define TRSM_CacheB	DCTRSM_CacheB
#define TRSMsB		DCTRSMsB
#define TRMMsB		DCTRMMsB
#define TRSMsBUnScal	DCTRSMsBUnScal
#else
#define SpreadL2TR_M	DZSpreadL2TR_M
#define TR_M		DZTR_M
#define TRSM_CacheB	DZTRSM_CacheB
#define TRSMsB		DZTRSMsB
#define TRMMsB		DZTRMMsB
#define TRSMsBUnScal	DZTRSMsBUnScal
#endif 


extern int TRSM_CacheB;		/* if set, next call to TR_M(1, ...) will 
				   try to cache spread of B; its rep. will be 
				   put int TRSMsB. at end of call TRSM_CacheB==0
				   and TRSMsB will either be B (if no caching 
				   performed) or represent its spread ow */
extern DistMat TRSMsB;
    
   
extern DistMat TRMMsB;		/* if non-NULL, next call to TR_M() assumes
				   DDTRMMsB = Spread(DM_Trans(B), ..., B) */

extern pLocMat TRSMsBUnScal;	/* if non-null, keep copy of (spread of) B
				   before scaling by diaognal lements of A.
				   In this case, its workspace should
				   be externally allocated and N corresponds
				   to the number of columns.
				   This is only active if Diag!=Unit */
				      
void SpreadL2TR_M STDC_PARAMS(
        (int DoTRSM, int UpLo, char Diag, int RowCol, int w, int w_loc, int n, 
	 pMatDim BDM, FDATA * AsMScale, pLocMat AsM, pLocMat BsM, pLocMat B, 
	 int brdB)
);
/* pre:  AsM is a spread matrix of a global w x w matrix, of form UpLo
 *       and diagonal Diag. BsM is the spread matrix of B in dimension w;
 *       BDM specifies the dimension of B in the spread dimension.
 *       n is the local width of B and BsM; note w is a global size.
 *       w_loc = G2Ld(w, BDM)
 *       pFD_VAL(AsMScale)=+/-1; (-1 only if DoTRSM)
 * post: performs distributed update of w x n B by AsM and BsM
 *       if (DoTRSM), b/c of BsM rows only if brdB
 */


void TR_M STDC_PARAMS( 
	(int DoTRSM, char UpLo, char Diag, int M, int N, 
	 FDATA *alpha, DistMat A, DistMat B)
);
/* pre:  all cells are executing TR_M() with consistent 
 *	 DoTRSM, UpLo, Diag M, N, alpha, A, & B;
 *	 UpLoOK(UpLo); DiagOK(Diag); 0<= M, N
 * post: performs B := *alpha f(A)B 
 *       where f(A) = DoTRSM? Inv(A): A, and A is of form UpLo.
 *       with MxN B, 
 *       if Diag=Unit, diagonal elements of A assumed unity and not accessed
 * 	 The `left' variant of parallel TRSM/TRMM is the normal one here;
 *	 to achieve the `right' variant, interchange M & N and transpose A & B
 */

#endif /*__TRSM_INCL__*/
