/************************************************************************/
/* TriMatSolve: DBLAS Tri Mat Solve alg					*/
/************************************************************************/
/* SCCS INFO @(#)TriMatSolve.h 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 97/03
*/

#ifndef __TriMatSolve_INCL__
#define __TriMatSolve_INCL__
   
#include "LocMat.h"
#include "DistMat.h"

#if FDATATYPE==SINGLE
# define TriMatSolve		DSTriMatSolve
# define TriMatSolve1Blk	DSTriMatSolve1Blk
#elif FDATATYPE==DOUBLE
# define TriMatSolve		DDTriMatSolve
# define TriMatSolve1Blk	DDTriMatSolve1Blk
#elif FDATATYPE==COMPLEX
# define TriMatSolve		DCTriMatSolve
# define TriMatSolve1Blk	DCTriMatSolve1Blk
#else
# define TriMatSolve		DZTriMatSolve
# define TriMatSolve1Blk	DZTriMatSolve1Blk
#endif

void TriMatSolve STDC_PARAMS( 
  (char UpLo, char Diag, int M, int N, FDATA *alpha, DistMat A, DistMat B)
);
/* pre:  all cells are executing TriMatSolve with consistent
 *       UpLo, Diag M, N, alpha, A, & B;
 *       UpLoOK(UpLo); DiagOK(Diag); 0<= M, N
 *	 Dim_BCompat(DM_DM(B), DM_DM(A)), Dim_BCompat(DM_DM(B), DM_DN(A))
 * post: performs B := *alpha A^-1 B
 *       where A is of form UpLo with MxN B,
 *       if Diag=Unit, diagonal elements of A assumed unity and not accessed
 * notes: efficient for N<<M
 */


void TriMatSolve1Blk STDC_PARAMS( 
  (char UpLo, char Diag, int M, int N, FDATA *alpha, DistMat A, DistMat B)
);
/* pre:  as for TriMatSolve(), except Dim_BCompat(DM_DM(B), DM_DM(A))
 * 	 not required
 * post: as for TriMatSolve()
 * notes: more efficient than TriMatSolve() for Blk1d_(N, DM_DN(B)), esp N=1
 */

#endif /*__TriMatSolve_INCL__*/

