/************************************************************************/
/* TriMatUpdate: DBLAS Tri Mat Update (Multiply) alg				*/
/************************************************************************/
/* SCCS INFO @(#)TriMatUpdate.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 97/03
*/

#include "TriMatUpdate.h"

#include "BLASOptions.h"
#include "FDATAOps.h"
#include "FDAssert.h"

#include "LocMat.h"
#include "MatDim.h"
#include "DistMat.h"
#include "MachineParams.h"
#include "CellMatrix.h"
#include "CellTri1Matrix.h"
#include "SpreadMatrix.h"

#define ADM DM_DM(A)
#define ADN DM_DN(A)
#define BDM DM_DM(B)
#define BDN DM_DN(B)  
#define lA DM_Loc(A)
#define lB DM_Loc(B)

void TriMatUpdate(UpLo, Diag, M, N, alpha, A, B)
  char UpLo, Diag; int M, N;
  FDATA *alpha; DistMat A, B; 
{ int n;                         	
  int RowCol = DM_Trans(A);
  DistMatStore BN;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%c%c,%dx%d,%s,%d a%d %d %d,%d a%d %d %d)",
     UpLo, Diag, M, N, pFD2Str(alpha), DM_Trans(A), FD_ADR(LM_ADR(lA)), LM_RInc(lA), 
     LM_CInc(lA), DM_Trans(B), FD_ADR(LM_ADR(lB)), LM_RInc(lB), LM_CInc(lB)),
     CompatBlk(BDM, ADN) && M>=0 && N>=0 , 0, "TriMatUpdate", 1, TRACE_ALL
  ); 
 
  if (M==0 || N==0) {
    TRACE_RETURN;
  }

  if (DM_Trans(B))
    UpLo = TransForm(UpLo);
  n=0;
  while (n < N) {
    int mA /*= G2Ld(M, ADM)*/, 		kA /*= G2Ld(M, ADN)*/;
    FDATA ScAlpha = *alpha;
    int dN = MIN(PAR_OUTER_VECSIZE, N-n);
    DistMat Bt = RowCol==0? Trans(B): B, Ct = RowCol==0? B: Trans(B);	
    LocMat BsN, CsN;                	/* spread workspaces of B */
    int BsNn, CsNn;

    LM_ADR(CsN) = NULL;
    CsNn = InitReduceWS(1-RowCol, dN, M, Ct, &FD_0, ADM, &CsN, 1, M, ADN);    
    mA = ABS(CsNn);

    LM_ADR(BsN) = NULL;
    BsNn = SpreadMat(RowCol, dN, M, Bt, &ScAlpha, ADN, &BsN, 0, M, ADM);
    kA = ABS(BsNn);			

    LM_Trans(1-RowCol, BsN);		/* ensure BsN is kA x dN */
    LM_Trans(RowCol, CsN);		/* ensure CsN is mA x dN */
    Tri1MatPartMult(mA, dN, kA, &ScAlpha, UpLo, Diag, A, &BsN, NULL, &CsN); 
    if (BsNn>0)
      MatFree(LM_ADR(BsN));
    LM_Trans(RowCol, CsN);		/* restore CsN */

    ReduceMat(1-RowCol, dN, M, Ct, &FD_0, ADM, &CsN, M, ADN);
    if (CsNn>0)
      MatFree(LM_ADR(CsN));
      
    n += dN;
    if (n < N)
      B = SubMat_(&BN, B, 0, dN);
  } /* while (n...) */
  TRACE_EXIT;
} /*TriMatUpdate()*/
