/************************************************************************/
/*	VPPStride: implementation of strided messaging in VPPLib 	*/
/************************************************************************/
/* SCCS INFO @(#)VPPStride.c 1.2         last modified P. Strazdins, 99/07/06
   first created: P. Strazdins 99/03
 */
/* compilation: if DEBUG specified, tracing is enabled.
 *		
 */
#include <stdio.h>
#include <stdlib.h>

#ifdef DEBUG
#define PRINTF_(printf_args)	printf printf_args /* for debugging */
#else
#define PRINTF_(printf_args)	
#endif

#include "FDATAOps.h"
#include "VPPStride.h"
#include "FDAssert.h"
#include "LocMat.h"

#include "FDMalloc.h"
#include "FDAssert.h"
#include "FDCommun.h"
#include "CellMatCopy.h"
#include "CellVector.h"
#include "CellBlkMatrix.h"

#include "vpp_mplib.h"

#if defined(MPI)
#include "mpi.h"
#define MPIComm MPI_COMM_WORLD	/* temporary: assume communicator is this */
#endif

/************************* mpl memory alloc declarations ********************/

#include <time.h>               /* caddr_t */
 
extern int mpl_inplace, mpl_chunk_size;
 
enum mpl_comm_medium {MPL_SHARED_MEM, MPL_TCPIP_MEM, MPL_APNET_MEM,
                      MPL_DEVZERO_MEM};
 
enum mpl_alloc_type {MPL_RECVMEM, MPL_SENDMEM, MPL_FLAGMEM, MPL_RBMEM};
 
/*
 * Structure used to record a contiguous area of special memory.
 */
typedef struct mpl_memory {
    enum mpl_comm_medium type;  /* Type of special memory */
    int id;                     /* Identifier associated with this memory */
    int offset;                 /* Offset from the start of the chunk */
    int size;                   /* Size of this memory area */
    caddr_t addr;               /* Address this memory is mapped in */  /*use*/
    struct mpl_vaddr *vaddr;    /* Virtual address space allocation */
    struct mpl_memory_list *mlist; /* Free list to deallocate to */
    struct mpl_memory *link;    /* Link to next element */ 
} mpl_memory;
 
struct mpl_memory *mpl_alloc_mem(int size, enum mpl_comm_medium transport,
                                 enum mpl_alloc_type type);
struct mpl_memory *mpl_spec_alloc(int size, void * mlist, void * 
		   (*allocfn)());
void *mpl_apnet_get_mlist(enum mpl_alloc_type alloc_type, int size);
void *mpl_apnet_alloc();
extern void * _mpl_sdram_mlist, * _mpl_kmem_mlist;

int mpl_free_mem(struct mpl_memory *mem);
 
int mpl_getcid(void);
enum mpl_comm_medium mpl_get_transport(int);
const char *mpl_get_transport_string(int cellid);

/************************* special memory alloc code ********************/

void SpecMatMalloc(Send, BufPtr, HndlePtr, size)
  int Send, size; FDATA **BufPtr; void **HndlePtr; 
{ struct mpl_memory *mem;
  int sizeb = size*sizeof(FDATA);
  enum mpl_alloc_type type =  Send? MPL_SENDMEM: MPL_RECVMEM; 
#if 1
  mem = mpl_alloc_mem(sizeb, MPL_APNET_MEM, type); 
/* becomes: mpl_apnet_alloc_usermem(sizeb, type) */
/* becomes: mpl_spec_alloc(sizeb, mpl_apnet_get_mlist(type, sizeb),
                           &mpl_apnet_alloc) 
   nb. requires sizeb <= mpl_chunk_size.
   with:	    mpl_apnet_get_mlist(type, sizeb) = &_mpl_sdram_mlist 
   	    		iff type = MPL_SENDMEM && sizeb >= mpl_chunk_size
 */                          
		     
#else
  mem = mpl_spec_alloc(sizeb,
#if 1		/* much faster for small msgs */
 		mpl_apnet_get_mlist(sizeb, type),
#else
		Send? &_mpl_sdram_mlist: &_mpl_kmem_mlist,
#endif
 		
		mpl_apnet_alloc);
#endif
  *BufPtr = (FDATA *) mem->addr;		 	
  *HndlePtr = (void *) mem;		    
  PRINTF_(("SpecMatMalloc(%d=%d,a%d,a%d,%d)\n", Send, (int) type, FD_ADR(*BufPtr), (int) *HndlePtr, size)); 
} /* SpecMatMalloc() */

void SpecMatFree(HndlePtr)
  void *HndlePtr;
{ 
  PRINTF_(("SpecMatFree(%d)\n", (int) HndlePtr));
  mpl_free_mem((struct mpl_memory *) HndlePtr);
} /* SpecMatFree() */

/******************** copying code **************************************/

static void * MsgHndles[32];
static void * MsgHndles_Alt[32];

/********************* tree b/c code ********************************/
  	
#define USE_VecCopy(sz, SrcType, DestType)				\
	((SrcType)==OrdMem && ((sz) <= TOP_CACHESIZE || ((DestType)==OrdMem)))

#define VecCopy_(n, a, inca, b, incb)   memcpy(b, a, (n)*sizeof(FDATA))
 
#define MatCopy_(m, n, a, alpha, b)                                     \
{ int i_;  FDATA * A_ = pLM_ADR(a), *B_ = pLM_ADR(b);                   \
  for (i_=0; i_< (m); i_++) {                                           \
      VecCopy_(n, A_, 1, B_, 1);                          		\
          A_ += pLM_RInc(a),  B_ += pLM_RInc(b);              		\
  }                                                                     \
}

void ChkCpy(int sz, int chk_offs, int chk_offs_nxt, int chk_m, int N, 
		   enum MemType SrcType, LocMat *Src, enum MemType DestType, LocMat *Dst)
{ int UseVecCpy = USE_VecCopy(sz, SrcType, DestType);
  PRINTF_(("ChkCpy(%d,%d %d,%dx%d,%d a%d %d,%d a%d %d): UseVecCpy=%d\n", sz, chk_offs, chk_offs_nxt,
  		chk_m, N, (int) SrcType, FD_ADR(pLM_ADR(Src)),  pLM_RInc(Src),
  		(int) DestType, FD_ADR(pLM_ADR(Dst)), pLM_RInc(Dst), UseVecCpy));
  if ((chk_m)==0) {							
     int n_ = (chk_offs_nxt) - (chk_offs);				
     if (UseVecCpy) 
       VecCopy(n_, pLM_ADR(Src), 1, pLM_ADR(Dst), 1);
     else
       VecCopy_(n_, pLM_ADR(Src), 1, pLM_ADR(Dst), 1);       
     pLM_ADR(Src) += n_; pLM_ADR(Dst) += n_;		    		
  } else { 								
    int m_ = (chk_m);							
    if ((chk_offs) > 0) {						
      if (UseVecCpy) 
        VecCopy(N - chk_offs, pLM_ADR(Src), 1, pLM_ADR(Dst), 1);
      else 
        VecCopy_(N - chk_offs, pLM_ADR(Src), 1, pLM_ADR(Dst), 1);		
      pLM_ADR(Src) += pLM_RInc(Src) - (chk_offs); 				
      pLM_ADR(Dst) += pLM_RInc(Dst) - (chk_offs);				
      m_--; 
    }
    if (UseVecCpy)								
      MatCopy(m_, N, Src, NULL, Dst); 					
    else
      MatCopy_(m_, N, Src, NULL, Dst);  
    pLM_ADR(Src) += m_*pLM_RInc(Src);	pLM_ADR(Dst) += m_*pLM_RInc(Dst); 	
    if ((chk_offs_nxt) > 0) {		
      if (UseVecCpy)				
        VecCopy(chk_offs_nxt, pLM_ADR(Src), 1, pLM_ADR(Dst), 1);		
      else
        VecCopy_(chk_offs_nxt, pLM_ADR(Src), 1, pLM_ADR(Dst), 1);  
      pLM_ADR(Src) += (chk_offs_nxt);	pLM_ADR(Dst) += (chk_offs_nxt); 
    }	
  }									
} /* ChkCpy() */	


#define RECV_LA				/* post receives in advance for more
					   potential overlapping */

#define NUM_BRD_CHKS 4			/* number of chunks for b/c; 
					   in theory, this should be
					   >= 3 for correct RECV_LA , >= 2 ow;
					   in practice, need to add 1 to be safe */

/* a problem observed when multiple consecutive  calls to VPPTreeStrideBrd() 
   occur is that some messages appear to come in out of order, if I_RECV()
   used. A solution is add a offset depending on the message size
   0..NUM_MSG_TYPE_OFFS-1 to the message tag, to avoid confusion.	
   Note: it should be that NUM_MSG_TYPE_OFFS < differences between the
   message types contatns (currently 5) in FDCommun.h.
   Note: a cycling offset will not work, as there is no guarantee
   that cells will keep the same value of such an offset.
 */
#define NUM_MSG_TYPE_OFFS 4
#define MSG_OFFSET(sz)   (sz < 10? 0: (sz < 100? 1: (sz < 1000? 2: 3)))

int VPPTreeStrideBrd(Src, NDests, Dests, M, N, A, IncA)			
  int Src, NDests, Dests[], M, N, IncA; FDATA A[];
{ FDATA *SpecBuf[NUM_BRD_CHKS]; 	void *SpecBufHdle[NUM_BRD_CHKS];
  LocMat SpecA, 			lA;
  int sz = M*N, 			recvsz=0;
  int chk_sz = CHUNKSIZE(sz), 		num_chk,
      chk_m, 				chk_offs=0, chk_offs_nxt;
  int NoSpecRecvBuffs = (M<=1 || IncA==N)/* message is contiguous */
  		        && (chk_sz*sizeof(FDATA) <= mpl_inplace);
  		        		/* they will go direct to ring buffer */
  /* if NoSpecRecvBuffs, use mpl to allocate any special memory.
     Test indicate this is only marginally faster */
  int MsgTag = BRDMSGTYPE + MSG_OFFSET(sz);    
  int c, i;
  
  if (sz==0) return;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%d %d %d,%dx%d,a%d %d)", Src, NDests, Dests[0], Dests[1], 
      Dests[2], M, N, FD_ADR(A), IncA), 
      1, 0, "VPPTreeStrideBrd",  6, TRACE_COMMUN 
  ); 	
          
  num_chk = NUM_CHUNKS(sz, chk_sz);
  for (c=0; c < MIN(num_chk, NUM_BRD_CHKS); c++) 
    if (NoSpecRecvBuffs)
      SpecBuf[c] = A, SpecBufHdle[c] = NULL;
    else  
      SpecMatMalloc_Snd(&SpecBuf[c], &SpecBufHdle[c], chk_sz);

  LM_Init(lA, A, IncA, 1);
#ifdef RECV_LA
  if (Src >= 0)
    I_RECV(Src, chk_sz, SpecBuf[0], MsgTag, MsgHndles_Alt[0]);
#endif    
  for (c=0; c < num_chk; c++) {
    chk_sz = MIN(sz, chk_sz); 		sz -= chk_sz;
    chk_m = (chk_sz + chk_offs) / N,	chk_offs_nxt = (chk_offs + chk_sz) % N; 
    LM_Init(SpecA, SpecBuf[c%NUM_BRD_CHKS], N, 1);	
    if (Src < 0) { 			/* root of broadcast */
      if (!NoSpecRecvBuffs)
        ChkCpy(chk_sz, chk_offs, chk_offs_nxt, chk_m, N, OrdMem, &lA, SendMem, &SpecA); 
    } else {
#ifndef RECV_LA
      CTG_RECV(Src, chk_sz, SpecBuf[c%NUM_BRD_CHKS], MsgTag, recvsz);
#else
      WAIT_RECV(MsgHndles_Alt[c%NUM_BRD_CHKS], recvsz);
      if (c+1 < num_chk) {
        I_RECV(Src, MIN(sz, chk_sz), SpecBuf[(c+1)%NUM_BRD_CHKS], MsgTag, 
        	MsgHndles_Alt[(c+1)%NUM_BRD_CHKS]);	
      }      
#endif      
    }
    if (c > 0)	/* this should allow both copies to be overlapped with uncompleted sends */
      for (i=0; i < NDests; i++) 
        WAIT_SEND(MsgHndles[i]);

    for (i=0; i < NDests; i++) 
      I_SEND(Dests[i], chk_sz, SpecBuf[c%NUM_BRD_CHKS], MsgTag, MsgHndles[i]);
    
    if (Src>=0) {			/* non-leaf; store result */
      if (!NoSpecRecvBuffs)
        ChkCpy(chk_sz, chk_offs, chk_offs_nxt, chk_m, N, SendMem, &SpecA, OrdMem, &lA);
    }
    
    chk_offs = chk_offs_nxt;
  } /* for (c..) */
  for (i=0; i < NDests; i++) 		/* allow sends of prev itn to complete*/
    WAIT_SEND(MsgHndles[i]);
  if (!NoSpecRecvBuffs)  
    for (c=0; c < MIN(num_chk, NUM_BRD_CHKS); c++)
      SpecMatFree(SpecBufHdle[c]); 

  if (Src>=0 && recvsz != M*N) TraceLevel = 10;
  COMMON_ASSERT_TR(
    (TraceStr, "(%d,%d,%d %d %d,%dx%d,a%d %d) recvsz=%d, num_chk=%d lst chk_sz=%d SpecBufs=%d", 
                Src, NDests, Dests[0], Dests[1], 
                Dests[2], M, N, FD_ADR(A), IncA, recvsz, num_chk, chk_sz, !NoSpecRecvBuffs), 
      1/*Src<0 || recvsz == M*N*/, 0, "VPPTreeStrideBrd",  6, TRACE_COMMUN 
  ); 	

  TRACE_EXIT;  
  
  return(recvsz);
} /* VPPTreeStrideBrd() */


/*************************** tree reduce code *****************************/

void VecAdd_(int n, FDATA *a, FDATA *b) 
{ int i;
#if 0					/* slower */
  for (i=0; i < n; i++)
    FD_SETOP(b[i], b[i], +, a[i]);	
#else
  FDATA b0, b1, a0, a1, c0, c1;
  i = 0;
  if (n>=2) {
    a0 = a[0], a1 = a[1];
    b0 = b[0], b1 = b[1];
    for (i=2; i+1 < n; i+=2) {	
      FD_SETOP(c0, a0, +, b0); 		a0 = a[i];
      FD_SETOP(c1, a1, + , b1); 	a1 = a[i+1];
      b0 = b[i], 			b1 = b[i+1];
      b[i-2] = c0;			b[i-1] = c1;
    }
    FD_SETOP(c0, a0, +, b0);
    FD_SETOP(c1, a1, +, b1);       
    b[i-2] = c0;	b[i-1] = c1;
  }
  for (; i < n; i++)
    FD_SETOP(b[i], b[i], +, a[i]);	  	
#endif    
}

#define MatAdd_(m, n, a, alpha, b)					\
{ int i_;  FDATA * A_ = pLM_ADR(a), *B_ = pLM_ADR(b);			\
  for (i_=0; i_< (m); i_++) {						\
    VecAdd_(n, A_, B_);				\
    A_ += pLM_RInc(a),	B_ += pLM_RInc(b);		\
  }  									\
}									

#define USE_VecAdd(sz, SrcType, DestType)				\
	((SrcType)==OrdMem && ((sz) <= TOP_CACHESIZE || ((DestType)==OrdMem)))

void ChkAdd(int sz, int chk_offs, int chk_offs_nxt, int chk_m, int N, 
		   enum MemType SrcType, LocMat *Src, enum MemType DestType, LocMat *Dst)
{ int UseVecAdd = USE_VecAdd(sz, SrcType, DestType) || pLM_CInc(Dst)!=1;
  PRINTF_(("ChkAdd(%d,%d %d,%dx%d,%d a%d %d,%d a%d %d): UseVecAdd=%d\n", sz, chk_offs, chk_offs_nxt,
  		chk_m, N, (int) SrcType, FD_ADR(pLM_ADR(Src)),  pLM_RInc(Src),
  		(int) DestType, FD_ADR(pLM_ADR(Dst)), pLM_RInc(Dst), UseVecAdd));
  if ((chk_m)==0) {							
     int n_ = (chk_offs_nxt) - (chk_offs);				
     if (UseVecAdd) 
       VecAdd(n_, pLM_ADR(Src), 1, pLM_ADR(Dst), 1);
     else
       VecAdd_(n_, pLM_ADR(Src), pLM_ADR(Dst));       
     pLM_ADR(Src) += n_; pLM_ADR(Dst) += n_;		    		
  } else { 								
    int m_ = (chk_m);							
    if ((chk_offs) > 0) {						
      if (UseVecAdd) 
        VecAdd(N - chk_offs, pLM_ADR(Src), 1, pLM_ADR(Dst), 1);
      else 
        VecAdd_(N - chk_offs, pLM_ADR(Src), pLM_ADR(Dst));		
      pLM_ADR(Src) += pLM_RInc(Src) - (chk_offs); 				
      pLM_ADR(Dst) += pLM_RInc(Dst) - (chk_offs);				
      m_--; 
    }
    if (UseVecAdd)								
      MatAdd(m_, N, Src, NULL, Dst); 					
    else
      MatAdd_(m_, N, Src, NULL, Dst);  
    pLM_ADR(Src) += m_*pLM_RInc(Src);	pLM_ADR(Dst) += m_*pLM_RInc(Dst); 	
    if ((chk_offs_nxt) > 0) {		
      if (UseVecAdd)				
        VecAdd(chk_offs_nxt, pLM_ADR(Src), 1, pLM_ADR(Dst), 1);		
      else
        VecAdd_(chk_offs_nxt, pLM_ADR(Src), pLM_ADR(Dst));  
      pLM_ADR(Src) += (chk_offs_nxt);	pLM_ADR(Dst) += (chk_offs_nxt); 
    }	
  }									
} /* ChkAdd() */	


int VPPTreeStrideSum(Dest, NSrcs, Srcs, M, N, A, RIncA, CIncA)			
  int Dest, NSrcs, Srcs[], M, N, RIncA, CIncA; FDATA A[];
{ FDATA *AddBuf[2]; 			void *AddBufHdle[2];
  FDATA *RecvBuf[2]; 			void *RecvBufHdle[2];
  LocMat SpecA, SpecAdd,		srcA, AddA;
  int sz = M*N, 			recvsz=0;
  int chk_sz[2], 			num_chks,
      chk_m[2], 			chk_offs[2], chk_offs_nxt[2];
  int c, i, j;
  void *SndHndles[2],			*RecvHndles[2];
  int MsgTag = SUMMSGTYPE + MSG_OFFSET(sz);
  
  if (sz==0) return;  
    
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%d,%d %d %d,%dx%d,a%d %d %d)", Dest, NSrcs, Srcs[0], Srcs[1], 
      Srcs[2], M, N, FD_ADR(A), RIncA, CIncA), 
      1, 0, "VPPTreeStrideSum",  6, TRACE_COMMUN 
  ); 	
          
  chk_sz[0] = chk_sz[1] = CHUNKSIZE(sz);	chk_offs[0] = 0;
  num_chks = NUM_CHUNKS(sz, chk_sz[0]);

  for (j=0; j < 2; j++) {  
    /* leaf node: can send from SDRAM if large enough, as we only
       need to copy into that buffer; note that for adding into buffer
       or receiving, KMEM should be faster */
    SpecMatMalloc(NSrcs==0, &AddBuf[j], &AddBufHdle[j], chk_sz[j]);
    if (NSrcs >= 1)
      /* receive buffers to be added into AddBuf[j] (non-root) or
         into A directly (root) */
      SpecMatMalloc_Recv(&RecvBuf[j], &RecvBufHdle[j], chk_sz[j]);
    else
      RecvBuf[j] = NULL;  
  }

  LM_Init(srcA, A, RIncA, CIncA);
  c=0;  
  while (c < num_chks) {	
    int chk_pairs = (num_chks - c >= 2);  
    int jj = 0;
    for (j=0; j<=chk_pairs; j++) {    	
      chk_sz[j] = MIN(sz, chk_sz[j]); 	sz -= chk_sz[j];
      chk_m[j] = (chk_sz[j] + chk_offs[j]) / N;
      chk_offs_nxt[j] = (chk_offs[j] + chk_sz[j]) % N;
       if (j==0)
        chk_offs[1] = chk_offs_nxt[j]; 
    } /* for (j...) */   

    if (Dest >= 0)			/* copy this cells chunks of A into 
    					   send buffer */
      for (j=0; j<=chk_pairs; j++) {	
      	LocMat SendA;			LM_Init(SendA, AddBuf[j], N, 1);
      	if (c > 0)			/* ensure AddBuf[j] free to be overwritten*/
      	  WAIT_SEND(SndHndles[j]);
	  ChkCpy(chk_sz[j], chk_offs[j], chk_offs_nxt[j], chk_m[j], N, OrdMem, &srcA, SendMem, &SendA);
        if (NSrcs == 0)  		/* leaf node; send immediately */
          I_SEND(Dest, chk_sz[j], AddBuf[j], MsgTag, SndHndles[j]);
      } /* for (j...) */ 

    if (NSrcs > 0) {			/* non-leaf; receive messages */
      for (j=MIN(c,1); j<=chk_pairs; j++) {
      	/* lookahead is over next chunk from 1st source */
      	FDATA *RecvPtr = /*Dest<0*/ 1 ? RecvBuf[j]: AddBuf[j];
        I_RECV(Srcs[0], chk_sz[j], RecvPtr, MsgTag, RecvHndles[j]);
      }
      if (!chk_pairs && NSrcs>1) {	
        /* otherwise, lookahead for 1st message of next chunk  */
        I_RECV(Srcs[1], chk_sz[1], RecvBuf[1], MsgTag, RecvHndles[1]);
      }
    }
    jj=0;
    
    for (i=0; i<NSrcs; i++) {
      int i_recv_la = chk_pairs? 1: 2;	/* degree of lookahead for receives */
      LocMat srcA_ = srcA;		/* to prevent premature update on root*/
      for (j=0; j<=chk_pairs; j++) {	      
        WAIT_RECV(RecvHndles[jj], recvsz);
        if (Dest < 0) {          	/* at root node; add msg to srcA */
      	  LocMat RecvA;			LM_Init(RecvA, RecvBuf[jj], N, 1);
  	  ChkAdd(chk_sz[j], chk_offs[j], chk_offs_nxt[j], chk_m[j], N, RecvMem, &RecvA, OrdMem, &srcA);      	  
      	} else				/* at intermediate node */
      	  if (USE_VecAdd(chk_sz[j], RecvMem, RecvMem))
      	    VecAdd(chk_sz[j], RecvBuf[jj], 1, AddBuf[j], 1);  
      	  else
      	    VecAdd_(chk_sz[j], RecvBuf[jj], AddBuf[j]);        	    
        if (i+i_recv_la < NSrcs) {      
          I_RECV(Srcs[i+i_recv_la], chk_sz[j], RecvBuf[jj], MsgTag, RecvHndles[jj]);	
        } else if (i == NSrcs-1) {
          if (Dest>=0) {		/* non-leaf; send on partial sum */
             I_SEND(Dest, chk_sz[j], AddBuf[j], MsgTag, SndHndles[j]); 
          } 
          if (j==0 && c+2<num_chks) {/* lookahead to next chunk */
      	    FDATA *RecvPtr = RecvBuf[jj];
            chk_sz[j] = MIN(chk_sz[j], sz);	  
            I_RECV(Srcs[0], chk_sz[j], RecvPtr, MsgTag, RecvHndles[jj]);
          }  
        }  
      jj++; jj %= 2;
      } /* for (j...) */
      if (i < NSrcs-1 && Dest<0)	/* reset for multiple messages */
        srcA = srcA_;  
    } /* for (i..) */
    
    chk_offs[0] = chk_offs_nxt[1];
    c += chk_pairs + 1;
  } /* while (c..) */

  if (Dest>=0)
    for (j=0; j < MIN(num_chks, 2); j++) {	
      WAIT_SEND(SndHndles[j]);
    }
    
  for (j=0; j < 2; j++) {    
    SpecMatFree(AddBufHdle[j]); 
    if (NSrcs >= 1)
      SpecMatFree(RecvBufHdle[j]);
  } /* for (j...) */
  TRACE_EXIT;
  return(NSrcs==0? M*N: recvsz/NSrcs);
} /* VPPTreeStrideSum() */




/*************************** ring combine code *****************************/

#include "MatDim.h"
#include "MachineParams.h"

#define REL2ABS1CID(RCIDI, RCIDJ)                                       \
        (CYCLIC_ADD(AbsCellId[0], (RCIDI), NumCells[0])*NumCells[1] +     \
         CYCLIC_ADD(AbsCellId[1], (RCIDJ), NumCells[1]))
	/* AbsCellId[], NumCells[] exported from FDCommun.h */

#define CALC_CHK_PARAMS(chunk_size, N, SegSz, ChkOffs, ChkSz, Chk_m, ChkOffsNxt) \
      (ChkSz) = MIN(SegSz, chunk_size), Chk_m = ((ChkSz) + (ChkOffs)) / (N),  \
      (ChkOffsNxt) = ((ChkOffs) + (ChkSz)) % (N)
#define UPDATE_CHK_PARAMS(SegSz, ChkOffs, SegPtr, ChkSz, ChkOffsNxt, NewPtr)  \
      (SegSz) -= (ChkSz), (ChkOffs) = (ChkOffsNxt), (SegPtr) = (NewPtr)
      
/*#define EXTRA_RECV_LA*/		/* more aggressive lookahead on posting
					   receives; on average, results in a
					   marginal improvement */

void StrideRingComb(RowCol, Add, D, M, n, A)
  int RowCol, Add, M, n;  pMatDim D; FDATA A[];
/* pre:  RCOK(RowCol); Dim_OK(D); Inc(A) = {n,1}; d_P(D) > 1 
 *	 all (d_P(D)) cells in scope of RowCol are participating in this call; 
 *	 M, n and D, are consistent across all these cells; 
 * post: if Add, in cell (I+i)%P, the ith row block of A contains the sum of 
 *	 that in all cells in the scope, 0 <= i <= NBLKS(M, B0, B)
 *       else, the ith row block of A the value of that in cell (I+i)%P
 */
{ LocMat lAj;				int c, i, j;
  int m_mx = G2Ld_Mx(M, D),		sz = m_mx*n;
  int chunk_sz = CHUNKSIZE(sz),         num_chk = NUM_CHUNKS(sz, chunk_sz);

  int N = n*d_B(D);
  int Src = REL2ABS1CID(RowCol==0? d_P(D)+RowCol-1: 0, RowCol==0? 0: d_P(D)-RowCol),
      Dst = REL2ABS1CID(1-RowCol, RowCol);
  int MsgTag = (Add? SUMMSGTYPE: BRDMSGTYPE) + MSG_OFFSET(sz);
  
#define NUM_SEND_BUFS 2			/* max # buffers need for pipelining */
  /* buffers and parameters refering to a window of messages */
  int SndChkIx, NumSndBufs;		/* 0<=SndChkIx<NumSndBufs<=NUM_SEND_BUFS*/
  FDATA *SndBuf[NUM_SEND_BUFS];		/* buffers for source messages and ...*/
  void *SndBufMsgHndl[NUM_SEND_BUFS]; 	/* ... assoc memory & message handles */
  void *SndBufMemHndl[NUM_SEND_BUFS];
  
  int NoSpecRecvBuffs = !Add && m_mx==1 /* all messages are contiguous */
  		        && (chunk_sz*sizeof(FDATA) <= mpl_inplace);
  		        		/* they will go direct to ring buffer */
  		        		/* In this case, receive directly into A */
  		        		  
#define NUM_CHKS 4			/* max # of special memory to use; >2*/
  int RecvChkIx, NumRecvBufs;		/* 0<=RecvChkIx<NumRecvBufs<NUM_CHKS */
  FDATA *SpecBuf[NUM_CHKS];		/* special memory buffers for msgs...*/
  void *SpecBufSndHdle[NUM_CHKS];	/* ... received then passed on .... */
  void *SpecBufMemHdle[NUM_CHKS];

  int *SegSz, *SegChkOffs; 		/* arrays [0..d_P(D)-1]; remaining...*/
  FDATA **SegA;  			/* ... size, chunk offset & pointers
  					  corresp. to each cells contribution */
  int ChkSz_SrcNxt;			/* records size of chunk for next source send */

  SegSz = (int *) alloca(d_P(D)*sizeof(int));
  SegChkOffs = (int *) alloca(d_P(D)*sizeof(int));
  SegA = (FDATA **) alloca(d_P(D)*sizeof(FDATA *));
  for (j=0; j < d_P(D); j++) {
    SegSz[j] = G2Ld_P0(M, D, j)*n;
    SegChkOffs[j] =  BlkB0(j, d_B0(D))*n;
    SegA[j] = A+(BlkIxd_(j, D))*n;
  }  	

  NumRecvBufs = MIN(NUM_CHKS, num_chk*(d_P(D)-1));
  for (RecvChkIx=0; RecvChkIx < NumRecvBufs; RecvChkIx++) {
    if (NoSpecRecvBuffs)
      SpecBuf[RecvChkIx] = NULL, SpecBufMemHdle[RecvChkIx] = NULL;	
    else  
      SpecMatMalloc(!Add, &SpecBuf[RecvChkIx], &SpecBufMemHdle[RecvChkIx], chunk_sz);
    SpecBufSndHdle[RecvChkIx] = NULL;
  }
  NumSndBufs = MIN(NUM_SEND_BUFS, num_chk);
  for (SndChkIx=0; SndChkIx < NumSndBufs; SndChkIx++) 
    SpecMatMalloc(1, &SndBuf[SndChkIx], &SndBufMemHndl[SndChkIx], chunk_sz);  

  SndChkIx = RecvChkIx = 0;
  LM_Init(lAj, NULL, N*d_P(D), 1);
  for (c=0; c < num_chk; c++) {
    int ChkSz, Chk_m, ChkOffsNxt;
    LocMat SpecA;		
    int LstRecvChkIx;		/* records RecvChkIx from last itn */
#ifdef EXTRA_RECV_LA
    void *NxtRecvHndle = NULL;
#endif
    j = Add? CYCLIC_DEC(d_P0(D), d_P(D)): d_P0(D); 	
    LM_ADR(lAj) = SegA[j];
      
    CALC_CHK_PARAMS(chunk_sz, N, SegSz[j], SegChkOffs[j], ChkSz, Chk_m, ChkOffsNxt);
    PRINTF_(("c=%d, chk sz=%d m=%d onxt=%d\n", c, ChkSz, Chk_m, ChkOffsNxt));
    for (i=1; i < d_P(D); i++) { 
      int Nxtj = CYCLIC_DEC(j, d_P(D));
      void *RecvHndle = NULL;
      FDATA *RecvBuf = NoSpecRecvBuffs? SegA[Nxtj]: SpecBuf[RecvChkIx];

#ifdef EXTRA_RECV_LA
      if (i==1) {
#endif
        /* post next receive in advance to promote overlapping */      
        if (c*(d_P(D)-1) + i-1 >= NumRecvBufs && SpecBufSndHdle[RecvChkIx]!=NULL){	
          /* ensure that the prev. send from this buffer has completed */
          WAIT_SEND(SpecBufSndHdle[RecvChkIx]);
        }
        SpecBufSndHdle[RecvChkIx] = NULL;
        I_RECV(Src, MIN(SegSz[Nxtj],chunk_sz), RecvBuf, MsgTag, RecvHndle);
        PRINTF_(("I_RECV(%d, %d, a%d, _, %d) j'=%d\n", Src, MIN(SegSz[Nxtj],chunk_sz),
      	         FD_ADR(RecvBuf), RecvHndle, Nxtj));  
#ifdef EXTRA_RECV_LA
      } else
        RecvHndle = NxtRecvHndle;

      if (i < d_P(D)-1) {
        /* post receive of following itn in advance to more aggressively 
           promote overlapping */      
        int Nxt2j = CYCLIC_DEC(Nxtj, d_P(D));
        int NxtRecvChkIx = (RecvChkIx+1)% NumRecvBufs;
        FDATA *NxtRecvBuf = NoSpecRecvBuffs? SegA[Nxt2j]: SpecBuf[NxtRecvChkIx];
        if (c*(d_P(D)-1) + i >= NumRecvBufs && SpecBufSndHdle[NxtRecvChkIx]!=NULL){	
          /* ensure that the prev. send from this buffer has completed */
          WAIT_SEND(SpecBufSndHdle[NxtRecvChkIx]);
        }
        SpecBufSndHdle[NxtRecvChkIx] = NULL;
        I_RECV(Src, MIN(SegSz[Nxt2j],chunk_sz), NxtRecvBuf, MsgTag, (int) NxtRecvHndle);
        PRINTF_(("I_RECV(%d, %d, a%d, _, %d) j''=%d\n", Src, MIN(SegSz[Nxt2j],chunk_sz),
      	         FD_ADR(NxtRecvBuf), NxtRecvHndle, Nxt2j));        	
      } 	         
#endif

      if (i==1) {
      	if (c==0) {
          LM_Init(SpecA, SndBuf[SndChkIx], N, 1);
          ChkCpy(ChkSz, SegChkOffs[j], ChkOffsNxt, Chk_m, N, OrdMem, &lAj, 
          	 SendMem, &SpecA);
	  ChkSz_SrcNxt = ChkSz;	 
          UPDATE_CHK_PARAMS(SegSz[j], SegChkOffs[j], SegA[j], 
          		    ChkSz, ChkOffsNxt, LM_ADR(lAj));
          CALC_CHK_PARAMS(chunk_sz, N, SegSz[j], SegChkOffs[j], ChkSz, Chk_m, ChkOffsNxt);
        }
        I_SEND(Dst, ChkSz_SrcNxt, SndBuf[SndChkIx], MsgTag, SndBufMsgHndl[SndChkIx]);
        PRINTF_(("i=%d I_SEND(%d, %d, a%d, _, h%d)\n", i, Dst, ChkSz_SrcNxt, FD_ADR(SndBuf[SndChkIx]), (int) SndBufMsgHndl[SndChkIx]));
        SndChkIx = (SndChkIx+1) % NUM_SEND_BUFS;
        if (c < num_chk-1) {
          LM_Init(SpecA, SndBuf[SndChkIx], N, 1);
          if (c+1 >= NUM_SEND_BUFS) {
            /*ensure prev. send from SndBuf[SndChkIx] compltete*/
            PRINTF_(("WAIT_SEND(a%d)\n", (int) SndBufMsgHndl[SndChkIx]));
            WAIT_SEND(SndBufMsgHndl[SndChkIx]);
          }
	  ChkCpy(ChkSz, SegChkOffs[j], ChkOffsNxt, Chk_m, N, OrdMem, &lAj, 
	  	 SendMem, &SpecA);
	  ChkSz_SrcNxt = ChkSz;	 
        }
      } else {
	FDATA *SndBuf = NoSpecRecvBuffs? SegA[j]: SpecBuf[LstRecvChkIx];
        I_SEND(Dst, ChkSz, SndBuf, MsgTag, SpecBufSndHdle[LstRecvChkIx]);
        PRINTF_(("i=%d I_SEND(%d, %d, a%d, _, h%d)\n", i, Dst, ChkSz, FD_ADR(SndBuf), (int) SpecBufSndHdle[LstRecvChkIx]));
        if (!Add && !NoSpecRecvBuffs) {
          LM_Init(SpecA, SpecBuf[LstRecvChkIx], N, 1);
	  ChkCpy(ChkSz, SegChkOffs[j], ChkOffsNxt, Chk_m, N, 
	         RecvMem, &SpecA, OrdMem, &lAj);
	}         
      } 
      PRINTF_(("UPDATE_CHK_PARAMS\n"));
      UPDATE_CHK_PARAMS(SegSz[j], SegChkOffs[j], SegA[j], 
      		        ChkSz, ChkOffsNxt, LM_ADR(lAj));
      PRINTF_(("c=%d j=%d, SegSz=%d, SegChkOffs=%d, SegA=a%d\n", 
    	       c, j ,SegSz[j], SegChkOffs[j], FD_ADR(SegA[j])));
     
      j = Nxtj;
      CALC_CHK_PARAMS(chunk_sz, N, SegSz[j], SegChkOffs[j], 
      		      ChkSz, Chk_m, ChkOffsNxt);
      PRINTF_(("c=%d i=%d, chk sz=%d m=%d onxt=%d\n", c, i, ChkSz, Chk_m, ChkOffsNxt));
      PRINTF_(("WAIT_RECV(%d, %d)\n", RecvHndle, ChkSz));
      ChkSz = 0;
      WAIT_RECV(RecvHndle, ChkSz);
      PRINTF_(("done: WAIT_RECV(%d)\n", ChkSz));

      LM_ADR(lAj) = SegA[j];
      LM_Init(SpecA, SpecBuf[RecvChkIx], N, 1);
      if (Add)
        if (i < d_P(D)-1)
          ChkAdd(ChkSz, SegChkOffs[j], ChkOffsNxt, Chk_m, N, OrdMem, &lAj, 
          	 RecvMem, &SpecA);
      LstRecvChkIx = RecvChkIx;
      RecvChkIx = (RecvChkIx + 1) % NumRecvBufs;
      
    } /* for (i...) */
    
    if (!NoSpecRecvBuffs)
      (Add ? ChkAdd: ChkCpy) (ChkSz, SegChkOffs[j], ChkOffsNxt, Chk_m, N, 
          	              Add? RecvMem: SendMem, &SpecA, OrdMem, &lAj);
           
    UPDATE_CHK_PARAMS(SegSz[j], SegChkOffs[j], SegA[j], 
    		      ChkSz, ChkOffsNxt, LM_ADR(lAj));
    PRINTF_(("end c=%d j=%d, SegSz=%d, SegChkOffs=%d, SegA=a%d\n", 
    	     c, j ,SegSz[j], SegChkOffs[j], FD_ADR(SegA[j])));		      
  } /* for (c...) */	

   
  SndChkIx = (SndChkIx+1) % NUM_SEND_BUFS;
  PRINTF_(("SndChkIx=%d; WAIT_SEND(%d)\n", SndChkIx, (int) SndBufMsgHndl[SndChkIx]));
  WAIT_SEND(SndBufMsgHndl[SndChkIx]);	/* wait for all sends to complete */ 
  for (c=0; c < NumRecvBufs; c++) { 	
    if (SpecBufSndHdle[RecvChkIx] != NULL) {
      PRINTF_(("RecvChkIx=%d, WAIT_SEND(%d)\n", RecvChkIx, (int) (SpecBufSndHdle[RecvChkIx])));
      WAIT_SEND(SpecBufSndHdle[RecvChkIx]);
    }  
    RecvChkIx = (RecvChkIx + 1) % NumRecvBufs;
  }  
  for (SndChkIx=0; SndChkIx < NumSndBufs; SndChkIx++) 
    SpecMatFree(SndBufMemHndl[SndChkIx]);  
  if (!NoSpecRecvBuffs)  
    for (RecvChkIx=0; RecvChkIx < NumRecvBufs; RecvChkIx++) 
      SpecMatFree(SpecBufMemHdle[RecvChkIx]);   
}  

