/************************************************************************/
/*	VPPStride: implementation of strided messaging in VPPLib 	*/
/************************************************************************/
/* SCCS INFO @(#)VPPStride.h 1.1         last modified P. Strazdins, 99/07/01
   first created: P. Strazdins 99/03
 */
/* compilation: if DEBUG specified, tracing is enabled.
 *		
 */

#include "FDATASetUp.h"
#include "LocMat.h"
#include "MatDim.h"

#if FDATATYPE==SINGLE
#define SpecMatMalloc 		SSpecMatMalloc
#define SpecMatFree		SSpecMatFree
#define ChkCpy			SChkCpy
#define VPPTreeStrideBrd 	SVPPTreeStrideBrd	
#define ChkAdd			SChkAdd
#define VPPTreeStrideSum	SVPPTreeStrideSum
#define StrideRingComb		SStrideRingComb
#elif FDATATYPE==DOUBLE
#define SpecMatMalloc 		DSpecMatMalloc
#define SpecMatFree		DSpecMatFree
#define ChkCpy			DChkCpy
#define VPPTreeStrideBrd 	DVPPTreeStrideBrd	
#define ChkAdd			DChkAdd
#define VPPTreeStrideSum	DVPPTreeStrideSum
#define StrideRingComb		DStrideRingComb
#elif FDATATYPE==COMPLEX
#define SpecMatMalloc 		CSpecMatMalloc
#define SpecMatFree		CSpecMatFree
#define ChkCpy			CChkCpy
#define VPPTreeStrideBrd 	CVPPTreeStrideBrd	
#define ChkAdd			CChkAdd
#define VPPTreeStrideSum	CVPPTreeStrideSum
#define StrideRingComb		CStrideRingComb
#else
#define SpecMatMalloc 		ZSpecMatMalloc
#define SpecMatFree		ZSpecMatFree
#define ChkCpy			ZChkCpy
#define VPPTreeStrideBrd 	ZVPPTreeStrideBrd	
#define ChkAdd			ZChkAdd
#define VPPTreeStrideSum	ZVPPTreeStrideSum
#define StrideRingComb		ZStrideRingComb
#endif


/************************* external mpl memory declarations ***************/

extern int mpl_inplace, mpl_chunk_size;
 
/************************* special memory alloc code ********************/

#define SpecMatMalloc_Snd(bufptr, hndlptr, sz)				\
	SpecMatMalloc(1, bufptr, hndlptr, sz)

#define SpecMatMalloc_Recv(bufptr, hndlptr, sz)				\
	SpecMatMalloc(0, bufptr, hndlptr, sz)

void SpecMatMalloc STDC_PARAMS(
  (int Send, FDATA **BufPtr, void **HndlePtr, int size)
);
/* pre:  size*sizeof(FDATA) <= mpl_chunk_size
 * post: allocate a chunk of special memory of size at least size words
 *	 start address at *BufPtr; *HndlePtr must be used to free this
 *	 chunk. If Send, the buffer is assumed to be used to send messages;
 *	 (and will possibly be SDRAM, depending on value of size); 
 *        otherwise it is intended for receiving messages (probably KMEM)
 */

void SpecMatFree  STDC_PARAMS((void *HndlePtr));
 /* post: deallocate a chunk of special memory of handle HndlePtr
  */
/************************* communication macros **************************/

#ifdef VPPLIB

# define WAIT_SEND(hndle)	 VPP_SendDone((TMSG *) (hndle))

# define BLK_SEND(DEST, N, A, M_TYPE) 					\
	WAIT_SEND(VPP_Send((PROC) DEST, M_TYPE, 			\
		     (ADDRP) (A), (size_t) (N)*sizeof(FDATA)))

# define I_SEND(DEST, N, A, M_TYPE, HNDLE) 				\
	(HNDLE) = (void *) (VPP_Send((PROC) (DEST), M_TYPE, 		\
		     (ADDRP) (A), (size_t) (N)*sizeof(FDATA)))


# define I_RECV(SRC, sz, A, M_TYPE, hndle) 				\
   	(hndle) = VPP_Recv((PROC) (SRC), M_TYPE, (ADDRP) (A),		\
	  		   (size_t) (sz)*sizeof(FDATA)) 

# define WAIT_RECV(hndle, r)						\
	{ int t_, sz__;							\
	  VPP_RecvDone((TMSG *) (hndle), (PROC*) &t_, &t_, (size_t *) &sz__);\
	  (r) += sz__/sizeof(FDATA);					\
	}	

# define CTG_RECV(SRC, sz, A, M_TYPE, r)                       		\
        { int t_, sz__;                                                 \
          VPP_RecvDone(                                                 \
            VPP_Recv((PROC) SRC, M_TYPE, (ADDRP) (A),			\
            (size_t) (sz)*sizeof(FDATA)), (PROC*) &t_, &t_, (size_t *)&sz__);\
          (r) += sz__/sizeof(FDATA);                                    \
        }


#elif defined(MPI)	

/*#include "mpi.h"*/			/* problematic when included in some
					   DBLAS files (DistMatix.c) as Makefile
					   currenntly does not add MPI include path */
extern int /*MPI_Comm*/ MPIComm;	/* set this before using any routine */

#   define WAIT_SEND(hndle)                                             \
        { MPI_Status status_; MPI_Wait((MPI_Request *) &(hndle), &status_); }
         
#   define I_SEND(DEST, N, A, M_TYPE, HNDLE)                    	\
        MPI_Isend((void *) (A), (N)*sizeof(FDATA), MPI_BYTE,            \
                  DEST, M_TYPE, MPIComm, (MPI_Request *) &(HNDLE))
                                                    
# define I_RECV(SRC, N, A, M_TYPE, HNDLE) 				\
   	MPI_Irecv((void *) (A), (N)*sizeof(FDATA), MPI_BYTE,		\
		  SRC, M_TYPE, MPIComm, (MPI_Request *) &(HNDLE))

# define WAIT_RECV(hndle, r)						\
        { MPI_Status status_; int r_; MPI_Wait((MPI_Request *) &(hndle), &status_);\
	  MPI_Get_count(&status_, MPI_BYTE, &r_);			\
	  (r) += r_/sizeof(FDATA);					\
	}	
                                                   
#   define CTG_RECV(SRC, sz, A, M_TYPE, r)                       	\
        { MPI_Status status_;                                           \
          MPI_Recv((void *) (A), (sz)*sizeof(FDATA), MPI_BYTE,          \
                   SRC, M_TYPE, (MPI_Comm) MPIComm, &status_);		\
          (r) += (sz);                                                  \
        }
#endif

/************************* chunk calculation macros *************************/
 
#define CHUNKSIZE(sz) 		MIN(sz, mpl_chunk_size/sizeof(FDATA))
	/* returns the maximum size of chunks for message of size sz */
	
#define NUM_CHUNKS(sz, chk_sz)	(((sz) + (chk_sz)-1)/(chk_sz))
	/* returns the number of chunks required for a message of size sz,
	   given a maximum chunk size chk_sz */

#define CALC_CHK_PARAMS(chk_sz, N, RemSz, ChkOffs, ChkSz, Chk_m, ChkOffsNxt) \
      (ChkSz) = MIN(RemSz, chk_sz), Chk_m = ((ChkSz) + (ChkOffs)) / (N),     \
      (ChkOffsNxt) = ((ChkOffs) + (ChkSz)) % (N)
	/* for chunks of (max) size chk_sz, a matrix width of N, a remaining
	   message size of RemSz, and a current chunk offset of ChkOffs,
	   caclulate the parameters required for passing to ChkCpy/Add() 
	*/	

#define UPDATE_CHK_PARAMS(RemSz, ChkOffs, /* FDATA* */ ChkPtr, ChkSz, ChkOffsNxt, /* FDATA* */ NewPtr)  \
      (RemSz) -= (ChkSz), (ChkOffs) = (ChkOffsNxt), (ChkPtr) = (NewPtr)
        /* update chunk remaining size, offset  for the next chunk;
           ChkPtr is the pointer used to access the current chunk of the
           message; NewPtr is a pointer pointing to the next chunk 
	*/

/******************** copying code **************************************/

enum MemType { OrdMem, SendMem /* eg SDRAM */, RecvMem /* eg KMEM */};

#define TOP_CACHESIZE		(/*16384*/8192 / sizeof(FDATA))


#define USE_VecCopy(sz, SrcType, DestType)				\
	((SrcType)==OrdMem && ((sz) <= TOP_CACHESIZE || ((DestType)==OrdMem)))

void ChkCpy STDC_PARAMS(
	(int sz, int chk_offs, int chk_offs_nxt, int chk_m, int N, 
         enum MemType SrcType, LocMat *Src, enum MemType DestType, LocMat *Dst)
);
/* pre:  0 <= chk_offs,chk_offs_nxt < N; 
 *	 if chk_m==0, sz = chk_offs_nxt - chk_offs
 *	 else sz = N-chk_offs + (chk_m-1)*N + chk_offs_nxt
 *	 LM_CInc(*Src) = LM_RInc(*Dst) = 1;  LM_RInc(*Src), LM_RInc(*Dst) >= N 
 * post: copies the next sz elements from *Src to *Dst, taking into account
 *	 the 1st element starts at column chk_offs. This is approximately
 *	 a chk_m x N matrix.
 *	 The address pointers in *Src and *Dst are advanced to start of
 *	 the next chunk.         
 * notes:tries to optimize depending on sz, SrcType & DstType,
 *	 using either DBLAS copy routines (using top-level cache lookahead
 *	 and hence faster at least if source is normal meory and small)
 *	 and Solaris memcpy().  
 */


int VPPTreeStrideBrd STDC_PARAMS(
	(int Src, int NDests, int Dests[], int M, int N, FDATA A[], int IncA)
);
/* pre:  in this cell's contribution to a tree broadcast, it receives
 *	 the message from cell Src and sends it to each of cells
 *	 Dests[0..NDests-1]. Leaf cells have Src<0 and the root has NDests=0.
 * post: the broadcast of M x N matrix {A, IncA, 1} is performed
 */

#define USE_VecAdd(sz, SrcType, DestType)				\
	((SrcType)==OrdMem && ((sz) <= TOP_CACHESIZE || ((DestType)==OrdMem)))

void ChkAdd STDC_PARAMS(
	(int sz, int chk_offs, int chk_offs_nxt, int chk_m, int N, 
         enum MemType SrcType, LocMat *Src, enum MemType DestType, LocMat *Dst)
);
/* pre/post: as for ChkCpy() except 
 *	 	LM_CInc(*Src), LM_RInc(*Dst) >= 1
 *       and Src is added onto Dst
 * notes:tries to optimize depending on sz, SrcType & DstType,
 *	 using either DBLAS copy routines (using top-level cache lookahead
 *	 and hence faster at least if source is normal memory and small)
 *	 and a minimally pipelined equivalent.
 */


int VPPTreeStrideSum STDC_PARAMS(
	(int Dest, int NSrcs, int Srcs[], int M, int N, FDATA A[], int RIncA,
	 int CIncA)
);
/* pre:  in this cell's contribution to a tree broadcast, it receives
 *	 messages from cell Srcs[0..NSrcs-1] adds  sends it to each of cells
 *	 Dests[0..NDests-1].
 *	 Leaf cells have NSrcs=0 and the root has Dest<0..
 * post: the reduction of an M x N matrix {A, IncA, 1} is performed,
 *	 with the result being held in the root cell.
 */

void StrideRingComb STDC_PARAMS(
	(int RowCol, int Add, pMatDim D, int M, int n, FDATA A[])
);
/* pre/post: as for RingComb() in DistMatrix.c; takes advantage of
 *	special memory to minimize memory copying, while trying to pipeline
 *	essential copies with sends and recieves.
 */
