/************************************************************************/
/*	common macro definitions and headers for files implementing 	*
 *      the selected LAPACK routines					*
/************************************************************************/
/* SCCS INFO %Z%%M% %I%         last modified P. Strazdins, %E%
   first created: P. Strazdins 99/02
*/
    
#ifndef __LAPACK_INCL__
#define __LAPACK_INCL__

#if (FDATATYPE==SINGLE)         /*define procedure identifiers and names*/
# define _GETRF 	sgetrf_
# define _GETRS 	sgetrs_
# define _GEQRF 	sgeqrf_
# define _GEQRS 	sgeqrs_
# define _POTRF 	spotrf_
# define _POTRS 	spotrs_
# define _SYTRF 	ssytrf_
# define _SYTRS 	ssytrs_
#elif (FDATATYPE==DOUBLE)
# define _GETRF 	dgetrf_
# define _GETRS 	dgetrs_
# define _GEQRF 	dgeqrf_
# define _GEQRS 	dgeqrs_
# define _POTRF 	dpotrf_
# define _POTRS 	dpotrs_
# define _SYTRF 	dsytrf_
# define _SYTRS 	dsytrs_
#elif (FDATATYPE==COMPLEX)
# define _GETRF 	cgetrf_
# define _GETRS 	cgetrs_
# define _GEQRF 	cgeqrf_
# define _GEQRS 	cgeqrs_
# define _POTRF 	cpotrf_
# define _POTRS 	cpotrs_
# define _SYTRF 	csytrf_
# define _SYTRS 	csytrs_
# else
# define _GETRF 	zgetrf_
# define _GETRS 	zgetrs_
# define _GEQRF 	zgeqrf_
# define _GEQRS 	zgeqrs_
# define _POTRF 	zpotrf_
# define _POTRS 	zpotrs_
# define _SYTRF 	zsytrf_
# define _SYTRS 	zsytrs_
#endif /*FDATATYPE*/

# define ILAENV		ilaenv_


#ifdef NoChange

# define sgetrf_	sgetrf
# define sgetrs_	sgetrs
# define sgeqrf_	sgeqrf
# define sgeqrs_	sgeqrs
# define spotrf_	spotrf
# define spotrs_	spotrs
# define ssytrf_	ssytrf
# define ssytrs_	ssytrs

# define dgetrf_	dgetrf
# define dgetrs_	dgetrs
# define dgeqrf_	dgeqrf
# define dgeqrs_	dgeqrs
# define dpotrf_	dpotrf
# define dpotrs_	dpotrs
# define dsytrf_	dsytrf
# define dsytrs_	dsytrs

# define cgetrf_	cgetrf
# define cgetrs_	cgetrs
# define cgeqrf_	cgeqrf
# define cgeqrs_	cgeqrs
# define cpotrf_	cpotrf
# define cpotrs_	cpotrs
# define csytrf_	csytrf
# define csytrs_	csytrs

# define zgetrf_	zgetrf
# define zgetrs_	zgetrs
# define zgeqrf_	zgeqrf
# define zgeqrs_	zgeqrs
# define zpotrf_	zpotrf
# define zpotrs_	zpotrs
# define zsytrf_	zsytrf
# define zsytrs_	zsytrs

# define ilaenv_	ilaenv
#endif /* NoChange */

#ifdef UpCase

# define sgetrf_	SGETRF
# define sgetrs_	SGETRS
# define sgeqrf_	SGEQRF
# define sgeqrs_	SGEQRS
# define spotrf_	SPOTRF
# define spotrs_	SPOTRS
# define ssytrf_	SSYTRF
# define ssytrs_	SSYTRS

# define dgetrf_	DGETRF
# define dgetrs_	DGETRS
# define dgeqrf_	DGEQRF
# define dgeqrs_	DGEQRS
# define dpotrf_	DPOTRF
# define dpotrs_	DPOTRS
# define dsytrf_	DSYTRF
# define dsytrs_	DSYTRS

# define cgetrf_	CGETRF
# define cgetrs_	CGETRS
# define cgeqrf_	CGEQRF
# define cgeqrs_	CGEQRS
# define cpotrf_	CPOTRF
# define cpotrs_	CPOTRS
# define csytrf_	CSYTRF
# define csytrs_	CSYTRS

# define zgetrf_	ZGETRF
# define zgetrs_	ZGETRS
# define zgeqrf_	ZGEQRF
# define zgeqrs_	ZGEQRS
# define zpotrf_	ZPOTRF
# define zpotrs_	ZPOTRS
# define zsytrf_	ZSYTRF
# define zsytrs_	ZSYTRS

# define ilaenv_	ILAENV
#endif /* UpCase */

F_VOID_FCT _GETRF STDC_PARAMS(
	(int *M, int *N, FDATA A[], int *LdA, int IPIV[], int *INFO)
);

F_VOID_FCT _POTRF STDC_PARAMS(
	(F_CHAR_T UPLO, int *N, FDATA A[], int *LdA, int *INFO)
);

F_VOID_FCT _GEQRF STDC_PARAMS(
	(int *M, int *N, FDATA A[], int *LdA, FDATA TAU[], 
	 FDATA WORK[], int *LWORK, int *INFO)
);

F_VOID_FCT _SYTRF STDC_PARAMS(
	(F_CHAR_T UPLO, int *N, FDATA A[], int *LdA, int IPIV[],
	 FDATA WORK[], int *LWORK, int *INFO)
);

F_VOID_FCT _GETRS STDC_PARAMS(
	(int *M, int *N, int *NRHS, FDATA A[], int *LdA, int IPIV[],
	 FDATA B[], int *LdB, int *INFO)
);

F_VOID_FCT _POTRS STDC_PARAMS(
	(F_CHAR_T UPLO, int *N, int *NRHS, FDATA A[], int *LdA, 
	 FDATA B[], int *LdB, int *INFO)
);


F_VOID_FCT _GEQRS STDC_PARAMS(
	(int *M, int *N, int *NRHS, FDATA A[], int *LdA, int IPIV[],
	 FDATA B[], int *LdB, int *INFO)
);


F_VOID_FCT _SYTRS STDC_PARAMS(
	(F_CHAR_T UPLO, int *N, int *NRHS, FDATA A[], int *LdA, int IPIV[],
	 FDATA B[], int *LdB, int *INFO)
);


F_INTG_FCT ILAENV STDC_PARAMS(
	(int *ISPEC, char *SUBNAME, F_CHAR_T UPLO, int *N1, int *N2, int *N3,
         int *N4, int LenSUBNAME)
);
#endif /* __LAPACK_INCL__ */
