/************************************************************************/
/* 	LLT: distributed Cholesky factorization of a global NxN Matrix A*/
/************************************************************************/
/* SCCS INFO @(#)LLT.c 1.5         last modified P. Strazdins, 00/10/18
   first created: P. Strazdins 94/09
*/

/* reduced startup version supporting algorithmic blocking 
 * (externally-defined) cpp FLAGs:
 *      PRINTMAT        print out intermediate matrices
 */
/*#define SEMI_REP*/

#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#ifdef ASSERT
# define PRINTMAT
#endif
#include "GetBlockFactor.h"
#include "MatDim.h"
#include "DistMat.h"
#include "D_BLAS.h"

#include "TRSM.h"
#include "BLASOptions.h"
#include "FDMalloc.h"
#include "FDCommun.h"
#include "CellMatrix.h"

#include "D_LAPACK.h"

double sqrt STDC_PARAMS( (double) );

#ifdef SEMI_REP
static DistMat Tf;
#endif

void D_POTF2(N, A)
  int N; DistMat A;
{ int j; 
  LocMat lAj0, lAjj;
    
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%s,_)", N, DM_TRACE(0,A)),
    DistMatOK(N, N, A),     0, "POTF2", 0, TRACE_OUTER
  );

  if (!(Blk1d(N, DM_D0(A)) && Blk1d(N, DM_D1(A)))) {
    for (j=0; j < N; j++) {
      DistMat_RVAlloc(6);        	
      DistMat Ajj = SubMat(A, j, j);	
      DistMat Aj1j = SubMat(Ajj, 1, 0),	Aj1j1 = SubMat(Aj1j, 0, 1);
      if (InColScope(Ajj)) {
        FDATA Ajje = D_GetElt(ColVec(Ajj)), AjjeInv;
        FD_SET_REAL(Ajje, (FD_BASE) sqrt((double) FD_REAL(Ajje)));	   
        FD_SET_IMAG(Ajje, 0.0);
        FD_ASSIGN_SC(AjjeInv, =, 1.0/FD_REAL(Ajje), 0.0);
        D_SCAL(N-(j+1), AjjeInv, Aj1j);
        D_SetElt(Ajj, Ajje);
      }
#ifdef SEMI_REP
      if (Tf != NULL) { 
        DistMat Tj = SubMat(Tf, j, j);
        D_COPY(N-j, Ajj, Tj);	
        Aj1j = SubMat(Tj, 1, 0);
      }  
#endif            
      D_SYR("Lower", N-(j+1), FD_M1, Aj1j, Aj1j1);
    } /* for (j...) */ 
    TRACE_RETURN;
  }
  
  if (!D_CellHolds(N, N, A)) {
    TRACE_RETURN;
  }
  
  LM_Copy(DM_Trans(A), lAj0, DM_Loc(A));
  LM_Copy(1, lAjj, lAj0);
  for (j=0; j < N; j++) {  	
    FDATA Ajj;
    LocMat lAj0t;    			LM_Copy(1, lAj0t, lAj0);
    VMMult(N-j, j, &FD_M1, &lAj0, &lAj0t, &FD_1, &lAjj);
    Ajj = *LM_ADR(lAjj);
    FD_SET_REAL(Ajj, (FD_BASE) sqrt((double) FD_REAL(Ajj)));
    FD_SET_IMAG(Ajj, 0.0);
    *LM_ADR(lAjj) =  Ajj;
    pLocSubMat(&lAjj, 0, 1);

    FD_RECIP(Ajj,  Ajj);	
    MatScale(1, N-j-1, &Ajj, &lAjj);  
    pLocSubMat(&lAjj, 1, 0);
    pLocSubMat(&lAj0, 1, 0);  
  } /* for (j...) */ 
#ifdef SEMI_REP
  if (Tf != NULL)  
    MatCopy(N, N, DM_pLoc(A), NULL, DM_pLoc(Tf));
#endif
  TRACE_EXIT;          	
} /* D_POTF2() */


void D_POTRF(N, A, MaxW)
  int N, MaxW;  DistMat A;
{ int j;
  FDATA *TRSMsBfree;
  char svBrdTopChar[2];

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%s,_)", N, DM_TRACE(0,A)),
    DistMatOK(N, N, A) && Dim_BCompat(DM_D0(A), DM_D1(A)), 
    0, "POTRF", -1, TRACE_OUTER
  );
  
  svBrdTopChar[0] = BrdTopChar[0],	svBrdTopChar[1] = BrdTopChar[1];
  BrdTopChar[DM_Trans(A)] = ' ',        BrdTopChar[1-DM_Trans(A)] = 'I';
 
  j=0;
  while (j < N) {
    DistMat_RVAlloc(8);
    DistMat Ajj = SubMat(A, j, j), 	T;
    int minW = MAX(d_B(DM_DN(A)), MAX(MIN_BLOCKING_FACTOR, -MaxW));
    int maxW = MIN(ABS(MaxW), N-j);
    int w = GenGetW(N-(j+1), 1.0, minW, maxW, DM_DN(Ajj), j, 1);
    int BlockT = Blk1d(w, DM_DM(Ajj)) && Blk1d(w, DM_DN(Ajj));   
    DistMat Ajwj = SubMat(Ajj, w, 0),   Ajwjw = SubMat(Ajwj, 0, w);

    DistMat cT;	 
      
    DM_PRINT("TL ", N-j, w, Ajj);        

#ifdef SEMI_REP
    Tf = NULL;
#endif    
    if (!BlockT) {			/* create fully replicated tri factor */
      cT = Trans(NewMat(w, w, Trans(ReduceDimMat(DM_Trans(Ajj), Ajj)), NULL));
      T = NewMat(w, w, ReduceDimMat(-1, Ajj), NULL);
#ifdef SEMI_REP
      Tf = T;	
      T = cT;
#endif
      D_ScaleAdd(w, w, FD_1, Ajj, FD_0, T);      
    } else
      T = Ajj;
    D_POTF2(w, T);
    DM_PRINT("T", w, w, T);	
    if (!BlockT) {			/* copy replicated T into original */
      D_ScaleAdd(w, w, FD_1, T, FD_0, Ajj);            	
      					/* copy sections needed for TRSM */
#ifdef SEMI_REP
      DM_PRINT("Tf", w, w, Tf);
      T = Tf;
#endif      
      D_ScaleAdd(w, w, FD_1, T, FD_0, Trans(cT));
      DisposeMat(T);            	
      T = Trans(cT);
      DM_PRINT("cT", w, w, T);
    }
             
    TRSM_CacheB = 1; 
    D_TRSM("Lower", "NonUnit", w, N-(j+w), FD_1, T, Trans(Ajwj));
    DM_PRINT("L", N-(j+w), w, Ajwj);
    TRSMsBfree = NULL; 
    if (TRSM_CacheB==0) { /* TRSM() attempted cache of sB */
      TRSMsBfree = LM_ADR(DM_Loc(TRSMsB));
      if (TRSMsBfree==LM_ADR(DM_Loc(Ajwj))) TRSMsBfree = NULL;  
      Ajwj = Trans(TRSMsB);   
    }  
    TRSM_CacheB=0;
    DM_PRINT("L*", N-(j+w), w, Ajwj);        

    D_SYRK("Lower", N-(j+w), w, FD_M1, Ajwj, FD_1, Ajwjw);
    
    if (TRSMsBfree!=NULL)
      MatFree(TRSMsBfree); 
    if (!BlockT)
      DisposeMat(T);  
    j += w;
  } /* while (j...)*/
  BrdTopChar[0] = svBrdTopChar[0],	BrdTopChar[1] = svBrdTopChar[1];
} /* D_POTRF()*/


void D_POTRM(N, A, MaxW)
  int N, MaxW; DistMat A;
{ int  j;  
  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%s,_,%d)", N, DM_TRACE(0,A), MaxW),
    DistMatOK(N, N, A) && Dim_BCompat(DM_D0(A), DM_D1(A)) && MaxW>0, 
    0, "POTRM", -1, TRACE_OUTER
  );    
  j = N-1;
  while (j>=0) {
    DistMat_RVAlloc(4);
    DistMat A0j = SubMat(A, 0, j);
    int minW = MAX(d_B(DM_DN(A)), MAX(MIN_BLOCKING_FACTOR, -MaxW)); 
    int maxW = MIN(ABS(MaxW), j+1);
    int w = GenGetW(N-(j+1), 1.0, minW, maxW, DM_DN(A0j), j, -1);
    int jw = j-w+1;
    DistMat Ajwjw, Aj1jw, Aj1j1;
    int i; 
  
    Ajwjw = SubMat(A, jw, jw);
#ifndef OPT_SUBMAT
    Aj1jw = SubMat(A, j+1, jw);		Aj1j1 = SubMat(A, j+1, j+1); 	
#else
    Aj1jw = SubMat(Ajwjw, w, 0);	Aj1j1 = SubMat(Aj1jw, 0, w); 	
#endif
    
    D_SYRK("Lower", N-(j+1), w, FD_1, Aj1jw, FD_1, Aj1j1);
    D_TRMM("Lower", "NonUnit", w, N-(j+1), FD_1, Ajwjw, Trans(Aj1jw));

    for (i=j; i>=jw; i--) {
      DistMat_RVAlloc(4);
      DistMat Aii = SubMat(A, i, i);	
#ifndef OPT_SUBMAT
      DistMat Ai1i = SubMat(A, i+1, i),	Ai1i1 = SubMat(A, i+1, i+1);
#else
      DistMat Ai1i = SubMat(Aii, 1, 0),	Ai1i1 = SubMat(Ai1i, 0, 1);
#endif      
      if (i<j)
        D_SYR("Lower", j-i, FD_1, Ai1i, Ai1i1);
      if (InColScope(Aii)) {
        FDATA Aiie = D_GetElt(ColVec(Aii));
        if (i<j)
          D_SCAL(j-i, Aiie, Ai1i);
        FD_MULSET(Aiie, =, Aiie, Aiie);  
	D_SetElt(Aii, Aiie);
      }
    }  
    j -= w;
  } /* while (j...) */
  TRACE_EXIT;
} /* D_POTRM() */
