/************************************************************************/
/* LU: distributed LU factorization of a global M x N Matrix A		*/
/************************************************************************/
/* SCCS INFO %Z%%M% %I%         last modified P. Strazdins, %E%
   first created: P. Strazdins 93/09/18					
*/
/* reduced startup version supporting algorithmic blocking
 * (externally-defined) cpp FLAGs:
 *	PRINTMAT	print out intermediate matrices
 *	BLOCK_P		block b/c of pivots for storage blocking (recommended)
 *	COMB_PIV	combine pivot-related communications
 *	
 */
/*#define DEMO */
#define BLOCK_P
/*#define COMB_PIV*/

#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#ifdef ASSERT
# define PRINTMAT
#endif 
#include "GetBlockFactor.h"
#include "DistMat.h"
#include "LocMat.h" 
#include "D_BLAS.h"
#include "FDMalloc.h"
#include "CellMatrix.h"
#include "FDCommun.h"

# include "TRSM.h"

#include "MachineParams.h"

#include "D_LAPACK.h"


#define IxPiv(P, j, A)        ((P)[PivColRep ? G2Ld(j, DM_DM(A)): (j)])
int PivColRep = 0; /* current limitation: if !=0, MaxW should be <= blocksize */
int LU_SwapL = 1;  /* if true, complete row swaps in L */

struct SwapStruct {
  FDATA *Ax;
  int Inc;
  int m, dm;
}; 
    	
static int NumSwapStructs=1;    
static struct SwapStruct SwapStructs[4]; 

static void D_LASWPAuxSet(A, M, dM)
  DistMat A; int M, dM;
{ 
  SwapStructs[NumSwapStructs].Ax = LM_ADR(DM_Loc(A));
  SwapStructs[NumSwapStructs].Inc = LM_Inc(DM_Trans(A), DM_Loc(A));
  SwapStructs[NumSwapStructs].m = G2Ld(M, DM_DM(A));
  SwapStructs[NumSwapStructs].dm = dM;  
  NumSwapStructs++;
} /* D_LASWPAuxSet() */
    

#include "CellVector.h"		/* VecSwap() */
#define NSW 128			/* max size M for packed swaps */ 

void D_LASWP(Direction, M, N, A, P)
  char *Direction; int M, N; DistMat A; int P[];
{ int j, dj, Lj;

  COMMON_ASSERT_TR_E(
   (TraceStr, "(%c,%dx%d,%s,a%d)", *Direction, M, N, DM_TRACE(0,A), ((int) P)/sizeof(int)),
   P!=NULL && DistMatOK(M, N, A), 0, "LASWP", 0, 
   TRACE_OUTER);
   
   if (*Direction=='F')
     j=0, dj=1, Lj=N;
   else
     j=N-1, dj=-1, Lj=-1;  
            
   if (PivColRep==0 || Blk1d(N, DM_DN(A))) {      
     int Inc = LM_Inc(DM_Trans(A), DM_Loc(A));
     LocMat lA;
     if (Inc==1 || *Direction!='F' || N > NSW) {
       for (; j!=Lj; j+=dj) {
         DistMat_RVAlloc(2);
         if (j!=P[j]) 
           D_SWAP(M, SubMat(A, 0, j), SubMat(A, 0, P[j]));
       } /* for (j..) */
     } else {
       int n=G2Ld(N, DM_DN(A)), NSw=0, SrcSw[NSW], DstSw[NSW], jG[NSW];
       int npk=n, pk2lj[2*NSW], lj=0;
       int m = G2Ld(M, DM_DM(A)),	i, dm;
       FDATA *pkA = NULL;    
       int isw, sm, ms;       
#ifdef GLOB_BLK_SEND
       FDATA *pkASnd = NULL;
#endif               
       for (i=0; i<n; i++)
         pk2lj[i] = -1;			/* if left -1, this row is not active */
       for (j=0;j!=Lj;j+=dj) {
         int p0j = SubMatP0d(j, DM_DN(A)), p0pj = SubMatP0d(P[j], DM_DN(A));
         if (P[j]!=j) {			/* swap is non-trivial */
       	   if (p0pj > 0) {		
       	     if (p0j==0)		/* source but not dest for this cell */
       	       jG[NSw] = j, SrcSw[NSw] = pk2lj[lj] = lj, DstSw[NSw] = -p0pj, NSw++;
       	   } else {			/* dest row on this cell */
       	     int lpj = G2Ld(P[j], DM_DN(A)), k = lpj;
       	     if (lpj >= n) {		/* dest row not one of the source rows */
       	       k=n;
       	       while (k<npk && pk2lj[k]!=lpj) k++;
       	       if (k==npk) npk++;	/* new row */   
       	     }       	        	         
       	     if (p0j==0)		/* source row also for this cell */
       	       SrcSw[NSw] = pk2lj[lj] = lj, DstSw[NSw] = k;
       	     else
       	       SrcSw[NSw] = k, DstSw[NSw] = -p0j;
       	     jG[NSw] = j, pk2lj[k] = lpj, NSw++;     
       	   }
       	 }  
       	 if (p0j==0) lj++;
       } /* for (j...) */
#if 0       
printf("m=%d, n=%d, Inc=%d, lj=%d, NSw=%d:: ", m, n, Inc, lj, NSw);
for (j=0; j < NSw; j++) 
  printf("%d:%d:%d  ", jG[j], SrcSw[j], DstSw[j]); 
printf("\n npk=%d: ", npk);
for (j=0; j < npk; j++) 
  printf("%d ", pk2lj[j]);
printf("\n");   	
#endif
       SwapStructs[0].Ax = LM_ADR(DM_Loc(A));
       SwapStructs[0].Inc = Inc;
       SwapStructs[0].m = m;
       SwapStructs[0].dm = 0;
      
       ms=0;
       for (isw=0; isw < NumSwapStructs; isw++) 
         ms += SwapStructs[isw].m;
#if 0
printf(" ms=%d:: ", ms);
for (isw=0; isw < NumSwapStructs; isw++)
  printf("%d ", SwapStructs[isw].m);
printf("\n");  
#endif       
       if (ms==0 || NSw==0) {
	 NumSwapStructs=1; 
	 TRACE_RETURN; /* fast exit */
       }	  	
       MatMalloc("LASWP", 1, &pkA, ms*npk);
       sm = 0; 
       for (isw = 0; isw < NumSwapStructs; isw++) {
       	 FDATA *Ax = SwapStructs[isw].Ax;
       	 m = SwapStructs[isw].m; 
       	 Inc = SwapStructs[isw].Inc;	
         dm = N<MIN_VECSIZE || TLBSIZE==0 || CACHESIZE==0? m: 
              MAX(TLBSIZE/2, CACHESIZE/2/Inc);            	
         for (i=0; i < m; i+=dm) {
       	   int mi = MIN(m-i, dm); 	        
           for (j=0; j < npk; j++) 
             if (pk2lj[j] >= 0)
       	       VecCopy(mi, Ax+i*Inc+pk2lj[j], Inc, pkA+i+sm+j*ms, 1);
         } /* for (i...) */
         sm += m;
       }  
#ifdef GLOB_BLK_SEND
       MatMalloc_FT("LASWP", 1, &pkASnd, ms);
#endif       
       for (j=0; j < NSw; j++) {
	 int m = ms;
	 if (SwapStructs[NumSwapStructs-1].dm==1)
	   m -= (SwapStructs[NumSwapStructs-1].m - jG[j]);
       	 if (DstSw[j] >= 0)
       	   VecSwap(m, pkA+SrcSw[j]*ms, 1, pkA+DstSw[j]*ms, 1);
       	 else {
       	   int DI = DM_Trans(A)? DstSw[j]: 0,	DJ = DM_Trans(A)? 0: DstSw[j];	
#ifndef GLOB_BLK_SEND
       	   SendMat(DI, DJ, 1, m, pkA+SrcSw[j]*ms, ms, GENMSGTTYPE);	
       	   RecvMat(DI, DJ, 1, m, pkA+SrcSw[j]*ms, ms, GENMSGTTYPE);	
#else
	   SendWaitMark();
	   VecCopy(m, pkA+SrcSw[j]*ms, 1, pkASnd, 1);
       	   SendMat(DI, DJ, 1, m, pkASnd, ms, GENMSGTTYPE);	
       	   RecvMat(DI, DJ, 1, m, pkA+SrcSw[j]*ms, ms, GENMSGTTYPE);	
       	   CompleteSends(-1);
#endif
       	 }   
       } /* for (j..) */
#ifdef GLOB_BLK_SEND
       MatFree(pkASnd);
#endif       
       sm = 0;
       for (isw = 0; isw < NumSwapStructs; isw++) {
         FDATA *Ax = SwapStructs[isw].Ax;  
         m = SwapStructs[isw].m;
         Inc = SwapStructs[isw].Inc;
         dm = N<MIN_VECSIZE || TLBSIZE==0 || CACHESIZE==0? m: 
              MAX(TLBSIZE/2, CACHESIZE/2/Inc);            	
         for (i=0; i < m; i+=dm) {
       	   int mi = MIN(m-i, dm); 	        
           for (j=0; j < npk; j++) 
             if (pk2lj[j] >= 0)
       	       VecCopy(mi, pkA+i+sm+j*ms, 1, Ax+i*Inc+pk2lj[j], Inc);
         } /* for (i...) */
         sm += m;
       }
       MatFree(pkA);
     }  
   } else {  
     for (; j!=Lj; j+=dj) {
       DistMat_RVAlloc(4);
       int Pj = IxPiv(P, j, Trans(A)); FDATA fPj;
       DistMat A0j = SubMat(A, 0, j); 
       FD_SET_REAL(fPj, (FD_BASE) Pj);
       ColBrdScalar(fPj, Trans(A0j));
       Pj = (int) FD_REAL(fPj);    
       if (j!=Pj) 
         D_SWAP(M, A0j, SubMat(A, 0, Pj));
     } /* for (j..) */
   }  
   NumSwapStructs = 1;
   TRACE_EXIT;
} /* D_LASWP() */



#ifdef COMB_PIV		/* use combined pivot find and swap routine */

static int nA = 0, nB = 0;
#define PivALen       	nA
#define PivBLen       	nB
#define PivAMxIx	2
#define PivBMxIx	(PivAMxIx + PivALen)
#define PivA0Ix		(2 + PivBMxIx + PivBLen)
#define PivB0Ix		(PivA0Ix +  PivALen)
 
static int D_AMAXSwap(M, X, Max, NA, A, NB, B, PivBuf)
 int M, NA, NB; DistMat X, A, B; FDATA *Max, **PivBuf;
/* pre:  N>= 0; DistMatOK(M, 1, X); DistMatOK(M, NA, A), DistMatOK(M, NB, B)
 *       PivBuf has space for 2*(1+nA+nB) elements
 * post: as for i = D_AMAX(M, X, Max) and *PivBuf[] has been set to the following:
 *	 *PivBuf[0] = i, *Max = *PivBuf[1] = X[i,1] (value of abs.max)
 *	 *PivBuf[PivAMxIx..+PivALen] = this_row(A[i,0..NA-1])
 *	 *PivBuf[PivBMxIx..+PivBLen] = this_row(B[i,0..NA-1])
 *	 Rows 0 and i of A[:,0..NA-1] and BA[:,0..NB-1] are swapped 
 *	 this is is only valid if InColScope(X)
 * note: 1. the motivation of this function is combine the broadcast of 
 *          of the maximum with that of the pivot rows
 *	 2. for algorithmic blocking, X must be row-replicated so that
 *	    all cells can know the position of the maximum
 */
{ FDATA *Buf; int imax=0;

  COMMON_ASSERT_TR_E(
    (TraceStr, "(%d,%s,_,%d,%s,%d,%s,_)", M, DM_TRACE(0, X), NA, DM_TRACE(1, A), NB, DM_TRACE(2, B)),
     M>=0 && DistMatOK(M, 1, X) && DistMatOK(M, NA, A) && DistMatOK(M, NB, B),
     0, "AMAXSwap", 1, TRACE_OUTER
  );
  
  nA = G2Ld(NA, DM_DN(A)),		nB = G2Ld(NB, DM_DN(B));
  MatMalloc_FT("D_AMAXAuxComm", 1, &Buf, 3*(nA+nB+2));
  CommSetUp(d_P(DM_D0(X)), d_P(DM_D1(X)));

  if (d_P0(DM_DN(X))==0) {  
    int nx = G2Ld(M, DM_DM(X));         int IncX = LM_Inc(DM_Trans(X), DM_Loc(X));
    int limax = VecAbsMaxPos(nx, LM_ADR(DM_Loc(X)), IncX);
    int P = d_P(DM_DM(X)),            P0 = d_P0(DM_DM(A));
    int RIncA = LM_Inc(DM_Trans(A), DM_Loc(A)),
	CIncA = LM_Inc(1-DM_Trans(A), DM_Loc(A));
    int RIncB = LM_Inc(DM_Trans(B), DM_Loc(B)),
	CIncB = LM_Inc(1-DM_Trans(B), DM_Loc(B));  
    if (limax==-1) {                  /* assert nx = 0 */
      Buf[0] = FD_0;                   limax = 0;
    } else
      Buf[0] = *(LM_ADR(DM_Loc(X)) + FAST_MUL2(limax, IncX));

    imax = L2Gd(limax, DM_DM(X));        /* convert to a global index */
    FD_SET_REAL(Buf[1], (FD_BASE) imax);

    VecCopy(nA, &(DM_Elt00(A))+FAST_MUL2(limax, RIncA), CIncA, &Buf[PivAMxIx], 1);
    VecCopy(nB, &(DM_Elt00(B))+FAST_MUL2(limax, RIncB), CIncB, &Buf[PivBMxIx], 1);

    if (P0==0) {
      VecCopy(nA, &DM_Elt00(A), CIncA, &Buf[PivA0Ix], 1);
      VecCopy(nB, &DM_Elt00(B), CIncB, &Buf[PivB0Ix], 1);
    }
    
    if (d_P(DM_DM(X)) > 1) {
      int DelP0 = 1, lgDelP0=0,         RowCol = DM_Trans(X);
      int n = 2 + nA + nB;	 	FDATA *RMxI = Buf + 2*n;
      while (DelP0 < P) {
        int SendMsg = (P0 & (2*DelP0-1)) == DelP0;
        int DelP0_ = SendMsg? -DelP0: DelP0;
        int DelIP0 = RowCol==0? DelP0_: 0,  DelJP0 = RowCol==0? 0: DelP0_;
        if (SendMsg) {
          SendMat(DelIP0, DelJP0, 1, n, Buf, n, SUMMSGTYPE);
        } else if ((P0 & (2*DelP0-1))==0 && P0+DelP0 < P) {
          RecvMat(DelIP0, DelJP0, 1, n, RMxI, n, SUMMSGTYPE);
        if (FD_ABS(RMxI[0]) > FD_ABS(Buf[0]))
          VecCopy(n, RMxI, 1, Buf, 1); 
        } /* if (SendMsg) */
        DelP0 *= 2, lgDelP0++;
      } /* while (DelP0 ...) */
      BrdMat(DM_Trans(X), P0, 1, 2*n, Buf, 2*n);
    }
    if (FD_EQ(Buf[0], FD_0))           	/* take top element as being the max */
      imax = 0, Buf[1] = FD_0;
    if (!FD_EQ(Buf[1], FD_0)) {        	/* non-trivial swap*/		
      if (P0==0) {		 	/* holds row 0 */
        VecCopy(nA, &Buf[PivAMxIx], 1, &DM_Elt00(A), CIncA);
        VecCopy(nB, &Buf[PivBMxIx], 1, &DM_Elt00(B), CIncB);        
      }
      if (imax == (int) FD_REAL(Buf[1])) { /* holds row i */
        VecCopy(nA, &Buf[PivA0Ix], 1, (&DM_Elt00(A))+FAST_MUL2(limax, RIncA), CIncA);
        VecCopy(nB, &Buf[PivB0Ix], 1, (&DM_Elt00(B))+FAST_MUL2(limax, RIncB), CIncB);                 
      }     
    }
    *Max = Buf[0], imax = (int) FD_REAL(Buf[1]);  
  } 
  
  *PivBuf = Buf;
  TRACE_RETURN(imax);
} /* D_AMAXSwap() */
#endif /*COMB_PIV*/

static DistMatStore T_;
static DistMat T = &T_;
static int BlockL = 0;



void D_GETF2(M, N, A, P)
  int M, N, P[]; DistMat A;
/* note: P is assumed to be fully replicated */  
{ int j;
  COMMON_ASSERT_TR_E(
   (TraceStr, "(%d,%d,%s,a%d)", M, N, DM_TRACE(0,A), ((int) P)/sizeof(int)),
   P!=NULL && DistMatOK(M, N, A), 0, "GETF2", 0, TRACE_OUTER);

  BlockL = Blk1d(N, DM_DN(A));
  T = &T_;
  T = BlockL? A: Trans(NewMat_(T, N, M, Trans(ReduceDimMat_(T, 1, A)), NULL));

#ifdef BLOCK_P
  if (!BlockL || InColScope(A)) 	/* only factor in cells holding A */
#endif
  for (j=0; j < MIN(M,N); j++) {
    DistMat_RVAlloc(11);	
    DistMat Aj0 = SubMat(A, j, 0),
	    Ajj = SubMat(Aj0, 0, j),	Aj1j = SubMat(Ajj, 1, 0),
    	    Ajj1 = SubMat(Ajj, 0, 1),	Aj1j1 = SubMat(Ajj1, 1, 0);	    
    DistMat PAjj = Ajj;	   		
    DistMat Tj0 = SubMat(T, j, 0),	Tjj = SubMat(Tj0, 0, j); 
    FDATA A_j_mx, *PivBuf;

#ifdef BLOCK_P
    if (BlockL)
      PAjj = ColVec(Ajj);
#endif       
#ifdef COMB_PIV
    if (!BlockL && M-j>0) {
      PAjj = Tjj;
      D_COPY(M-j, Ajj, PAjj);		/* horiz b/c before pivot */
    }

    P[j] = j + D_AMAXSwap(M-j, PAjj, &A_j_mx, N, Aj0, BlockL? 0: j+1, Tj0, &PivBuf);
    Ajj1 = NewMat(1, N-(j+1), RowVec(Ajj1), &PivBuf[PivAMxIx+G2Ld(j+1, DM_DN(Aj0))]);
					
    Aj1j = SubMat(PAjj, 1, 0);
    if (!FD_EQ(A_j_mx, FD_0)) {
      FD_RECIP(A_j_mx, A_j_mx); 	
      D_SCAL(M-(j+1), A_j_mx, Aj1j);
      if (!BlockL && M-j>0) 		/* copy scaled column back into A */
        D_ScaleAdd(M-j, 1, FD_1, PAjj, FD_0, Ajj); 
    }  
    D_GER(M-(j+1), N-(j+1), FD_M1, Aj1j, Trans(Ajj1), Aj1j1);

    MatFree(PivBuf);
#else
    P[j] = j + D_AMAX(M-j, PAjj, &A_j_mx);
    if (P[j] != j)
      D_SWAP(N, Trans(Aj0), Trans(SubMat(A, P[j], 0)));

    if (!FD_EQ(A_j_mx, FD_0)) {
      FD_RECIP(A_j_mx, A_j_mx); 	
      D_SCAL(M-(j+1), A_j_mx, Aj1j);
    }  
    if (!BlockL && M-(j+1)>0) {
      D_COPY(M-j, Ajj, Tjj);
      Aj1j = SubMat(Tjj, 1, 0); 	
    }
    D_GER(M-(j+1), N-(j+1), FD_M1, Aj1j, Trans(Ajj1), Aj1j1);
#endif
    DM_PRINT("a", M-j, N-j, Ajj);  	  
  } /* for (j...) */

#ifdef BLOCK_P
  if (BlockL && d_P(DM_DN(A)) > 0 && N>0) {
     FDATA *fP; int i;
     MatMalloc("GETF2", 1,  &fP, N);
     if (InColScope(A))
       for (i=0; i < N; i++)
         FD_SET_REAL(fP[i], (FD_BASE) P[i]);
     CommSetUp(d_P(DM_D0(A)), d_P(DM_D1(A)));    
     BrdMat(1-DM_Trans(A), d_P0(DM_DN(A)), 1, N, fP, N); 
     if (!InColScope(A))
       for (i=0; i < N; i++)
         P[i] = (int) FD_REAL(fP[i]);         
     MatFree(fP);    
  }	
#endif

  TRACE_EXIT;
} /*D_GETF2()*/   


int PivRep2Col(P, D, w)
  int P[]; pMatDim D;
{ int i, li = 0, p0 = d_P0(D), b0 = d_B0(D);
  for (i=0; i<w; i++) {
    if (p0==0) 
      P[li] = P[i], li++;	      
    b0++;
    if (b0==d_B(D)) {
      b0 = 0, p0--;
      if (p0 == -1)
        p0 += d_P(D);      	            
    }	
  } /* for (i..) */ 	
} /* PivRep2Col() */ 


void D_GETRF(M, N, A, P, MaxW)
  int M, N, P[], MaxW; DistMat A;
{ int  j;
  char svBrdTopChar[2];
#ifdef DEMO
  double lastratiodone = 0.0, ratiodone;
  double totalops = 2.0/3.0 * N *N *N;
#endif  
  
  COMMON_ASSERT_TR_E(
   (TraceStr, "(%d,%d,%s,_,%d)", M, N, DM_TRACE(0,A), MaxW),
   P!=NULL && DistMatOK(M, N, A) &&
   LM_Inc(DM_Trans(A), DM_Loc(A))==1, 0, "GETRF", -1, TRACE_OUTER);

  svBrdTopChar[0] = BrdTopChar[0],      svBrdTopChar[1] = BrdTopChar[1];    
  BrdTopChar[DM_Trans(A)] = ' ',        BrdTopChar[1-DM_Trans(A)] = 'I';
    j=0; 
  while (j < MIN(M,N)) {
    DistMat_RVAlloc(7);
    DistMat Ajj = SubMat(A, j, j);
    int minW = MAX(d_B(DM_DN(A)), MAX(MIN_BLOCKING_FACTOR, -MaxW));
    int maxW = MIN(ABS(MaxW), MIN(M,N)-j);
    int w = GenGetW(N-(j+1), 1.0, minW, maxW, DM_DN(Ajj), j, 1);
    int *Pj = &IxPiv(P, j, A);
    DistMat Ajjw, Ajwj, Ajwjw;	
    int i;
    int PivColRepV = PivColRep;

#ifdef DEMO
    ratiodone = (double) (N-j) / (double) (N);
    ratiodone *= ratiodone * ratiodone; ratiodone = 1.0 - ratiodone;
    if (ratiodone >= lastratiodone + 0.05) {
       if (GetAbsCId(0)==0 && GetAbsCId(1)==0)
          printf("LINPACK: eliminated %d of the %d columns; done %2.1f%% of the %1.2e FP ops\n", 
		 j, N, ratiodone*100.0, totalops);
       lastratiodone += 0.05;
    }
#endif

    Ajjw = SubMat(Ajj, 0, w);		Ajwjw = SubMat(Ajjw, w, 0);		

    D_GETF2(M-j, w, Ajj, Pj); 
    DM_PRINT("L", M-j, w, Ajj);  	  

    PivColRep = 0;			/* Pj at this point is fully replicated */
    if (LU_SwapL)
      D_LASWPAuxSet(Trans(SubMat(A, j, 0)), j, 0);
#ifndef COMB_PIV			/* swaps in T were not done in D_GETF2() */
    if (!BlockL)
      D_LASWPAuxSet(Trans(T), w, 1);
#endif
    DM_PRINT("U", w, N-(j+w), Ajjw); 	      
    D_LASWP("F", N-(j+w), w, Trans(Ajjw), Pj);
    DM_PRINT("U", w, N-(j+w), Ajjw); 	      
    PivColRep = PivColRepV;
    if (!BlockL)
      DM_PRINT("sL", M-j, w, T);  	  
 
    for (i=j; i<(j+w); i++)		/* make P[j..j+w-1] relative to A */
      Pj[i-j] += j;
    if (PivColRep)  
      PivRep2Col(Pj, DM_DM(Ajj), w);      

    Ajj = T;
    Ajwj = SubMat(Ajj, w, 0);    
    
    TRSM_CacheB = 1;
    D_TRSM("Lower", "Unit", w, N-(j+w), FD_1, Ajj, Ajjw);
    DM_PRINT("U", w, N-(j+w), Ajjw); 	      
    if (TRSM_CacheB==0)
      Ajjw = TRSMsB;
    TRSM_CacheB=0;  
    
    DM_PRINT("L'", M-(j+w), w, Ajwj);
    DM_PRINT("U'", w, N-(j+w), Ajjw);
    DM_PRINT("A", M-(j+w), N-(j+w), Ajwjw);  	  
    
    D_GEMM(M-(j+w), N-(j+w), w, FD_M1, Ajwj, Ajjw, FD_1, Ajwjw);
    DM_PRINT("A", M-(j+w), N-(j+w), Ajwjw);  	  

    if (d_P(DM_DM(A))!=d_P(DM_DM(Ajjw))) DisposeMat(Ajjw);       
    if (!BlockL) DisposeMat(T);
  
    j += w;
  } /* for (j...)*/ 
  BrdTopChar[0] = svBrdTopChar[0],      BrdTopChar[1] = svBrdTopChar[1];    
  TRACE_EXIT;
} /* D_GETRF */

  
void D_LUTRM(M, N, A, P, MaxW)
  int M, N, P[], MaxW; DistMat A;
{ int  j;
  COMMON_ASSERT_TR_E(
   (TraceStr, "(%d,%d,%s,_,%d)", M, N, DM_TRACE(0,A), MaxW),
   P!=NULL && DistMatOK(M, N, A), 0, "LUTRM", -1, TRACE_OUTER);

  j = MIN(M,N)-1; 
  while (j>=0) {
    DistMat_RVAlloc(5);
    DistMat Ajj = SubMat(A, j, j);
    int minW = MAX(d_B(DM_DN(A)), MAX(MIN_BLOCKING_FACTOR, -MaxW));
    int maxW = MIN(ABS(MaxW), j+1);
    int w = GenGetW(N-(j+1), 1.0, minW, maxW, DM_DN(Ajj), j, -1);
    int jw = j-w+1;
    DistMat Aj1j1, Aj1jw, Ajwj1, 	Ajwjw = SubMat(A, jw, jw);
    int i;	

#ifndef OPT_SUBMAT
    Aj1jw = SubMat(A, j+1, jw);		Aj1j1 = SubMat(A, j+1, j+1);
    Ajwj1 = SubMat(A, jw, j+1);		
#else    
    Ajwj1 = SubMat(Ajwjw, 0, w);
    Aj1jw = SubMat(Ajwjw, w, 0);	Aj1j1 = SubMat(Aj1jw, 0, w);
#endif 
    
    D_GEMM(M-(j+1), N-(j+1), w, FD_1, Aj1jw, Ajwj1, FD_1, Aj1j1);
    D_TRMM("Lower", "Unit", w, N-(j+1), FD_1, Ajwjw, Ajwj1);

    D_TRMM("Lower", "NonUnit", w, M-(j+1), FD_1, Trans(Ajwjw), Trans(Aj1jw));
    for (i=j-1; i>=jw; i--) {
      DistMat_RVAlloc(5);
      DistMat Aii, Ai1i, Aii1, Ai1i1;	
#ifndef OPT_SUBMAT
      Aii = SubMat(A, i, i);			Ai1i = SubMat(A, i+1, i);
      Aii1 = SubMat(A, i, i+1);			Ai1i1 = SubMat(A, i+1, i+1);
#else
      Aii = SubMat(A, i, i);			Ai1i = SubMat(Aii, 1, 0);
      Aii1 = SubMat(Aii, 0, 1);			Ai1i1 = SubMat(Aii1, 1, 0);
#endif      
      D_GER(j-i, j-i, FD_1, Ai1i, Trans(Aii1),  Ai1i1);
      if (InColScope(Aii))
        /* ColVec(Aii) reduces scope of D_GetElt() to (this) cell column */
        D_SCAL(j-i, D_GetElt(ColVec(Aii)), Ai1i);
    }
    j -= w;
  } /* while (j...)*/ 

  D_LASWP("B", N, MIN(M,N), Trans(A), P);
  
  TRACE_EXIT;
} /* D_LUTRM() */
