/************************************************************************/
/* QR: distributed QR factorization of a global M x N Matrix A		*/
/************************************************************************/
/* SCCS INFO @(#)QR.c 1.4         last modified P. Strazdins, 00/10/18
   first created: P. Strazdins 95/01
*/
/* reduced startup version supporting algorithmic blocking. 
 * (externally-defined) cpp FLAGs:
 *      PRINTMAT        print out intermediate matrices
 */
#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "FDATAOps.h"
#include "FDAssert.h"
#ifdef ASSERT 
#define PRINTMAT
#endif
#include "MachineParams.h"
#include "LocMat.h"
#include "DistMat.h"
#include "GetBlockFactor.h"
#include "D_BLAS.h"

#include "FDCommun.h"		/* CommSetUp(), BrdMat() */
#include "FDMalloc.h" 
#include "CellMatrix.h" 
#include "D_LAPACK.h"

int TauRowRep = 1;		/* =1 if Tau is to be row-replicated */
#define IxTau(Tau, j, A)	(Tau[TauRowRep ? G2Ld(j, DM_DN(A)): j])

#include "FtnBLAS.h"

void D_LARFTComplete(Direction, K, T)
  int K; char *Direction;   DistMat T; 
{ int j; 
  COMMON_ASSERT_TR_E(
     (TraceStr, "(%c,%d,%s)", *Direction, K, DM_TRACE(0, T)),
      *Direction=='F' && DistMatOK(K, K, T), 0, "LARFTComplete", 1, TRACE_OUTER
  );

  
#if 0
  if (Blk1d(K, DM_D0(T)) && Blk1d(K, DM_D1(T))) { 
    if (InEltScope(T)) {
      int LdT = LM_Ld(DM_Loc(T));
      int Trans = LM_RInc(DM_Loc(T))!=1;
      int IncT = LM_CInc(DM_Loc(T));
      char TrC[] = {'N', 'T'}, UpLoC[] = {'L', 'U'};
      for (j=1; j < K; j++) {
#if 0    	
        printf("dtrmv(%c,%c,N,%d,a%d,%d,a%d,%d)\n",
      	      UpLoC[Trans], TrC[0], j, FD_ADR(LM_ADR(DM_Loc(T))), LdT,
      	      FD_ADR(LM_Index(DM_Loc(T), j, 0)), IncT);       	     
#endif      	       
        FTN_TRMV(C2F_CHAR(&UpLoC[Trans]), C2F_CHAR(&TrC[0]), C2F_CHAR("N"), 
      	         &j, LM_ADR(DM_Loc(T)), &LdT, LM_Index(DM_Loc(T), j, 0), &IncT);
      }  	
    }
  } else 
#endif  
  for (j=1; j < K; j++) {
    DistMat_RVAlloc(2);	
    DistMat Tj0 = SubMat(T, j, 0),	Tjj = SubMat(Tj0, 0, j);

    DM_PRINT("Tj", 1, j, Tj0);
    D_TRMV("Upper" ,"NonUnit", j, Trans(T), Trans(Tj0));
  } /* for (j...) */ 	

  DM_PRINT("T", K, K, T);
  TRACE_EXIT;
} /* D_LARFTComplete() */


void D_LARFB(Direction, M, N, K, V, T, C)   
  int M, N, K; char *Direction;  DistMat V, T, C; 
{ DistMat_RVAlloc(1);			
  DistMat W;

  COMMON_ASSERT_TR_E(
     (TraceStr, "(%c,%d,%d,%d,%s,%s,%s)", 
      *Direction, M, N, K,  DM_TRACE(0, V), DM_TRACE(1, T), DM_TRACE(2, C)),
      *Direction=='F' && DistMatOK(M, K, V) /*&& DistMatOK(K, K, T)*/, 0, 
      "LARFB", 1, TRACE_OUTER
  );

  if (N==0 || K==0) return;	/* quick exit */

  W = NewMat(K, N, C, NULL);
  D_GEMM(K, N, M, FD_1, Trans(V), C, FD_0, W);
  D_TRMM("Lower", "NonUnit", K, N, FD_1, T, W);
  D_GEMM(M, N, K, FD_M1, V, W, FD_1, C);
  DisposeMat(W);
  TRACE_EXIT;
} /* D_LARFB() */
    

static DistMat Tri;	/* ifdef OPT_VM, set by D_GEQRF(), used by D_GEQF2() */
static DistMatStore sV_;
static DistMat sV = &sV_;

void D_GEQF2(M, N, A, Tau)
  int M, N;
  DistMat A; FDATA Tau[]; 
{ int j; 				
  int BlockV = Blk1d(N, DM_DN(A));
  FDATA *WS;				int n = G2Ld(N, DM_DN(A));
  DistMat_RVAlloc(2);
  DistMat sVj, sVj1;      
      
  COMMON_ASSERT_TR_E(
     (TraceStr, "(%d,%d,%s,a%d)", M, N, DM_TRACE(0, A), FD_ADR(Tau)), 
      DistMatOK(M, N, A), 0, "GEQF2", 1, TRACE_OUTER
  );
  
  sV = Trans(NewMat_(&sV_, N, M, Trans(ReduceDimMat_(&sV_, 1, A)), NULL));
  				/*ensures LM_Inc(DM_Trans(sV), DM_Loc(sV))==1*/
  D_ScaleAdd(N, N, FD_0, sV, FD_0, sV);

  MatMalloc("GEQF2", 1, &WS, n+1); 
  sVj = SubMat(sV, 0, 0);  		sVj1 = SubMat(sV, 1, 0);

  for (j = 0; j < MIN(M,N); j++) {	
    FDATA Tauj, Ajscale; 
    DistMat_RVAlloc(12);
    DistMat Aj0 = SubMat(A, j, 0);
    DistMat Ajj = SubMat(Aj0, 0, j), 	Ajj1 = SubMat(Ajj, 0, 1),
    	    Aj1j = SubMat(Ajj, 1, 0);
    DistMat W = NewMat(1, N, Aj0, WS),  rW = RowVec(W);	    
    DistMat Wj = SubMat(rW, 0, j), 	Wj1 = RowVec(SubMat(Wj, 0, 1));
    DistMat AjjCol = ColVec(Ajj);	

    /* D_LARFG*/
    if (!BlockV)
      D_COPY(M-(j+1), Aj1j, sVj1);
    else if (InColScope(Ajj))		/* suppress communication */
      D_COPY(M-j, AjjCol, sVj);
    D_SetElt(sVj, FD_1);

    D_GEMV(N, M-(j+1), FD_1, Trans(SubMat(Aj0, 1, 0)), sVj1, FD_0, Trans(W));

    if (InEltScope(Ajj)) {
      double sqrt(); 
      FDATA Ajje = DM_Elt00(Ajj);
      FD_BASE beta = FD_REAL(DM_Elt00(Wj)) + FD_REAL(Ajje) * FD_REAL(Ajje) +
      		     FD_IMAG(Ajje) * FD_IMAG(Ajje);
      beta = (FD_BASE) sqrt((double) beta);
      if (FD_REAL(Ajje) > 0.0)
        beta = -beta;	
      FD_ASSIGN_SC(Tauj, =, (beta - FD_REAL(Ajje))/beta, FD_IMAG(Ajje)/beta);
      FD_ASSIGN_SC(Ajscale, =, FD_REAL(Ajje) - beta, FD_IMAG(Ajje));
      FD_ASSIGN_SC(Ajje, =, beta, 0.0);
      FD_RECIP(Ajscale, Ajscale);
      D_SetElt(Ajj, Ajje);
    }
    /* D_LARF*/

    if (!BlockV && InRowScope(Ajj)) {
      FDATA TmpBrd[2];
      CommSetUp(d_P(DM_D0(A)), d_P(DM_D1(A)));
      if (InColScope(Ajj)) 
        TmpBrd[0] = Tauj, TmpBrd[1] = Ajscale;
      BrdMat(1-DM_Trans(Ajj), d_P0(DM_DN(Ajj)), 1, 2, TmpBrd, 2);
      Tauj = TmpBrd[0], Ajscale = TmpBrd[1];
    }  
    { FDATA MTauj, MTaujAjscale;
      FD_NEG(MTauj, Tauj);		
      FD_MULSET(MTaujAjscale, =, MTauj, Ajscale);  
      D_ScaleAdd(1, N, MTauj, Aj0, MTaujAjscale, W);
    }  
    if (InEltScope(Ajj))		
      D_SetElt(Wj, Tauj);   

    if (!BlockV || InColScope(Ajj)) {
      CommSetUp(d_P(DM_D0(A)), d_P(DM_D1(A)));
      WS[n] = Ajscale;
      BrdMat(DM_Trans(Ajj), d_P0(DM_DM(Ajj)), 1, n+1, WS, n+1);
      Ajscale = WS[n];              
      D_SCAL(M-(j+1), Ajscale, Aj1j);
      DM_PRINT("Ajj", M-j, 1, Ajj);
      D_SCAL(M-(j+1), Ajscale, sVj1);
      
      if (InColScope(Ajj))
        Tauj = DM_Elt00(Wj);
    }  
    
    DM_PRINT("wj", 1, N-(j+1), Wj1);
    D_GER(M-j, N-(j+1), FD_1, sVj, Trans(Wj1), Ajj1);
     
    D_COPY(j+1, Trans(rW), Trans(SubMat(Tri, j, 0)));

    DM_PRINT("a", M-j, N-j, Ajj);

    if (!TauRowRep || InColScope(Ajj))     
      IxTau(Tau, j, A) = Tauj;

    sVj = SubMat_(sVj, sVj, 1, 1);	sVj1 = SubMat_(sVj1, sVj1, 1, 1);
  } /* for (j...) */

  MatFree(WS);
  
  if (BlockV) {
    int Ld = LM_Inc(1-DM_Trans(sV), DM_Loc(sV));	
    int m = G2Ld(M, DM_DM(sV));
    CommSetUp(d_P(DM_D0(A)), d_P(DM_D1(A)));
    BrdMat(1-DM_Trans(A), d_P0(DM_DN(A)), N, m, LM_ADR(DM_Loc(sV)), Ld);
    if (!TauRowRep)
      BrdMat(1-DM_Trans(A), d_P0(DM_DN(A)), 1, N, Tau, N);
  }

  TRACE_EXIT;
} /* D_GEQF2() */
  

void D_GEQRF(M, N, A, Tau, MaxW)
  int M, N, MaxW;
  DistMat A; FDATA Tau[];
{ int j;  				char svBrdTopChar[2];

  COMMON_ASSERT_TR_E(
     (TraceStr, "(%d,%d,%s,a%d,%d)", M, N, DM_TRACE(0, A), FD_ADR(Tau), MaxW), 
      DistMatOK(M, N, A), 0, "GEQRF", 0, TRACE_OUTER
  );

  svBrdTopChar[0] = BrdTopChar[0],      svBrdTopChar[1] = BrdTopChar[1];  
  BrdTopChar[DM_Trans(A)] = ' ',	BrdTopChar[1-DM_Trans(A)] = 'I'; 
  j=0;  
  while (j < MIN(M,N)) {
    DistMat_RVAlloc(8);
    DistMat Ajj = SubMat(A, j, j);	
    int minW = MAX(d_B(DM_DN(A)), MAX(MIN_BLOCKING_FACTOR, -MaxW));
    int maxW = MIN(ABS(MaxW), MIN(M,N)-j);
    int w = GenGetW(M-(j+1), 1.0, minW, maxW, DM_DN(Ajj), j, 1);
    DistMat Ajjw = SubMat(Ajj, 0, w), sT, rsT;
    FDATA *Tau_j = &IxTau(Tau, j, A);
    int BlockT = Blk1d(w, DM_DM(Ajj)) && Blk1d(w, DM_DN(Ajj));
    
    Tri = NewMat(w, w, ReduceDimMat(DM_Trans(Ajj), Ajj), NULL);	
    /* row-replicated triangular factor, set by D_GEQF2() */
    D_GEQF2(M-j, w, Ajj, Tau_j);    
    if (!BlockT) {			/* make it fully repliocated */
      sT = NewMat(w, w, ReduceDimMat(1-DM_Trans(Tri), Tri), NULL);
      D_ScaleAdd(w, w, FD_1, Tri, FD_0, sT);
    } else {
      /* only need workspace in col-replicated sT for cell row owning Ajj;
         `forge' sT to have this workspace */ 
      sT = SubMat(Ajj, 0, 0);
      LM_Copy(0, DM_Loc(sT), DM_Loc(Tri));    	
    }
    
    if (j+w < N) {
      D_LARFTComplete("F", w, sT);
      DM_PRINT("T", w, w, sT);
      if (BlockT) 
        rsT = sT;	
      else {      	
        /* select bits of fully replicated sT for this processor row */
        rsT = NewMat(w, w, ReduceDimMat(1-DM_Trans(Ajj), Ajj), NULL);
        D_ScaleAdd(w, w, FD_1, sT, FD_0, rsT);  	     
      }              
      D_LARFB("F", M-j, N-(j+w), w, sV, rsT, Ajjw);
      if (!BlockT)
        DisposeMat(rsT);	
    } /* if (j+w < N) */
    
    if (!BlockT)
      DisposeMat(sT);
    DisposeMat(sV);
    DisposeMat(Tri);
      
    DM_PRINT("a", M-j, N-j, Ajj);
    j += w;
  } /* for (j...) */
  BrdTopChar[0] = svBrdTopChar[0],      BrdTopChar[1] = svBrdTopChar[1];
  TRACE_EXIT;
} /* D_GEQRF() */
  

void D_GEQRM(M, N, A, Tau)
  int M, N;
  DistMat A; FDATA Tau[];
{ int j;	
  COMMON_ASSERT_TR_E(
     (TraceStr, "(%d,%d,%s,_)", M, N, DM_TRACE(0, A)),
      DistMatOK(M, N, A), 0, "GEQRM", 0, TRACE_OUTER
  );

  for (j = MIN(M,N)-1; j >= 0; j--) {
    FDATA Ajje, Tauj;
    DistMat_RVAlloc(7);    
    DistMat Ajj = SubMat(A, j, j), 	Ajj1 =  SubMat(A, j, j+1), 
    	    W;
    if (InColScope(Ajj)) {
      Ajje = D_GetElt(ColVec(Ajj));
      D_SetElt(Ajj, FD_1);
    }  

    W = NewMat(1, N, Ajj1, NULL);
#ifndef NO_LEVEL2
    D_GEMV(N-(j+1), M-j, FD_1, Trans(Ajj1), Ajj, FD_0, Trans(W)); 
#else    
    D_GEMM(1, N-(j+1), M-j, FD_1, Trans(Ajj), Ajj1, FD_0, W);
#endif    
    DM_PRINT("W", 1, N-(j+1), W);
 
    Tauj = IxTau(Tau, j, A);            /* corresp. global Tau[j]*/
    if (TauRowRep)
      RowBrdScalar(Tauj, Ajj);
    
    FD_NEG(Tauj, Tauj);          	/* nb. Tauj = -Tau[j] from here on */
#ifndef NO_LEVEL2
    D_GER(M-j, N-(j+1), Tauj, Ajj, Trans(W), Ajj1);
#else    
    D_GEMM(M-j, N-(j+1), 1, Tauj, Ajj, W, FD_1, Ajj1);
#endif
    DisposeMat(W);

    if (InColScope(Ajj)) {
      FDATA Ajscale;
      FD_MULSET(Ajscale, =, Ajje, Tauj);
      D_SCAL(M-(j+1), Ajscale, SubMat(A, j+1, j));
      FD_ASSIGN(Ajje, +=, Ajscale);
      D_SetElt(Ajj, Ajje);
    } 
    
    DM_PRINT("a", M-j, N-j, Ajj);
  } /* for (j...) */
  TRACE_EXIT;
} /* D_GEQRM() */

void D_GEORQM(M, K, N, A, Tau, C)
  int M, K, N;
  DistMat A; FDATA Tau[]; DistMat C;
{ int j;	
  COMMON_ASSERT_TR_E(
     (TraceStr, "(%d,%d,%d,%s,_,%s)", M, K, N, DM_TRACE(0, A), DM_TRACE(1, C)),
      DistMatOK(M, K, A) && DistMatOK(M, N, C), 0, "GEORQM", 0, TRACE_OUTER
  );

  for (/*j = MIN(M,K)-1; j >= 0; j--*/j=0; j<MIN(M,K); j++) {
    FDATA Ajje, Tauj;
    DistMat_RVAlloc(6);    
    DistMat Ajj = SubMat(A, j, j), 	Cj = SubMat(C, j, 0), W;
    
    if (InEltScope(Ajj)) {
      Ajje = D_GetElt(UnitVec(Ajj));
      D_SetElt(Ajj, FD_1);
    }  
    
    W = NewMat(1, N, Cj, NULL);

    if (N==1) 
      D_SetElt(W, D_DOT(M-j, Cj, Ajj));
    else 
      D_GEMV(N, M-j, FD_1, Trans(Cj), Ajj, FD_0, Trans(W));
    DM_PRINT("W", 1, N, W);

 
    Tauj = IxTau(Tau, j, A);            /* corresp. global Tau[j]*/
    if (TauRowRep)
      RowBrdScalar(Tauj, Ajj);  
    FD_NEG(Tauj, Tauj);              
    D_GER(M-j, N, Tauj, Ajj, Trans(W), Cj);
    DM_PRINT("C", M-j, N, Cj);
    DisposeMat(W);

    D_SetElt(Ajj, Ajje);
    
  } /* for (j...) */
  TRACE_EXIT;
} /* D_GEORQM() */


