 /************************************************************************/
/* 	DLAPACK: DBLAS-based distributed LAPACK routines		*/
/************************************************************************/
/* SCCS INFO @(#)D_LAPACK.h 1.3         last modified P. Strazdins, 99/05/03
   first created: P. Strazdins 96/06/12			
*/
/*
for compilation, the macro FDATATYPE must be defined.

Below, the underscores in the procedure name should be replaced by 'S',
if FDATATYPE=SINGLE (the DSBLAS procedures), and 'D', if FDATATYPE=DOUBLE
(the DDBLAS procedures). eg. for FDATATYPE=SINGLE, D_GETRF becomes DDGETRF.

See D_BLAS.h for a description of the DistMat matrices.
*/

#ifndef __D_LAPACK_INCL__
#define __D_LAPACK_INCL__

#if (FDATATYPE==SINGLE)
#define D_LASWP		DSLASWP
#define	D_GETF2		DSGETF2
#define D_GETRF		DSGETRF
#define D_LUTRM		DSLUTRM
#define D_POTF2		DSPOTF2
#define D_POTRF		DSPOTRF
#define D_POTRM		DSPOTRM
#define D_LARFT		DSLARFT
#define D_LARFB		DSLARFB
#define D_GEQF2		DSGEQF2
#define D_GEQRF		DSGEQRF
#define D_GEQRM		DSGEQRM
#define TauRowRep	STauRowRep
#define PivColRep	SPivColRep
#elif (FDATATYPE==DOUBLE) 
#define D_LASWP		DDLASWP
#define	D_GETF2		DDGETF2
#define D_GETF2		DDGETF2
#define D_GETRF		DDGETRF
#define D_LUTRM		DDLUTRM
#define D_POTF2		DDPOTF2
#define D_POTRF		DDPOTRF
#define D_POTRM		DDPOTRM
#define D_LARFT		DDLARFT
#define D_LARFB		DDLARFB
#define D_GEQF2		DDGEQR2
#define D_GEQRF		DDGEQRF
#define D_GEQRM		DDGEQRM
#define TauRowRep	DTauRowRep
#define PivColRep	DPivColRep
#elif (FDATATYPE==COMPLEX) 
#define D_LASWP		DCLASWP
#define	D_GETF2		DCGETF2
#define D_GETF2		DCGETF2
#define D_GETRF		DCGETRF
#define D_LUTRM		DCLUTRM
#define D_POTF2		DCPOTF2
#define D_POTRF		DCPOTRF
#define D_POTRM		DCPOTRM
#define D_LARFT		DCLARFT
#define D_LARFB		DCLARFB
#define D_GEQF2		DCGEQR2
#define D_GEQRF		DCGEQRF
#define D_GEQRM		DCGEQRM
#define TauRowRep	CTauRowRep
#define PivColRep	CPivColRep
#else
#define D_LASWP		DZLASWP
#define	D_GETF2		DZGETF2
#define D_GETF2		DZGETF2
#define D_GETRF		DZGETRF
#define D_LUTRM		DZLUTRM
#define D_POTF2		DZPOTF2
#define D_POTRF		DZPOTRF
#define D_POTRM		DZPOTRM
#define D_LARFT		DZLARFT
#define D_LARFB		DZLARFB
#define D_GEQF2		DZGEQR2
#define D_GEQRF		DZGEQRF
#define D_GEQRM		DZGEQRM
#define TauRowRep	ZTauRowRep
#define PivColRep	ZPivColRep
#endif


#include "FDATASetUp.h"
#include "DistMat.h"


#define RowBrdScalar(alpha, A)        \
        (alpha) = D_GetElt(RowVec(NewMat(1, 1, A, &(alpha))))
#define ColBrdScalar(alpha, A)        \
        (alpha) = D_GetElt(ColVec(NewMat(1, 1, A, &(alpha))))         

#define MIN_BLOCKING_FACTOR 4	/* minimum blocking factor applied to level 3 routines */

void D_LASWP STDC_PARAMS( 
	(char *Direction, int M, int N, DistMat A, int P[])
);
/* pre:	 0 <= P[0..N-1] and P[0..N-1] consistent across all cells
 *       A' is a global M x N matrix, DistMatOK(M, N, A')
 * post: performs column swaps A[0..M-1, j] <--> A[0..M-1, P[j]], 
 *	 for j=0,1,..,N-1 for *Direction='F', and j=N-1,N-2,...,0 ow.
 * note: 1. to use this procedure to swap K rows of A of length N, use 
 *		D_LASWP(Direct, N, K, Trans(A), P)
 *	 2. this differs from the ScaLAPACK equivalent in that 
 *	    the base address & values of P[] corresponds to column 0 of A
 *	 3. A may have N' >= N columns asociated with it; 
 *	    the values P[j] should be < N'  
 */
 
extern int PivColRep;	/* set to 1 of P[] is a col-replicated vector
			   (aligned with A; ow a row&column replicated vector*/



void D_GETF2 STDC_PARAMS( (int M, int N, DistMat A, int P[]) );
/* pre:  P has storage for at least N elements replicated over all cells
 *	 A is a global M x N matrix 
 *	 DistMatOK(M, N, A) && P!=NULL
 * post: A is overwritten by L\U, where L is stored in the proper lower
 *	 triagular portion of A (with implicit unit diagonal) 
 *	 and U is in upper triangular portion of A and
 *		P' A_pre = LU
 *	 where P' is matrix made from permutations P'[0], ... P'[N-1]
 *	 and P'[i] signifies to interchange rows i and P[i]
 */

void D_GETRF STDC_PARAMS( (int M, int N, DistMat A, int P[], int MaxW) );
/* pre/post: as for D_GETF2(M, N, A, P)
 * but with MaxW is maximum blocking factor
 */

  
void D_LUTRM STDC_PARAMS( (int M, int N, DistMat A, int P[], int  MaxW) );
/* pre:  P contains at least N elements and is replicated over all cells
 *	 A is a global M x N matrix; 
 *	 A = L\U, where L is stored in the proper lower
 *       triagular portion of A (with implicit unit diagonal)
 *       and U is in upper triangular portion of A 
 *	 DistMatOK(M, N, A) && P!=NULL
 * post: A = P' * L * U
 *	 where P' is matrix made from permutations P'[0], ... P'[N-1]
 *	 and P'[i] signifies to interchange rows i and P[i].
 *	 uses maximum blocking factoor of MaxW.
 */


void D_POTF2 STDC_PARAMS( (int N, DistMat A) );
/* pre:  a +ve definite global N x N symmetric matrix is stored in lower tri 
 *	 portion of A 
 *	 DistMatOK(N, N, A)
 * post: A is overwritten by L, where L is stored in the lower
 *	 triagular portion of A and:
 *		A_pre = L L^T
 */

void D_POTRF STDC_PARAMS( (int N, DistMat A, int MaxW) );
/* pre/post: as for D_POTF2(N, A)
 * but with MaxW is maximum blocking factor
 */

void D_POTRM STDC_PARAMS( (int N, DistMat A, int  MaxW) );
/* pre:  A is a global N x N lower triangular matrix; 
 *	 DistMatOK(N, N, A)
 * post: A = A*A^T	(lower half only)
 */

extern int TauRowRep;	/* set to 1 of Tau is a row-replicated vector
			   (aligned with A; ow a row&column replicated vector*/
 
void D_LARFT STDC_PARAMS( 
	(char *Direction, int N, int K, DistMat V, FDATA Tau[], DistMat T)
);
/* pre:	DistMatOK(N, K, V) && DistMatOK(K, K, T)
 * post: T is a KxK lower triagular factor of a reflector H = H[0]..H[K-1]
 *       where H = I - V^T * T^T * V
 */

void D_LARFB STDC_PARAMS(
	(char *Direction, int M, int N, int K, DistMat V, DistMat T, DistMat C)
);	 
/* post: apply block reflector H  where H = I - V^T*T*V to C
 */
 
void D_GEQF2  STDC_PARAMS( (int M, int N, DistMat A, FDATA Tau[] ) );
/* pre: Tau is a replicated across all cells & has storage for at least N elts.
 *      A is a global M x N square block-distributed block matrix
 *	DistMatOK(M, N, A) && Tau!=NULL
 * post A is overwritten by Q and R, where Q is in the proper lower
 *      triagular portion of A (with implicit unit diagonal)
 *      and R is in upper triangular portion of A and
 *      	A_pre = Orthogonal(Q, Tau)*R 
 */          

void D_GEQRF  STDC_PARAMS( (int M, int N, DistMat A, FDATA Tau[], int MaxW) );
/* pre/post: D_GEQR2() 
 * but with MaxW is maximum blocking factor
 */

void D_GEQRM STDC_PARAMS(  (int M, int N, DistMat A, FDATA Tau[]) );
/* pre: Tau is a replicated across all cells & contains at least N elts.
 *      A is a global M x N square block-distributed block matrix
 *	A = Q\R where Orthogonal(Q, Tau) represents an orthogonal matrix
 *	and R us upper triangular.
 *	DistMatOK(M, N, A) && Tau!=NULL
 * post A is overwritten by Orthogonal(Q, Tau)*R
 */          

void D_SYTRF STDC_PARAMS(
	(int N, DistMat A, FDATA *DInv, int P[], int *info, int MaxW)
);
/* pre:  P is a vector replicated over all cells, 
 *	 and has storage for at least N elements.
 *	 DInv is a column-replicated vector, and has storage for 3*n
 *	 elements, where n = G2Ld(N, DM_DM(A)) is the local column length of A.
 *	 A is a global symmetric N x N matrix 
 *	 DistMatOK(N, N, A) && *FUpLo == 'L' && d_B(DM_D0(A)) ==  d_B(DM_D1(A)) 
 * post: the lower half of A is overwritten by L\D, 
 *	 where L is stored in the proper lower
 *	 triangular portion of A (with implicit unit diagonal) 
 *	 and D is a diagonal matrix of 1x1 or 2x2 sub-blocks.
 *       (the latter occurs only where P[i] < 0) where 
 *		P' A_pre P'^T = L*D*L^T
 *	 where P' is matrix made from permutations P'[0], ... P'[N-1]
 *	 and P'[i] signifies to interchange rows and columns i and |P[i]|.
 *	 DInv contains elements of D^-1 for all diagonal elements
 *	 in this processor row. This enables optimizations in D_SYTRS(). 
 *       If *info==0, the inverse of A exists, otherwise *info = the column 
 *       index of the first zero column of D.
 * note: this uses the Bunch-Kaufman algorithm
 */

void D_SYTRS STDC_PARAMS(
	(int N, int K, DistMat A, FDATA *DInv, int P[], DistMat B, int *info)
);
/* pre:  P has storage for at least N elements replicated over all cells
 *	 A is a global N x N matrix, B is a N x K matrix. 
 *	 DistMatOK(N, N, A) && DistMatOK(N,K,B), P!=NULL, MDimAligned(X, A);
 *	 MDimAligned(X, Trans(A)).
 *       The lower half of A  contains L\D, 
 *	 where L is stored in the proper lower
 *	 triangular portion of A (with implicit unit diagonal) 
 *	 and D is a diagonal matrix of 1x1 or 2x2 sub-blocks
 *       (the latter occurs where P[i] < 0).
 *	 DInv contains elements of D^-1 for all diagonal elements
 *	 in this processor row. It must have been initialized by D_SYTRF().  
 * post: B <- P L^_T D^-1 L^-1 P^T B 
 *	 where P' is matrix made from permutations P'[0], ... P'[N-1]
 *	 and P'[i] signifies to interchange rows and columns i and |P[i]|
 */
  
int DILAENV STDC_PARAMS(
   (int FDATAsize, char *DLAPACK_proc_basename, int DBLAS_MachParamVec[], 
    int B, int P, int Q)
);
/* pre:  B, P, Q >=1; FDATAsize=2^k, where k=2..5
 * post: returns the platform-specific blocking factor  
 *	 for datatype FDATA where sizeof(FDATA) == FDATAsize
 *	 for DLAPACK procedure basename (``D_'' removed) DLAPACK_proc_basename,
 *	 for a block-cyclic matrix with B x B blocks distributed over
 *	 a P x Q grid.
 *	 DBLAS_MachParamVec[] should be the array for machine-dept
 *	 tuning parameters for the DBLAS library (file MachineParams.h);
 *	 if non-NULL, it may be used by this routine to compute the return 
 *	 value.
 * note: the infomration in the parameters to DILAENV() should be regarded
 *	 as a `hint' to the implementation of the routine.
 */
 
/*************************** misc macros ******************************/

#define DM_Elt00(A)		(*LM_ADR(DM_Loc(A)))
#define DM_TRACE(n, A)  PrintDistMat(A, TmpTraceStrs[n])

#ifdef PRINTMAT
# define DM_PRINT(name, M, N, A)                                        \
        if (PrintRes)                                                   \
          MatPrint(name, MIN(PrintRes, G2Ld(!DM_Trans(A)? M: N, DM_D0(A))), \
                         MIN(PrintRes, G2Ld(!DM_Trans(A)? N: M, DM_D1(A))), \
                         DM_Loc(A))
#else
# define DM_PRINT(name, M, N, A)
#endif
   	
   
#endif /*__D_LAPACK_INCL__*/
