/************************************************************************/
/*         PBLAS interface to the DBLAS                                 */
/************************************************************************/
/* SCCS INFO @(#)PBLAS.h 1.2         last modified P. Strazdins, 99/05/03
   first created: P. Strazdins 96/10
 */
/* 
Header file for PBLAS interface to the DBLAS
for compilation, the macro FDATATYPE (having the values SINGLE or DOUBLE)
must be defined:
	if FDATATYPE=SINGLE, FDATA=float
	if FDATATYPE=DOUBLE, FDATA=double
Below, the underscores in the procedure name should be replaced by 'S',
if FDATATYPE=SINGLE (the PSBLAS procedures), and 'D', if FDATATYPE=DOUBLE
(the PDBLAS procedures). eg. for FDATATYPE=SINGLE, P_GEMM becomes psgemm_.

See ScaL:APACK documentation on the meaning of the procedures.
*/   

#ifndef __P_BLAS_INCL__
#define __P_BLAS_INCL__

#include "FDATASetUp.h"
#include "F2C.h"

#ifndef ALTNAMES			/* use standard PBLAS names */
#if (FDATATYPE==SINGLE)		/*define procedure identifiers and names*/
# define P_GEMM  	psgemm_
# define P_SYMM  	pssymm_
# define P_SYRK  	pssyrk_
# define P_SYR2K 	psyr2k_
# define P_TRMM  	pstrmm_
# define P_TRSM  	pstrsm_

# define P_GEMV	 	psgemv_	 
# define P_SYMV	 	pssymv_	 
# define P_TRMV	 	pstrmv_	 
# define P_TRSV	 	pstrsv_	 
# define P_GER	 	psger_	 
# define P_SYR	 	pssyr_	 
# define P_SYR2	 	pssyr2_	 

# define P_SWAP	 	psswap_	 
# define P_SCAL	 	psscal_	 
# define P_COPY	 	pscopy_	 
# define P_AXPY	 	psaxpy_	 
# define P_DOT	 	psdot_	 
# define P_NRM2	 	psnrm2_	 
# define P_ASUM	 	psasum_	 
# define P_AMAX		psamax_	 

# define P_TRAN		pstran_

# define P_ELSET	pselset_
# define P_ELGET	pselget_
 
#elif (FDATATYPE==DOUBLE)
# define P_GEMM  	pdgemm_
# define P_SYMM  	pdsymm_
# define P_SYRK  	pdsyrk_
# define P_SYR2K 	pdsyr2k_
# define P_TRMM  	pdtrmm_
# define P_TRSM  	pdtrsm_

# define P_GEMV	 	pdgemv_	 
# define P_SYMV	 	pdsymv_	 
# define P_TRMV	 	pdtrmv_	 
# define P_TRSV	 	pdtrsv_	 
# define P_GER	 	pdger_	 
# define P_SYR	 	pdsyr_	 
# define P_SYR2	 	pdsyr2_	 

# define P_SWAP	 	pdswap_	 
# define P_SCAL	 	pdscal_	 
# define P_COPY	 	pdcopy_	 
# define P_AXPY	 	pdaxpy_	 
# define P_DOT	 	pddot_	 
# define P_NRM2	 	pdnrm2_	 
# define P_ASUM	 	pdasum_	 
# define P_AMAX		pdamax_	 

# define P_TRAN		pdtran_

# define P_ELSET	pdelset_
# define P_ELGET	pdelget_
 
#elif (FDATATYPE==COMPLEX)

# define P_GEMM  	pcgemm_
# define P_SYMM  	pcsymm_
# define P_SYRK  	pcsyrk_
# define P_SYR2K 	pcsyr2k_
# define P_TRMM  	pctrmm_
# define P_TRSM  	pctrsm_

# define P_GEMV	 	pcgemv_	 
# define P_SYMV	 	pcsymv_	 
# define P_TRMV	 	pctrmv_	 
# define P_TRSV	 	pctrsv_	 
# define P_GER	 	pcgeru_	 
# define P_SYR	 	P_HYR
# define P_SYR2	 	P_HYR2

# define P_SWAP	 	pcswap_	 
# define P_SCAL	 	pcscal_	 
# define P_COPY	 	pccopy_	 
# define P_AXPY	 	pcaxpy_	 
# define P_DOT	 	pcdotu_	 
# define P_NRM2	 	pscnrm2_	 
# define P_ASUM	 	pscasum_	 
# define P_AMAX		pcamax_	 

# define P_TRAN		pctran_

# define P_ELSET	pcelset_
# define P_ELGET	pcelget_
 
#else

# define P_GEMM  	pzgemm_
# define P_SYMM  	pzsymm_
# define P_SYRK  	pzsyrk_
# define P_SYR2K 	pzsyr2k_
# define P_TRMM  	pztrmm_
# define P_TRSM  	pztrsm_

# define P_GEMV	 	pzgemv_	 
# define P_SYMV	 	pzsymv_	 
# define P_TRMV	 	pztrmv_	 
# define P_TRSV	 	pztrsv_	 
# define P_GER	 	pzgeru_	 
# define P_SYR	 	P_HYR
# define P_SYR2	 	P_HYR2

# define P_SWAP	 	pzswap_	 
# define P_SCAL	 	pzscal_	 
# define P_COPY	 	pzcopy_	 
# define P_AXPY	 	pzaxpy_	 
# define P_DOT	 	pzdotu_	 
# define P_NRM2	 	pdznrm2_	 
# define P_ASUM	 	pdzasum_	 
# define P_AMAX		pzamax_	 

# define P_TRAN		pztran_

# define P_ELSET	pzelset_
# define P_ELGET	pzelget_
 
#endif /*FDATATYPE*/

#ifdef NoChange
# define psgemm_	psgemm
# define pssymm_	pssymm
# define pssyrk_	pssyrk
# define pssyr2k_	pssyr2k
# define pstrmm_	pstrmm
# define pstrsm_	pstrsm
# define psgemv_	psgemv
# define pssymv_	pssymv
# define pstrmv_	pstrmv
# define psger_		psger
# define pssyr_		pssyr
# define pssyr2_	pssyr2
# define psswap_	psswap
# define psscal_	psscal
# define pscopy_	pscopy
# define psaxpy_	psaxpy
# define psdot_		psdot
# define psnrm2_	psnrm2
# define psasum_	psasum
# define psamax_	psamax
# define pstran_	pstran
# define pselset_	pselset
# define pselget_	pselget

# define pdgemm_	pdgemm
# define pdsymm_	pdsymm
# define pdsyrk_	pdsyrk
# define pdsyr2k_	pdsyr2k
# define pdtrmm_	pdtrmm
# define pdtrsm_	pdtrsm
# define pdgemv_	pdgemv
# define pdsymv_	pdsymv
# define pdtrmv_	pdtrmv
# define pdger_		pdger
# define pdsyr_		pdsyr
# define pdsyr2_	pdsyr2
# define pdswap_	pdswap
# define pdscal_	pdscal
# define pdcopy_	pdcopy
# define pdaxpy_	pdaxpy
# define pddot_		pddot
# define pdnrm2_	pdnrm2
# define pdasum_	pdasum
# define pdamax_	pdamax
# define pdtran_	pdtran
# define pdelset_	pdelset
# define pdelget_	pdelget

# define pcgemm_	pcgemm
# define pcsymm_	pcsymm
# define pcsyrk_	pcsyrk
# define pcsyr2k_	pcsyr2k
# define pctrmm_	pctrmm
# define pctrsm_	pctrsm
# define pcgemv_	pcgemv
# define pcsymv_	pcsymv
# define pctrmv_	pctrmv
# define pcger_		pcger
# define pcsyr_		pcsyr
# define pcsyr2_	pcsyr2
# define pcswap_	pcswap
# define pcscal_	pcscal
# define pccopy_	pccopy
# define pcaxpy_	pcaxpy
# define pcdot_		pcdot
# define pscnrm2_	pscnrm2
# define pscasum_	pscasum
# define pcamax_	pcamax
# define pctran_	pctran
# define pcelset_	pcelset
# define pcelget_	pcelget

# define pzgemm_	pzgemm
# define pzsymm_	pzsymm
# define pzsyrk_	pzsyrk
# define pzsyr2k_	pzsyr2k
# define pztrmm_	pztrmm
# define pztrsm_	pztrsm
# define pzgemv_	pzgemv
# define pzsymv_	pzsymv
# define pztrmv_	pztrmv
# define pzger_		pzger
# define pzsyr_		pzsyr
# define pzsyr2_	pzsyr2
# define pzswap_	pzswap
# define pzscal_	pzscal
# define pzcopy_	pzcopy
# define pzaxpy_	pzaxpy
# define pzdot_		pzdot
# define pdznrm2_	pdznrm2
# define pdzasum_	pdzasum
# define pzamax_	pzamax
# define pztran_	pztran
# define pzelset_	pzelset
# define pzelget_	pzelget

#endif /* NoChange */

#ifdef UpCase
# define psgemm_	PSGEMM
# define pssymm_	PSSYMM
# define pssyrk_	PSSYRK
# define pssyr2k_	PSSYR2K
# define pstrmm_	PSTRMM
# define pstrsm_	PSTRSM
# define psgemv_	PSGEMV
# define pssymv_	PSSYMV
# define pstrmv_	PSTRMV
# define psger_		PSGER
# define pssyr_		PSSYR
# define pssyr2_	PSSYR2
# define psswap_	PSSWAP
# define psscal_	PSSCAL
# define pscopy_	PSCOPY
# define psaxpy_	PSAXPY
# define psdot_		PSDOT
# define psnrm2_	PSNRM2
# define psasum_	PSASUM
# define psamax_	PSAMAX
# define pstran_	PSTRAN
# define pselset_	PSELSET
# define pselget_	PSELGET

# define pdgemm_	PDGEMM
# define pdsymm_	PDSYMM
# define pdsyrk_	PDSYRK
# define pdsyr2k_	PDSYR2K
# define pdtrmm_	PDTRMM
# define pdtrsm_	PDTRSM
# define pdgemv_	PDGEMV
# define pdsymv_	PDSYMV
# define pdtrmv_	PDTRMV
# define pdger_		PDGER
# define pdsyr_		PDSYR
# define pdsyr2_	PDSYR2
# define pdswap_	PDSWAP
# define pdscal_	PDSCAL
# define pdcopy_	PDCOPY
# define pdaxpy_	PDAXPY
# define pddot_		PDDOT
# define pdnrm2_	PDNRM2
# define pdasum_	PDASUM
# define pdamax_	PDAMAX
# define pdtran_	PDTRAN
# define pdelset_	PDELSET
# define pdelget_	PDELGET

# define pcgemm_	PCGEMM
# define pcsymm_	PCSYMM
# define pcsyrk_	PCSYRK
# define pcsyr2k_	PCSYR2K
# define pctrmm_	PCTRMM
# define pctrsm_	PCTRSM
# define pcgemv_	PCGEMV
# define pcsymv_	PCSYMV
# define pctrmv_	PCTRMV
# define pcger_		PCGER
# define pcsyr_		PCSYR
# define pcsyr2_	PCSYR2
# define pcswap_	PCSWAP
# define pcscal_	PCSCAL
# define pccopy_	PCCOPY
# define pcaxpy_	PCAXPY
# define pcdot_		PCDOT
# define pscnrm2_	PSCNRM2
# define pscasum_	PSCASUM
# define pcamax_	PCAMAX
# define pctran_	PCTRAN
# define pcelset_	PCELSET
# define pcelget_	PCELGET

# define pzgemm_	PZGEMM
# define pzsymm_	PZSYMM
# define pzsyrk_	PZSYRK
# define pzsyr2k_	PZSYR2K
# define pztrmm_	PZTRMM
# define pztrsm_	PZTRSM
# define pzgemv_	PZGEMV
# define pzsymv_	PZSYMV
# define pztrmv_	PZTRMV
# define pzger_		PZGER
# define pzsyr_		PZSYR
# define pzsyr2_	PZSYR2
# define pzswap_	PZSWAP
# define pzscal_	PZSCAL
# define pzcopy_	PZCOPY
# define pzaxpy_	PZAXPY
# define pzdot_		PZDOT
# define pdznrm2_	PDZNRM2
# define pdzasum_	PDZASUM
# define pzamax_	PZAMAX
# define pztran_	PZTRAN
# define pzelset_	PZELSET
# define pzelget_	PZELGET

#endif /* UpCase */

#else				/* use names prepended with a `d' to 
				   avoid clashes with the UTK PBLAS */
#if (FDATATYPE==SINGLE)		/* define procedure identifiers and names*/
# define P_GEMM  	dpsgemm_
# define P_SYMM  	dpssymm_
# define P_SYRK  	dpssyrk_
# define P_SYR2K 	dpsyr2k_
# define P_TRMM  	dpstrmm_
# define P_TRSM  	dpstrsm_

# define P_GEMV	 	dpsgemv_	 
# define P_SYMV	 	dpssymv_	 
# define P_TRMV	 	dpstrmv_	 
# define P_TRSV	 	dpstrsv_	 
# define P_GER	 	dpsger_	 
# define P_SYR	 	dpssyr_	 
# define P_SYR2	 	dpssyr2_	 

# define P_SWAP	 	dpsswap_	 
# define P_SCAL	 	dpsscal_	 
# define P_COPY	 	dpscopy_	 
# define P_AXPY	 	dpsaxpy_	 
# define P_DOT	 	dpsdot_	 
# define P_NRM2	 	dpsnrm2_	 
# define P_ASUM	 	dpsasum_	 
# define P_AMAX		dpsamax_	 

# define P_TRAN		dpstran_

# define P_ELSET	dpselset_
# define P_ELGET	dpselget_
 
#elif (FDATATYPE==DOUBLE)
# define P_GEMM  	dpdgemm_
# define P_SYMM  	dpdsymm_
# define P_SYRK  	dpdsyrk_
# define P_SYR2K 	dpdsyr2k_
# define P_TRMM  	dpdtrmm_
# define P_TRSM  	dpdtrsm_

# define P_GEMV	 	dpdgemv_	 
# define P_SYMV	 	dpdsymv_	 
# define P_TRMV	 	dpdtrmv_	 
# define P_TRSV	 	dpdtrsv_	 
# define P_GER	 	dpdger_	 
# define P_SYR	 	dpdsyr_	 
# define P_SYR2	 	dpdsyr2_	 

# define P_SWAP	 	dpdswap_	 
# define P_SCAL	 	dpdscal_	 
# define P_COPY	 	dpdcopy_	 
# define P_AXPY	 	dpdaxpy_	 
# define P_DOT	 	dpddot_	 
# define P_NRM2	 	dpdnrm2_	 
# define P_ASUM	 	dpdasum_	 
# define P_AMAX		dpdamax_	 

# define P_TRAN		dpdtran_

# define P_ELSET	dpdelset_
# define P_ELGET	dpdelget_
 
#endif /*FDATATYPE*/

#ifdef NoChange
# define dpsgemm_	dpsgemm
# define dpssymm_	dpssymm
# define dpssyrk_	dpssyrk
# define dpssyr2k_	dpssyr2k
# define dpstrmm_	dpstrmm
# define dpstrsm_	dpstrsm
# define dpsgemv_	dpsgemv
# define dpssymv_	dpssymv
# define dpstrmv_	dpstrmv
# define dpsger_	dpsger
# define dpssyr_	dpssyr
# define dpssyr2_	dpssyr2
# define dpsswap_	dpsswap
# define dpsscal_	dpsscal
# define dpscopy_	dpscopy
# define dpsaxpy_	dpsaxpy
# define dpsdot_	dpsdot
# define dpsnrm2_	dpsnrm2
# define dpsasum_	dpsasum
# define dpsamax_	dpsamax
# define dpstran_	dpstran
# define dpselset_	dpselset
# define dpselget_	dpselget

# define dpdgemm_	dpdgemm
# define dpdsymm_	dpdsymm
# define dpdsyrk_	dpdsyrk
# define dpdsyr2k_	dpdsyr2k
# define dpdtrmm_	dpdtrmm
# define dpdtrsm_	dpdtrsm
# define dpdgemv_	dpdgemv
# define dpdsymv_	dpdsymv
# define dpdtrmv_	dpdtrmv
# define dpdger_	dpdger
# define dpdsyr_	dpdsyr
# define dpdsyr2_	dpdsyr2
# define dpdswap_	dpdswap
# define dpdscal_	dpdscal
# define dpdcopy_	dpdcopy
# define dpdaxpy_	dpdaxpy
# define dpddot_	dpddot
# define dpdnrm2_	dpdnrm2
# define dpdasum_	dpdasum
# define dpdamax_	dpdamax
# define dpdtran_	dpdtran
# define dpdelset_	dpdelset
# define dpdelget_	dpdelget
#endif /* NoChange */

#ifdef UpCase
# define dpsgemm_	DPSGEMM
# define dpssymm_	DPSSYMM
# define dpssyrk_	DPSSYRK
# define dpssyr2k_	DPSSYR2K
# define dpstrmm_	DPSTRMM
# define dpstrsm_	DPSTRSM
# define dpsgemv_	DPSGEMV
# define dpssymv_	DPSSYMV
# define dpstrmv_	DPSTRMV
# define dpsger_	DPSGER
# define dpssyr_	DPSSYR
# define dpssyr2_	DPSSYR2
# define dpsswap_	DPSSWAP
# define dpsscal_	DPSSCAL
# define dpscopy_	DPSCOPY
# define dpsaxpy_	DPSAXPY
# define dpsdot_	DPSDOT
# define dpsnrm2_	DPSNRM2
# define dpsasum_	DPSASUM
# define dpsamax_	DPSAMAX
# define dpstran_	DPSTRAN
# define dpselset_	DPSELSET
# define dpselget_	DPSELGET

# define dpdgemm_	DPDGEMM
# define dpdsymm_	DPDSYMM
# define dpdsyrk_	DPDSYRK
# define dpdsyr2k_	DPDSYR2K
# define dpdtrmm_	DPDTRMM
# define dpdtrsm_	DPDTRSM
# define dpdgemv_	DPDGEMV
# define dpdsymv_	DPDSYMV
# define dpdtrmv_	DPDTRMV
# define dpdger_	DPDGER
# define dpdsyr_	DPDSYR
# define dpdsyr2_	DPDSYR2
# define dpdswap_	DPDSWAP
# define dpdscal_	DPDSCAL
# define dpdcopy_	DPDCOPY
# define dpdaxpy_	DPDAXPY
# define dpddot_	DPDDOT
# define dpdnrm2_	DPDNRM2
# define dpdasum_	DPDASUM
# define dpdamax_	DPDAMAX
# define dpdtran_	DPDTRAN
# define dpdelset_	DPDELSET
# define dpdelget_	DPDELGET
#endif /* UpCase */


#endif /* ALTNAMES */				   			   

#ifdef NoChange
# define pberror_	pberror
# define ptopget_	ptopget
#endif /* NoChange */

#ifdef UpCase
# define pberror_	PBERROR
# define ptopget_	PTOPGET
#endif /* UpCase */




/*********************** PBLAS descriptor indices ************************/

#define    DLEN_        9                     /* Length of a descriptor */
#define    DT_          0                     /* Descriptor Type        */
#define    CTXT_        1                              /* BLACS context */
#define    M_           2                      /* Global Number of Rows */
#define    N_           3                   /* Global Number of Columns */
#define    MB_          4                          /* Row Blocking Size */
#define    NB_          5                       /* Column Blocking Size */
#define    RSRC_        6                     /* Starting Processor Row */
#define    CSRC_        7                  /* Starting Processor Column */
#define    LLD_         8                    /* Local Leading Dimension */

/*********************** level 3 procedures ******************************/

void P_GEMM STDC_PARAMS(
	(F_CHAR_T TransA, F_CHAR_T TransB, int *m, int *n, int *k, 
	 FDATA *alpha, FDATA A[], int *ia, int *ja, int desc_A[],
	 FDATA B[], int *ib, int *jb, int desc_B[], FDATA *beta, 
	 FDATA C[], int *ic, int *jc, int desc_C[])
);

void P_SYMM STDC_PARAMS(
	(F_CHAR_T Side, F_CHAR_T UpLo, int *m, int *n, FDATA *alpha, 
	 FDATA A[], int *ia, int *ja, int desc_A[], 
	 FDATA B[], int *ib, int *jb, int desc_B[], FDATA *beta, 
	 FDATA C[], int *ic, int *jc, int desc_C[])
);	

void P_SYRK STDC_PARAMS(
	(F_CHAR_T UpLo, F_CHAR_T Trans, int *n, int *k, FDATA *alpha, 
	 FDATA A[], int *ia, int *ja, int desc_A[], FDATA *beta, 
	 FDATA C[], int *ic, int *jc, int desc_C[])
);	

void P_SYR2K STDC_PARAMS(
	(F_CHAR_T UpLo, F_CHAR_T Trans, int *n, int *k, FDATA *alpha, 
	 FDATA A[], int *ia, int *ja, int desc_A[], 
	 FDATA B[], int *ib, int *jb, int desc_B[], FDATA *beta, 
	 FDATA C[], int *ic, int *jc, int desc_C[])
);	


void P_TRMM STDC_PARAMS(
	(F_CHAR_T Side, F_CHAR_T UpLo, F_CHAR_T TransA, F_CHAR_T Diag, int *m, int *n, 
	 FDATA *alpha, FDATA A[], int *ia, int *ja, int desc_A[], 
	 FDATA B[], int *ib, int *jb, int desc_B[])
); 


void P_TRSM STDC_PARAMS(
	(F_CHAR_T Side, F_CHAR_T Uplo, F_CHAR_T TransA, F_CHAR_T Diag, int *m, int *n, 
	 FDATA *alpha, FDATA A[], int *ia, int *ja, int desc_A[], 
	 FDATA B[], int *ib, int *jb, int desc_B[])
); 

/*********************** level 2 procedures ******************************/ 

void P_GEMV STDC_PARAMS(
	(F_CHAR_T Trans, int *m, int *n, FDATA *alpha, 
	 FDATA A[], int *ia, int *ja, int desc_A[], 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx, FDATA *beta, 
	 FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy)
);

void P_SYMV STDC_PARAMS(
	(F_CHAR_T UpLo, int *n, FDATA *alpha, 
	 FDATA A[], int *ia, int *ja, int desc_A[], 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx, FDATA *beta, 
	 FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy)
);

void P_TRMV STDC_PARAMS(
        (F_CHAR_T UpLo, F_CHAR_T Trans, F_CHAR_T Diag, int *n, 
         FDATA A[], int *ia, int *ja, int desc_A[], 
         FDATA X[], int *ix, int *jx, int desc_X[], int *incx)
);

void P_TRSV STDC_PARAMS(
        (F_CHAR_T UpLo, F_CHAR_T Trans, F_CHAR_T Diag, int *n, 
        FDATA A[], int *ia, int *ja, int desc_A[], 
        FDATA X[], int *ix, int *jx, int desc_X[], int *incx)
);

void P_GER STDC_PARAMS(
	(int *m, int *n, FDATA *alpha, 
	FDATA X[], int *ix, int *jx, int desc_X[], int *incx, 
	FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy, 
	FDATA A[], int *ia, int *ja, int desc_A[])
);

void P_SYR STDC_PARAMS(
	(F_CHAR_T UpLo, int *n, FDATA *alpha, 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx, 
	 FDATA A[], int *ia, int *ja, int desc_A[])
);	

void P_SYR2 STDC_PARAMS(
	(F_CHAR_T UpLo, int *n, FDATA *alpha, 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx, 
	 FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy, 
	 FDATA A[], int *ia, int *ja, int desc_A[])
);	

/*********************** level 1 procedures ******************************/ 

void P_SWAP STDC_PARAMS( 
	(int *n, FDATA X[], int *ix, int *jx, int desc_X[], int *incx, 
		 FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy) 
);

void P_SCAL STDC_PARAMS( 
	(int *n, FDATA *alpha, 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx) 
);

void P_COPY STDC_PARAMS(
	(int *n, FDATA X[], int *ix, int *jx, int desc_X[], 
	 int *incx, FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy)
);

void P_AXPY STDC_PARAMS(  
	(int *n, FDATA *alpha, 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx, 
	 FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy)
);

void P_DOT STDC_PARAMS( 
	(int *n, FDATA *Dot, 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx, 
	 FDATA Y[], int *iy, int *jy, int desc_Y[], int *incy) 
);

void P_NRM2 STDC_PARAMS( 
	(int *n, FD_BASE *Nrm2, 
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx) 
);	

void P_ASUM STDC_PARAMS( 
	(int *n, FD_BASE *ASum,
	 FDATA X[], int *ix, int *jx, int desc_X[], int *incx) 
);

void P_AMAX STDC_PARAMS( 
	(int *n, FDATA *Max, int *MaxIndx,
	FDATA X[], int *ix, int *jx, int desc_X[], int *incx) 
);

/********************** new parallel level 1 procedures  ********************/

void P_TRAN STDC_PARAMS(
	(int *m, int *n, FDATA *alpha, 
	 FDATA A[], int *ia, int *ja, int desc_A[], FDATA *beta, 
	 FDATA C[], int *ic, int *jc, int desc_C[])
);


/********************** PBLAS tools  procedures  ********************/

void pberror_ STDC_PARAMS( (int *cntxt, F_CHAR_T Msg, int *info ) );

void ptopget_ STDC_PARAMS( 
	(int *cntxt, F_CHAR_T toptype, F_CHAR_T rowcol, F_CHAR_T top ) 
);

void P_ELGET STDC_PARAMS(
	(F_CHAR_T Scope, F_CHAR_T Top, FDATA *Aij,
	FDATA A[], int *ia, int *ja, int desc_A[]) 
);

void P_ELSET STDC_PARAMS(
	(FDATA A[], int *ia, int *ja, int desc_A[], FDATA *Aij) 
);


#endif /*__P_BLAS_INCL__*/
