/************************************************************************/
/*	common macro definitions and headers for files implementing 	*
 *      the PBLAS interface to the DBLAS. Includes error checking macros etc*/ 
/************************************************************************/
/* SCCS INFO @(#)PBLASCheck.h 1.2         last modified P. Strazdins, 99/05/03
   first created: P. Strazdins 96/12
*/
/* compilation: if USE_PBERROR_ defined, use pberror_() as the PBLAS error
 *		routine (default: use GenAssert.Error()).
 */ 
/*#define USE_PBERROR_*/
    
#ifndef __PBLASCheck_INCL__
#define __PBLASCheck_INCL__

/******************************  common include files *********************/

#include "BLASOptions.h"
#include "FDATASetUp.h"
#include "IntOps.h"
# undef CALLER_PREFIX
#if (FDATATYPE==SINGLE)
#define  CALLER_PREFIX "PS"
#elif (FDATATYPE==DOUBLE)
#define  CALLER_PREFIX "PD"
#elif (FDATATYPE==COMPLEX)
#define  CALLER_PREFIX "PC"
#else
#define  CALLER_PREFIX "PZ"
#endif
#include "GenAssert.h"
 
#include "FtnBLACS.h"
#include "LocMat.h"
#include "FDCommun.h"
#include "MatDim.h"
#include "DistMat.h"

#include "PBLAS.h"
#include "D_BLAS.h"
#include "D_BLASCheck.h"

/******************************  common procedures *********************/


#if (FDATATYPE==SINGLE)         /*define procedure identifiers and names*/
# define PDescChkConv	PSDescChkConv
# define PBError_	PSBError_
# define PDescCacheOn   psdesccacheon_	
# define PDescCacheOff  psdesccacheoff_	
# define DBLASSetGet	dsblassetget_
#elif (FDATATYPE==DOUBLE)
# define PDescChkConv	PDDescChkConv
# define PBError_	PDBError_
# define PDescCacheOn   pddesccacheon_	
# define PDescCacheOff  pddesccacheoff_	
# define DBLASSetGet	ddblassetget_
#elif (FDATATYPE==COMPLEX)
# define PDescChkConv	PCDescChkConv
# define PBError_	PCBError_
# define PDescCacheOn   pcdesccacheon_	
# define PDescCacheOff  pcdesccacheoff_	
# define DBLASSetGet	dcblassetget_
#else
# define PDescChkConv	PZDescChkConv
# define PBError_	PZBError_
# define PDescCacheOn   pzdesccacheon_	
# define PDescCacheOff  pzdesccacheoff_	
# define DBLASSetGet	dzblassetget_
#endif /*FDATATYPE*/


#ifdef NoChange
# define psdesccacheon_		psdesccacheon
# define psdesccacheoff_	psdesccacheoff
# define dsblassetget_		dsblassetget
# define pddesccacheon_		pddesccacheon
# define pddesccacheoff_	pddesccacheoff
# define ddblassetget_		ddblassetget
# define pcdesccacheon_		pcdesccacheon
# define pcdesccacheoff_	pcdesccacheoff
# define dcblassetget_		dcblassetget
# define pzdesccacheon_		pzdesccacheon
# define pzdesccacheoff_	pzdesccacheoff
# define dzblassetget_		dzblassetget
#endif 

#ifdef UpCase
# define psdesccacheon_		PSDESCCACHEON
# define psdesccacheoff_	PSDESCCACHEOFF
# define dsblassetget_		DSBLASSETGET
# define pddesccacheon_		PDDESCCACHEON
# define pddesccacheoff_	PDDESCCACHEOFF
# define ddblassetget_		DDBLASSETGET
# define pcdesccacheon_		PCDESCCACHEON
# define pcdesccacheoff_	PCDESCCACHEOFF
# define dcblassetget_		DCBLASSETGET
# define pzdesccacheon_		PZDESCCACHEON
# define pzdesccacheoff_	PZDESCCACHEOFF
# define dzblassetget_		DZBLASSETGET
#endif 


int PDescChkConv STDC_PARAMS(
	( int Pix, int TrA, int M, int N, FDATA *Ax, int iA, int jA, 
	  int DescA[],  int ParamInfo, DistMat *dA)
);
/* pre:  A' = (Ax, iA, jA, DescA) is an M x N matrix.
 *	 ParamNums is the parameter position of DescA (those of iA & jA are 
 *	 assumed to immediately precede that of DescA, etc).
 *	 TrA in {0, 1} indicates whether A' is transposed.
 *	 Pix in {0, 1, 2} is a PBLAS call parameter index for the matrix.
 *	 For a PBLAS call with 3 matrix parameters, the first call to this
 *	 routine must have Pix=0; subsequent calls must have mutually 
 *	 different values of Pix (this is not checked).
 * post: checks for validity of A'; if invalid, returns the 
 *	 parameter number of the first faulty parameter found; o.w. returns 0.
 *	 *dA is the DistMat equivalent of A'
 * note: (inter call) PBLAS descriptor caching is performed for Pix=0;
 *	 for Pix>0, caching is performed with respect to the last call of this
 *	 routine for Pix=0.  Any one of matrix parameters in a PBVLAS call
 *       may be converted with Pix=0; however, it is recommended that it
 *	 be used for the proper matrix parameter in level 2 calls, and the
 *	 triangular/symmetric matrix parameter in level 3 calls. 
 */
 

void PBError_ STDC_PARAMS(
        ( char *CallerPrefix, char *Caller, int ParamNum )
);
/* pre:	CallerPrefix##Caller is the procedure name
 *	GenAssert.TraceStr has the vaues of the procedure's paramters.
 * post:call the PBLAS (PTOOLS) error routine, with the current BLACS context, 
 *	procedure name, and ParamNum.
 *      A trace of the procedure call is printed before this.
 */

/****** Fortran callable procedures for controlling tracing/tuning *******/

void PDescCacheOn STDC_PARAMS( (void) );
/* post: turn on PBLAS matrix descriptor - DistMat cache between PBLAS calls
 */
         
void PDescCacheOff STDC_PARAMS( (void) );
/* post: turn off PBLAS matrix descriptor - DistMat cache between PBLAS calls
 */

void DBLASSetGet STDC_PARAMS( 
	(int *GetSet, int *ParamCat, int *ParamNum, int *Val)
);
/* post: case (*ParamCat, *ParamNum) of:
 *	 	(0, 0..MachineParams.NUM_MACH_PARAMS): 
 *			v = &MachineParams.MachParamVec[*ParamNum]
 *		(1, 0): v = &GenAssert.TraceLevel 		
 *		(1, 1): v = &GenAssert.TraceType		
 *		(2, 0): v = &GenAssert.PrintRes		
 *		(3, 0): v = &GenAssert.AlgVers
 *	 else v = NULL; 
 *	 if (v = NULL) *Val = -999;
 *	 else if (*GetSet==1) *Val = *v else *v = *Val
 * notes: 1. *ParamCat==0 only effective for DBLAS compiled with -DTUNE.
 *	     This can be used for tuning architecture-depedent parameters
 *	     used in DBLAS algorithms. See MachineParams.h.
 *        2. *ParamCat==1 only effective for DBLAS compiled with -DTRACE.
 *	     Setting TraceLevel = 0 will cause tracing of outer-level 
 *	     PBLAS calls; setting TraceLevel > 0 will also cause tracing
 *	     of internal DBLAS procedures up to depth TraceLevel.
 *	     TraceType can be used for restriciting which categories of 
 *	     routines are to be traced.	See GenAssert.h for details.
 *	  3. *ParamCat==2 can be used to cause printing of up to *Val x *Val
 *	     matrices in the DBLAS SpreadMatrix module. This can display
 *	     useful intermediate results in most computations.
 *	  4. ParamCat==3 can be used to force a particular algorithm variant
 *	     within the DBLAS to be executed.
 */  	     
      
/************** macros used in error checking ********************************/

#define DMLocSetUp(Trans, Ax, desca, Aval, A)				\
	(A) = &(Aval), LM_Init(DM_Loc(A), Ax,  1, (desca)[LLD_]), 	\
	DM_SetTrans(Trans, A)
#endif /* __PBLASCheck_INCL__ */

#define PBalphaval(alpha)	(alpha)
#define PBbetaval(beta)		(beta)

#define VIncOK(inc, desc)	(((inc)==1) | ((inc)==(desc)[M_]))

#ifdef USE_PBERROR_
# undef Error_
# define Error_(pref, call, v)	PBError_(pref, call, v); return
#endif

#define DBLAS1_CHECK_FLAG	1	
/* whether to suppress outer level checking in the DBLAS level 1 routines,
   called by the PBLAS level 1 implementation */ 

#define PBRowVec(inc, desc)	((inc)==(desc)[M_])
#define PBColVec(inc, desc)	((inc)!=(desc)[M_])
