/************************************************************************/
/*         DBLAS interface to P_GEMM() 					*/
/************************************************************************/
/* SCCS INFO @(#)P_GEMM.c 1.1         last modified P. Strazdins, 99/04/23
   first created: P. Strazdins 96/10
 */

#include "PBLASCheck.h"
#include "GenMult.h"

#ifndef ALTNAMES
static int UseDBLAS_P_GEMM=0;		/* to indicate if P_GEMM is linked in*/
#endif

void P_GEMM( FTransA, FTransB, m, n, k, alpha, A, ia, ja, desc_A,
             B, ib, jb, desc_B, beta, C, ic, jc, desc_C )              
  F_CHAR_T FTransA, FTransB; 		int *m, *n, *k;
  FDATA *alpha, A[], B[], *beta, C[];  	int *ia, *ja, *ib, *jb, *ic, *jc;
  int desc_A[], desc_B[], desc_C[];
{ char *TransA = F2C_CHAR(FTransA); 	char *TransB = F2C_CHAR(FTransB); 
  char TransAu = UpperCase(*TransA), 	TransBu = UpperCase(*TransB);
  int ParamErr = 0;			
  DistMat dA, dB, dC;
  int AlignM, AlignN, AlignK;

  if (!TransOK(TransAu))		ParamErr = 1;
  else if (!TransOK(TransBu))		ParamErr = 2; 
  else if (!DimOK(*m)) 			ParamErr = 3; 
  else if (!DimOK(*n)) 			ParamErr = 4; 
  else if (!DimOK(*k)) 			ParamErr = 5; 
  else {
    int TrA =  TransAu!=NoTranspose,	TrB = TransBu!=NoTranspose;	 
    ParamErr =   PDescChkConv(0, 0,   *m, *n, C, *ic, *jc, desc_C, 19, &dC);
    if (ParamErr==0)
      ParamErr = PDescChkConv(1, TrA, *m, *k, A, *ia, *ja, desc_A, 10, &dA);
    if (ParamErr==0)
      ParamErr = PDescChkConv(2, TrB, *k, *n, B, *ib, *jb, desc_B, 14, &dB);

    if (ParamErr==0) {
      AlignM = Dim_BCompat(DM_DM(dA), DM_D0(dC));  
      AlignN = Dim_BCompat(DM_DN(dB), DM_D1(dC));
      AlignK = Dim_BCompat(DM_DN(dA), DM_DM(dB));	                
      if (AlignM+AlignN+AlignK < 2) 	ParamErr = 10;    
    }
  }
  
  ERROR_CHECK_TR(
    (TraceStr, "(%c,%c,%d,%d,%d,%1.1f,a%d,%d,%d,_,a%d,%d,%d,_,%1.1f,a%d,%d,%d,_)", 
      *TransA, *TransB, *m, *n, *k, *alpha, FD_ADR(A), *ia, *ja,  
      FD_ADR(B),*ib, *jb, *beta, FD_ADR(C), *ic, *jc), 
    -ParamErr, "GEMM", 0
  );

  if (AlignM+AlignN+AlignK==3)          /* chose version simply on dim. size */
    if (*m < *n && *m < *k)
      AlignM = 0;  
    else if (*n < *m && *n < *k)
      AlignN = 0;
    else
     AlignK = 0;

  alpha = PBalphaval(alpha);		beta = PBbetaval(beta);
  if (!AlignK)
    DistRankKMult(*m, *n, *k, alpha, dA, dB, beta,  dC);
  else if (!AlignM)
    DistVecMatMult(*m, *n, *k, alpha, dA, dB, beta, dC);
  else
    DistVecMatMult(*n, *m, *k, alpha, Trans(dB), Trans(dA), beta, Trans(dC));

  TRACE_EXIT;
} /* P_GEMM */

