/************************************************************************/
/*         DBLAS interface to P_GEMV() 					*/
/************************************************************************/
/* SCCS INFO @(#)P_GEMV.c 1.1         last modified P. Strazdins, 99/04/23
   first created: P. Strazdins 97/04
 */

#include "PBLASCheck.h"
#include "GenMult.h"

#ifndef ALTNAMES
static int UseDBLAS_P_GEMV=0;		/* to indicate if P_GEMV is linked in*/
#endif

void P_GEMV( FTransA, m, n, alpha, A, ia, ja, desc_A,
             B, ib, jb, desc_B, incb, beta, C, ic, jc, desc_C, incc )
  F_CHAR_T FTransA;  			int *m, *n;
  FDATA *alpha, A[], B[], *beta, C[];  	int *ia, *ja, *ib, *jb, *ic, *jc;
  int desc_A[], desc_B[], *incb, desc_C[], *incc;
{ char *TransA = F2C_CHAR(FTransA);
  char Trans = UpperCase(*TransA);
  int ParamErr = 0;			
  int TrA = Trans!=NoTranspose,		TrB = PBRowVec(*incb, desc_B),
      TrC = PBRowVec(*incc, desc_C);
  DistMat dA, dB, dC;
  int M = TrA? *n: *m,                	N = TrA? *m: *n;
  
  if (!TransOK(Trans))	ParamErr = 1;
  else if (!DimOK(*m))	ParamErr = 2; 
  else if (!DimOK(*n))	ParamErr = 3; 
  else if (!VIncOK(*incb, desc_B))	ParamErr = 13; 
  else if (!VIncOK(*incc, desc_C))	ParamErr = 19; 
  else {
    ParamErr = PDescChkConv(0, TrA, M, N, A, *ia, *ja, desc_A,  8, &dA);
    if (ParamErr==0)
      ParamErr = PDescChkConv(1, TrB, M, 1, B, *ib, *jb, desc_B, 12, &dB);
    if (ParamErr==0)
      ParamErr = PDescChkConv(2, TrC, N, 1, C, *ic, *jc, desc_C, 18, &dC);
    if (ParamErr==0) {
      int AlignM = Dim_BCompat(DM_DM(dA), DM_DM(dC));  
      int AlignK = Dim_BCompat(DM_DN(dA), DM_DM(dB));
      if (!AlignM)			ParamErr = 12;    
      else if (!AlignK)			ParamErr = 8;    
    }
  }

#if 0
#define DESC_PRINT(desc) for (i=0;i<DLEN_;i++) printf("%d ", desc[i]);
  if (ParamErr) {
    char pA[64], pB[64], pC[64]; int i;	
    printf("%d.%d: pdgemv: Align =%d %d, dA=%s, dB=%s, dC=%s\n", 
    	    getcidy(), getcidx(), Dim_BCompat(DM_DM(dA), DM_D0(dC)), Dim_BCompat(DM_DN(dA), DM_DN(dB)), 
    	    PrintDistMat(dA, pA), PrintDistMat(dB, pB), PrintDistMat(dC, pC));
     printf("%d.%d: pdgemv: desc A=[", getcidy(), getcidx());
     DESC_PRINT(desc_A); printf("], B=["); DESC_PRINT(desc_B);
     printf("], C=["); DESC_PRINT(desc_C); printf("\n");      
  }
#endif

  ERROR_CHECK_TR(
      (TraceStr, "(%c,%d,%d,%1.1f,a%d,%d,%d,_,a%d,%d,%d,_,%d,%1.1f,a%d,%d,%d,_,%d)",
      *TransA, *m, *n, *alpha, FD_ADR(A), *ia, *ja,  
      FD_ADR(B),*ib, *jb, *incb, *beta, FD_ADR(C), *ic, *jc, *incc), 
    -ParamErr, "GEMV", 0
  );

  alpha = PBalphaval(alpha);		beta = PBbetaval(beta);
  DistVecMatMult(1, M, N, alpha, Trans(dB), Trans(dA), beta, Trans(dC));

  TRACE_EXIT;
} /* P_GEMV */

