/************************************************************************/
/*         DBLAS interface to P_GETRF() 				*/
/************************************************************************/
/* SCCS INFO @(#)P_GETRF.c 1.3         last modified P. Strazdins, 00/08/01
   first created: P. Strazdins 96/10
 */

#include "PBLASCheck.h"
#include "MachineParams.h" 
#include "D_LAPACK.h"
#include "ScaLAPACKMisc.h"

#ifndef ALTNAMES
static int UseDBLAS_P_GETRF=0;		/* to indicate if P_GETRF is linked in*/
#endif

void P_GETRF( m, n, A, ia, ja, desc_A, ipiv, info )
  int *m, *n;
  FDATA A[];  				int *ia, *ja;
  int desc_A[], ipiv[], *info;
{ int ParamErr = 0;
  DistMat dA;
  int i, nl, w;

  if (!DimOK(*m))                       ParamErr = 1;
  else if (!DimOK(*n))                  ParamErr = 2;
  else {
    ParamErr = PDescChkConv(0, 0, *m, *n, A, *ia, *ja, desc_A, 6, &dA);
  }

  /* note: info will be ignored */

  ERROR_CHECK_TR(
    (TraceStr, "(%d,%d,a%d,%d,%d)",
     *m, *n, FD_ADR(A), *ia, *ja), -ParamErr, "GETRF", 0
  );
  
  w = DILAENV(sizeof(FDATA), "GETRF", MachParamVec, desc_A[ MB_ ], 
	      d_P(DM_D0(dA)), d_P(DM_D1(dA)));
  PivColRep = 1;		/* regard ipiv[] as a column-replicated vector*/

  D_GETRF(*m, *n, dA, ipiv, w);

  nl = G2Ld(MIN(*m, *n), DM_DM(dA));  
  for (i=0; i < nl; i++)	/* compensate for Ftn indices */
    ipiv[i]++; 
  
  TRACE_EXIT;    
} /* P_GETRF */


void P_GETRS( ftrans, n, nrhs, A, ia, ja, desc_A, ipiv, B, ib, jb, desc_B, info )
  int *n, *nrhs;			F_CHAR_T ftrans;
  FDATA A[], B[];  			int *ia, *ja, *ib, *jb;
  int desc_A[], desc_B[], ipiv[], *info;
{ int ParamErr = 0;
  DistMat dA, dB;
  char TransA = UpperCase(*F2C_CHAR(ftrans));
  int TrA = (TransA!=NoTranspose),	   i, nl;

  if (!TransOK(TransA))			   ParamErr = 1;
  if (!DimOK(*n))                 	   ParamErr = 2;
  else if (!DimOK(*nrhs))                  ParamErr = 3;
  else {
    ParamErr = PDescChkConv(0, TrA, *n, *n, A, *ia, *ja, desc_A, 7, &dA);
    if (ParamErr==0)
      ParamErr = PDescChkConv(1, 0, *n, *nrhs, B, *ib, *jb, desc_B, 12, &dB);
    if (ParamErr==0) {
      int AlignM = Dim_BCompat(DM_DM(dA), DM_DM(dB));
      if (AlignM==0)			  ParamErr = 12;	      
    }  
  }

  ERROR_CHECK_TR(
    (TraceStr, "(%c,%d,%d,a%d,%d,%d,a%d,%d,%d)",
     TransA, *n, *nrhs, FD_ADR(A), *ia, *ja, FD_ADR(B), *ib, *jb), 
     -ParamErr, "GETRS", 0
  );
  
  PivColRep = 1;		/* regard ipiv[] as a column-replicated vector*/
  nl = G2Ld(*n, DM_DM(dA));  
  for (i=0; i < nl; i++)	/* comvert to C indices */
    ipiv[i]--;  
  if (TrA==0) { 
    D_LASWP("Forward", *nrhs, *n, Trans(dB), ipiv);
    D_TRSM("Lower", "Unit", *n, *nrhs,  FD_1, dA, dB);
    D_TRSM("Upper", "NonUnit", *n, *nrhs, FD_1, dA, dB); 
  } else {
    D_TRSM("Upper", "NonUnit", *n, *nrhs, FD_1, dA, dB);
    D_TRSM("Lower", "Unit", *n, *nrhs,  FD_1, dA, dB);  
    D_LASWP("Backward", *nrhs, *n, Trans(dB), ipiv);         	 
  }
  for (i=0; i < nl; i++)	/* restore indices */
    ipiv[i]++; 
  TRACE_EXIT;    
} /* P_GETRS */


