/************************************************************************/
/*         DBLAS interface to P_SYMM()                                  */
/************************************************************************/
/* SCCS INFO @(#)P_SYMM.c 1.1         last modified P. Strazdins, 99/04/23
   first created: P. Strazdins 96/10
 */

#include "PBLASCheck.h"
#include "SYMMult.h"

#ifndef ALTNAMES
static int UseDBLAS_P_SYMM=0;		/* to indicate if P_SYMM is linked in*/
#endif

void P_SYMM(FSide, FUpLo, m, n, alpha, A, ia, ja, desc_A, B, ib, jb,
            desc_B, beta, C, ic, jc, desc_C)
  F_CHAR_T FSide, FUpLo; 		int *m, *n; 
  FDATA *alpha, A[], B[], *beta, C[]; 	int  *ia, *ja, *ib, *jb, *ic, *jc;	
  int desc_A[], desc_B[], desc_C[];
{ char *Side = F2C_CHAR(FSide); 	char *UpLo = F2C_CHAR(FUpLo); 
  int ParamErr = 0;
  DistMat dA, dB, dC;
  char SideA = UpperCase(*Side),	UpLoA = UpperCase(*UpLo);
  int Trans = SideA!=Left;		
  int M = Trans? *n: *m,		N = Trans? *m: *n;

  if (!SideOK(SideA)) 			ParamErr = 1;
  else if (!UpLoOK(UpLoA)) 		ParamErr = 2;
  else if (!DimOK(*m)) 			ParamErr = 3;
  else if (!DimOK(*n)) 			ParamErr = 4;
  else {
    ParamErr = PDescChkConv(0, Trans, M, M, A, *ia, *ja, desc_A,  9, &dA);
    if (ParamErr==0)
      ParamErr = PDescChkConv(1, Trans, M, N, B, *ib, *jb, desc_B, 13, &dB);
    if (ParamErr==0)
      ParamErr = PDescChkConv(2, Trans, M, N, C, *ic, *jc, desc_C, 18, &dC);
    if (ParamErr==0) {
      int AlignM = Dim_BCompat(DM_DM(dA), DM_DM(dC));
      int AlignK = Dim_BCompat(DM_DN(dA), DM_DM(dB));
      if (!Dim_BCompat(DM_D0(dA), DM_D1(dA))) 	ParamErr = 9;
      else if (!AlignK) 		ParamErr = 13;
      else if (!AlignM) 		ParamErr = 18;
    }
  }
    
  ERROR_CHECK_TR(
    (TraceStr, "(%c,%c,%d,%d,%1.1f,a%d,%d,%d,_,a%d,%d,%d,_,%1.1f,a%d,%d,%d)",
      *Side, *UpLo, *m, *n,  *alpha, FD_ADR(A), *ia, *ja, 
      FD_ADR(B), *ib, *jb,  *beta, FD_ADR(C), *ic, *jc),
    -ParamErr, "SYMM", 0
  );

  if (Trans) 
    UpLoA = TransForm(UpLoA);
  alpha = PBalphaval(alpha);		beta = PBbetaval(beta);
  SYMMult(M, N, alpha, UpLoA, dA, dB, beta, dC);
  TRACE_EXIT;

} /* P_SYMM() */


