/************************************************************************/
/*         DBLAS interface to P_SYMV()                                  */
/************************************************************************/
/* SCCS INFO @(#)P_SYMV.c 1.1         last modified P. Strazdins, 99/04/23
   first created: P. Strazdins 97/04
 */

#include "PBLASCheck.h"
#include "SYMMult.h"

#ifndef ALTNAMES
static int UseDBLAS_P_SYMV=0;		/* to indicate if P_SYMV is linked in*/
#endif

void P_SYMV(FUpLo, n, alpha, A, ia, ja, desc_A, B, ib, jb, 
            desc_B, incb, beta, C, ic, jc, desc_C, incc)
  F_CHAR_T FUpLo; 				int *n; 
  FDATA *alpha, A[], B[], *beta, C[]; 	int  *ia, *ja, *ib, *jb, *ic, *jc;	
  int desc_A[], desc_B[], *incb, desc_C[], *incc;
{ char *UpLo = F2C_CHAR(FUpLo);
  int ParamErr = 0;
  DistMat dA, dB, dC;  
  char UpLoA = UpperCase(*UpLo);
  int TrB = PBRowVec(*incb, desc_B),	TrC = PBRowVec(*incc, desc_C);

  if (!UpLoOK(UpLoA))			ParamErr = 1;
  else if (!DimOK(*n))			ParamErr = 2;
  else if (!VIncOK(*incb, desc_B))	ParamErr = 12;
  else if (!VIncOK(*incc, desc_C))	ParamErr = 18;
  else {
    ParamErr = PDescChkConv(0,   0, *n, *n, A, *ia, *ja, desc_A,  7, &dA);  
    if (ParamErr==0)
      ParamErr = PDescChkConv(1, TrB, *n,  1, B, *ib, *jb, desc_B, 11, &dB);  
    if (ParamErr==0)
      ParamErr = PDescChkConv(2, TrC, *n,  1, C, *ic, *jc, desc_C, 17, &dC);
    if (ParamErr == 0) { 
      int AlignM = Dim_BCompat(DM_DM(dA), DM_DM(dC));
      int AlignK = Dim_BCompat(DM_DN(dA), DM_DM(dB));
      if (!Dim_BCompat(DM_D0(dA), DM_D1(dA))) ParamErr = 7;
      else if (!AlignK)			ParamErr = 11;
      else if (!AlignM)			ParamErr = 17;
    }
  }

#if 0
#define DESC_PRINT(desc) for (i=0;i<DLEN_;i++) printf("%d ", desc[i]);
  if (ParamErr) {
    char pA[64], pB[64], pC[64]; int i;
    printf("%d.%d: pdsymv: Align =%d %d, dA=%s, dB=%s, dC=%s\n",
            getcidy(), getcidx(), 
	    Dim_BCompat(DM_DM(dA), DM_D0(dC)), Dim_BCompat(DM_DN(dA), DM_DN(dB)),
            PrintDistMat(dA, pA), PrintDistMat(dB, pB), PrintDistMat(dC, pC));
     printf("%d.%d: pdsymv: desc A=[", getcidy(), getcidx());
     DESC_PRINT(desc_A); printf("], B=["); DESC_PRINT(desc_B);
     printf("], C=["); DESC_PRINT(desc_C); printf("\n");
  }
#endif 
    
  ERROR_CHECK_TR(
    (TraceStr, "(%c,%d,%1.1f,a%d,%d,%d,_,a%d,%d,%d,_,%d,%1.1f,a%d,%d,%d,_,%d)",
     *UpLo, *n, *alpha, FD_ADR(A), *ia, *ja, FD_ADR(B), *ib, *jb, *incb, 
     *beta, FD_ADR(C), *ic, *jc, *incc),
    -ParamErr, "SYMV", 0
  );

  alpha = PBalphaval(alpha);		beta = PBbetaval(beta);
  SYMMult(*n, 1, alpha, UpLoA, dA, dB, beta, dC);
  TRACE_EXIT;

} /* P_SYMV() */


