/************************************************************************/
/*         DBLAS interface to P_SYTRF() 				*/
/************************************************************************/
/* SCCS INFO @(#)P_SYTRF.c 1.3         last modified P. Strazdins, 00/08/28
   first created: P. Strazdins 99/04
 */

#include "FDATASetUp.h"
#include "FDAssert.h"
#include "MachineParams.h"
#include "PBLASCheck.h"
#include "D_LAPACK.h"
#include "ScaLAPACKMisc.h"

#ifndef ALTNAMES
static int UseDBLAS_P_SYTRF=0;		/* to indicate if P_SYTRF is linked in*/
#endif

void P_SYTRF( FUpLo, n, A, ia, ja, desc_A, diaginv, ipiv, info )
  F_CHAR_T FUpLo;			int *n;
  FDATA A[], diaginv[];  		int *ia, *ja;
  int desc_A[], ipiv[], *info;
{ int ParamErr = 0;
  DistMat dA;
  char *UpLo = F2C_CHAR(FUpLo);
  int i, w;

  if (!UpLoOK(*UpLo) || UpperCase(*UpLo)!=Lower)              
  				     	ParamErr = 1;
  else if (!DimOK(*n))			ParamErr = 2;
  else {
    ParamErr = PDescChkConv(0, 0, *n, *n, A, *ia, *ja, desc_A, 6, &dA);
  }

  ERROR_CHECK_TR(
    (TraceStr, "(%c,%d,a%d,%d,%d,_,_,_)",
     *UpLo, *n, FD_ADR(A), *ia, *ja), -ParamErr, "SYTRF", 0
  );
  
  w = DILAENV(sizeof(FDATA), "SYTRF", MachParamVec, desc_A[ MB_ ], 
	      d_P(DM_D0(dA)), d_P(DM_D1(dA)));
  PivColRep = 0;		/* regard ipiv[] as a fully-replicated vector*/

  D_SYTRF(*n, dA, diaginv, ipiv, info, -w);

  for (i=0; i < *n; i++)		/* compensate for Ftn indices */
    if (ipiv[i]>=0)
      ipiv[i]++; 
    else
      ipiv[i]--;
  TRACE_EXIT;    
} /* P_SYTRF */


void P_SYTRS( FUpLo, n, nrhs, A, ia, ja, desc_A, diaginv, ipiv, 
	      B, ib, jb, desc_B, info )
  F_CHAR_T FUpLo;			int *n, *nrhs;
  FDATA A[], diaginv[], B[];  		int *ia, *ja, *ib, *jb;
  int desc_A[], ipiv[], desc_B[], *info;
{ int ParamErr = 0;
  DistMat dA, dB;
  char *UpLo = F2C_CHAR(FUpLo);
  int i;

  if (!UpLoOK(*UpLo) || UpperCase(*UpLo)!=Lower)              
  				     	ParamErr = 1;
  else if (!DimOK(*n))			ParamErr = 2;
  else if (!DimOK(*nrhs))		ParamErr = 3;
  else if (*info != 0)			ParamErr = 13;
  else {
    ParamErr = PDescChkConv(0, 0, *n, *n, A, *ia, *ja, desc_A, 7, &dA);
    ParamErr = PDescChkConv(1, 0, *n, *nrhs, B, *ib, *jb, desc_B, 13, &dB);
    if (ParamErr==0 && (!Dim_BCompat(DM_DM(dA), DM_DM(dB)) ||
    	                !Dim_BCompat(DM_DN(dA), DM_DM(dB))))
    	                  		ParamErr = 13;
  }

  ERROR_CHECK_TR(
    (TraceStr, "(%c,%d,%d,a%d,%d,%d,_,_,a%d,%d,%d,_)",
     *UpLo, *n, *nrhs, FD_ADR(A), *ia, *ja, FD_ADR(B), *ib, *jb), 
     -ParamErr, "SYTRS", 0
  );
  
  for (i=0; i < *n; i++)		/* compensate for Ftn indices */
    if (ipiv[i]>0)
      ipiv[i]--; 
    else
      ipiv[i]++;

  D_SYTRS(*n, *nrhs, dA, diaginv, ipiv, dB, info);

  for (i=0; i < *n; i++)		/* compensate for Ftn indices */
    if (ipiv[i]>=0)
      ipiv[i]++; 
    else
      ipiv[i]--;
  TRACE_EXIT;    
} /* P_SYTRS */


