/************************************************************************/
/*         DBLAS interface to P_TRSM()                                  */
/************************************************************************/
/* SCCS INFO @(#)P_TRSM.c 1.1         last modified P. Strazdins, 99/04/23
   first created: P. Strazdins 96/10
 */

#include "PBLASCheck.h"
#include "TRSM.h"
#include "TriMatSolve.h"

#ifndef ALTNAMES
static int UseDBLAS_P_TRSM=0;		/* to indicate if P_TRSM is linked in*/
#endif

void P_TRSM( FSide, FUpLo, FTransA, FDiag, m, n, alpha, A, ia, ja, desc_A,
              B, ib, jb, desc_B )              
  F_CHAR_T FSide, FUpLo, FTransA, FDiag;int *m; int *n;
  FDATA *alpha, A[], B[];		int *ia, *ja, *ib, *jb; 
  int desc_A[], desc_B[];
{ char *Side = F2C_CHAR(FSide),         *UpLo = F2C_CHAR(FUpLo),
       *TransA = F2C_CHAR(FTransA),     *Diag = F2C_CHAR(FDiag);
  char SideA = UpperCase(*Side), 	UpLoA = UpperCase(*UpLo), 
       DiagA = UpperCase(*Diag), 	Trans = UpperCase(*TransA); 
  int ParamErr = 0;
  DistMat dA, dB;  			
  int TrB = (SideA!=Left); 
  int M = TrB? *n: *m,                  N = TrB? *m: *n;           
  int AlignM, AlignK, NoCommA;
  
  if (!SideOK(SideA))                   ParamErr = 1;
  else if (!UpLoOK(UpLoA))              ParamErr = 2;
  else if (!TransOK(Trans))             ParamErr = 3;
  else if (!DiagOK(DiagA))              ParamErr = 4;
  else if (!DimOK(*m))                  ParamErr = 5;
  else if (!DimOK(*n))                  ParamErr = 6;
  else {
    int TrA = TrB^(Trans!=NoTranspose);
    ParamErr = PDescChkConv(0, TrA, M, M, A, *ia, *ja, desc_A, 11, &dA);
    if (ParamErr==0)
      ParamErr = PDescChkConv(1, TrB, M, N, B, *ib, *jb, desc_B, 15, &dB);
 
    if (ParamErr==0) {
      AlignM = Dim_BCompat(DM_DM(dA), DM_DM(dB));
      AlignK = Dim_BCompat(DM_DN(dA), DM_DM(dB));
      if ((AlignM | AlignK) == 0)       ParamErr = 15;
    }
  }

  ERROR_CHECK_TR(   
    (TraceStr, "(%c,%c,%c,%c,%d,%d,%1.1f,a%d,%d,%d,_,a%d,%d,%d,_)",
     *Side, *UpLo, *TransA, *Diag, *m, *n, *alpha, FD_ADR(A), *ia, *ja, 
     FD_ADR(B), *ib, *jb),
    -ParamErr, "TRSM", 0
  );
  
  NoCommA = (DM_Trans(dA)==DM_Trans(dB)) && (d_P(DM_DN(dA))==1);   
  alpha = PBalphaval(alpha);
  if (TrB)
    UpLoA = TransForm(UpLoA);
    
  if (!AlignM || (!NoCommA && AlignK && 4*N<=M))
    if (Blk1d(N, DM_DN(dB)) || !AlignM)
      TriMatSolve1Blk(UpLoA, DiagA, M, N, alpha, dA, dB);
    else
      TriMatSolve(UpLoA, DiagA, M, N, alpha, dA, dB);
  else
    TR_M(1, UpLoA, DiagA, M, N, alpha, dA, dB);
  TRACE_EXIT;
} /* P_TRSM() */

