# SCCS INFO %Z%%M% %I%         last modified P. Strazdins, %E%
#   first created: P. Strazdins 98/11
# 
# System Level Makefile definitions for DBLAS and associated Libraries
# This is based on the SCALAPACK definitions, as similar libraries
# (BLAS, BLACS) and similar range of platforms are being used.
#
############################################################################
# # defintions for a supersparc sun5, single processor (eg silver)

# The complete path to the top level DBLAS directory
home         	= /home/peter/dblas_rel
LIB          	= $(home)/LIB
DBLASSRC	= $(home)/DBLAS
DLAPACKSRC	= $(home)/DLAPACK
PBLASSRC	= $(home)/PBLAS
BLASSRC		= $(home)/BLAS
BLACSSRC	= $(home)/BLACS

# shell to be used by make
SHELL		= /bin/sh

# The platform identifier to suffix to the end of library names
#
PLAT      	= bunyip
# The cell architecture identifier.
#
ARCH		= PIII

# Host program compiler; needed for AP1000/AP+
#
HOSTCC=cc.hc7	

# The Fortran and C compilers, loaders, and their flags
#
CC            	= hcc 			# /opt/SUNWspro/bin/cc -Dinline=
# for compiling most library codes

# include flags for communication libraries
COMM_INCLUDE    = 

MAINCC	      	= $(CC)  $(COMM_INCLUDE)
# for compiling test programs

COMMUN_CC     	= $(CC) $(COMM_INCLUDE)
# for communication library interface module
# 	(MAINCC for -DAPLIB*, otherwise CC)

INNERLOOP_CC  	= $(CC)
# compiler + flags for compiling optimized inner loops
# 	set to $(CC) $(CCFLAGS) 

CCNOOPT	      	= 	# -xarch=v8plusa -xchip=ultra
# architecture-dept C flags

CCFLAGS       	= $(CCNOOPT) -O2 	# $(CCNOOPT) -xO2 
# CCLOADER 	= $(MAINCC) 		
CCLOADER	= $(CC) # needed to link in PerfLibrary
CCLOADFLAGS   	=

F77           	= g77
F77NOOPT      	= 
F77FLAGS      	= $(F77NOOPT) -O2
F77LOADER     	= $(F77)
F77LOADFLAGS  	=

# The archiver and the flag(s) to use when building archive (library)
# Also the ranlib utility.  If your system has no ranlib, set RANLIB = echo
#
AR          	= ar
ARFLAGS     	= cr
RANLIB        	= ranlib

# libraries needed for linking in Fortran BLAS/BLACS to C programs
# this will generally include -lF77, possibly with some secondary archives
# (also if for example gcc was used to compile some modules, but is not
# the loader, -lgcc should also be added). 
#
# LIBF77 		= -L/opt/SUNWspro/lib -lF77 -lsunmath  
LIBF77 	= -lg2c # -lG77 -lM77 

# BLAS library archive (needed only if -DBLAS flag used)
#
BLAS_NAME     	= -lf77blas 
# BLASLIB       	= $(LIB)/$(BLAS_NAME) 
BLASLIB	= -L $(HOME)/blas/Linux_PIII -latlas -llapack 
#BLASLIB = -lf77blas -latlas


# Macros creating C Compilation FLAGS
#

TESTFLAGS	= # -Dsun5	
# TEST program FLAGS (system-dependent):
#	HOST		for platforms where standard output (generated by cell 0)
# 		     	must get passed by messages to a host machine  
#			eg. AP1000, AP+ using BLACS / APLIB
# 	sun5		for platforms able to use Solaris high resolution timer
	
FTNCOMP 	= -DAdd_ 
#  C preprocessor defs for compilation for Fortran-C interfaces
#	Add_		add _ for Fortran procedure name
#	UpCase		Fortran procedure name is in upper case
#	NoChange	Fortran procedure name is the same
#	f77IsF2C	use a Fortran to C compiler

COMMON		= # -DTRACE -DTUNE
# COMMON compilation switches:
#       ASSERT          turn on extensive assertion checking (and call tracing)
#       TRACE           incorporate parameter value tracing (implies ASSERT)
#                       Note: these both are reasonably efficient in run-time
#                       (except on very small computations). Larger code size.
#       TUNE            allow run-time tuning of MachineParams `constants'  

CELLPARAMS	= -DCACHE=256 

COMPUTE		= -DBLAS
# COMPUTE (cell COMPUTation): 
#	<none>		use non-standard ANU matrix-multiply SPARC tuned kernels
#		        -DULTRA for UltraSPARC-tuned loops 
#       VANILLA         use unoptimized versions of procedures 
#       BLAS    	use external Fortan BLAS wherever possible
#       BLAS_MM  	use BLAS only for matrix-multiply-like routines
#			(level-1 routines use SPARC- optimized C codes)
#       NO_PART    	suppress partitioning of (rectangular) matrix multiply
#                       (if there is no cache or -DBLAS* and
#                        the external Fortran BLAS performs good partitioning)

BLAS_LOOPS_LIB		= # ${LIB}/BLASInnerLoops_$(FD_i)_$(ARCH).a
BLAS_INNER_LOOPS_LINK	= 						\
	# ld -r -o x_.o CellMatrix.o $(BLAS_LOOPS_LIB); mv x_.o CellMatrix.o
# command to integrate  ANU matrix-multiply SPARC tuned kernels in DBLAS/BLAS
# 	non-empty only if not -DBLAS* or not -DVANILLA 

COMMUN		= -DMPI
# COMMUNication-oriented compilation switches (choose one only) 
#       BLACS       	use Fortan BLACS for communication. BLACS libraries
#		        need only be defined (below) in this case.		
#       APLIB_STRIDE  	use customized AP1000 communication library with
#                       stride send / ring buffer receive routines.
#       APLIB_PLUS_STRIDE use customized AP+ communication library with
#                       stride send / ring buffer receive routines.
#       APLIB_LINUX   	use AP+ Linux communication library
#                       (has some stride routines)
#       APLIB_STD     	use standard AP libraries only
#	<none>		compile for single processor workstation

XY_STRIDE_LIB	= 
# needed only for COMMUN=-DAPLIB_STRIDE*; otherwise should leave empty


# BLACS setup.  All versions need the debug level (0 or 1),
# and the directory where the BLACS libraries are
#
BLACSDBGLVL   	= 0
BLACSdir      	= $(LIB)

# MPI setup; uncomment and tailor to your system if using MPIBLACS
# Will need to comment out the default native BLACS setup below below
#
#USEMPI       	= -DUsingMpiBlacs
#SMPLIB        	= $(HOME)/mpich/lib/sun4/ch_p4/libmpi.a
#BLACSFINIT    	= $(BLACSdir)/blacsF77init_MPI-$(PLAT)-$(BLACSDBGLVL).a
#BLACSCINIT    	= $(BLACSdir)/blacsCinit_MPI-$(PLAT)-$(BLACSDBGLVL).a
#BLACSLIB      	= $(BLACSdir)/blacs_MPI-$(PLAT)-$(BLACSDBGLVL).a
#TESTINGdir    	= $(home)/TEST

# PVMBLACS setup, comment out next 6 lines if using MPI
#
USEMPI        	=
SMPLIB        	= #$(PVM_ROOT)/lib/$(PLAT)/libpvm3.a
BLACSFINIT    	= 
BLACSCINIT    	=
BLACS_NAME    	= blacs_$(PLAT)-$(BLACSDBGLVL).a
BLACSLIB_      	= $(BLACSdir)/$(BLACS_NAME)
TESTINGdir    	= $(DBLAS)/TEST		# $(HOME)/pvm3/bin/$(PLAT)

CBLACSLIB     	= # $(BLACSCINIT) $(BLACSLIB_) $(BLACSCINIT)
BLACSLIB     	= # $(BLACSFINIT) $(BLACSLIB_) $(BLACSFINIT)


# The name of the libraries to be created/linked to:
#
DBLASLIB_      	= dblas_$(FD_i)_$(PLAT).a
PBLASLIB_      	= pblas_$(FD_i)_$(PLAT).a
DLAPACKLIB_    	= dlapack_$(FD_i)_$(PLAT).a
SCALAPACKLIB_	= scalapack_$(FD_i)_$(PLAT).a
BLACSLT_	= blacslt_$(FD_i)_$(PLAT).a
BLASLT_		= blaslt_$(FD_i)_$(ARCH).a

DBLASLIB	= $(LIB)/$(DBLASLIB_)
PBLASLIB	= $(LIB)/$(PBLASLIB_)
DLAPACKLIB	= $(LIB)/$(DLAPACKLIB_)
SCALAPACKLIB	= $(LIB)/$(SCALAPACKLIB_)
BLACSLT		= # $(LIB)/$(BLACSLT_)
BLASLT		= # $(LIB)/$(BLASLT_)

#
