/************************************************************************/
/*                              Cell.h 	                                */
/************************************************************************/
/* SCCS INFO @(#)Cell.h 1.5         last modified P. Strazdins, 99/06/02
   first created: P. Strazdins 95/09
*/
/* set up definitions for cell programs wrt communication with host (if any)
 * and other cells.
 * compilatation:
 *	if BLACS defined, set up cell for Fortran BLACS communication.
 *	if APLIB_* defined, set up cell for APLIB (APx000) communication.
 *	if none defined, set up as for single workstation.
 *	if HOST defined (valid only if BLACS or APLIB_*),
 *	  (argv, argc) and date strings are b/c from a host program
 *	  and overall output (in string form) is sent to host for printing
 *	else (assumes Unix I/O calls implemented on cells with main(argv, argc))
 *	  if APLIB_* || BLACS defined, print main stdout into a file
 *	  else print to default stdout.
 */
#ifndef __Cell_Incl__
#define __Cell_Incl__

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#if defined(APLIB_STD) || defined(APLIB_STRIDE) || defined(APLIB_LINUX) || defined(APLIB_PLUS_STRIDE)
# define APLIB
#endif 


#ifdef BLACS
# define CELL_INIT			{ MPI_Init(&argc, &argv); FTN_BLACS_PINFO(&CID, &NP); }
# define CELL_FINALIZE			{ /*VPP_Release();*/ MPI_Finalize(); }

# ifdef HOST				/* these 2 ap1000-dept*/
#  include <ccell.c7.h>			/* needed for host-cell messaging*/
#  define main cell_main
# else
int getncelx(),  getncely(); 
# endif

static int BLACSCntxt_ = -1;	
static int CID, CID_I, CID_J, NP = 1, NP_I = 1, NP_J = 1;
# include       "FtnBLACS.h"

# define COMMUN_INIT(Ncely, Ncelx)					\
    { int BLACSDefHandle=0;						\
      NP_I = (Ncely); NP_J = (Ncelx);					\
      FTN_BLACS_PINFO(&CID, &NP);					\
      if (NP < 0) {                                                     \
        if (CID==0) NP = NP_I*NP_J;                                     \
        FTN_BLACS_SETUP(&CID, &NP); }                                   \
      FTN_BLACS_GET(&BLACSDefHandle, &BLACSDefHandle, &BLACSCntxt_);	\
      FTN_BLACS_GRIDINIT(&BLACSCntxt_, "Row", &NP_I, &NP_J);		\
      FTN_BLACS_GRIDINFO(&BLACSCntxt_, &NP_I, &NP_J, &CID_I, &CID_J);	\
      BLACSCntxtSet(BLACSCntxt_); CommSetUp(NP_I, NP_J);		\
      /*above 2 needed if FDCommun procedures are called directly*/	\
    }                            

# include "FDCommun.h"
# include <string.h>
static FDATA a_comm_[16];	/* typically CELL_SEND/RECV() have size <= 8 */
# define CELL_COMM_MTYPE 1010
# define CELL_SEND(i, j, a, size)					\
	/*send to cell (i,j) size bytes from address a*/		\
	{ int n_ = ((size)+sizeof(FDATA)-1)/sizeof(FDATA);		\
	  memcpy((void *) a_comm_, (void *) (a), (size_t) (size));	\
	  SendMat((NP_I+(i)-CID_I)%NP_I, (NP_J+(j)-CID_J)%NP_J, 1, n_, 	\
	  	   a_comm_, n_, CELL_COMM_MTYPE); 			}
# define CELL_RECV(i, j, a, size)					\
	/*receive from cell (i,j) size bytes at address a */		\
	{ int n_ = ((size)+sizeof(FDATA)-1)/sizeof(FDATA);		\
	  RecvMat((NP_I+(i)-CID_I)%NP_I, (NP_J+(j)-CID_J)%NP_J, 1, n_, 	\
	  	   a_comm_, n_, CELL_COMM_MTYPE);			\
	  memcpy((void *) (a), (void *) a_comm_, (size_t) (size));	}


#elif defined(VPPLIB)
# include "vpp_mplib.h"

static int CID, CID_I, CID_J, NP = 1, NP_I = 1, NP_J = 1;

# define CELL_INIT			CID = VPP_MyProc()
# define CELL_FINALIZE			VPP_Release()
#include "FDCommun.h"   		/* GridInit() */

# define COMMUN_INIT(Ncely, Ncelx)					\
    { CID = VPP_MyProc();						\
      NP_I = (Ncely); NP_J = (Ncelx); NP = NP_I*NP_J;			\
      CID_I = CID / NP_J, CID_J = CID % NP_J;				\
      if (CID==0) {							\
        PEn *PeList; int npe; 						\
        VPP_Init(MASTER);						\
        PeList=VPP_GetPeList(&npe);	        	             	\
        if (npe+1 < NP) {						\
          printf("error: cant create %dx%d grid: only %d cells avail.\n", NP_I, NP_J, npe+1);\
          exit(2); };							\
        if (NP > 1) {  							\
          PROC *ProcList = (PROC*)malloc((NP-1)*sizeof(PROC));		\
          VPP_StartNodes(PeList, NP-1, ProcList, NP-1, argv[0]);	\
          VPP_NodesReady(ProcList, NP-1);				\
          free((void *) ProcList);					\
        }								\
      } else								\
        VPP_Init(NODE);  						\
      GridInit(Ncely, Ncelx);						\
    }

static FDATA a_comm_[16];	/* typically CELL_SEND/RECV() have size <= 8 */
# define CELL_COMM_MTYPE 1010
# define CELL_SEND(i, j, a, size)					\
	/*send to cell (i,j) size bytes from address a*/		\
	{ int n_ = ((size)+sizeof(FDATA)-1)/sizeof(FDATA);		\
	  memcpy((void *) a_comm_, (void *) (a), (size_t) (size));	\
	  SendMat((NP_I+(i)-CID_I)%NP_I, (NP_J+(j)-CID_J)%NP_J, 1, n_, 	\
	  	   a_comm_, n_, CELL_COMM_MTYPE); 			}
# define CELL_RECV(i, j, a, size)					\
	/*receive from cell (i,j) size bytes at address a */		\
	{ int n_ = ((size)+sizeof(FDATA)-1)/sizeof(FDATA);		\
	  RecvMat((NP_I+(i)-CID_I)%NP_I, (NP_J+(j)-CID_J)%NP_J, 1, n_, 	\
	  	   a_comm_, n_, CELL_COMM_MTYPE);			\
	  memcpy((void *) (a), (void *) a_comm_, (size_t) (size));	}


#elif defined(MPI)
# include "mpi.h"

static int CID, CID_I, CID_J, NP = 1, NP_I = 1, NP_J = 1;

# define CELL_INIT			{ MPI_Init(&argc, &argv); 	\
					  MPI_Comm_rank(MPI_COMM_WORLD, &CID);}
# define CELL_FINALIZE			MPI_Finalize()
#include "FDCommun.h"   		/* GridInit() */

# define COMMUN_INIT(Ncely, Ncelx)					\
    { MPI_Comm_rank(MPI_COMM_WORLD, &CID);				\
      NP_I = (Ncely); NP_J = (Ncelx); NP = NP_I*NP_J;			\
      CID_I = CID / NP_J, CID_J = CID % NP_J;				\
      MPICommSet((int) MPI_COMM_WORLD);					\
      GridInit(Ncely, Ncelx);						\
    }

static FDATA a_comm_[16];	/* typically CELL_SEND/RECV() have size <= 8 */
# define CELL_COMM_MTYPE 1010
# define CELL_SEND(i, j, a, size)					\
	/*send to cell (i,j) size bytes from address a*/		\
	{ int n_ = ((size)+sizeof(FDATA)-1)/sizeof(FDATA);		\
	  memcpy((void *) a_comm_, (void *) (a), (size_t) (size));	\
	  SendMat((NP_I+(i)-CID_I)%NP_I, (NP_J+(j)-CID_J)%NP_J, 1, n_, 	\
	  	   a_comm_, n_, CELL_COMM_MTYPE); 			}
# define CELL_RECV(i, j, a, size)					\
	/*receive from cell (i,j) size bytes at address a */		\
	{ int n_ = ((size)+sizeof(FDATA)-1)/sizeof(FDATA);		\
	  RecvMat((NP_I+(i)-CID_I)%NP_I, (NP_J+(j)-CID_J)%NP_J, 1, n_, 	\
	  	   a_comm_, n_, CELL_COMM_MTYPE);			\
	  memcpy((void *) (a), (void *) a_comm_, (size_t) (size));	}


#elif defined(APLIB)
# ifdef APLIB_LINUX	
#   define CELL_INIT		{ ap_init(); ap_use_signals(0); sync(0,0);}
# else		
#   define CELL_INIT		
# endif
# define CELL_FINALIZE

# include <ccell.c7.h>
# define COMMUN_INIT(Ncely, Ncelx)	GET_PNUM(Ncely, Ncelx)
# ifndef APLIB_LINUX
# define main cell_main
# endif
# define CID     		getcid()
# define CID_I   		getcidy()
# define CID_J   		getcidx()
# define NP			getncel()
# define NP_I    		getncely()
# define NP_J    		getncelx()
# define GET_PNUM(Ncely, Ncelx)	{ (Ncely) = NP_I; (Ncelx) = NP_J; }
# define COMMUN_INIT(Ncely, Ncelx) GET_PNUM(Ncely, Ncelx)	
# define CELL_SEND(i, j, a, size) r_asend(j, i, gettid(), 0, (MSGP) a, size)
# define CELL_RECV(i, j, a, size)					\
	{ r_arecv(j, i, gettid(), ANY_TYPE);	readmsg((MSGP) a, size);}

#else				/* serial */
# define CELL_INIT		
# define CELL_FINALIZE		
# define CID     		0
# define CID_I   		0
# define CID_J   		0
# define NP_I    		1
# define NP_J    		1
# define NP			1
# define GET_PNUM(Ncely, Ncelx)	{ (Ncely) = 1; (Ncelx) = 1; }	
# define CELL_RECV(i, j, a, size)
# define CELL_SEND(i, j, a, size)					
# define COMMUN_INIT(Ny, Nx)	GET_PNUM(Ny, Nx)
#endif


#define EXTERN extern
#include "LogIO.h"			/* defines GetDate, STDOUT_MSGTYPE */


#if defined(HOST)
					/* modify next 3 dirs. for non APx000*/
# define HostSend(type, a, size) 	/* send size bytes at addr. a to host*/\
	{ h_send(type, a, size); }
# define COMMUN(v)        		/* receive value of var. v from host*/\
	{ h_recv(ANY_TYPE); readmsg(&v, sizeof(v)); }
# define COMMUN_ARR(a, r) 		/*receive r bytes from host to addr. a*/\
	{ h_recv(ANY_TYPE); if (r < 0) r = getmsize(); readmsg(a, r); }

static char HostString[512];
# define InitArgs(argc, argv)						\
	{ int r=-1;  char *argvcpy;					\
	  argv = (char **) malloc(128*sizeof(char *));			\
	  COMMUN_ARR(HostString, r);					\
	  argvcpy = malloc(r);	memcpy(argvcpy, HostString, r);		\
	  Str2Chv(argvcpy, r, argc, argv);				\
	}
# undef GetDate
# define GetDate(Date)							\
	{ int r=-1; COMMUN_ARR(Date, r); Date[r] = '\0'; }

# define GatherSend(v, size) 		CELL_SEND(0, 0, &(v), size)
# define GatherMsg(Ny, Nx, arr, size) 	GatherMessage(Ny, Nx, arr, size)

# define STDOUT 			HostString
# define STDOUT_SETUP(Stdout)
# define HPRINT(FPRINTF_ARGS)						\
	{ if (CID==0) {							\
	    sprintf FPRINTF_ARGS;					\
	    HostSend( STDOUT_MSGTYPE, STDOUT, strlen(STDOUT)+1);	\
	} }
# define LOGFILE_APPEND(LogFile, Msg)					\
	{ if (CID==0) 							\
	   HostSend( LOGFILE_MSGTYPE, Msg, strlen(Msg)+1 );		\
	}
#else

# define HostSend(type, v, size) 
# define COMMUN(v)        
# define COMMUN_ARR(v, r)        


# ifndef VPPLIB
#   define InitArgs(argc,argv)
# else	
#   define InitArgs(argc,argv) 		if (VPP_MyProc()==0) VPP_SetArgv(argv)
# endif
# if !defined(APLIB)  && !defined (BLACS) && !defined(VPPLIB) && !defined(MPI) /* serial */
#   define GatherMsg(Ny, Nx, arr, size)
#   define GatherSend(v, size) 		HostSend(STDOUT_MSGTYPE, &(v), size)

#   define STDOUT 			stdout
#   define STDOUT_SETUP(Stdout)		
# else
#   define GatherSend(v, size) 		if (CID!=0) CELL_SEND(0, 0, &(v), size)
#   define GatherMsg(Ny, Nx, arr, size) GatherMessage(Ny, Nx, arr, size)
#   ifndef APLIB_LINUX
#     define STDOUT 			stdout
#     define STDOUT_SETUP(Stdout)		
#   else  	
static FILE * stdoutLog;
#     define STDOUT 			stdoutLog 
#     define STDOUT_SETUP(Stdout)	stdoutLog = fopen("stdout.cell", "a")
#   endif
# endif
# define HPRINT(FPRINTF_ARGS)		{if (CID==0) fprintf FPRINTF_ARGS;}
# define LOGFILE_APPEND(LogFile, Msg)					\
	{ if (CID==0) 							\
	   LogFileAppend(LogFile, Msg)	}				
#endif /*HOST*/


#if defined(HOST) || (!defined(BLACS) && !defined(APLIB) && !defined(VPPLIB) &&!defined(MPI))
# define CPRINTF(PRINTF_ARGS)		printf PRINTF_ARGS
#else
# define CPRINTF(PRINTF_ARGS)		
#endif

#endif /*__Cell_Incl__*/
