/************************************************************************/
/*                              Commun.cell.c                           */
/************************************************************************/
/* SCCS INFO @(#)Commun.cell.c 1.4         last modified P. Strazdins, 99/06/09
   first created: P. Strazdins 93/01/20
*/
/* cell program for testing Ny x Nx distributed communication functions:
 * see  Commun.Parameters.c for meaning.
 *
 *  (externally-defined) cpp FLAGs:
 *      FDATATYPE       SINGLE, DOUBLE, 
 *	TEST_NATIVE	if defined, test equiv. AP native routine (m=1 or inc=n)
 *	TEST_PACK_SEND 	if defined, pack messages before sending
 */
/************************************************************************/
#include "Cell.h"
#include "Timer.c"

#include "FDATASetUp.h"
#include "LocMat.h"
#include "CellMatrix.h"
#include "FDCommun.h"

#include "HostMsg.c"
#define PROGNAME "Commun.cell"
#include "Commun.Parameters.c"

#include "AuxMatrix.c"
#define Cid(I, J) ((10.0*(I)+(J))/100.0)

static int P[2];

static FDATA AIMULT = FD_MKVAL(10.0, -10.0),
	     AJMULT = FD_MKVAL(1.0, -1.0),
	     AOFFS;

static int MatCheckR(r, m, n, i0, rc, I0, J0, B)
  int r, m, n, i0, rc, I0, J0; LocMat B;
{ int r1; int nerr=0; 
  for (r1=0; r1<r; r1++) {
    FD_BASE v = ((FD_BASE) i0) + Cid(I0, J0);	
    FDATA cid; FD_ASSIGN_SC(cid, =, v, -v);
    nerr += MatCheck(0, AIMULT, AJMULT, cid, m, n, LocSubMat(B, r1*m, 0));
    nerr += MatCheck(0, FD_0, FD_0, FD_0, m, LM_RInc(B)-n, LocSubMat(B, r1*m, n));
    if (rc==0) I0 = (I0+1)%P[0];
    else J0 = (J0+1)%P[1];
  } /* for (r1...) */
  return(nerr);
} /* MatCheckR() */  

static double Max(a, n)
 double a[]; int n;
{ int i; double m=a[0];
  for (i=1; i < n; i++)
   if (a[i] > m) m = a[i];
  return (m);
}

static double Min(a, n)
 double a[]; int n;
{ int i; double m=a[0];
  for (i=1; i < n; i++)
   if (a[i] < m) m = a[i];
  return (m);
}

static double Avg(a, n)
 double a[]; int n;
{ int i; double m=a[0];
  for (i=1; i < n; i++)
   m += a[i];
  return (m/n);
}


int main(argc, argv) 
  int argc; char *argv[];
{ FDATA *Ax, *Bx;		/* source & destination matrices */
  LocMat A, B;
  int NBytes;			/* number of NBytes in operation*/
  double tim[2];		/* timing & result checking variables */	
  double *tcomp[2]; int Ti;	/* timing */
  FDATA FTi; 
  int NErr;
  int r1, R1;
  char Date[32], ResultMsg[128];
  int P0, Rj;
  LocMat Bi, Ai;
  int SpecTop=0, VP[2], I[2];

  CELL_INIT;
  STDOUT_SETUP(STDOUT);
     
  InitArgs(argc, argv);
  GetDate(Date);

  GetParameters(argc, argv);
  PrintParameters(); 

  COMMUN_INIT(Ncely, Ncelx);
  Chv2Str(argv, 0, argc, ' ', ResultMsg, r1/*dummy*/);
  CPRINTF(("%d.%d: config %dx%d: %s\n", CID_I, CID_J, NP_I, NP_J, ResultMsg));

  P[0] = Ncely; P[1] = Ncelx;
  VP[0] = P[0], VP[1] = P[1];
  I[0] = CID_I /*GetAbsCId(0)*/, I[1] = CID_J /*GetAbsCId(1)*/;
#ifndef BLACS		/* CommSetUp() is not called in COMMUN_INIT() */
  if (rc > 1)
    SpecTop=1, rc -= 2;
  if (SpecTop)
    VP[rc] *= VP[1-rc], VP[1-rc] = 1; 
  CommSetUp(VP[0], VP[1]);
#endif
#define CID_rc(rc) (SpecTop? GetAbsCId(1-rc)*P[rc] + GetAbsCId(rc): \
			     GetAbsCId( rc )) 


  MatMalloc(PROGNAME, 1, &Ax, m*inc);
  MatMalloc(PROGNAME, 1, &Bx, R*m*inc);
  tcomp[0] = (double *) memalign(sizeof(double), sizeof(double)*TReps);
  tcomp[1] = (double *) memalign(sizeof(double), sizeof(double)*TReps);
  LM_Init(A, Ax, inc, 1);	LM_Init(B, Bx, inc, 1);

  MatZero(R*m, inc, &B);
  FD_ASSIGN_SC(AOFFS, =, Cid(GetAbsCId( 0 ), GetAbsCId(1)),
		        -Cid(GetAbsCId( 0 ), GetAbsCId(1)));
  MatInit(0, AIMULT, AJMULT, AOFFS, m, inc, A);
  MatPrint("A", pm, pn, A);

  NBytes=r*R*m*n*sizeof(FDATA);

  for (Ti=0; Ti < TReps; Ti++) {

  P0 = (I0*VP[rc] + CID_rc(rc) - I0)%VP[rc];
  Rj = rj; 
  ri = ri % P[0];
  pLocSubMat_(&Bi, &B, 0, 0);
  pLocSubMat_(&Ai, &A, 0, i0);
  Bx = LM_ADR(B);		/* reset */

  BarrierSync(-1);
  start_timer();

  for (R1=0; R1 < R; R1++) {
#ifdef TEST_PACK_SEND
    LocMat Bc;
    LM_Init(Bc, Bx, n, 1);
#endif 
    for (r1=0; r1 < r; r1++) {  
      if (DoBrd) {
#ifdef TEST_PACK_SEND
        if (P0 == 0)
          MatCopy(m, n, &Ai, NULL, &Bc); 		
        BrdMat(rc, P0, m, n, Bx, P0==0? n: inc);
#else
        /*BrdBlkMat( rc, P0, m, 0, 1, 1, n, P0==0? Ax+i0: Bx, inc);*/
        BrdMat( rc, P0, m, n, P0==0? Ax+i0: Bx, inc);
#endif 
        if (P0==0 && m>0 && n>0)
	  MatCopy(m, n, &Ai, NULL, &Bi); 
      } else if (DoPingPong) {
	if (CID_rc( rc )%2 == 0) {
	  if (VP[rc]>1 && CID_rc( rc )+1<VP[rc]) { 
            SendMat( 1-rc, rc, m, n, Ax+i0, inc, TRANSMSGTYPE);
            RecvMat( -(1-rc), -rc, m, n, Bx, inc, TRANSMSGTYPE);
          }
	} else {
          RecvMat( -(1-rc), -rc, m, n, Bx, inc, TRANSMSGTYPE);
          SendMat( 1-rc, rc, m, n, Ax+i0, inc, TRANSMSGTYPE);
	}
      } else {
	int r_[2];
#define CONV1DIM(r_)							\
	if (SpecTop) {							\
	   if (r_[rc] + I[rc] < 0) r_[rc] += P[rc];			\
	   else if (r_[rc] + I[rc] >= P[rc])  r_[rc] -= P[rc];		\
	   r_[rc] += r_[1-rc]*P[rc], r_[1-rc] = 0;			\
	}
	
        r_[0] = - ri, r_[1] = - Rj;
	CONV1DIM(r_);
#ifdef TEST_PACK_SEND
        MatCopy(m, n, &Ai, NULL, &Bc);
        SendMat( r_[0], r_[1], m, n, Bx, n, TRANSMSGTYPE);
#else
	SendWaitMark();		/* ISend mode needed for this pattern */
# ifdef ALT_SEND
	if (CID_J+r1)%2==0)
# endif
        SendMat( r_[0], r_[1], m, n, Ax+i0, inc, TRANSMSGTYPE);
#endif 
	r_[0] = ri, r_[1] = Rj;
	CONV1DIM(r_);
#ifdef ALT_SEND
	if (CID_J+r1)%2!=0)
#endif
        RecvMat(  r_[0], r_[1], m, n, Bx, inc, TRANSMSGTYPE);
	CompleteSends(-1);	/* correctness relies on send buffer Ax does not overlap bx */
      }
    } /* for (r1...) */ 
    Bx += FAST_MUL1(m, inc);    LM_ADR(Bi) = Bx;
    P0--; 			if (P0 < 0) P0 = VP[rc]-1;
    Rj++;			if (Rj >= P[1]) Rj = 0;
  } /* for (R1...) */
  
  tcomp[0][Ti] = end_timer(0);
  tcomp[1][Ti] = end_timer(1);	/* synchronizes cells */
  
  } /* for (Ti... ) */

  CommSetUp(Ncely, Ncelx);		/* restore to normal topology */

  if (CID == 0) {		/* cell 0 selects minimum timing */
    tim[0] = Min(tcomp[0], TReps);
    Ti = 0;
    while (tcomp[0][Ti] > tim[0]) Ti++;
  } 
  FD_SET_REAL(FTi, (FD_BASE) (Ti));
  BrdMat(0, CID_I, 1, 1, &FTi, 1);
  BrdMat(1, CID_J, 1, 1, &FTi, 1);
  Ti = (int) FD_REAL(FTi); 
      
  tim[0] = tcomp[0][Ti];
  tim[1] = tcomp[1][Ti];	/* time after synchroniztion */

  CPRINTF(("%d.%d: timing %d: before (after) sync, time is %G (%G)s @ %G (%G) MB/s\n", 
	CID_I, CID_J, Ti, tim[0], tim[1], NBytes * 1.0E-6 / tim[0], NBytes * 1.0E-6 / tim[1]));

  MatPrint("B", R*pm, pn, B);

  if (!DoPingPong) {
    NErr =  
    MatCheckR(R, m, n, i0, DoBrd? rc: 1,
	      DoBrd? (!rc||SpecTop ? I0%VP[0]: GetAbsCId( 0 )): 
		     (P[0]+GetAbsCId(0)+ri)%P[0],
 	      DoBrd? (rc||SpecTop?  I0%VP[1]: GetAbsCId( 1 )):  
		     (P[1]+GetAbsCId(1)+rj)%P[1],
	      B);	  
    CPRINTF(("%d.%d: %d errors in %d x %d (%d) B\n", CID_I, CID_J, NErr, m, n, LM_RInc(B)));
  } else
    NErr = 0;

  GatherSend(NErr, sizeof(int));
  GatherSend(tim[0], sizeof(double)); 
  GatherSend(tim[1], sizeof(double)); 
  if (CID == 0) {
    int i, *NErrs;  double *Tim0, *Tim1;
    double minTime0, minTime1, avgTime0, avgTime1, maxTime0, maxTime1;
    NErrs = (int *) malloc(Ncely*Ncelx*sizeof(int));
    Tim0 = (double *)  memalign(sizeof(double), 2*Ncely*Ncelx*sizeof(double));
    Tim1 =  Tim0 + Ncely*Ncelx;

    GatherMsg(Ncely, Ncelx, NErrs, sizeof(int));
    NErrs[0] = NErr;
    NErr = SumIntMat(Ncely, Ncelx, NErrs);
    if (!DoPingPong) {
      HPRINT((STDOUT, "Total errors over %d x %d cells is %d\n", Ncely, Ncelx, NErr));
    }
    if (NErr)
      PrintIntMat(Ncely, Ncelx, NErrs);

    GatherMsg(Ncely, Ncelx, Tim0, sizeof(double));
    Tim0[0] = tim[0];
 
    minTime0 = Min(Tim0, Ncely*Ncelx);
    maxTime0 = Max(Tim0, Ncely*Ncelx);
    avgTime0 = Avg(Tim0, Ncely*Ncelx);
    GatherMsg(Ncely, Ncelx, Tim1, sizeof(double));
    Tim1[0] = tim[1];
    minTime1 = Min(Tim1, Ncely*Ncelx);
    maxTime1 = Max(Tim1, Ncely*Ncelx);
    avgTime1 = Avg(Tim1, Ncely*Ncelx);
  
    HPRINT((STDOUT, "timing %d: for %1.3f MByte xfer, time1: min %f, max %f, avg %f; \n\t time2: min %f, max %f, avg %f\n",
	  Ti, NBytes/1000000.0, minTime0, maxTime0, avgTime0, minTime1, maxTime1, avgTime1));

    for (i=0; i < Ncely; i++) { 
      int j;
      HPRINT((STDOUT, "%s", i==0? "R1: " :"    "));
      for (j=0; j < Ncelx; j++) {
 	if (NBytes == 0)
	  HPRINT((STDOUT, "%1.2e ", Tim0[i*Ncelx+j]))
	else 
          HPRINT((STDOUT, "%2.2f ", NBytes * 1.0E-6 / Tim0[i*Ncelx+j]));
      }
      HPRINT((STDOUT, "\n"));
    }

    for (i=0; i < Ncely; i++) { 
      int j;
      HPRINT((STDOUT, "%s", i==0? "R2: " :"    "));
      for (j=0; j < Ncelx; j++) {
 	if (NBytes == 0)
	  HPRINT((STDOUT, "%1.2e ", Tim1[i*Ncelx+j]))
	else 
          HPRINT((STDOUT, "%2.2f ", NBytes * 1.0E-6 / Tim1[i*Ncelx+j]));
      }
      HPRINT((STDOUT, "\n"));
    }
    HPRINT((STDOUT, "\n\n"));

    free((char *) NErrs);		free((char *) Tim0);
  } /* if (CID==0) */

  CELL_FINALIZE;
  exit(NErr != 0);
} /* main() */
