/************************************************************************/
/*                              DistMat.Parameters.c                     */
/************************************************************************/
/* SCCS INFO @(#)DistMat.Parameters.c 1.1         last modified P. Strazdins, 98/12/24
   first created: P. Strazdins 97/01
*/
/* input command line parameters into variables for timing elementary 
 * integer operations, with the usage: 					
 *	PROGNAME		
 * followed optionally by following USAGE string:			*/
#define USAGE \
"[-X TReps] [-r r] [-f f] [-B B] [-p P] [-q Q] [-v v] [-n] [-D] [-c] M N"
/* with default values:							*/
#define DEFAULTS \
"f='S' (also 'P' 'g' 'l' 'i'), NReps=r=1, B=1 v=0 P=ncely, Q=Ncelx"
/* 
 * TReps x NReps repetitions are performed, with timings taken over NReps
 */

/* command line parameters; see header comment*/
  static int TReps=1, r=1, B=1, P=1, Q=1, v=0, M, N;
  static char f='S'; static int NegSubMat = 0, DiagComp = 0, PBCacheOn = 0;


static void PrintUsage(), PrintParameters(), OptionError();

static void GetParameters(argc, argv)
  int argc; char *argv[];
/* post: set DistMat command line variables according to argv[0..argc-1] */
{ extern char *optarg;
  extern int optind;
  int c;
  int SetQ = 0, SetP = 0;

  P = NP_I, Q = NP_J;
  while ((c = getopt(argc, argv, "r:X:f:B:p:q:v:nDc")) != -1) /* parse options */
    switch (c) {
      case 'r':
        r = atoi(optarg);                                           	break;
      case 'f':
        f = *optarg;                                           		break;
      case 'B':
        B = atoi(optarg);                                           	break;
      case 'p':
        P = atoi(optarg); SetP = 1;                         			break;
      case 'q':
        Q = atoi(optarg); SetQ = 1;                         			break;
      case 'v':
        v = atoi(optarg);                                           	break;
      case 'n':
	NegSubMat = 1;							break;
      case 'X':
        TReps = atoi(optarg);                                           break;
      case 'D':
	DiagComp = 1;							break;
      case 'c':
	PBCacheOn = 1;							break;
      case '?':
        PrintUsage();
        exit (2);
    }
  if (SetP && !SetQ)
    Q = P;

  if (optind+1 == argc) {
    PrintUsage();
    exit (0);
  }
  if (optind + 1 >= argc || argc<=2) { 
      HPRINT((STDOUT, "%s: insufficient parameters M N given.\n", PROGNAME));
    exit (2);
  }

  /* get non-optional parameters */
  M = atoi(argv[optind]);
  N = atoi(argv[optind+1]);
} /* GetParameters() */


static void PrintUsage()
/*post: prints program usage to stdout*/
{ HPRINT((STDOUT, "usage : %s %s\n", PROGNAME, USAGE));
  HPRINT((STDOUT, "with default values: %s\n", DEFAULTS));
  PrintParameters();
} /* PrintUsage() */

static void PrintParameters()
/*post: print out current values of variables set by command line to stdout*/
{ HPRINT((STDOUT, "%s: %dx%d reps of fn %c (v=%d %s%s%s) with M=%d N=%d  B=%d PxQ=%dx%d\n",
	 PROGNAME, TReps, r, f, v, f=='S'&&NegSubMat? "-": "",
	 (f=='P'||f=='S') && DiagComp? "Diag": "", f=='P' && PBCacheOn ? "c":"",
	 M, N, B, P, Q)); 
} /* PrintParameters() */
