/************************************************************************/
/*                              DistMat.cell.c                           */
/************************************************************************/
/* SCCS INFO @(#)DistMat.cell.c 1.2         last modified P. Strazdins, 99/02/25
   first created: P. Strazdins 97/01
*/
/* cell program for testing Ny x Nx distributed communication functions:
 * see  DistMat.Parameters.c for meaning.
 *
 *  (externally-defined) cpp FLAGs:
 *      PBLAS		add checking for (old) PBLAS interface
 *	PBLAS_OLD
 */
/************************************************************************/
#include "Cell.h"
#include "Timer.c"

#include "FDATASetUp.h"
#include "MatDim.h"
#include "DistMat.h"
#include "FDAssert.h"
#include "D_BLASCheck.h"		/* DM_TRACE() */

#include "HostMsg.c"
#define PROGNAME "DistMat.cell"
#include "DistMat.Parameters.c"

#if defined(PBLAS) || defined(PBLAS_OLD)
# include "../PBLAS/PBLAS.h"
# include "../PBLAS/PBLASCheck.h"

void FTN_BLACS_GRIDINFO( int *Cntxt, int *pP, int *pQ, int *I, int *J)
{
  *pP = P;  *pQ=Q; *I=CID_I; *J=CID_J; 
}
void ptopget_ ( int *Cntxt, char *Which, char *RowC, char *RowTop)
{ *RowTop = ' ';
}

void pberror_ (  int *Cntxt, char *Msg,  int *Param )
{
}
#endif

/****************checking routines ***********************************/

#define G2L1(i, P0, P)  (((i) + P - 1 - (P0))/(P))
 
#define L2G1(i, P0, P)  ((P0) + (i)*(P))

#define NBLKS(N, B0, B)         (((N) + (B0) + (B) - 1)/ (B))

int G2L_(N, B0, B, P0, P)
  int N, B0, B, P0, P;
{ register int NBlks, NB, NBr;
  if (N==1)
    return( (P0==0) );                  /* fast exit */
  if (B==1)
    return( G2L1(N, P0, P) );           /* fast exit */
  NBlks = NBLKS(N, B0, B);              /* #full distributed blocks */
  NBr = (N + B0) % B;                   /* size of last block */
  NB = G2L1(NBlks, P0, P)*B;            /* #elts on cell;,assuming full blocks*/
  if ((P0==0) && (B0 > 0))              /* adjust for partial block 1st cell */
    NB -= B0;
  if ((P0 == ((NBlks-1)%P)) && (NBr > 0))/* adjust for partial block last cell */
    NB -= (B-NBr);
  return (NB);
} /* G2L() */

int L2G_(i, B0, B, P0, P)
  int i, B0, B, P0, P;
/* pre:  0 <= B0 < B; 0 <= P0 < P
 *       a global vector A of block size B is distributed across P cells;
 *       starts at relative cell -P0; has block offset B0 for 1st block
 * post: returns the global index corresponding to local index i on cell P0
 */
{ if (P0==0) i += B0;
  return( ((i/B)*P              /* number of block boundaries passed */
           + P0)*B              /* number of cell boundaries passed */
          + (i%B) - B0 );
} /* L2G() */


#define SMOD 0x10000

static int IterateConvFn(f, v, r, D, N)
  char f; int v, r, N; pMatDim D;
{ int i, s=0; int B=d_B(D), P=d_P(D);
  int dog2l= (f=='g'), dol2g = (f=='l');
  for (;r>0;r--) {
    if (dog2l) {
      if (v==0)  
        for (i=0; i<N; i++)
	  s += G2Ld(i, D), s %= SMOD;
      else if (v==1) {
        unsigned int B0=d_B0(D), B=d_B(D), P0=d_P0(D), P=d_P(D), 
		     lgB=d_lgB(D), lgP = d_lgP(D);
        for (i=0; i<N; i++)
	  s += G2Llg(i, B0, B, P0, P, lgB, lgP), s %= SMOD;
      } else
        for (i=0; i<N; i++)
	  s += G2L_(i, 0, B, 0, P), s %= SMOD;
    } else if (dol2g) {
      if (v==0)  
        for (i=0; i<N; i++)
	  s += L2Gd(i, D);
      else if (v==1) {
	unsigned int B0=d_B0(D), B=d_B(D), P0=d_P0(D), P=d_P(D),
                     lgB=d_lgB(D), lgP = d_lgP(D);
        for (i=0; i<N; i++)
	  s += L2Glg(i, B0, B, P0, P, lgB, lgP), s %= SMOD;
      } else
        for (i=0; i<N; i++)
	  s += L2G_(i, 0, B, 0, P), s %= SMOD;        
    } else {
      unsigned int B0=d_B0(D), B=d_B(D), P0=d_P0(D), P=d_P(D);
      for (i=0; i<N; i++)
        pInitDim(D, B0, B, P0, P), s+= d_B0(D) + d_P0(D), s %= SMOD; 
    }  
  }
  return (s);
}

static int CheckConvFn(dog2l, v, D, N)
  int dog2l, v, N; pMatDim D;
{ int i, j, s=0; int B=d_B(D), P=d_P(D); 
  int B0, P0, act, exp;
  int err=0;
  for (i=0; i<N; i++)
    for (B0=0; B0<B; B0++)
      for (P0=0; P0<P; P0++) {
        if (dog2l) {
          if (v==0)
	    act = G2Ld_(i, D, B0, P0);
	  else
            act = G2L(i, B0, B, P0, P);
	  exp = G2L_(i, B0, B, P0, P);
	} else {
          if (v==0)
	    act = L2Gd_(i, D, B0, P0);
	  else
            act = L2G(i, B0, B, P0, P);
	  exp = L2G_(i, B0, B, P0, P);
        }
        if (exp != act) {
	  if (!err) 
	    HPRINT((STDOUT, "%c%d(%d,%d,%d,%d,%d) = %d, exp %d\n", 
		    f, v, j, B0, B, P0, P, act, exp)); 
	   err++;
	}
     }
  return (err);
}

#define DistMat_Eq(A, B)						\
	((d_B0(DM_D0(A))==d_B0(DM_D0(B))) &&		\
	 (d_P0(DM_D0(A))==d_P0(DM_D0(B))) &&		\
	 (d_B0(DM_D1(A))==d_B0(DM_D1(B))) &&		\
	 (d_P0(DM_D1(A))==d_P0(DM_D1(B))) &&		\
	 (LM_ADR(DM_Loc(A))==LM_ADR(DM_Loc(B)))) 

/************************************************************************/
static double Min(a, n)
 double a[]; int n;
{ int i; double m=a[0];
  for (i=1; i < n; i++)
   if (a[i] < m) m = a[i];
  return (m);
}

#define LARGEINC 10000

void main(argc, argv) 
  int argc; char *argv[];
{ LocMat lA;
  DM_DeclareAlloc(A_, A);
  int NOps;			/* number of operations in the timing */
  double tim[2];		/* timing & result checking variables */	
  int NErrs;
  char Date[32], ResultMsg[128];
  int s, s2, r1, i_, i, j;
  double *tcomp; int Ti;/* timing */

  CELL_INIT;
  STDOUT_SETUP(STDOUT);
     
  InitArgs(argc, argv);
  GetDate(Date);

  GetParameters(argc, argv);
  PrintParameters(); 

  COMMUN_INIT(P, Q);

  LM_Init(lA, (FDATA *) ((N*LARGEINC + M)*sizeof(FDATA)), 1, LARGEINC);	
  DM_Init(A, 0, lA, 0, B, CID_I, P, 0, B, CID_J, Q);

  s = 0;
  tcomp = (double *) memalign(sizeof(double), sizeof(double)*TReps);
  for (Ti=0; Ti < TReps; Ti++) {
  start_timer();
  if (f!='S' && f!='P') {
    s = IterateConvFn(f, v, r, DM_D0(A), M);  
    s += IterateConvFn(f, v, r, DM_D1(A), N);
  } else if (f=='S') {
    for (r1=0; r1 < r; r1++) {
      for (i=0; i < (DiagComp? 1: M); i++) { 
	if (NegSubMat)
          if (DiagComp)
            for (j=0; j < N; j++) {	    
	      DistMat_RVAlloc(1);		DistMat B;
              B = SubMat(A, -j, -j);
	      s += d_P0(DM_D0(B)) + d_P0(DM_D1(B));
            }
	  else 
            for (j=0; j < N; j++) {	    
	      DistMat_RVAlloc(1);		DistMat B;
              B = SubMat(A, -i, -j);
	      s += d_P0(DM_D0(B)) + d_P0(DM_D1(B));
            }
	else
	  if (DiagComp)
            for (j=0; j < N; j++) {
	      DistMat_RVAlloc(1);		DistMat B;
              B = SubMat(A, j, j);
	      s += d_P0(DM_D0(B)) + d_P0(DM_D1(B));
            } 
	  else
            for (j=0; j < N; j++) {
	      DistMat_RVAlloc(1);		DistMat B;
              B = SubMat(A, i, j);
	      s += d_P0(DM_D0(B)) + d_P0(DM_D1(B));
            } 
      }
    } /* for (r...) */
  } 
#if defined(PBLAS) || defined(PBLAS_OLD)
    else if (v==0) {
    int ParamErr = 0;
    DistMat A_[3]; 
    int desc[10]; 

    desc[DT_] = 1; desc[CTXT_]=0; desc[M_]=M+1; desc[N_]=N+1; 
    desc[MB_] = desc[NB_] = B; desc[RSRC_] = desc[CSRC_] = 0; 
    desc[LLD_] = LM_CInc(DM_Loc(A));

    if (PBCacheOn)
      PDescCacheOn();
    for (r1=0; r1 < r; r1++) {
      for (i_=1; i_ < (DiagComp? 1: M)+1; i_++) { 
        for (j=1; j < N+1; j++) {
	  i = DiagComp? j: i_;
	  s += PDescChkConv(0, 0, M-i, N-j, LM_ADR(DM_Loc(A)), i+1, j+1, 
		 		  desc, 17, &A_[2]);
	  s += PDescChkConv(1, 0, M-i, 1, LM_ADR(DM_Loc(A)), i+1, j, 
		 		  desc, 7, &A_[1]);
	  s += PDescChkConv(2, 0, 1, N-j,LM_ADR(DM_Loc(A)), i, j+1, 
		 		  desc, 12, &A_[0]);
        } 
      }
    } /* for (r...) */

    if (PBCacheOn)     
      PDescCacheOff();
  } else {
    int info = 0;
    int desc[10]; 

    desc[DT_] = 1; desc[CTXT_]=0; desc[M_]=M+1; desc[N_]=N+1; 
    desc[MB_] = desc[NB_] = B; desc[RSRC_] = desc[CSRC_] = 0; 
    desc[LLD_] = LM_CInc(DM_Loc(A));

    for (r1=0; r1 < r; r1++) {
      for (i_=1; i_ < (DiagComp? 1: M)+1; i_++) { 
        for (j=1; j < N+1; j++) {
#ifdef PBLAS_OLD
 	  int Cntxt, nprow, npcol, myrow, mycol;
	  int iix, jjx, iiy, jjy, iia, jja, ixrow, ixcol, iyrow, iycol,
	      iarow, iacol, nra, nca, iroffa, icoffa;
	  i = DiagComp? j: i_;
          info = 0;
	  FTN_BLAC_GRIDINFO(&Cntxt, &nprow, &npcol, &myrow, &mycol);
	  iroffa = i % desc[MB_];       icoffa = j % desc[NB_];
      	  pbchkvect( M-i, 1, i+1, j, desc, 1, 7, &iix, &jjx,
                 &ixrow, &ixcol, nprow, npcol, myrow, mycol, &info );
          pbchkvect( N-j, 2, i, j+1, desc, desc[M_], 12, &iiy, &jjy,
                 &iyrow, &iycol, nprow, npcol, myrow, mycol, &info );
          pbchkmat( M-i, 1, N-j, 2, i+1, j+1, desc, 17, &iia, &jja,
                &iarow, &iacol, nprow, npcol, myrow, mycol,
                &nra, &nca, &info );
	  s += info + (iroffa != i % desc[MB_]) + (icoffa != j % desc[NB_]) +
	       (jjx*desc[LLD_]==iix) + (jjy*desc[LLD_]==iiy) +
	       (jja*desc[LLD_]==iia); 
#else
#undef MAX
#undef MIN
#undef ABS
#undef DLEN_
#undef MB_
#undef NB_
#undef RSRC_
#undef CSRC_
#undef LLD_
#include "PBLAS/PBtools.h"
	  int i = DiagComp? j: i_;
	    int IX = i+1, JX = j, IY = i, JY = j +1, IA = i+1, JA = j+1;
	    int nprow, npcol, myrow, mycol; 
	    int  Acol, Ai, Aii, Aimb, Aimb1, Ainb, Ainb1, Aj, Ajj, Ald, Amb,
                 Amp, Anb, Anq, Arow, XAfr, Xi, Xj, YAfr, Yi, Yj, ctxt, info;
	    int  Ad[DLEN_], Ad0[DLEN_], XAd[DLEN_], Xd[DLEN_], YAd[DLEN_],
                  Yd[DLEN_];

   	    MIndxTrans( &IX, &JX, Xi, Xj );
   	    MIndxTrans( &IY, &JY, Yi, Yj );
   	    MIndxTrans( &IA, &JA, Ai, Aj );
   	    MDescTrans( desc, Xd );
   	    MDescTrans( desc, Yd );
   	    MDescTrans( desc, Ad );

	    ctxt = Xd[CTXT_];
	    FTN_BLACS_GRIDINFO(&ctxt, &nprow, &npcol, &myrow, &mycol);
	    info = 0;

           PB_Cchkvec( ctxt, "PDGER", "X", M-i, 1, Xi, Xj, Xd, 1,  7, &info );
           PB_Cchkvec( ctxt, "PDGER", "Y", N-j, 2, Yi, Yj, Yd, 1, 12, &info );
           PB_Cchkmat( ctxt, "PDGER", "A", M-i, 1, N-j, 2, Ai, Aj, Ad, 17, &info );

	   Aimb = Ad[IMB_]; Ainb = Ad[INB_]; Amb = Ad[MB_]; Anb = Ad[NB_];
	   Ald  = Ad[LLD_];
	   Mfirstnb( Aimb1, M-i, Ai, Aimb, Amb );
	   Mfirstnb( Ainb1, N-j, Aj, Ainb, Anb );
	   Minfog2l( Ai, Aj, Ad, nprow, npcol, myrow, mycol, Aii, Ajj, Arow, Acol );
	   s += Aii-Arow + Ajj-Acol + Aimb1 + Ainb1;
#if 1
	   MDescSet( Ad0, M-i, N-j, Aimb1, Ainb1, Amb, Anb, Arow, Acol, ctxt, Ald );
	   Mnumroc( Amp, M-i, 0, Aimb1, Amb, myrow, Arow, nprow );
   	   Mnumroc( Anq, N-j, 0, Ainb1, Anb, mycol, Acol, npcol );
	   s += Amp + Anq;
#endif	

	  { int  Xii, Xjj, Yii, Yjj, Xrow, Xcol, Yrow, Ycol;
	   Minfog2l( Xi, Xj, Xd, nprow, npcol, myrow, mycol, Xii, Xjj, Xrow, Xcol );
	   Minfog2l( Yi, Yj, Yd, nprow, npcol, myrow, mycol, Yii, Yjj, Yrow, Ycol );
	   s += Xii-Xrow + Xjj-Xcol + Yii-Yrow + Yjj-Xcol;
          }
	
#undef DLEN_
#undef MB_
#undef NB_
#undef RSRC_
#undef CSRC_
#undef LLD_
#define    DLEN_        9                     /* Length of a descriptor */
#define    DT_          0                     /* Descriptor Type        */
#define    CTXT_        1                              /* BLACS context */
#define    M_           2                      /* Global Number of Rows */
#define    N_           3                   /* Global Number of Columns */
#define    MB_          4                          /* Row Blocking Size */
#define    NB_          5                       /* Column Blocking Size */
#define    RSRC_        6                     /* Starting Processor Row */
#define    CSRC_        7                  /* Starting Processor Column */
#define    LLD_         8                    /* Local Leading Dimension */
# endif
        } 
      }
    } /* for (r...) */     
  }  
#endif /*PBLAS*/

  tcomp[Ti] = end_timer(0); 
  } /* for (Ti...) */
  tim[0] = Min(tcomp, TReps);

  if (f!='S' && f!='P') {
    NOps = r * (M + N);	
  } else {
    NOps = r * ((DiagComp? 1: M) * N);
  }
  HPRINT((STDOUT, "for %d %c operations (s=%d), (min) time=%0.2e, time/op=%0.2e\n",
	  NOps, f, s, tim[0], tim[0]/NOps));

  NErrs = 0;
  if (f!='S' && f!='P') {
    if (f!='i')
      NErrs = CheckConvFn(f=='g', v, DM_D0(A), MAX(M, N));
  } else if (f=='S' && !NegSubMat) {
    char pB[256];
    HPRINT((STDOUT, "SubMat tests on *%d=%s from 0..%d,0..%d\n",
           (int) A, PrintDistMat(A, pB), 2*M-1, 2*N-1));
    for (i=0; i<(DiagComp? 1 : M); i++)
      for (j=0; j<N; j++) {
        DistMat_RVAlloc(2);
    	DistMat A1, A2;		
	int d;
    	A1 = SubMat(A, i, j);	A2 = SubMat(A1, M, N);
	for (d=0; d<2;d++) {
	  int x = d==0? i+M: j+N;
	  int B0 = SubMatB0d(x, DM_D(d, A)),
	      P0 = SubMatP0d(x, DM_D(d, A)),
	      il = G2Ld(x, DM_D(d, A));  
	  int B0_ = d_B0(DM_D(d, A2)),	
	      P0_ = d_P0(DM_D(d, A2)),
	      il_ = LM_ADR(DM_Loc(A2)) - LM_ADR(DM_Loc(A));
	  if (d==1) il_ /= LM_CInc(DM_Loc(A)); else il_ %= LM_CInc(DM_Loc(A));
	  if (B0!=B0_ || P0!=P0_ || il!=il_) {
	    if (NErrs == 0) {
    	      CPRINTF(("1st error: dim %d of *%d=%s [exp B0=%d, P0=%d, il=%d(!=%d)]\n", 
		      d, (int) A2, PrintDistMat(A2, pB), B0, P0, il, il_));
    	      CPRINTF(("	(derived from offset %d *%d=%s,\n", 
		      x, (int) A, PrintDistMat(A, pB)));
    	      CPRINTF(("	via offset %d of *%d=%s)\n", 
		      d==0? i: j, (int) A1, PrintDistMat(A1, pB)));
	    }
	    NErrs++; break;
	  } 
        } /* for (d...) */
      } /* for (j...) */
  } else if (f=='S' && NegSubMat) {
    HPRINT((STDOUT, "SubMat tests on -ve indices *%d=%s from 0..%d,0..%d\n",
           (int) A, DM_TRACE(0, A), 2*M-1, 2*N-1));
    for (i=0; i<M; i++)
      for (j=0; j<N; j++) {
        DistMat_RVAlloc(5);
    	DistMat A_[3], B_[2];		
	int x;
	A_[0] = A;
    	A_[1] = SubMat(A, i, j);	B_[0] = SubMat(A_[1], -i, -j);	
	A_[2] = SubMat(A_[1], M, N);	B_[1] = SubMat(A_[2], -M, -N);	
	for (x=0; x < 1; x++) {
	  if (!DistMat_Eq(A_[x], B_[x])) {
	    if (NErrs == 0) {
    	      CPRINTF(("1st error: from %s displacement (%d,%d) %s != %s\n",
		     DM_TRACE(2, A_[x+1]), x==0? -i: -M, x==0? -j: -N,  
		     DM_TRACE(0, B_[x]), DM_TRACE(1, A_[x])));
	    }	
	    NErrs++; break;
	  } 
        } /* for (x...) */
      } /* for (j..) */
  } 
#if defined(PBLAS) || defined(PBLAS_OLD)
    else if (f=='P' && v==0) {
    int info = 0; 
    int desc[10]; 
    DistMat A_[3], B_[3];

    desc[DT_] = 1; desc[CTXT_]=0; desc[M_]=M+1; desc[N_]=N+1; 
    desc[MB_] = desc[NB_] = B; desc[RSRC_] = desc[CSRC_] = 0; 
    desc[LLD_] = LM_CInc(DM_Loc(A));
    
    HPRINT((STDOUT, "PB descriptor conversion tests from *%=%s \n", DM_TRACE(0, A)));

    if (PBCacheOn)
      PDescCacheOn();
    for (i_=1; i_ < (DiagComp? 1: M)+1; i_++)  
      for (j=1; j < N+1; j++) {	  
	DistMat_RVAlloc(3); int x;
	i = DiagComp? j: i;
	info += PDescChkConv(0, 0, M-i, N-j, LM_ADR(DM_Loc(A)), i+1, j+1, 
		 		  desc, 17, &A_[0]);
	info += PDescChkConv(1, 0, M-i, 1, LM_ADR(DM_Loc(A)), i+1, j, 
		 		  desc, 7, &A_[1]);
	info += PDescChkConv(2, 0, 1, N-j, LM_ADR(DM_Loc(A)), i, j+1, 
		 		  desc, 12, &A_[2]);
	B_[1] = SubMat(A, i, j-1); B_[2] = SubMat(A, i-1, j);
	B_[0] = SubMat(A, i, j); 
	for (x=0; x < 3; x++) { 
	  if (!DistMat_Eq(A_[x], B_[x])) {
	    if (NErrs == 0) {
    	      CPRINTF(("1st error: (%d,%d) from displacement %d (%d,%d) %s != %s\n",
		    i, j, x, x==1? i-1: i, x==0? j-1: j,  
		    DM_TRACE(0, B_[x]), DM_TRACE(1, A_[x])));
	     }	
	     exit (1);
	     NErrs++; break;
	  } 
        } /* for (x...) */
     } /* for (j..) */
  }
#endif /*PBLAS*/

  if (NErrs) {
    HPRINT((STDOUT, "%d ERRORS reported cell 0 under exhaustive tests\n", NErrs));
  }

  exit(NErrs!=0);
} /* main() */

