/************************************************************************/
/*                              Gen.host                                 */
/************************************************************************/
/* SCCS INFO @(#)Gen.host.c 1.1         last modified P. Strazdins, 98/12/24
   first created: P. Strazdins 95/05/13
*/
/* generic host program, having the following usage:			*/
#define USAGE "Gen.host Ncely Ncelx CellProgram <cell program parameters>"
/* This program configures a Ncely x Ncelx AP1000 cell program CellProgram,
 * and briadcasts the date and <cell program parameters> to all cells.
 * It then simply waits from messages from cells, assumed to be strings,
 * and prints them either to stdout or a logfile (see "LogIO.h").
 * It terminates shortly after abnormal termination detected on cells,
 * or all cells have exited.
 */

#include "Host.h"
#include <unistd.h>			/* for sleep() */
#define MAX_ERR_WAIT	2		/* wait in s for stray messages after */
					/* cell error 1st detected */
#define MAX_EXIT_WAIT	1		/* wait after all cells exited */

#define MAX_ERR_WAIT_ENV	"HOST_MAX_ERR_WAIT"
main(argc, argv) 
  int argc; char *argv[]; {  	

  time_t t_start = time(NULL);		/* start time */
  time_t t_exit = 0;			/* time 1st error/all cells exit detected */
  int ErrExit = 0;			/* true if cell error 1st occured >= */
  					/* maxErrWait s ago */  
  int maxErrWait = MAX_ERR_WAIT; 				
  char Date[64], LogFileName[64];
  int Ncelx, Ncely; char *CellProg;
  

  char CellMsg[512];			/* messages read from cell */
  int mtype; int mlen;
  
  if (argc < 4) {
    printf("usage: %s", USAGE);
  }   
  Ncely = atoi(argv[1]);		Ncelx = atoi(argv[2]);
  CellProg = argv[3];
  if (getenv(MAX_ERR_WAIT_ENV) !=NULL)
    maxErrWait = atoi(getenv(MAX_ERR_WAIT_ENV));  

  GetDate(Date);
  GetLogFileName(CellProg, Date, LogFileName)
  printf("Gen.host: setting up %d x %d cell program %s at %s", 
          Ncely, Ncelx, CellProg, Date);
  SetupProcess(Ncely, Ncelx, CellProg);

  Chv2Str(argv, 3, argc, '\0', CellMsg, mlen);
  COMMUN_ARR(CellMsg, mlen);
  COMMUN_ARR(Date, strlen(Date)+1);

  while (!ErrExit) {
    if (QueryCellRecv()!=NULL) {
      mtype = GetMsgType();	      	ReadStrMsg(CellMsg);
      if (mtype == LOGFILE_MSGTYPE) {
	LogFileAppend(LogFileName, CellMsg);
      } else
        printf("%s", CellMsg);
    } else { 
      if (CellError()) {
        if (t_exit == 0)
          t_exit = time(NULL);
        else  
          ErrExit = ((int) (time(NULL) - t_exit)) > maxErrWait;
      } else if (AllCellsExit()) {
        if (t_exit == 0)
          t_exit = time(NULL);
        else  
          ErrExit = ((int) (time(NULL) - t_exit)) > MAX_EXIT_WAIT;
        if (ErrExit) 
	  ErrExit = -1;
      }
      sleep(1);
    }   
  } /* while (...) */
  if (ErrExit < 0) ErrExit = 0;
  printf("Gen.host: exiting with status %d after %d s\n", 
  	  ErrExit, (int) time(NULL) - t_start);   
  exit(ErrExit);
} /* main() */

