/************************************************************************/
/*                              HostMsg.c                               */
/************************************************************************/
/* SCCS INFO @(#)HostMsg.c 1.2         last modified P. Strazdins, 99/03/19
   first created: P. Strazdins 95/09
*/
/* host library used for processing error and timing values gathered from cell 
 * programs, and for their display. Here all matrices must be packed in 
 * row-major order, ie A[i,j] = *(A+i*N+j) if A is M x N.
 */


static void PrintCoords(M, N, MxDiffI, MxDiffJ)
/* post: prints MxDiffI,MxDiffJ[0..M-1][0..N-1] in row-major tuple format
 */
int N, M, MxDiffI[], MxDiffJ[];
{ int i, j;
  for (i=0; i < M; i++) {
    for (j=0; j < N; j++)
      HPRINT((STDOUT, "(%d,%d) ", MxDiffI[i*N+j], MxDiffJ[i*N+j]));
    HPRINT((STDOUT, "\n"));
  }
} /* PrintCoords() */


static void PrintIntMat(M, N, A)
int N, M, A[];
/* post: prints A[0..M-1][0..N-1] in row-major format
 */
{ int i, j;
  for (i=0; i < M; i++) {
    for (j=0; j < N; j++)
      HPRINT((STDOUT, " %4d", A[i*N+j]));
    HPRINT((STDOUT, "\n"));
  }
} /* PrintIntMat() */

static int SumIntMat(M, N, A)
  int N, M, A[];
/* post: returns Sum: 0 <= i < M 0 <= j , N A[i,j]
 */
{ int i, j, s=0; 
  for (i=0; i < M; i++) 
    for (j=0; j < N; j++)
      s += A[i*N+j];
  return (s);    
} /* SumIntMat() */
 

static void GatherMessage(Ncely, Ncelx, Array, SizeElts)
int Ncelx, Ncely, SizeElts; char Array[];
/* pre:  SizeElts > 0
 *  post: Array[i*Ncelx+j] = message of size SizeElts received from task tid 
 *        on cell (i,j)
 */
{ int i, j;
  for (i=0; i < Ncely; i++)
    for (j=0; j < Ncelx; j++) {
#ifndef HOST 	    /* avoid sending to self as could deadlock under VPPLIB*/
      if (i!=0 || j!=0)
#endif           
      CELL_RECV(i, j, &Array[(Ncelx*i+j)*SizeElts], SizeElts);
    }
} /* GatherMessage() */


static void PrintMatrix(Name, Array, M, N, Ld)
  char Name; int M, N, Ld; FD_BASE Array[];
{ int i, j;
  for (i=0; i < M; i++) {
    HPRINT((STDOUT, "%c ", Name));	
    for (j=0; j < N; j++)
      HPRINT((STDOUT, "%2.3f ", Array[i*Ld+j]));
    HPRINT((STDOUT, "\n")); 
    Name = ' ';
  }
} /* PrintMatrix() */


static FD_BASE MaxMatrix(Array, Ncely, Ncelx)  
  int Ncelx, Ncely; FD_BASE Array[];
{ int i; FD_BASE Mx=Array[0];
  for (i=1; i < Ncelx*Ncely; i++) 
    if (Array[i] > Mx)
         Mx = Array[i];
  return(Mx);  
} /* MaxMatrix() */


static FD_BASE SumMatrix(Array, Ncely, Ncelx)  
  int Ncelx, Ncely; FD_BASE Array[];
{ int i; FD_BASE Sum=0.0;
  for (i=0; i < Ncelx*Ncely; i++) 
    Sum += Array[i];
  return(Sum);  
} /* SumMatrix() */
