/************************************************************************/
/*                              IntOp.Parameters.c                     */
/************************************************************************/
/* SCCS INFO @(#)IntOp.Parameters.c 1.1         last modified P. Strazdins, 98/12/24
   first created: P. Strazdins 97/01
*/
/* input command line parameters into variables for timing elementary 
 * integer operations, with the usage: 					
 *	PROGNAME		
 * followed optionally by following USAGE string:			*/
#define USAGE \
"[-X R] [-r r] [-s] [vl iop vr n]"
/* with default values:							*/
#define DEFAULTS \
"R=r=1"
/* performs R timings of r x n (vl iop vr) integer operations
 * if -s specified, corresp. time for integer addition is subtracted.
 */

/* command line parameters; see header comment*/
  static int R=1, r=1, SubAddTim=0, vl=1, vr=1, n=1;
  static char iop='*';

static void PrintUsage(), PrintParameters();

static void GetParameters(argc, argv)
  int argc; char *argv[];
/* post: set IntOp command line variables according to argv[0..argc-1] */
{ extern char *optarg;
  extern int optind;
  int c;

  while ((c = getopt(argc, argv, "X:r:s")) != -1) /* parse options */
    switch (c) {
      case 'r':
        r = atoi(optarg);                                           	break;
      case 'X':
        R = atoi(optarg);                                           	break;
      case 's':
        SubAddTim = 1;                                           	break;
      case '?':
        PrintUsage();
        exit (2);
    }

  if (optind+1 == argc) {
    PrintUsage();
    exit (0);
  }
  if (optind + 3 >= argc || argc<=4) { 
      HPRINT((STDOUT, "%s: insufficient parameters vl iop vr n given.\n", PROGNAME));
    exit (2);
  }

  /* get non-optional parameters */
  vl = atoi(argv[optind]);
  iop = *argv[optind+1];
  vr = atoi(argv[optind+2]);
  n = atoi(argv[optind+3]);
} /* GetParameters() */


static void PrintUsage()
/*post: prints program usage to stdout*/
{ HPRINT((STDOUT, "usage : %s %s\n", PROGNAME, USAGE));
  HPRINT((STDOUT, "with default values: %s\n", DEFAULTS));
  PrintParameters();
} /* PrintUsage() */

static void PrintParameters()
/*post: print out current values of variables set by command line to stdout*/
{ HPRINT((STDOUT, "%s: %d trials of time %dx%d  of integer  %d%c%d caclculations \n",
	 PROGNAME, R, r, n, vl, iop, vr)); 
} /* PrintParameters() */

