/************************************************************************/
/*                              IntOp.cell.c                           */
/************************************************************************/
/* SCCS INFO @(#)IntOp.cell.c 1.1         last modified P. Strazdins, 98/12/24
   first created: P. Strazdins 97/01
*/
/* cell program for testing elementary integer operation speed (*, /, %)
 * see  IntOp.Parameters.c for meaning.
 *
 * compilation:
 *	Solaris-1 supersparc: 	
 *		gcc -O -mv8 -o IntOp IntOp.cell.c 
 *      Solaris-2 ultrasparc: 
 *:	gcc -O -DSOL2 -mv8 -o IntOp IntOp.cell.c 
 *		cc  -O -DSOL2 -xarch=v8plus -xchip=ultra IntOp IntOp.cell.c
 */
/************************************************************************/
/*#define UNSIGNED unsigned*/		/* may be faster */
#define UNSIGNED
#include "Cell.h"
#include "Timer.c"
#include <stdlib.h>

#define PROGNAME "IntOp.cell"
#include "IntOp.Parameters.c"

static double min(v, n)
  double v[]; int n;
{ int i; double m = v[0]; 
  for (i=1; i < n; i++)
    if (v[i] < m)
      m = v[i];
  return (m); 
}

static void IntVecInit(A, n, v) 
  int A[], n, v;
{
  int i;
  for (i=0; i<n; i++)
    A[i] = v;
}

void IntOp(n, A, op, v, B) 
  UNSIGNED int A[], n, v, B[];  char op;
{ 
  int i;
  if (op=='*') 
    for (i=0; i<n; i++)
      B[i] = A[i] * v;
  else if (op=='/') 
    for (i=0; i<n; i++)
      B[i] = A[i] / v;
  else if (op=='%') 
    for (i=0; i<n; i++)
      B[i] = A[i] % v;
  else 
    for (i=0; i<n; i++)
      B[i] = A[i] + v;
}  

void main(argc, argv) 
  int argc; char *argv[];
{ int *A, *B;			/* source & destination vectors */
  int NOps;			/* number of elemntary operations in computation*/
  double mtim1, mtim2, *tim1, *tim2;		/* timing variables */	
  int i, r1, R1;
  char Date[32], ResultMsg[128];

  CELL_INIT;
  STDOUT_SETUP(STDOUT);
     
  InitArgs(argc, argv);
  GetDate(Date);

  Chv2Str(argv, 0, argc, ' ', ResultMsg, r1/*dummy*/);
  CPRINTF(("%s cell (%d,%d) of %dx%d: %s\n", Date, CID_I, CID_J, NP_I, NP_J, ResultMsg));
  GetParameters(argc, argv);
  PrintParameters(); 

  tim1 = malloc(R*sizeof(double));
  tim2 = malloc(R*sizeof(double));
  A = malloc(n*sizeof(int));
  B = malloc(n*sizeof(int));

  IntVecInit(A, n, vl);
  IntVecInit(B, n, 0);

  for (R1=0; R1 < R; R1++) {
    start_timer();
    for (r1=0; r1 < r; r1++) 
      IntOp(n, A, iop, vr, B);
    tim1[R1] = end_timer(0);
  }
  mtim1 = min(tim1, R);

  if (SubAddTim) {
    for (R1=0; R1 < R; R1++) {
      start_timer();
      for (r1=0; r1 < r; r1++) 
        IntOp(n, A, '+', vr, B);
      tim2[R1] = end_timer(0);
    }
    mtim2 = min(tim2, R);
    mtim1 -= mtim2;
  }

  NOps=r*n;

  if (CID == 0) {
    HPRINT((STDOUT, "for %1.3fM %d%c%d operations, time: %f, time/op %e\n",
	  NOps/1000000.0, vl, iop, vr, mtim1, mtim1 / NOps));
    if (SubAddTim) 
    HPRINT((STDOUT, "for the same +operations, time: %f, time/op %e\n",
	    mtim2, mtim2 / NOps));
  }
} /* main() */
