/************************************************************************/
/*                              LogIO                                   */
/************************************************************************/
/* SCCS INFO %Z%%M% %I%         last modified P. Strazdins, %E%
 * first created: P. Strazdins 93/12/17
 */
/* general macros for string and file manipulation;
   used for IO communication for cell and host programs
*/

#ifndef __LogIO_Incl__
#define __LogIO_Incl__

#include <stdio.h>
#include <string.h>
#define APPEND_FILE "a"

#include <time.h>

#define GetDate(Date)                                                   \
        { time_t t = time(NULL);         char * date = ctime (&t);      \
          strcpy(Date, date);                                           }

					/* types for cell -> host messages */
#define STDOUT_MSGTYPE          0	/* to be printed on stdout */
#define LOGFILE_MSGTYPE         7	/* to be printed on log file */

#define GetLogFileName(CellProgName, Date, LogFileName)			\
	/* form a name of a log file out of the cell program name and date */\
	{ char month[4], day[4];   					\
	  sscanf(Date, "%*s %s %s", month, day);			\
	  sprintf(LogFileName, "LOG/%s.%s%s", CellProgName, month, day);\
	}

#define LogFileAppend(LogFileName, str)					\
	/* append str on to the file of name LogFileName */		\
	{ FILE *logfile;						\
  	  logfile = fopen (LogFileName, APPEND_FILE);			\
  	  if (logfile != NULL) {					\
    	    fputs(str, logfile);					\
    	    fclose(logfile);						\
  	  }								\
	}								

#define Chv2Str(chv, i1, i2, sep, str, len)				\
	/* str[0..len-1] = concat. of chv[i1..i2-1], using separator sep*/\
	{ int i_; (len) = 0; 						\
	  for (i_ = (i1); i_ < (i2); i_++) {				\
	    int li = strlen((chv)[i_]);					\
	    strcpy(&(str)[len], (chv)[i_]);				\
	    (str)[(len)+li] = (sep);					\
	    (len) += li+1;						\
	  } 								\
	  (str)[len] = '\0';						\
	}
	     	
#define Str2Chv(str, len, chc, chv)					\
	/* chv[0..chc-1] points to the chc strings in str[0..len-1] */	\
	{ int r_ = (len);		char *s_ = (str);		\
	  (chc) = 0;							\
	  while (r_ > 0) { 						\
	    (chv)[chc] = s_;		(chc)++;			\
	    r_ -= strlen(s_) + 1;	s_ += strlen(s_) + 1;		\
	  }								\
	}

#endif /*__LogIO_Incl__*/
