/************************************************************************/
/*                              LogResults.c                               */
/************************************************************************/
/* SCCS INFO @(#)LogResults.c 1.1         last modified P. Strazdins, 98/12/24
 * first created: P. Strazdins 93/12/17
 */
/* small library for logging execution summaries of DBLAS test programs
   (host); this is required for exhaustive testing, especially via
   cap_batch, where the essential information for each run:
   	the date,
   	the program's command line parameters (determining the test case),  
	the overall number, maximum and averafge errors, and
	the MFLOP rating per cell and overall execution time
   is recorded on a single line of a logfile. If the test did not terminate 
   normally, only the first two of these will be recorded.
   The name of the log file is given by:
   	DBLASLOG_<month><day>	
   where DBLASLOG is an externally-defined cpp FLAG.
*/

#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/time.h>
#include <stdio.h>

#define NULL 0               
#define APPEND_FILE "a"

LogDateParams(argc, argv)
  int argc; char *argv[];
/* post: append '\n' date ' # ' command line parameters to log file
 *	 of name DBLASLOG_<month><day>; nb. the line is not closed
 */
{ char *date, ignore[16], month[8], day[8], ttime[16], logfilename[128], logline[256];
  FILE *logfile;
  int i;
  time_t t;
  t = time(NULL);
  date = ctime (&t);  /* date contains date in 26 char format) */
  sscanf (date, "%s %s %s %s", ignore, month, day, ttime);
  
  sprintf (logfilename, "%s_%s%s", DBLASLOG, month, day);

  sprintf (logline, "\n%s %s %s # ", month, day, ttime);
  for (i=0 ; i < argc; i++) {
    strcat (logline, argv[i]); 
    strcat (logline, " ");
  }

  logfile = fopen (logfilename, APPEND_FILE);
  if (logfile != NULL) {
    fputs (logline, logfile);
    fclose (logfile);
  }
}

LogResults(nerrs, maxerr, averr, flopcount, seconds)
  int nerrs, seconds; double maxerr, averr, flopcount;
/* post: append '#e" ' nerrs maxerr averr '#s: flopcount seconds\n' to log file
 *	 of name DBLASLOG_<month><day>
 */
{ char *date, ignore[16], month[8], day[8], logfilename[128], logline[256];
  FILE *logfile;

  time_t t;
  t = time(NULL);
  date = ctime (&t);  /* date contains date in 26 char format) */
  date = ctime (&t);  /* date contains date in 26 char format) */
  sscanf (date, "%s %s %s", ignore, month, day);
  
  sprintf (logfilename, "%s_%s%s", DBLASLOG, month, day);

  sprintf (logline, " #e: %d %.5f %.5f #s: %.2f %d\n", nerrs, maxerr, averr, flopcount, seconds);

  logfile = fopen (logfilename, APPEND_FILE);
  if (logfile != NULL) {
    fputs (logline, logfile);
    fclose (logfile);
  }
}

LogResultsApp(head, n, f1, f2, tail)
  char *head, *tail; int n; double f1, f2;
/* post: append ' *head: n f1 f2 *tail\n' to log file
 *	 of name DBLASLOG_<month><day>
 */
{ char *date, ignore[16], month[8], day[8], logfilename[128], logline[256];
  FILE *logfile;

  time_t t;
  t = time(NULL);
  date = ctime (&t);  /* date contains date in 26 char format) */
  date = ctime (&t);  /* date contains date in 26 char format) */
  sscanf (date, "%s %s %s", ignore, month, day);
  
  sprintf (logfilename, "%s_%s%s", DBLASLOG, month, day);

  sprintf (logline, " %s: %d %.5f %.5f %s", head, n, f1, f2, tail);

  logfile = fopen (logfilename, APPEND_FILE);
  if (logfile != NULL) {
    fputs (logline, logfile);
    fclose (logfile);
  }
}

int SumIntVec(a, n)
  int a[], n;
/* post: returns sum i: 0<=i<n: a[i]
 */
{ int i, s=0;
  for (i=0; i < n; i++)
    s += a[i];
  return (s);
}
 

