/************************************************************************/
/*			 	MF.cell.c				*/
/************************************************************************/
/* program for testing level1 matrix functions correstness & performance
 * performs B = f(A), A is M x N 	(see MF.Parameters.c)
 * (externally-defined) cpp FLAGs:
 *	FDATATYPE	SINGLE, DOUBLE
 */
/* SCCS INFO @(#)MF.cell.c 1.3         last modified P. Strazdins, 99/03/19
   first created: P. Strazdins 92/01
 */
/************************************************************************/
#include "FDATASetUp.h"

#include "Cell.h"
#include "Timer.c"

#include "LocMat.h"
#include "FDAssert.h"
#include "FDMalloc.h"  
#include "FDCommun.h"  
#include "CellMatrix.h"
#include "CellVector.h"

#define NO_ALIGN
#undef __FDATAMatrixCopy_INCL__
#include "CellMatCopy.h"
#undef NO_ALIGN

#define PROGNAME "MF.cell"
#include "MF.Parameters.c"

#include "FDATAOps.h"
#ifdef BLAS
# include "FtnBLAS.h"
#else
# include "BLASInnerLoops.h"
#endif /*BLAS*/

#include "AuxMatrix.c"


static double Min(a, n)
 double a[]; int n;
{ int i; double m=a[0];
  for (i=1; i < n; i++)
   if (a[i] < m) m = a[i];
  return (m);
}
 
static double Avg(a, n)
 double a[]; int n;
{ int i; double m=a[0];
  for (i=1; i < n; i++)
   m += a[i];
  return (m/n);
}

static double var(a, n)
  double a[]; int n;
{ int i; double s=0.0, v=Avg(a, n), sqrt();
  for (i=1; i < n; i++) {
    double d = a[i] - v;
    s += d*d;
  }
  s = sqrt(s/(n-1));
  return(s);
}  

/*********************** specialized checking routines *******************/ 

static int VecAbsMaxPosS(m, A, IncA)
  FDATA A[];  int m, IncA;
/* post: returns 1st. 0 <= i < m: A[i] and |A[i]| >= |A[j]| for 0 <= j < m
 */
{ int  i, M=-1; FDATA M0 = FD_0; FD_BASE Max=0.0;
  for (i=0; i < m; i++) {
    M0 = *A;
    	A += IncA;   
                     if (FD_ABS(M0) > Max) { Max = FD_ABS(M0); M=i; };
  } 
  return (M);
} /* VecAbsMaxPosS() */


static FD_BASE VecAbsSumS(m, A, IncA)
  FDATA A[];  int m, IncA;
/* post: returns Sum: 0 <= i < m: |A[i]|
 */
{ int i; FDATA M0=FD_0; FD_BASE Sum=0.0;
  for (i=0; i < m; i++) {
    M0 = *A;
                        Sum += FD_ABS(M0);
    A += IncA;
  }
  return (Sum);   
} /* VecAbsSumS() */

/**************************************** main program ******************/
#define ASIZE 		(LdA*(!TransA? M: N))
#define BSIZE 		(LdB*(!TransB? M: Nb))
static FDATA 
	AIJOFFS	= 	FD_MKVAL(1.0, -1.0),
	AIMULT =  	FD_MKVAL(2.0, -1.0),
	AJMULT =  	FD_MKVAL(1.0, -2.0),
	BIJOFFS = 	FD_MKVAL(0.5, -0.5),
	BIMULT =  	FD_MKVAL(2.0, -1.0),
	BJMULT =  	FD_MKVAL(1.0, -2.0);

#ifndef APLIB
#define bcopy(a, b, r)	memcpy(b, a, r)
#endif

main(argc,argv) 
  int argc; char *argv[]; 
{
	LocMat A, B;	int Nb;	     		FDATA FD_alpha;
	int NErrs=0;  	FD_BASE es = 0.0;	/* error values */
	int i, j;
        double speed, *tcomp, *utcomp;		/* timing */
	char Date[64];

	CELL_INIT;
        STDOUT_SETUP(STDOUT);
	
	InitArgs(argc, argv);
        GetDate(Date);

        for (i=0; i < argc; i++)
          HPRINT((STDOUT, "%s ", argv[i]));
        HPRINT((STDOUT, "%s", Date));

  	GetParameters(argc, argv);
        FD_alpha = alpha;

  	PrintParameters(); 
	CheckParameters(); 

        COMMUN_INIT(Ncely, Ncelx);

	Nb = (f<=2 || f==5? N: 1);
	LM_Init(A, NULL, TransA? 1: LdA, !TransA? 1: LdA);
	LM_Init(B, NULL, TransB? 1: LdB, !TransB? 1: LdB);

	MatMalloc(PROGNAME, 1, &(LM_ADR(A)), ASIZE+BSIZE+SingleWordBoundary);
        LM_ADR(A) += SingleWordBoundary;
	LM_ADR(B) = LM_ADR(A) + (IdenticalAB? 0: ASIZE); 

	MatInit(RandElts, BIMULT, BJMULT, BIJOFFS, M, Nb, B);
	MatInit(RandElts, AIMULT, AJMULT, AIJOFFS, M, N, A);
	MatPrint("A", MIN(M, PrintM) , MIN(N, PrintN), A);
	if (f==1||f==2)
	  MatPrint("B", MIN(M, PrintM) , MIN(N, PrintN), B);

	if (f==5) 
          MatCopy(M, Nb, &A, &FD_1, &B);

	if (f==2 && NReps%2 == 0) NReps++;	/* ensure odd for checking */
	if (f==2 && TReps%2 == 0) TReps++;	/* ensure odd for checking */

        tcomp = (double *) memalign(sizeof(double), sizeof(double)*TReps);  
 	utcomp = (double *) memalign(sizeof(double), sizeof(double)*TReps);
        for (i=0; i < TReps; i++) {
	start_timer();
  	for (j=NReps; j!=0; j--) {
  	  int m;
	  if (f==0)
            if (AlgVers==0 || !FD_EQ(FD_alpha, FD_1) || TransA || TransB)
	      MatCopy(M, N, &A, &FD_alpha, &B);
	    else {
	      for (m=0; m<M; m++)
	        if (AlgVers==1)
		  memcpy((char *) LM_Index(B, m, 0), 
			 (char *) LM_Index(A, m, 0), N*sizeof(FDATA));
		else if (AlgVers==2)
 		  bcopy((char *) LM_Index(A, m, 0), (char *) LM_Index(B, m, 0),
		        N*sizeof(FDATA));
		else if (AlgVers==3)
 		  VecCopy(N, LM_Index(A, m, 0), 1, LM_Index(B, m, 0), 1);
		else {
		  int k; 
 		  FDATA *Ax = LM_Index(A, m, 0), *Bx = LM_Index(B, m, 0);
		  for (k=N-1; k > 0; k--)
		    Bx[k] = Ax[k];
	       }	
	    }
	  else if (f==1)
	    MatAdd(M, N, &A, &FD_alpha /*NULL*/, &B);
	  else if (f==2)
	    for (m=0; m < M; m++)
	      VecSwap(N, LM_Index(A, m, 0), LM_CInc(A), LM_Index(B, m, 0), LM_CInc(B));
	  else if (f==3) {
	    for (m=0; m < M; m++) {
	      FD_SET_REAL(*LM_Index(B, m, 0), VecAbsSum(N, LM_Index(A, m, 0), LM_CInc(A)));
	    }
	  } else if (f==4) 
	    for (m=0; m < M; m++) {
              int mj = VecAbsMaxPos(N, LM_Index(A, m, 0), LM_CInc(A));
	      *LM_Index(B, m, 0) = *LM_Index(A, m, mj);
	    }
 	  else if (f==5)
	    MatScale(M, N, &FD_alpha, &B);
	} /* for (j...) */
	tcomp[i] = end_timer(0); 		utcomp[i] = UserEndTimerVal;
        } /* for (i...) */

        speed = NReps * ((double) M*N) / 1.0E6 / Min(tcomp, TReps);
        HPRINT((STDOUT, "Computed %d reps of %dx%d f=%d (%d OPs) in %Gs at %G MW/s (%G MB/s)\n",
                NReps, M, N, f, NReps*M*N, Min(tcomp, TReps), speed, 
		speed*sizeof(FDATA)));
        speed = NReps * ((double) M*N) / 1.0E6 / Min(utcomp, TReps);
        if (Verbose) {
          HPRINT((STDOUT, "	user time in %Gs at %G MW/s (%G MB/s)\n",
                  Min(utcomp, TReps), speed, speed*sizeof(FDATA)));
          if (TReps>1) {  
            HPRINT((STDOUT, "%%var in t=%.3f (ut=%.3f); speed for 1st %d reps:\n",
                  var(tcomp, TReps)/Avg(tcomp, TReps), 
		  var(utcomp, TReps)/Avg(utcomp, TReps), MIN(10, TReps)));
            for (i=0; i < MIN(10, TReps); i++)
              HPRINT((STDOUT, "%.2f (%.2f)  ", 
			NReps * ((double) M*N) / 1.0E6 / tcomp[i], 
			NReps * ((double) M*N) / 1.0E6 / utcomp[i]));
            HPRINT((STDOUT, "\n"));
          }
        }

	if (f<=1) {
	  if (!RandElts) {
	    FDATA AR = FD_alpha;
	    if (f==1)
    	      FD_ASSIGN_SC(FD_alpha, *=, TReps*NReps, TReps*NReps);
	    FD_MULSET(AIMULT, =, AIMULT, FD_alpha);
	    FD_MULSET(AJMULT, =, AJMULT, FD_alpha);
	    FD_MULSET(AIJOFFS, =, AIJOFFS, FD_alpha);
	    if (f==1) {
	      FD_ASSIGN(AIMULT, +=, BIMULT);
	      FD_ASSIGN(AJMULT, +=, BJMULT);
	      FD_ASSIGN(AIJOFFS, +=, BIJOFFS);
	    } 
	    NErrs = MatCheck(RandElts, AIMULT, AJMULT, AIJOFFS,  M, N, B);
	   } else {
	     HPRINT((STDOUT, "warning: cant check matrix copy / add for random matrices\n"));
 	   }
	} else if (f==2) {
	  NErrs = MatCheck(RandElts, AIMULT, AJMULT, AIJOFFS, M, N, B);
	  NErrs += MatCheck(RandElts, BIMULT, BJMULT, BIJOFFS, M, N, A);
	} else if (f==3)
	  for (i=0; i < M; i++)	{   
	    FD_BASE s = VecAbsSumS(N, LM_Index(A, i ,0), LM_CInc(A));
	    NErrs += (FD_REAL(*LM_Index(B, i, 0)) != s);
	    s = FD_REAL(*LM_Index(B, i, 0)) - s; s = ABS(s);
	    es += s; 
	  }
	else if (f==4) 
	  for (i=0; i < M; i++) {
	    int mj = RandElts? VecAbsMaxPosS(N, LM_Index(A, i, 0), LM_CInc(A)): 
		     ((FD_ABS(*LM_Index(A, i, 0)) > FD_ABS(*LM_Index(A, i, N-1)))?
		      0: N-1);
	    NErrs += !FD_EQ(*LM_Index(B, i, 0), *LM_Index(A, i, mj));
	  }
	else if (f==5) 
	  if (!RandElts) {
	    int i;
	    FDATA R = FD_1;
	    for (i=0; i < TReps*NReps; i++)
	      FD_MULSET(R, =, R, FD_alpha);
	    FD_MULSET(AIMULT, =, AIMULT, R);
	    FD_MULSET(AJMULT, =, AJMULT, R);
	    FD_MULSET(AIJOFFS, =, AIJOFFS, R);	    
	    NErrs = MatCheck(RandElts, AIMULT, AJMULT, AIJOFFS, M, N, B);
	   } else {
	     HPRINT((STDOUT, "warning: cant check matrix scale for random matrices\n"));
 	   }
	HPRINT((STDOUT, "NErrs=%d (%G)\n", NErrs, es)); 
        HPRINT((STDOUT, "\n\n"));

	MatPrint("B", MIN(M, PrintM) , MIN(Nb, PrintN), B);

        CELL_FINALIZE;
	exit(NErrs!=0);
} /* main() */

