/************************************************************************/
/*                              Timer.c                                 */
/************************************************************************/
/* SCCS INFO %Z%%M% %I%         last modified P. Strazdins, %E%
 *  first created: P. Strazdins 95/09
 */
/* module to provide simplified timer interface for Solaris 1 & 2
 * workstations or AP cells.   
 * note: currently BLACS versions must run on AP cells
 */
  
#if defined (BLACS)
# include 	"FtnBLACS.h"
# define 	SYNCH 	FTN_BLACS_BARRIER(&BLACSCntxt_, "A")
# define	DTIMER	FTN_DWALLTIME00
#elif defined(APLIB)
# define 	SYNCH 	sync(0,0)
# define	DTIMER	dgettime
#elif defined(MPI)
double MPI_Wtime();
# define	SYNCH	MPI_Barrier(MPI_COMM_WORLD)
# define	DTIMER	MPI_Wtime
#else
# define 	SYNCH
#endif
#ifndef APLIB
# ifdef sun5
#   include 	<sys/time.h>
#   include 	<sys/times.h>
#   include 	<limits.h>
# else
#   include 	<sys/types.h>
#   include 	<sys/time.h>
#   include 	<sys/timeb.h>
#   include 	<sys/times.h>
#   define HZ 60		/* unit for time in times.h */
# endif /*sun5*/
#endif /*APLIB*/

double UserEndTimerVal;		/* records user time for last EndTimer() call */

#if defined(APLIB) || defined (BLACS) || defined (MPI)
static double tim[3]; 
#endif
#ifndef APLIB
# ifdef sun5
static hrtime_t t0, t1, t2;
# else
static struct tms ut0, ut1;
# endif /*sun5*/
static struct tms ut0, ut1;
#endif /*APLIB*/


static void start_timer()
{ SYNCH;	
#ifndef APLIB
  times(&ut0);
#endif
#if defined(APLIB) || defined (BLACS) || defined (MPI)
  tim[0] = DTIMER();  
#elif defined(sun5)  
  t1 = gethrtime();  
#else  
  ftime(&t1);
#endif
} /* start_timer() */


static double end_timer(s) 
  int s;     
{ double wt;	
  if (s)
    SYNCH; 
#if defined(APLIB) || defined (BLACS) || defined(MPI)
  tim[1] = DTIMER();
  if (s)
    SYNCH;
  tim[2] = DTIMER();
  wt = tim[1]-tim[0]-(tim[2]-tim[1]);
#elif defined(sun5)
  t2 = gethrtime();
  t0 = gethrtime() - t2;  
  wt = (double)(t2 - t1 - t0) / 1.0E9;
# else  
  ftime(&t2);  
  wt = (1000.0*t2.time + t2.millitm - 1000.0*t1.time - t1.millitm)/ 1000.0;
#endif  
#ifdef APLIB 
  UserEndTimerVal = tim[1] - tim[0];
#else
  times(&ut1);
  UserEndTimerVal = (double) (ut1.tms_utime - ut0.tms_utime) / (double) CLK_TCK;  
  return (wt);
#endif /*APLIB*/
} /* end_timer() */

