/************************************************************************/
/* SCCS INFO %Z%%M% %I%         last modified P. Strazdins, %E%
   first created: P. Strazdins 01/01
 */
   
/* --------------------------------------------------------------------- */
/* addfit_solve - Solve the linear system from				 */

   int addfit_solve(double *a, double *b, int N, double *x, int P, int Q, 
   		    int r, int w);					 

/* - a is a upper triangular matrix, stored row-major			 */
/*   This matrix is allocated on the first P*Q nodes, but not replicated */
/*   I.e. a = sum(A_0 + a_1 + ... a_{Q*P-1}				 */
/* - b is a double vector						 */
/*   This vector is allocated on the first P*Q nodes, but not replicated */
/*   I.e. b = sum(b_0 + b_1 + ... b_{Q*P-1}				 */
/* - N is the matrix and vector dimension, N>=0				 */
/* - P, Q are the size of the processor grid, P,Q>0			 */
/* - r is the matrix block distribution size, r>0		 	 */
/* - w is the logical blocking factor (generally = r, can be > r)	 */
/* - Returns 0 if OK or a negative number if something went wrong	 */
/*   result vector x is replicated on all p*q cells.			 */
/* - upon memory allocation failure, a cell will abort with an error message;*/
/*   other cells will hang. 						 */
/* --------------------------------------------------------------------- */
/* error codes: */
#define SOLVE_COMM_ERROR	-1	/* could not set up P*Q sub-grid */ 
#define NON_INV_A		-2	/* A was non-invertible */
#define RESID_FAIL		-3	/* failed residual check */

/* note: residual checking can be suppressed by not defining ADDFIT_RESID_CHECK
 *	 in addfit_solve.c. In this case, RESID_FAIL cannot be returned */

