/************************************************************************/
/*      common macro definitions and headers for files implementing     *
 *      the PBLAS interface to the DBLAS. Includes error checking prcoedures*/
/************************************************************************/
/* SCCS INFO @(#)PBLASCheck.c 1.2         last modified P. Strazdins, 99/05/03
   first created: P. Strazdins 96/10
 */
/* compilation switches:
 *	DEBUG:	     print info to help debug the descriptor conversion process
 *	DEBUG_CACHE: similar, but specific inof for the descriptor cache
 * 	CACHE_STATS: if defined, collect & dump cache hit statistics
 *		     every time PDescCacheOff() is called
 */ 
#include "PBLASCheck.h"
#include "FtnBLACS.h"

/******************** error checking area ******************************/

#define ParamNumI                    	((ParamInfo)-2)
#define ParamNumJ                    	((ParamInfo)-1)
#define ParamNumDesc(x) 		((ParamInfo)*100 + (x) + 1)

static int PBCheckMat(Cntxt, P, desc, ParamInfo)
  int Cntxt, desc[], P[], ParamInfo;
{ 
#ifdef DEBUG
  printf("desc a%d=%d %d %d %d %d %d %d %d %d\n", (int) desc, desc[0], desc[1], desc[2], desc[3],
	  desc[4], desc[5], desc[6], desc[7], desc[8], desc[9]);
#endif	 
  if (P[0]==-1)				/* invalid context */	 
    return(ParamNumDesc(CTXT_));
  if (desc[DT_] != 1)
    return(ParamNumDesc(DT_)); 
  if (desc[CTXT_] != Cntxt)  
    return(ParamNumDesc(CTXT_));
  if (desc[MB_] < 1)  
    return(ParamNumDesc(MB_)); 
  if (desc[NB_] < 1)  
    return(ParamNumDesc(NB_));
  if ((desc[RSRC_] < 0) | (desc[RSRC_] >= P[0]))
    return(ParamNumDesc(RSRC_));
  if ((desc[CSRC_] < 0) | (desc[CSRC_] >= P[1]))
    return(ParamNumDesc(CSRC_));
  if (desc[LLD_] < 1)
    return(ParamNumDesc(LLD_));
  return(0);  
} /* PBCheckMat() */
 
static inline int PBCheckIx(M, N, Trans, i, j, desc, ParamInfo)
  int M, N, Trans, i, j, desc[], ParamInfo; 
{ 
  if (desc[M_] < 1)
    return(ParamNumDesc(M_)); 
  if (desc[N_] < 1)  
    return(ParamNumDesc(N_)); 
  if (Trans)
    SWAP(int, M, N);
  if ((i < 0) | (i+M > desc[M_]))
    return(ParamNumI);  
  if ((j < 0) | (j+N > desc[N_]))
    return(ParamNumJ);
  return(0);  
} /* PBCheckIx() */


#define PBLLDCheck(dC, desc)	\
	if (G2Ld((desc)[M_], DM_D0(dC)) > (desc)[LLD_]) return (ParamNumDesc(LLD_));

#define InfoCheckRet(v)   { int info_ = (v); if (info_ > 0) return (info_); }


/********************** inter PBLAS call cache definition *******************/
/* cache recently used PBLAS descriptors & their DistMat equivalents.
 * This can be used to avoid relatively costly MatDim construction, 
 * and also exploit locality in SibDim operations.
 * PBLASDescCacheOn() can be called within a computation such as a matrix
 * factorization where all PBLAS operands share the same descriptor.
 * At the end of this computation, PBLASDescCacheOff() should be called.
 */
static DistMatStore DM_cache[3];	/* storage for 3 DistMat opernads for 
					   all PBLAS operations */
static struct {				
  int On; 				/* only cache between calls if = 1 */
  int *Desc;				/* last PBLAS desciptor cached */
  int i, ii, j, jj, jjLd; 		/* gloabl & local offsets */
  DistMat DM;				/* corresp DistMat representation */
} MatDescCache = { 0, NULL, 0, 0, 0, 0, 0, &DM_cache[0]};    

void PDescCacheOn()
{ MatDescCache.On = 1;
  MatDescCache.Desc = NULL;		/* invalidate the cache, so that
  					   next call to PDescChkConv(0,...)
  					   must get a miss */
} /* PDescCacheOn() */

#ifdef CACHE_STATS    
static int NHits = 0, adi = 0, adj = 0, nsubd = 0, n1subd = 0, nm1subd=0; 
#endif
  
void PDescCacheOff()
{ MatDescCache.On = 0; 			MatDescCache.Desc = NULL;
#ifdef CACHE_STATS
  printf("PB Cache: Hits=%d, a(%d,%d), subdim: %d, +1=%d -1=%d\n",
         NHits, adi, adj, nsubd, n1subd, nm1subd);
#endif        
} /* PDescCacheOff() */


#define PDESC_EQ(DescA, DescB)	((DescA)==(DescB))

/********************** data structure conversion macros **************/

#define Cntxt 	BLACSCntxt
#define I_	AbsCellId
#define P_	NumCells

#ifndef BLACS				/* keep impl. indpt of BLACS library */
# undef ptopget_			/* may not exist */
# define ptopget_(a, b, c, d) 
# undef FTN_BLACS_GRIDINFO		/* assume grid is set up externally */
# define FTN_BLACS_GRIDINFO(Cntxt, I0, I1, P0, P1) 
#endif

#define P2DBLAS_COMMSETUP(desc_A)  			\
        { char BroadCastC = 'B', CombineC = 'C', RowC = 'R', ColC = 'C';\
          (Cntxt) = (desc_A)[CTXT_];                       		\
          FTN_BLACS_GRIDINFO(&(Cntxt), &P_[0], &P_[1], &I_[0], &I_[1]); \
          AssertCId[0] = I_[0], AssertCId[1] = I_[1];			\
	  VNumCells[0] = P_[0]; VNumCells[1] = P_[1];			\
/*printf("Cntxt=%d P=%d %d, I=%d %d\n", Cntxt, P_[0], P_[1], I_[0], I_[1]);*/\
          if ((Cntxt)!=-1 && P_[0]!=-1) {				\
            ptopget_(&(Cntxt), &BroadCastC, &RowC, &BrdTopChar[1]);     \
            ptopget_(&(Cntxt), &BroadCastC, &ColC, &BrdTopChar[0]);     \
            ptopget_(&(Cntxt), &CombineC, &RowC, &CombineTopChar[1]);   \
            ptopget_(&(Cntxt), &CombineC, &ColC, &CombineTopChar[0]); 	\
          }         /*TraceLevel = -1;*/				\
        }

#define ConvDesc2DistMat(Pix, TrA, DescA, dA)				\
  (dA) = &DM_cache[Pix], 						\
  LM_RInc(DM_Loc(dA)) = 1, LM_CInc(DM_Loc(dA)) = DescA[LLD_],		\
  pInitDim(DM_D0(dA), 0, DescA[MB_], CYCLIC_SUB(I_[0],DescA[RSRC_],P_[0]), P_[0]),\
  pInitDim(DM_D1(dA), 0, DescA[NB_], CYCLIC_SUB(I_[1],DescA[CSRC_],P_[1]), P_[1]), \
  DM_SetTrans(TrA, dA)

#define CacheDistMat(Pix, TrA, dA)					\
  (dA) = &DM_cache[Pix];						\
  if ((Pix) > 0) *DM_Ptr(dA) = *DM_Ptr(MatDescCache.DM);		\
  DM_SetTrans(TrA, dA);							

int PDescChkConv(Pix, TrA, M, N, Ax, iA, jA, DescA, ParamInfo, dA)
  int Pix, TrA, M, N; FDATA *Ax; int iA, jA, DescA[]; int ParamInfo; 
  DistMat *dA; 
{ int iiA, jjA, djjA, jjALd, Hit;
#ifdef DEBUG
  printf("Pix=%d, TrA=%d, M=%d N=%d, iA=%d, jA=%d, ParamInfo=%d\n",
	 Pix, TrA, M, N, iA, jA, ParamInfo);
#endif  
  if (Pix==0) {
    P2DBLAS_COMMSETUP(DescA);
  }

  iA = iA-1, jA = jA-1;	
  InfoCheckRet(PBCheckIx(M, N, TrA, iA, jA, DescA, ParamInfo));
  if ((MatDescCache.On==0 && Pix==0) || !PDESC_EQ(DescA, MatDescCache.Desc)) {
    InfoCheckRet(PBCheckMat(Cntxt, P_, DescA, ParamInfo));
    ConvDesc2DistMat(Pix, TrA, DescA, *dA);    
    PBLLDCheck(*dA, DescA);
    iiA = 0, 				jjA = 0;
    Hit = 0;
  } else {
    CacheDistMat(Pix, TrA, *dA);	    
#ifdef DEBUG_CACHE
    printf("Hit: iA,jA =%d, %d", iA, jA);    
#endif    
    iA -=  MatDescCache.i, 		jA -= MatDescCache.j; 	
    iiA = MatDescCache.ii, 		jjA = MatDescCache.jj;
    Hit = 1;	
#ifdef DEBUG_CACHE    		
    printf("-> %d %d\n", iA, jA);
#endif    
  }

  if (iA>0)
    iiA += G2LSubDim(iA, DM_D0(*dA));	
  else if (iA < 0)  
    iiA += G2LNSubDim(iA, DM_D0(*dA));	
  djjA = 0;  
  if (jA>0)  
    djjA = G2LSubDim(jA, DM_D1(*dA));
  else if (jA < 0)  
    djjA = G2LNSubDim(jA, DM_D1(*dA));
  jjA += djjA;
  jjALd = djjA==0 && Hit? MatDescCache.jjLd: DescA[LLD_] * jjA;
  LM_ADR(DM_Loc(*dA)) = Ax + iiA + jjALd;
 
  if (Pix==0) {
    if (Hit) {
      if (iA != 0)	
         MatDescCache.i += iA, 		MatDescCache.ii = iiA;
      if (jA != 0) 
         MatDescCache.j += jA, 		MatDescCache.jj = jjA,
         MatDescCache.jjLd = jjALd;   
    } else {    
      /*nb. if !MatDescCache.On && only 1 parameter, this will be unnecessary*/	
      MatDescCache.Desc = DescA;	
      MatDescCache.i = iA, 		MatDescCache.ii = iiA;
      MatDescCache.j = jA, 		MatDescCache.jj = jjA,
      MatDescCache.jjLd = jjALd;   
    }  
#ifdef DEBUG_CACHE    
    printf("MatDescCache:%d, %d %d, %d %d %d %s\n", MatDescCache.Desc,
    	   MatDescCache.i, MatDescCache.ii, MatDescCache.j, 
    	   MatDescCache.jj, MatDescCache.jjLd, DM_TRACE(0, MatDescCache.DM));
#endif    	   
  }  
  
#ifdef CACHE_STATS    
  NHits += Hit;
  if (iA != 0) {
    if (iA==1) n1subd++;		if (iA==-1) nm1subd++;
    if (iA < 0) iA = -iA; nsubd++; adi += iA; 
  }	
  if (jA != 0) {
    if (jA==1) n1subd++;		if (jA==-1) nm1subd++;
    if (jA < 0) jA = -jA; nsubd++; adj += jA; 
  }	
#endif
    
  return (0);
} /* PDescChkConv() */


#ifdef USE_PBERROR_
void PBError_(CallerPrefix, Caller, ParamNum )
  char *CallerPrefix, *Caller; int ParamNum;
{ char PBCaller[16];
  PrintLinePrefix();
  sprintf(PBCaller, "%s%s", CallerPrefix, Caller);
  printf("*** Error detected in call %s%s (cntxt=%d, err=%d)\n",  PBCaller, TraceStr, Cntxt, ParamNum);
  pberror_(&Cntxt, PBCaller, &ParamNum);
} /* PBError_() */
#endif

/************************************************************************/
/*      procedure for getting/setting DBLAS internal parameters from    */
/*      Fortran programs. These control tracing & tuning values         */
/************************************************************************/
 
#include "MachineParams.h"

void DBLASSetGet(GetSet, ParamCat, ParamNum, Val)
  int *GetSet, *ParamCat, *ParamNum, *Val;
{ int *v = NULL;
  if (*ParamCat==0) {
    if (0 <= *ParamNum  &&  *ParamNum < NUM_MACH_PARAMS)
      v = &MachParamVec[*ParamNum]; 		
  } else if (*ParamCat==1) {
    if (*ParamNum==0)
      v = &TraceLevel;
    else if (*ParamNum==1) 
      v = &TraceType;
  } else if (*ParamCat==2) {
    if (*ParamNum==0)
      v = &PrintRes;
  } else if (*ParamCat==3) {
    if (*ParamNum==0)
      v = &ForceAlgVers;
  }
  if (v == NULL) *Val = -999;
  else if (*GetSet) *Val = *v;
  else *v = *Val;
} /* dblasgetset */ 
